// Rerank.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

#include "ModelLoader.hpp"
#include "TGTReader.hpp"
#include "TGTWriter.hpp"
#include "AbstractRankedReader.hpp"
#include "StackBasedDecoder.hpp"

#include <ppl.h>
#include <unordered_map>

struct MatchPathSeparator
{
	bool operator()( char ch ) const
	{
		return ch == '\\' || ch == '/';
	}
};

std::string
	basename( std::string const& pathname )
{
	return std::string( 
		std::find_if( pathname.rbegin(), pathname.rend(),
		MatchPathSeparator() ).base(),
		pathname.end() );
}

int main(int argc, char* argv[])
{	
	std::cerr << "GENERATION STARTED ######################" << std::endl;
	
	TGTWriter* w = nullptr;
	AbstractRankedReader* r = nullptr;
	AbstractModel* m = nullptr;
	
	try
	{
		Config conf = Config::ReadCommandLine(argc, argv);

		int stackSize = conf.GetInt("stackSize", 100);
		int maxLength = conf.GetInt("maxLength", 15);
		int enforceDiversity = conf.GetInt("enforceDiversity", 0);
		int nbestSize = conf.GetInt("nbestSize", 1000);
		std::string lambdaFile = conf.GetString("lambdaFile");
		m = ModelLoader::LoadFromDisk(conf.GetString("model"));

		std::string output = "";
		if (conf.GetBool("output"))
		{
			output = conf.GetString("output");
		}
		else
		{
			std::string tail = basename(conf.GetString("nbest"));
			output = conf.GetString("model") + "/" + tail + "." + m->GetModelName() + ".tgt";
		}

		r = new TGTReader(conf.GetString("nbest"));
		w = new TGTWriter(output);

		StackBasedDecoder decoder(lambdaFile, m, stackSize, maxLength, 0, 3, enforceDiversity);

		std::ofstream outputFile;

		int numRef = 0;
		HeadRecord x;
		while (r->ReadNextCandidate(x)) {
			std::cerr << "================================" << std::endl;
			std::cerr << " Generating (" << numRef++ << ")" << std::endl;
			std::cerr << " - C = " << x.ctext << std::endl;
			std::cerr << " - M = " << x.mtext << std::endl;
			std::cerr << "--------------------------------" << std::endl;

			std::vector<std::string> mtextTokens = Tokenize(x.mtext);
			std::vector<int> mtextIds = m->GetVocabulary()->ConvertToIds(x.mtext);
			std::vector<int> ctextIds = m->GetVocabulary()->ConvertToIds(x.ctext);

			std::vector<float> scores;

			std::mutex lck;

			// Build triple
			Triple t; t.resize(3); 
			t[0] = ctextIds; t[1] = mtextIds;
			//

			decoder.SetContext(t);
			std::vector<Hypothesis*> generated = decoder.StartDecoding(maxLength);
			std::sort(generated.begin(), generated.end(), ReversedHypothesisComparator());

			// Build output record
			HeadRecord y;
			y.ctext = x.ctext;
			y.mtext = x.mtext;
			y.rtext = x.rtext;

			y.featureNames.push_back(m->GetModelName());
			y.featureNames.push_back("c-1-grammatch");
			y.featureNames.push_back("c-2-grammatch");
			y.featureNames.push_back("c-3-grammatch");
			y.featureNames.push_back("c-4-grammatch");
			y.featureNames.push_back("m-1-grammatch");
			y.featureNames.push_back("m-2-grammatch");
			y.featureNames.push_back("m-3-grammatch");
			y.featureNames.push_back("m-4-grammatch");
			y.featureNames.push_back("WordPenalty");

			for (int i = 0; i < MIN(generated.size(), nbestSize); i++)
			{				
				TailRecord z;
				for (int j = 0; j < generated[i]->featureValues.size(); j++)
					z.featureValues.push_back(generated[i]->featureValues[j]);

				z.rtext = m->GetVocabulary()->ConvertToString(generated[i]->path);
				y.rlist.push_back(z);
			}

			for (int i = 0; i < generated.size(); i++)
			{
				if (generated[i]) delete generated[i];
				generated[i] = nullptr;
			}

			w->WriteRecord(y);
			std::cerr << "All good. " << std::endl;
			std::cerr << "================================" << std::endl;
		}

		delete w;
		delete m;
		if (r) delete r;
		std::cerr << std::endl;
	
	} catch (std::exception e) {
		std::cerr << e.what() << std::endl;

		if (m) delete m;
		if (r) delete r;
	}

	return 0;
}