#pragma once

#include "AbstractModel.hpp"
#include "AbstractReader.hpp"
#include "ThreadedFixedPQ.hpp"
#include <ppl.h>
#include <vector>

// High negative log-likelihood means lowest score
#define MIN_SCORE 1e7

struct CMMatcher {
	std::map<std::string, int> ctxGrams;
	std::map<std::string, int> messGrams;
	
	std::map<std::string, int> BuildNGrams(const Sequence& seq)
	{
		std::map<std::string, int> nGrams;
		for (int i = 0; i < seq.size(); i++)
		{
			// skip <s> && </s>
			if (i == 0 || i == seq.size() - 1) continue;

			std::string oneGram = std::to_string(seq[i]);
			nGrams.insert(std::make_pair(oneGram, 1));

			std::string twoGram = "";
			std::string threeGram = "";
			std::string fourGram = "";

			if (i+1 < seq.size()) {
				twoGram = oneGram + "_" + std::to_string(seq[i+1]);
				nGrams.insert(std::make_pair(twoGram, 2));
			}
			if (i+2 < seq.size()) {
				 threeGram = twoGram + "_" + std::to_string(seq[i+2]);
				nGrams.insert(std::make_pair(threeGram, 3));
			}
			if (i+3 < seq.size()) {
				fourGram = threeGram + "_" + std::to_string(seq[i+3]);
				nGrams.insert(std::make_pair(fourGram, 4));
			}
		}

		return nGrams;
	}

	CMMatcher(const Sequence& ctx, const Sequence& mess) {
		std::cerr << "Initializing CMMatcher ... ";

		ctxGrams = BuildNGrams(ctx);
		messGrams = BuildNGrams(mess);
		
		std::cerr << "Done." << std::endl;
	}

	void UpdateMarginalFeatures(Sequence& seq, std::vector<double>& matches) {
		size_t end = seq.size();
		if (matches.size() != 8) 
			throw std::exception("Allocate more space for cmmatches!");
		
		std::map<std::string, int> rGrams = BuildNGrams(seq);

		for (auto it = rGrams.begin(); it != rGrams.end(); it++) {
			if (ctxGrams.find(it->first) != ctxGrams.end()) {
				matches[it->second - 1]++; 
			}
			
			if (messGrams.find(it->first) != messGrams.end()) {
				matches[it->second + 3]++; 
			}
		}
	}
};

struct LambdaScorer {
	std::vector<float> lambdas;
	std::set<std::string> includeFeatures;

	LambdaScorer() {}
	LambdaScorer(const std::string& lambdaPath) {
		// DCGM score
		std::ifstream lambdaF(lambdaPath);
		if (!lambdaF.good())
			throw std::exception("Lambda file not found!");
		
		std::string line;
		while (std::getline(lambdaF, line))
		{
			if (line.find("//") != std::string::npos)
				continue;

			if (line.find("s v ") != std::string::npos) {
				int pos = line.find_last_of(" ");	
				std::string featName = line.substr(0, pos);
				featName = featName.substr(featName.find_last_of(" ")+1);
				float featValue = std::atof(line.substr(pos+1).c_str());

				std::cerr << "- Generating with: " << featName << " = " << featValue << std::endl;
				lambdas.push_back(featValue);
			}
		}
		
		if (lambdas.size() != 10)
			throw std::exception("> 10 feature values are not supported for the moment!");
		lambdaF.close();
	}

	double Score(std::vector<double> features) {
		if (features.size() != lambdas.size())
			throw std::exception("Nb of features different than number of parameters!");
		
		// Compute the score
		double score = 0;
		
		for (int i = 0; i < features.size(); i++)
		{
			double f = features[i];
			score += f*lambdas[i];
		}
		return exp(score);
	}
};

struct Hypothesis
{
	std::vector<int> path;
	std::vector<double> featureValues;

	PastInformation* infos;
	double score;

	Hypothesis() : score(0) {
		infos = new PastInformation();
		// Cm-matches + rnn + wp
		featureValues.resize(10);
	};

	Hypothesis(const Hypothesis& hypo) {
		path = hypo.path;
		infos = new PastInformation(*hypo.infos);
		score = hypo.score;
		featureValues = hypo.featureValues;
	}

	// Update the internal state of this hypothesis
	// Could be done in a better way but - lack of time
	void UpdateState(CMMatcher& c, LambdaScorer& r, Candidate& candidate) 
	{
		// Update CmMatches
		path.push_back(candidate.id);
		
		std::vector<double> cmMatches(8);
		c.UpdateMarginalFeatures(path, cmMatches);
				
		for (int i = 1; i < 9; i++)
		{
			featureValues[i] = cmMatches[i-1];
		}
					
		// Update word penalty
		featureValues[9] = -((int) path.size());
		featureValues[0] += -candidate.score;
		score = r.Score(featureValues);
	}
	
	~Hypothesis() {
		if (infos != nullptr) 
			delete infos;
		infos = nullptr;
	}
};

struct SuffixComparator : public std::equal_to<Hypothesis*>
{
	int m_nGram;
	
	SuffixComparator() { m_nGram = 3; }
	bool operator()(const Hypothesis* l, const Hypothesis* r)
	{
		size_t lSize = l->path.size();
		size_t rSize = r->path.size();

		if (lSize != rSize) return false;
		for (int i = 0; i < MIN(m_nGram, lSize); ++i) {
			if (l->path[lSize-i-1] != r->path[rSize-i-1])
				return false;
		}

		return true;
	}
};

struct ReversedHypothesisComparator : public std::less<Hypothesis*>
{
	bool operator()(const Hypothesis* l, const Hypothesis* r)
	{
		return l->score > r->score;
	}
};

struct HypothesisComparator : public std::less<Hypothesis*>
{
	HypothesisComparator() {}
	HypothesisComparator(LambdaScorer l) {}

	bool operator()(const Hypothesis* l, const Hypothesis* r)
	{
		return l->score < r->score;
	}
};

/*class DecoderPQ : public ThreadedFixedPQ<Hypothesis*, std::vector<Hypothesis*>, HypothesisComparator> {
public:
	std::vector<Hypothesis*> PopAll() {
		m_lock.lock();
		size_t l = m_queue.size();
		std::vector<Hypothesis*> v(l);
		while (m_queue.size() != 0)
		{
			v[--l] = m_queue.top();
			m_queue.pop();
		}

		m_keys.clear();
		m_lock.unlock();
		return v;
	}

	bool Push(Hypothesis*& value) {
		m_lock.lock();
		if (m_keys.find(value) != m_keys.end()) {
			throw std::exception("This element should not be in the heap!");
		}

		bool insert = true;
		for (auto it = m_keys.begin(); it != m_keys.end(); it++)
		{
			insert = !m_selComparator(value, *it);
			
			/// We cannot insert the node (same suffix) - so just keep the oldest one.
			if (!insert) {
				delete value;
				value = nullptr;
				m_lock.unlock();
				return false;
			}
		}
		
		m_keys.insert(value);		
		m_queue.push(value);
		
		if (m_queue.size() > m_maxSize)
		{
			Hypothesis*& toDelete = m_queue.top();
			
			// by construction it must be in the set
			m_keys.erase(m_keys.find(toDelete));
			if (toDelete == value)
				insert = false;

			delete toDelete; toDelete = nullptr;
			m_queue.pop();
		}
		
		m_lock.unlock();
		return insert;
	}

	DecoderPQ(size_t maxSize, int nGramSuffix) : ThreadedFixedPQ(maxSize) {
		m_selComparator.m_nGram = nGramSuffix;	
	}
	DecoderPQ() : ThreadedFixedPQ(100) {}

private:
	SuffixComparator m_selComparator;
	HypothesisComparator m_dataCompare;

	std::set<Hypothesis*> m_keys;
};*/

class DecoderPQ : public ThreadedFixedPQ<Hypothesis*, std::vector<Hypothesis*>, HypothesisComparator> {
public:
	std::vector<Hypothesis*> PopAll() {
		size_t l = m_maxSize;
		std::vector<Hypothesis*> v;
		v.reserve(l);

		while (m_queue.size() != 0)
		{
			Hypothesis* k = m_queue.top();
			auto flaggedIt = m_flagged.find(k);
			if (flaggedIt != m_flagged.end())
			{
				delete k; k = nullptr;
				m_flagged.erase(flaggedIt);
			}
			else
			{
				v.push_back(m_queue.top());
			}
			
			m_queue.pop();
		}

		if (v.size() > m_maxSize)
		{
			std::cerr << v.size() << std::endl;
			throw std::exception("Not expected size in pop");
		}

		if (m_flagged.size() != 0)
			throw std::exception("Expected flagged size of 0");

		m_flagged.clear();
		m_keys.clear();
		return v;
	}

	bool Push(Hypothesis*& value) {
		if (m_keys.find(value) != m_keys.end()) {
			throw std::exception("This element should not be in the heap!");
		}

		bool insert = true;
		for (auto it = m_keys.begin(); it != m_keys.end(); it++)
		{
			insert = !m_selComparator(value, *it);
			
			/// We cannot insert the node (same suffix) - so just keep the oldest one.
			if (!insert) {

#ifdef DEBUG
				std::cerr << "Detected same suffix:" << std::endl;
				std::cerr << value << " with score " << value->score << std::endl;
				for (int i = 0; i < value->path.size(); i++)
				{
					std::cerr << value->path[i] << " ";
				}
				std::cerr << std::endl;
				
				std::cerr << *it << " with score " << (*it)->score << std::endl;
				for (int i = 0; i < value->path.size(); i++)
				{
					std::cerr << (*it)->path[i] << " ";
				}
				std::cerr << std::endl;
#endif DEBUG

				if (m_dataCompare(value, *it)) {
					// if value score is greater than *it score
					// assign a lower bound score
					if (m_flagged.find(*it) != m_flagged.end())
					{
						throw std::exception("Reflagging flagged element?");
					}
#ifdef DEBUG
					std::cerr << "Flagging " << *it << std::endl;
					(*it)->score = MIN_SCORE;
					std::cerr << "Pushing " << value << std::endl;
#endif DEBUG

					m_flagged.insert(*it);
					m_keys.erase(it);
					
					// push the bigger element
					m_queue.push(value);
					m_keys.insert(value);
					return true;
				}
				else 
				{
					// Just delete the new element
					delete value;
					value = nullptr;
					return false;
				}
			}
		}
		
		m_keys.insert(value);		
		m_queue.push(value);
		
		if (m_queue.size() > m_maxSize + m_flagged.size())
		{
			while (1) {
				Hypothesis*& toDelete = m_queue.top();

				auto flaggedIt = m_flagged.find(toDelete);
				auto keysIt = m_keys.find(toDelete);
				if (toDelete == value)
				{
					// We are removing the item we just inserted
					insert = false;
				}
				
				delete toDelete; toDelete = nullptr;
				m_queue.pop();
			
				if (flaggedIt != m_flagged.end())
				{
					// by construction it must be in the set
					if (keysIt != m_keys.end())
					{
						throw std::exception("It should NOT be in the set!");
					}
					// Remove it from the flagged set
					m_flagged.erase(flaggedIt);
				}
				else
				{
					// by construction it must be in the set
					if (keysIt == m_keys.end()) 
					{
						throw std::exception("It should be in the set!");
					}
					
					m_keys.erase(keysIt);
					break;
				}
			}
		}
		
		return insert;
	}

	DecoderPQ(size_t maxSize, int nGramSuffix) : ThreadedFixedPQ(maxSize) {
		m_selComparator.m_nGram = nGramSuffix;	
	}
	DecoderPQ() : ThreadedFixedPQ(100) {}

private:
	SuffixComparator m_selComparator;
	HypothesisComparator m_dataCompare;

	// This is because i cannot remove an element from the heap with c++ std
	std::set<Hypothesis*> m_flagged; 
	std::set<Hypothesis*> m_keys;
};

class StackBasedDecoder 
{
private:
	LambdaScorer m_lambdaScorer;

	int m_maxLen;
	int m_maxSize;
	int m_suffixLength;
	int m_enforceDiversity;

public:
	StackBasedDecoder(std::string& lambdaFile, AbstractModel* m, size_t maxSize, 
		size_t maxLen, double thresholdPruning, int suffixPruning,
		int enforceDiversity) : m_model(m)
	{
		m_lambdaScorer = LambdaScorer(lambdaFile);
		m_suffixLength = suffixPruning;
		m_enforceDiversity = enforceDiversity;
		m_maxSize = maxSize;
		m_maxLen = maxLen;
	}

	void SetContext(Triple& ctx) {
		m_ctx = ctx;
	}

	std::vector<Hypothesis*> StartDecoding(size_t length) {
		return Decode(length, m_maxSize);	
	}

private:
	std::vector<Hypothesis*> Decode(size_t maxLength, size_t pqSize) 
	{
		/// Past stack
		std::vector<Hypothesis*> generated;
		std::vector<Hypothesis*> m1;

		/// Current stack
		DecoderPQ tQ(pqSize, m_suffixLength);

		size_t currLength = 0;

		Hypothesis* start = new Hypothesis();
		start->path.push_back(0);
		start->score = 0;
		m1.push_back(start);
		
		std::mutex lck;

		// Build our cmMatcher structure throughout this decoding process
		CMMatcher cmMatcher(m_ctx[0], m_ctx[1]);

		while (m1.size() != 0)
		{
			currLength++;
#ifdef DEBUG
			std::cerr << "-- Decoding length = " << currLength << std::endl;
			std::cerr << "-- Generated size = " << generated.size() << std::endl;
#endif
			// Last token = we artificially add <s>
			if (currLength == maxLength)
			{
				concurrency::parallel_for_each(m1.begin(), m1.end(), [&](Hypothesis* r)
				{
					r->path.push_back(1);
					
					Triple ctx; 
					ctx.push_back(m_ctx[0]); 
					ctx.push_back(m_ctx[1]);
					ctx.push_back(r->path);
					
					// Final score (with end of sentence included)
					r->featureValues[9] = -((int) r->path.size());
					r->featureValues[0] = m_model->Test(ctx, true);
					r->score = m_lambdaScorer.Score(r->featureValues);

					lck.lock();
					generated.push_back(r);
					lck.unlock();
				});

				// End everything
				break;
			}
			else
			{
				/// Concurrency is here possible
				concurrency::parallel_for_each(m1.begin(), m1.end(), [&](Hypothesis* r)
				{
#ifdef DEBUG
	std::cerr << "-- Generating candidates for = " << m_model->GetVocabulary()->ConvertToString(r->path);
#endif
					if (r == nullptr)
						throw std::exception("Major Error: Null not expected!");
					if (r->path.size() != currLength) {
						std::cerr << "Hypothesis : " << r << std::endl;
						std::cerr << r->path.size() << " instead of " << currLength << std::endl;
						throw std::exception("Major Error: Size mismatch in decoder!");
					}
					
					std::vector<Candidate> candidates = m_model->GenerateNextWord(m_ctx, r->infos, r->path[currLength-1], pqSize);
					
					int cpt = 0;
					for (int i = 0; i < candidates.size(); i++)
					{
						// No more than 10 candidates from each node (enforce diversity)
						if (cpt >= m_enforceDiversity && m_enforceDiversity != 0 && currLength != 1)
							break;

						Candidate& candidate = candidates[i];
												
						// Copy hypothesis + past information
						// Concatenate the candidate and update the score
						Hypothesis* next = new Hypothesis(*r);
						next->UpdateState(cmMatcher, m_lambdaScorer, candidate);

						if (candidate.id == m_model->GetVocabulary()->eId)
						{
							// Stop generation for this branch
							// when we hit the end of token
							Triple ctx; 
							ctx.push_back(m_ctx[0]);
							ctx.push_back(m_ctx[1]);
							ctx.push_back(next->path);
							
							// NCE
							next->featureValues[9] = -((int) next->path.size());
							next->featureValues[0] = m_model->Test(ctx, true);
							next->score = m_lambdaScorer.Score(next->featureValues);
							
							lck.lock();
							generated.push_back(next);
							lck.unlock();
						}
						else 
						{
							// Continue generation
							lck.lock();
							bool in = tQ.Push(next);
							lck.unlock();
							if (in) cpt++;
						}
					}

					delete r;
					r = nullptr;
				});

				m1 = tQ.PopAll();
				std::cerr << "(" << currLength << "," << m1.size() << ") ";
			}
		}
		
		std::cerr << "  Done. Generated size: " << generated.size() << std::endl;
		return generated;
	}

	Triple m_ctx;
	AbstractModel* m_model;
};