#pragma once

#include "GradientUpdater.hpp"
#include "NnetLib.hpp"
#include "ThreadedFixedPQ.hpp"

struct PastInformation {
	std::map<std::string, Matrix*> infos;
	
	PastInformation() {}
	PastInformation(const PastInformation& info) {
		for (auto it = info.infos.begin(); it != info.infos.end(); it++) {
			Matrix* copy = new Matrix(*it->second);
			infos.insert(std::make_pair(it->first, copy));
		}
	}

	void ReplaceFrom(PastInformation& info) {
		for (auto it = infos.begin(); it != infos.end(); it++) {
			delete it->second;
		}
		infos.clear();

		for (auto it = info.infos.begin(); it != info.infos.end(); it++) {
			Matrix* copy = new Matrix(*it->second);
			infos.insert(std::make_pair(it->first, copy));
		}
	}
	
	~PastInformation() {
		for (auto it = infos.begin(); it != infos.end(); it++) 
			if (it->second != nullptr) { delete it->second; }
		infos.clear();
	}
};

struct Candidate {
	int id;
	double score;
	Candidate() : score(0), id(-1) {};
	Candidate(int i, double s) : score(s), id(i) {};
};
struct CandidateComparator : public std::less<Candidate>
{
	bool operator()(const Candidate& l, const Candidate& r) {
		return l.score < r.score;
	}
};

class AbstractModel {
protected:
	std::string m_modelName;
	std::unordered_map<std::string, Matrix*> m_weights;
	RandomGenerator* m_generator;
	Vocabulary* m_vocabulary;
	IGradientUpdater* m_updater;
	GradientVisitor v;
	bool m_proprietaryMemory;

public:
	AbstractModel() {
		m_proprietaryMemory = false;
		m_updater = nullptr;
		m_vocabulary = nullptr;
		m_generator = nullptr;
		m_modelName = "";
	};

	AbstractModel(bool proprietaryMemory) {
		m_proprietaryMemory = proprietaryMemory;
		m_updater = nullptr;
		m_vocabulary = nullptr;
		m_generator = nullptr;
		m_modelName = "";
	};

	~AbstractModel() {
		if (m_proprietaryMemory)
		{
			if (m_vocabulary) delete m_vocabulary;
			if (m_generator) delete m_generator;
			if (m_updater) delete m_updater;
		}
	}

	std::string GetModelName() { return m_modelName; }
	void SetVocabulary(Vocabulary* o) { m_vocabulary = o; }
	Vocabulary* GetVocabulary() { return m_vocabulary; }
	void SetUpdater(IGradientUpdater* upd) { m_updater = upd; }
	void SetGenerator(RandomGenerator* g) { m_generator = g; }

	virtual double Test(const Triple& data, bool stoch=false) { NOT_IMPLEMENTED; };
	virtual double Test(const Sequence& data, bool stoch=false) { NOT_IMPLEMENTED; };
	
	virtual std::vector<Candidate> 
		GenerateNextWord(const Triple& ctx, 
		PastInformation* info, 
		int currentInput, 
		int nK=100, 
		bool sample=false) { NOT_IMPLEMENTED; };

	virtual double Train(const Triple& data) { NOT_IMPLEMENTED; };
	virtual double Train(const Sequence& data) { NOT_IMPLEMENTED; };
};