
#pragma once

#include "ThreadedQueue.hpp"
#include "AbstractReader.hpp"
#include "SequenceReader.hpp"
#include "TripleReader.hpp"
#include <chrono>
#include <iostream>
#include <string>
#include <iomanip>
#include <thread>
#include <mutex>

typedef std::vector<std::thread*> WorkerPool;

template<class ReaderClass>
class AbstractTrainer {
protected:
	std::mutex m_updMutex;
	int m_delay;
	
	RandomGenerator* m_generator;
	Vocabulary* m_vocabulary;

	std::string m_trainPath;
	std::string m_validPath;
	std::string m_testPath;
	int m_nThreads;

	std::vector<double> m_histTrainErr;
	std::vector<double> m_histValidErr;

	int m_epochs;
	int m_epochSize;

	clock_t m_startTime;

	double m_alpha;
	double m_beta;
	double m_eta;
	double m_startAlpha;
	double m_trainError;
	double m_testError;
	double m_validError;

	// Threads will keep track in these variables
	double m_sharedError;
	long long m_sharedPredictions;
	long long m_sharedBatchesDone;
	// 

	ThreadedQueue<typename ReaderTraits<ReaderClass>::DataType> m_trainQueue;
	ThreadedQueue<typename ReaderTraits<ReaderClass>::DataType> m_testQueue;
	ThreadedQueue<typename ReaderTraits<ReaderClass>::DataType> m_validQueue;

	WorkerPool m_trainPool;
	WorkerPool m_validPool;
	WorkerPool m_testPool;
	std::thread m_printer;

	ReaderClass m_trainReader;
	ReaderClass m_validReader;
	ReaderClass m_testReader;

	virtual void PrintStatus(bool training=true) = 0;
	
	static void __dispatchPrintJob(void* obj, int delay, bool& stopPrinting) {
		AbstractTrainer<ReaderClass>* caller = static_cast<AbstractTrainer<ReaderClass>*>(obj);
		
		while (!stopPrinting) {
			std::this_thread::sleep_for(std::chrono::seconds(delay));			
			caller->PrintStatus();
		}
	}

public:
	// This method should be re-implemented by the chilren workers!
	virtual void DoJob(AbstractTrainer<ReaderClass>& parentCaller, 
		ThreadedQueue<typename ReaderTraits<ReaderClass>::DataType>& workingQueue,
		bool isTraining=false) = 0;

	void Train() {
		int epoch = 0;

		m_trainReader.SetQueue(&m_trainQueue);
		m_trainReader.OpenStream(m_trainPath);
				
		bool stillReading = true;
		double minImprovement = 1.003;

		m_histValidErr.push_back(1e6);
		m_histTrainErr.push_back(1e6);

		bool notImproved = false; 

		// Set up workers
		for (int i = 0; i < m_nThreads; i++) {
			m_trainPool.push_back(new std::thread(&AbstractTrainer::__dispatchTrainJob, static_cast<void*>(this)));
		}

		m_startAlpha = m_alpha;

		double m_saC = m_startAlpha;
		bool schedule = false;
		m_eta = 0;

		while (epoch < m_epochs) {
			m_sharedError = 0;
			m_sharedPredictions = 0;
			m_sharedBatchesDone = 0;
			
			// if (schedule) m_alpha /= 2;

			// Start Epoch
			std::cout << std::endl << "- Starting Epoch #" << std::setw(2) << epoch << " / Alpha = " << m_alpha << std::endl;
			// Dispatch the printer
			bool stopPrinting = false;
			m_startTime = clock();

			if (m_delay != 0)
				m_printer = std::thread(&AbstractTrainer::__dispatchPrintJob, 
					static_cast<void*>(this), m_delay, std::ref(stopPrinting));

			m_trainReader.ResetCount();
			while (m_trainReader.GetCount() < m_epochSize && stillReading) {
				stillReading = m_trainReader.ReadNextBatch();
			}
			
			std::this_thread::sleep_for(std::chrono::milliseconds(1000)); 

			// Wait for all the workers to starve
			while (!(m_trainQueue.GetWaiting() == m_nThreads)) { 
				std::this_thread::sleep_for(std::chrono::milliseconds(1000)); 
			}
			
			m_trainError = (m_sharedError/(double)m_sharedPredictions)/log(2);

			// Kill the printer
			if (m_delay != 0) 
			{
				stopPrinting = true;
				m_printer.join(); 
			}
			// 
			
			// Print the current status
			std::cout << "- End Epoch #" << std::setw(2) << epoch << ", Error = " << pow(2, m_trainError) << std::endl;
			
			// Start validation job; fill m_validError
			Validate();

			m_validError = (m_sharedError/(double)m_sharedPredictions)/log(2);
			
			// Print the current status
			std::cout << "- Validation Error = " << m_validError << " / Perpl. = " << pow(2, m_validError) << std::endl;
			
			// End Epoch
			if (!stillReading) {
				// Reset count and restart stream
				m_trainReader.CloseStream();
				m_trainReader.OpenStream(m_trainPath);
				stillReading = true;
			}

			if (m_validError*minImprovement > m_histValidErr.back()) {
				m_histValidErr.push_back(m_validError);
				m_histTrainErr.push_back(m_trainError);
				break;
			}

			m_histValidErr.push_back(m_validError);
			m_histTrainErr.push_back(m_trainError);

			epoch++;
		}
		
		m_trainReader.SendKillBatch(m_nThreads);
		m_trainReader.CloseStream();

		// Join workers
		for (int i = 0; i < m_nThreads; i++) {
			m_trainPool[i]->join();
			delete m_trainPool[i];
		}
	}

	void Test() {
		m_sharedPredictions = 0;
		m_sharedError = 0;

		m_testReader.SetQueue(&m_testQueue);
		m_testReader.OpenStream(m_testPath);

		// Set up workers
		for (int i = 0; i < m_nThreads; i++) {
			m_testPool.push_back(new std::thread(&AbstractTrainer::__dispatchTestJob, static_cast<void*>(this)));
		}

		while (m_testReader.ReadNextBatch()) {}
		
		std::this_thread::sleep_for(std::chrono::milliseconds(1000)); 
		// Wait for all the workers to starve
		while (!(m_trainQueue.GetWaiting() == m_nThreads)) { 
			std::this_thread::sleep_for(std::chrono::milliseconds(1000)); 
		}
		m_testReader.SendKillBatch(m_nThreads);

		m_testError = (m_sharedError/(double)m_sharedPredictions)/log(2);
		
		std::cout << "- Test Error = " << m_testError << " / Perpl. = " << pow(2, m_testError) << std::endl;
		
		// Join workers
		for (int i = 0; i < m_nThreads; i++) {
			m_testPool[i]->join();
			delete m_testPool[i];
		}

		m_testPool.clear();

		std::cerr << "Train finished." << std::endl;
	}

	void SetAlpha(double alpha) { m_alpha = alpha; }
	void SetBeta(double beta) { m_beta = beta; }
	void SetEpochSize(int n) { m_epochSize = n; }	
	void SetEpochs(int n) { m_epochs = n; }
	void SetNThreads(int n) { m_nThreads = n; }
	void SetTrainDataset(const std::string& path) { m_trainPath = path; }
	void SetTestDataset(const std::string& path) { m_testPath = path; }
	void SetValidDataset(const std::string& path) { m_validPath = path; }
	void UpdateError(double e) { m_updMutex.lock(); m_sharedError += e; m_updMutex.unlock(); }
	void UpdatePredictions(int p) { m_updMutex.lock(); m_sharedPredictions += p; m_updMutex.unlock(); }
	void UpdateBatchesDone(int p) {m_updMutex.lock(); m_sharedBatchesDone += p; m_updMutex.unlock(); }
	void SetDelay(int delay) { m_delay = delay; }
		
	double GetStartAlpha() { return m_startAlpha; }
	double GetAlpha() { return m_alpha; }
	double GetBeta() { return m_beta; }
	bool IsTrain() { return m_trainPath != ""; }
	bool IsTest() { return m_testPath != ""; }
	bool IsValid() { return m_validPath != ""; }

	AbstractTrainer(Vocabulary* v, RandomGenerator* g) {
		m_alpha = 0.1;
		m_beta = 1e-7;

		m_trainPath = "";
		m_validPath = "";
		m_testPath = "";
		m_delay = 5;

		m_nThreads = 1;
		m_epochs = 20;
		m_epochSize = 136000;
		m_vocabulary = v;
		m_generator = g;
	}

	~AbstractTrainer() {
		if (m_vocabulary) delete m_vocabulary;
		if (m_generator) delete m_generator;
	}
	
	void WriteErrors(const std::string& output) 
	{
		std::ofstream error(output + "/train.err");
		std::for_each(m_histTrainErr.begin(), m_histTrainErr.end(), [&](double e) {
			error << e << std::endl;
		});
		error.close();
	
		error = std::ofstream(output + "/valid.err");
		std::for_each(m_histValidErr.begin(), m_histValidErr.end(), [&](double e) {
			error << e << std::endl;
		});
		error.close();

		error = std::ofstream(output + "/test.err");
		error << m_testError << std::endl;
		error.close();
	}

	void WriteVocabulary(const std::string& output) 
	{
		m_vocabulary->WriteToFile(output + "/vocab");
	}

	virtual void WriteModel(const std::string& outputPath) = 0;

private:
	static void __dispatchTrainJob(void* obj) {
		AbstractTrainer<ReaderClass>* caller = static_cast<AbstractTrainer<ReaderClass>*>(obj);
		caller->DoJob(*caller, caller->m_trainQueue, true);
	}
	
	static void __dispatchValidJob(void* obj) {
		AbstractTrainer<ReaderClass>* caller = static_cast<AbstractTrainer<ReaderClass>*>(obj);
		caller->DoJob(*caller, caller->m_validQueue, false);
	}
	
	static void __dispatchTestJob(void* obj) {
		AbstractTrainer<ReaderClass>* caller = static_cast<AbstractTrainer<ReaderClass>*>(obj);
		caller->DoJob(*caller, caller->m_testQueue, false); 
	}

	void Validate() {
		m_sharedError = 0;
		m_sharedPredictions = 0;

		m_validReader.SetQueue(&m_validQueue);
		m_validReader.OpenStream(m_validPath);

		// Set up workers
		for (int i = 0; i < m_nThreads; i++) {
			m_validPool.push_back(new std::thread(&AbstractTrainer::__dispatchValidJob, static_cast<void*>(this)));
		}

		while (m_validReader.ReadNextBatch()) {}
		m_validReader.SendKillBatch(m_nThreads);

		// Join workers
		for (int i = 0; i < m_nThreads; i++) {
			m_validPool[i]->join();
			delete m_validPool[i];
		}

		m_validPool.clear();
		m_validReader.CloseStream();
	}
};