#pragma once 

#include <vector>
#include <string>
#include <unordered_map>
#include <fstream>
#include <map>
#include <iostream>

class LengthPrior {

public:
	std::map<int, float> poissonMeans;

	LengthPrior() 
	{
	}
	
	void Load(std::string file) {
		std::ifstream r(file);
		bool isLength=true;
		int length=0;
		std::string num;
		while (r >> num)
		{
			if (isLength)
			{
				isLength = !isLength;
				length=std::atoi(num.c_str());
			}
			else
			{
				poissonMeans.insert(std::make_pair(length, (float) std::atof(num.c_str())));
				isLength = true;
			}
		}
		r.close();
	}

	float GetPrior(int lengthT, int lengthS)
	{
		auto it = poissonMeans.find(lengthS);
		if (it == poissonMeans.end())
			it = poissonMeans.find(0);

		double ff = 0;
		for (int i = 1; i <= lengthT; i++)
			ff += log(i);
		
		return (float) (-it->second + lengthT*log(it->second) - ff);
	}
};

struct TailRecord {
public:
	std::string rid;
	std::string rtext;
	std::list<double> featureValues;

	TailRecord() { 
		rid=""; rtext="";
	}
};

struct HeadRecord {
public:
	std::string ctext;
	std::string cid;

	std::string mtext;
	std::string mid;
	
	std::vector<std::string> rtext;
	std::string rid;
	
	std::string best;

	std::list<std::string> featureNames;
	std::vector<TailRecord> rlist;

	HeadRecord() {
		best="";ctext="";cid="";mtext="";mid="";rid="";
	}
};

class FrequencyMap 
{
public:
	std::map<int, double> m_data;

public:
	void Store(int x)
	{
		auto it=m_data.find(x);
		if (it == m_data.end())
			it = m_data.insert(std::make_pair(x, 0.f)).first;
		it->second++;
	}
	
	std::vector<int> GetKeys()
	{
		std::vector<int> keys; keys.reserve(m_data.size());
		for (auto it=m_data.begin();it!=m_data.end();it++)
			keys.push_back(it->first);
		return keys;
	}
	
	std::vector<double> GetValues()
	{
		std::vector<double> val; val.reserve(m_data.size());
		for (auto it=m_data.begin();it!=m_data.end();it++)
			val.push_back(it->second);
		return val;
	}
};


class Config {
public:
	std::unordered_map<std::string, std::string> m_config;
public:
	static Config ReadConfigFile(const std::string& file) {
		Config config;
		std::ifstream fileRead(file);
	 
		std::string line;
		while (std::getline(fileRead, line)) {
			size_t pos = line.find("=");
			if (pos == std::string::npos)
				continue;

			std::string option = line.substr(0, pos);
			std::string value = line.substr(pos+1, line.size());
			config.InsertValue(option, value);
		}
		return config;	
	}

	static Config ReadCommandLine(int argc, char* argv[]) {
		Config config;
		
		std::string line;
		
		int arg = 1;
		if ((argc - 1) % 2 != 0) 
		{
			std::cerr<<"Wrong format for command line"<<std::endl;
			exit(-1);
		}

		while (arg < argc) {
			std::string current(argv[arg++]);
			size_t pos = current.find("--");
			if (pos == std::string::npos)
				continue;

			std::string option = current.substr(pos + 2, current.size());
			std::string value = std::string(argv[arg]);
			config.InsertValue(option, value);
		}

		return config;	
	}

	void InsertValue(const std::string& option, const std::string& value) {
		m_config.insert(std::make_pair(option, value));
	}

	double GetFloat(const std::string& key, double defaultValue=-1) {
		if (m_config.find(key) != m_config.end()) {
			return (double) atof(m_config[key].c_str());
		}

		if (defaultValue == -1) {
			throw std::exception(("Option [" + key + "] is compulsory!").c_str());
		}
		return defaultValue;
	}

	std::string GetString(const std::string& key, std::string defaultValue="") {
		if (m_config.find(key) != m_config.end()) {
			return m_config[key];
		}

		if (defaultValue == "") {
			throw std::exception(("Option [" + key + "] is compulsory!").c_str());
		}

		return defaultValue;
	}

	std::vector<int> GetIntVector(const std::string& key, std::string defaultValue="") {
		std::string toParse = "";
		std::vector<int> keys;

		if (m_config.find(key) != m_config.end()) {
			// Comma separated values
			toParse = m_config[key];
		} else 
			toParse = defaultValue;

		if (toParse.size() == 0)
			throw std::exception(("Option " + key + " was not found!").c_str());

		size_t lastPos = 0, pos = 0;
		while ((pos = toParse.find(',', lastPos)) != std::string::npos) {
			keys.push_back(std::atoi(toParse.substr(lastPos, pos).c_str()));	
			lastPos = pos + 1;
		}
		
		keys.push_back(std::atoi(toParse.substr(lastPos).c_str()));	
		return keys;
	}

	int GetInt(const std::string& key, int defaultValue=0, bool needed=false) {
		if (m_config.find(key) != m_config.end()) {
			return atoi(m_config[key].c_str());
		}
		return defaultValue;
	}

	bool GetBool(const std::string& key) {
		return m_config.find(key) != m_config.end();
	}
};