#include "DcgmI.hpp"

Layer* DcgmI::BuildContext(const Triple& triple, NeuralNet& rnn) {
	// Build the context vector
	SparseInputLayer* cmInput = rnn.AddLayer(new SparseInputLayer(m_nIn, "cm_i"));

	// Paste C&M together as BOW
	FrequencyMap m;
	std::for_each(triple[0].begin(), triple[0].end(), [&](int id) {
		// Skip ss and se tokens
		if (id != m_vocabulary->sId && id != m_vocabulary->eId) m.Store(id); 
	});
	std::for_each(triple[1].begin(), triple[1].end(), [&](int id) {
		// Skip ss and se tokens
		if (id != m_vocabulary->sId && id != m_vocabulary->eId) m.Store(id); 
	});

	std::vector<int>& keys = std::move(m.GetKeys());
	std::vector<double>& values = std::move(m.GetValues());
	
	double norm = 0;
	for (int j = 0; j < keys.size(); j++)
	{
		double tfIdf = values[j]*m_vocabulary->GetIdf(keys[j]);
		values[j] = tfIdf;
		norm += tfIdf*tfIdf;
	}
	norm = sqrt(norm);

	for (int j = 0; j < keys.size(); j++)
	{
		values[j] /= norm;
	}
	
	cmInput->SetData(m_nIn, keys, values);

	Layer* lastLayer = cmInput;
	for (size_t i = 0; i < m_layerWeights.size(); i++) {
		std::string id = std::to_string(i);
		HiddenLayer* h = rnn.AddLayer(new HiddenLayer(m_layerDims[i], (i==0) ? ID : SIGMOID, "d_" + id));
		rnn.Connect(lastLayer, h, new DynamicWeight(m_layerWeights[i]));

		if (m_layerBias.size() > i)
			rnn.Connect(rnn.AddLayer(new BiasLayer("b_" + id)), h, new DynamicWeight(m_layerBias[i]));

		lastLayer = h;
	}

	return lastLayer;
}

DcgmI* DcgmI::LoadFromDisk(const std::string& dirPath) {
	DcgmI* a = new DcgmI(true);

	Matrix* Win = new Matrix();
	Matrix* Whh = new Matrix();
	Matrix* Wout = new Matrix();
	Matrix* bhh = new Matrix();
	Matrix* bout = new Matrix();
	Matrix* O = new Matrix();
	Matrix* H = new Matrix();
	Matrix* dO = new Matrix();
	Vocabulary *vO = new Vocabulary(dirPath + "\\vocab");
	RandomGenerator* g = new RandomGenerator(*vO);
	
	bool bo_found = bout->LoadFromExternalBinary(dirPath + "\\b_out.bin");
	bool bhh_found = bhh->LoadFromExternalBinary(dirPath + "\\b_hh.bin");
	bhh->InPlaceTranspose();
	bout->InPlaceTranspose();

	if (bo_found && bhh_found)
		a->SetExternalBias(bhh, bout);

	Whh->LoadFromExternalBinary(dirPath + "\\w_hh.bin");
	Whh->InPlaceTranspose();

	std::cerr << Whh->GetRows() << " " << Whh->GetCols() << " " << Whh->Norm2() << std::endl;

	Win->LoadFromExternalBinary(dirPath + "\\w_in.bin");
	Win->InPlaceTranspose();

	std::cerr << Win->GetRows() << " " << Win->GetCols() << " " << Win->Norm2() << std::endl;

	Wout->LoadFromExternalBinary(dirPath + "\\w_out.bin");
	Wout->InPlaceTranspose();

	std::cerr << Wout->GetRows() << " " << Wout->GetCols() << " " << Wout->Norm2() << std::endl;

	bool hfound = H->LoadFromExternalBinary(dirPath + "\\h.bin");
	H->InPlaceTranspose();
	if (hfound) 
		std::cerr << H->GetRows() << " " << H->GetCols() << " " << H->Norm2() << std::endl;
	a->SetIncludeH(hfound);

	std::ifstream f(dirPath + "\\d_num.txt");
	int numLayers = 0;
	f >> numLayers;
	f.close();
	std::cerr << numLayers << std::endl;

	std::vector<Matrix*> layerWeights;
	std::vector<Matrix*> layerBias;
	std::vector<int> layerDims;

	for (int i = 0; i < numLayers; i++) {
		Matrix* D = new Matrix();
		D->LoadFromExternalBinary(dirPath + "\\d_" + std::to_string(i) + ".bin");
		D->InPlaceTranspose();

		std::cerr << D->GetName() << " " << D->GetRows() << " " << D->GetCols() << " " << D->Norm2() << std::endl;
		layerWeights.push_back(D);
		

		Matrix* b = new Matrix();
		b->LoadFromExternalBinary(dirPath + "\\b_" + std::to_string(i) + ".bin");
		b->InPlaceTranspose();
		
		std::cerr << b->GetName() << " " << b->GetRows() << " " << b->GetCols() << " " << b->Norm2() << std::endl;
		layerBias.push_back(b);

		layerDims.push_back(D->GetRows());
	}

	Config c = Config::ReadConfigFile(dirPath + "\\config");
	a->SetAct(c.GetString("act", "sigmoid"));
	a->SetContextAct(c.GetString("ctxAct", "sigmoid"));

	std::cerr << "RnnActivation = " << c.GetString("act", "sigmoid") << std::endl;
	std::cerr << "ContextActivation = " << c.GetString("ctxAct", "sigmoid") << std::endl;

	a->SetExternalRnnWeights(Win, Whh, Wout);
	a->SetContextRnnWeights(H, O);
	a->SetGenerator(g);
	a->SetExternalWeights(layerDims, layerWeights, layerBias);
	a->SetDeepOutputWeight(dO);
	a->SetVocabulary(vO);
	return a;
}