#include "DcgmII.hpp"

Layer* DcgmII::BuildContext(const Triple& triple, NeuralNet& rnn) {
	// Build the context vector
	SparseInputLayer* mInput = rnn.AddLayer(new SparseInputLayer(m_nIn, "m_i"));
	SparseInputLayer* cInput = rnn.AddLayer(new SparseInputLayer(m_nIn, "c_i"));
	
	// Paste C&M together as BOW
	FrequencyMap c;
	std::for_each(triple[0].begin(), triple[0].end(), [&](int id) {
		// Skip ss and se tokens
		if (id != m_vocabulary->sId && id != m_vocabulary->eId) c.Store(id); 
	});

	std::vector<int> keys = c.GetKeys();
	std::vector<double> values = c.GetValues();
	
	float norm = 0;
	for (int j = 0; j < keys.size(); j++)
	{
		double tfIdf = values[j]*m_vocabulary->GetIdf(keys[j]);
		values[j] = tfIdf;
		norm += tfIdf*tfIdf;
	}
	norm = sqrt(norm);
	for (int j = 0; j < keys.size(); j++)
	{
		values[j] /= norm;
	}
	cInput->SetData(m_nIn, keys, values);

	FrequencyMap m;
	std::for_each(triple[1].begin(), triple[1].end(), [&](int id) {
		// Skip ss and se tokens
		if (id != m_vocabulary->sId && id != m_vocabulary->eId) m.Store(id);
	});

	keys = m.GetKeys();
	values = m.GetValues();	
	norm = 0;
	for (int j = 0; j < keys.size(); j++)
	{
		double tfIdf = values[j]*m_vocabulary->GetIdf(keys[j]);
		values[j] = tfIdf;
		norm += tfIdf*tfIdf;
	}
	norm = sqrt(norm);
	
	for (int j = 0; j < keys.size(); j++)
	{
		values[j] /= norm;
	}
	mInput->SetData(m_nIn, keys, values);
	
	// Continue
	Layer* lastLayer;
	for (size_t i = 0; i < m_layerWeights.size(); i++) {
		std::string id = std::to_string(i);
		if (i == 0) 
		{
			HiddenLayer* mh = rnn.AddLayer(new HiddenLayer(m_layerDims[i], "m_h"));
			HiddenLayer* ch = rnn.AddLayer(new HiddenLayer(m_layerDims[i], "c_h"));

			rnn.Connect(mInput, mh, new DynamicWeight(m_layerWeights[i]));
			rnn.Connect(rnn.AddLayer(new BiasLayer("bm_" + id)), mh, new DynamicWeight(m_layerBias[i]));

			rnn.Connect(cInput, ch, new DynamicWeight(m_layerWeights[i]));
			rnn.Connect(rnn.AddLayer(new BiasLayer("bc_" + id)), ch, new DynamicWeight(m_layerBias[i]));
			
			ConcatLayer* concatInputs = rnn.AddLayer(new ConcatLayer(m_layerDims[i], 2, ID, "concat_cm"));
			concatInputs->RegisterParent(ch);
			concatInputs->RegisterParent(mh);
			
			// Connect now
			rnn.Connect(mh, concatInputs, new FixedWeight());
			rnn.Connect(ch, concatInputs, new FixedWeight());
			lastLayer = concatInputs;
		}
		else 
		{
			HiddenLayer* h = rnn.AddLayer(new HiddenLayer(m_layerDims[i], SIGMOID, "d_" + id));
			rnn.Connect(lastLayer, h, new DynamicWeight(m_layerWeights[i]));
			rnn.Connect(rnn.AddLayer(new BiasLayer("b_" + id)), h, new DynamicWeight(m_layerBias[i]));
			lastLayer = h;
		}
	}
	
	return lastLayer;
}

DcgmII* DcgmII::LoadFromDisk(const std::string& dirPath) {
	DcgmII* a = new DcgmII(true);

	Matrix* Win = new Matrix();
	Matrix* Whh = new Matrix();
	Matrix* Wout = new Matrix();
	Matrix* bhh = new Matrix();
	Matrix* bout = new Matrix();
	Matrix* O = new Matrix();
	Matrix* H = new Matrix();
	Matrix* dO = new Matrix();
	Vocabulary* vO = new Vocabulary(dirPath + "\\vocab");
	RandomGenerator* g = new RandomGenerator(*vO);

	Whh->LoadFromExternalBinary(dirPath + "\\w_hh.bin");
	Whh->InPlaceTranspose();
	std::cerr << "- " << Whh->GetRows() << " " << Whh->GetCols() << " " << Whh->Norm2() << std::endl;

	Win->LoadFromExternalBinary(dirPath + "\\w_in.bin");
	Win->InPlaceTranspose();
	std::cerr << "- " << Win->GetRows() << " " << Win->GetCols() << " " << Win->Norm2() << std::endl;
	
	Wout->LoadFromExternalBinary(dirPath + "\\w_out.bin");
	Wout->InPlaceTranspose();
	std::cerr << "- " << Wout->GetRows() << " " << Wout->GetCols() << " " << Wout->Norm2() << std::endl;
	
	bool hfound = H->LoadFromExternalBinary(dirPath + "\\h.bin");
	H->InPlaceTranspose();
	if (hfound) 
		std::cerr << "- " << "H = " << H->GetRows() << " " << H->GetCols() << " " << H->Norm2() << std::endl;
	a->SetIncludeH(hfound);
	
	bool bo_found = bout->LoadFromExternalBinary(dirPath + "\\b_out.bin");
	bout->InPlaceTranspose();
	
	bool bhh_found = bhh->LoadFromExternalBinary(dirPath + "\\b_hh.bin");
	bhh->InPlaceTranspose();

	if (bo_found&&bhh_found)
		a->SetExternalBias(bhh, bout);

	std::ifstream f(dirPath + "\\d_num.txt");
	int numLayers = 0;
	f >> numLayers;
	f.close();

	std::vector<Matrix*> layerWeights;
	std::vector<Matrix*> layerBias;
	std::vector<int> layerDims;

	for (int i = 0; i < numLayers; i++) {
		Matrix* D = new Matrix();
		D->LoadFromExternalBinary(dirPath + "\\d_" + std::to_string(i) + ".bin");
		D->InPlaceTranspose();
		
		std::cerr << "- " << D->GetName() << " " << D->GetRows() << " " << D->GetCols() << " " << D->Norm2() << std::endl;
		layerWeights.push_back(D);

		Matrix* b = new Matrix();
		b->LoadFromExternalBinary(dirPath + "\\b_" + std::to_string(i) + ".bin");
		b->InPlaceTranspose();

		std::cerr << "- " << b->GetName() << " " << b->GetRows() << " " << b->GetCols() << " " << b->Norm2() << std::endl;
		layerBias.push_back(b);
		
		layerDims.push_back(D->GetRows());
	}
	
	Config c = Config::ReadConfigFile(dirPath + "\\config");
	a->SetAct(c.GetString("act", "sigmoid"));
	std::cerr << "- " << "RnnActivation = " << a->GetAct() << std::endl;
	
	a->SetExternalRnnWeights(Win, Whh, Wout);
	a->SetContextRnnWeights(H, O);
	a->SetGenerator(g);
	a->SetExternalWeights(layerDims, layerWeights, layerBias);
	a->SetDeepOutputWeight(dO);
	a->SetVocabulary(vO);
	return a;
}