#include "DcgmIII.hpp"

void DcgmIII::AddStep(const Sequence& s, size_t end, NeuralNet& rnn, 
				   Layer*& contextLayer, 
				   Layer*& lastHiddenLayer,
				   bool training) {
	if (end >= s.size())
		throw std::exception("Out of bound step in RNN!");
	
	// delete to the last time step
	rnn.DeleteOutputs();
	
	Layer* hl = lastHiddenLayer;
	std::string id = std::to_string(end);

	// set current input
	SparseInputLayer* x_next = rnn.AddLayer(new SparseInputLayer(m_nIn, "x_" + id));
	std::vector<int> ids(1, s[end-1]);
	std::vector<double> values(1, 1);
	x_next->SetData(m_nIn, ids, values);

	HiddenLayer * h_next = rnn.AddLayer(new HiddenLayer(m_nHid, m_act, "h_" + id));
	
	if (hl != 0) rnn.Connect(hl, h_next, new DynamicWeight(m_Whh));
	rnn.Connect(x_next, h_next, new DynamicWeight(m_Win));
	rnn.Connect(rnn.AddLayer(new BiasLayer("bhh_" + id)), h_next, new DynamicWeight(m_Bhh)); 
	
	LossLayer* y_t;
	if (!training) {
		CrossEntropyLossLayer* yCe_t = rnn.AddLayer(new CrossEntropyLossLayer(m_nOut, "pred_" + id));
		yCe_t->SetOutput(1, s[end]);
		y_t = yCe_t;
	} else {
		std::vector<int> samples = m_generator->SampleNoise(25, s[end]);
		std::vector<double> probs; probs.reserve(25);
		std::for_each(samples.begin(), samples.end(),
			[&](int x) { probs.push_back(m_generator->GetLogProb(x)); }); 

		double target_prob = -m_generator->GetLogProb(s[end]);
		samples.push_back(s[end]);
		probs.push_back(target_prob);

		NCELossLayer* yNce_t = rnn.AddLayer(new NCELossLayer(m_nOut, "pred_" + id));
		yNce_t->SetOutput(samples, probs);
		y_t = yNce_t;
	}
		
	ConcatLayer* concatBeforeOutput = rnn.AddLayer(new ConcatLayer(m_nHid, 2, ID, "concat_cm_" + id));
	concatBeforeOutput->RegisterParent(contextLayer);
	concatBeforeOutput->RegisterParent(h_next);
	
	// Connect the inputs of the concatenation layer
	rnn.Connect(contextLayer, concatBeforeOutput, new FixedWeight());
	rnn.Connect(h_next, concatBeforeOutput, new FixedWeight());
		
	// Connect the output now
	rnn.Connect(concatBeforeOutput, y_t, new DynamicWeight(m_Wout));
	rnn.Connect(rnn.AddLayer(new BiasLayer("bout_" + id)), y_t, new DynamicWeight(m_Bout));

	lastHiddenLayer = h_next;
}

DcgmIII* DcgmIII::LoadFromDisk(const std::string& dirPath) {
	DcgmIII* a = new DcgmIII(true);

	Matrix* Win = new Matrix();
	Matrix* Whh = new Matrix();
	Matrix* Wout = new Matrix();
	Matrix* H = new Matrix();
	Matrix* bhh = new Matrix();
	Matrix* bout = new Matrix();
	Vocabulary* vO = new Vocabulary(dirPath + "\\vocab");
	RandomGenerator* g = new RandomGenerator(*vO);

	Whh->LoadFromExternalBinary(dirPath + "\\w_hh.bin");
	Whh->InPlaceTranspose();
	std::cerr << "- " << Whh->GetRows() << " " << Whh->GetCols() << " " << Whh->Norm2() << std::endl;

	Win->LoadFromExternalBinary(dirPath + "\\w_in.bin");
	Win->InPlaceTranspose();
	std::cerr << "- " << Win->GetRows() << " " << Win->GetCols() << " " << Win->Norm2() << std::endl;
	
	Wout->LoadFromExternalBinary(dirPath + "\\w_out.bin");
	Wout->InPlaceTranspose();
	std::cerr << "- " << Wout->GetRows() << " " << Wout->GetCols() << " " << Wout->Norm2() << std::endl;
	
	H->LoadFromExternalBinary(dirPath + "\\h.bin");
	H->InPlaceTranspose();
	std::cerr << "- " << H->GetRows() << " " << H->GetCols() << " " << H->Norm2() << std::endl;
	
	bool bo_found = bout->LoadFromExternalBinary(dirPath + "\\b_out.bin");
	bout->InPlaceTranspose();
	
	bool bhh_found = bhh->LoadFromExternalBinary(dirPath + "\\b_hh.bin");
	bhh->InPlaceTranspose();

	if (bo_found&&bhh_found)
		a->SetExternalBias(bhh, bout);

	std::ifstream f(dirPath + "\\d_num.txt");
	int numLayers = 0;
	f >> numLayers;
	f.close();
	std::cerr << "- NumLayers = " << numLayers << std::endl;

	std::vector<Matrix*> layerWeights;
	std::vector<Matrix*> layerBias;
	std::vector<int> layerDims;

	for (int i = 0; i < numLayers; i++) {
		Matrix* D = new Matrix();
		D->LoadFromExternalBinary(dirPath + "\\d_" + std::to_string(i) + ".bin");
		D->InPlaceTranspose();
		
		std::cerr << " : " << D->GetName() << " " << D->GetRows() << " " << D->GetCols() << " " << D->Norm2() << std::endl;
		layerWeights.push_back(D);

		Matrix* b = new Matrix();
		b->LoadFromExternalBinary(dirPath + "\\b_" + std::to_string(i) + ".bin");
		b->InPlaceTranspose();

		std::cerr << " : " << b->GetName() << " " << b->GetRows() << " " << b->GetCols() << " " << b->Norm2() << std::endl;
		layerBias.push_back(b);
		
		layerDims.push_back(D->GetRows());
	}
	
	Config c = Config::ReadConfigFile(dirPath + "\\config");
	a->SetAct(c.GetString("act", "sigmoid"));
	std::cerr << " : " << " rnn_activation = " << a->GetAct() << std::endl;
	
	a->SetExternalRnnWeights(Win, Whh, Wout);
	a->SetContextRnnWeights(H, nullptr);
	a->SetGenerator(g);
	a->SetExternalWeights(layerDims, layerWeights, layerBias);
	a->SetVocabulary(vO);
	return a;
}