#include "DcgmTrainer.hpp"

void DcgmTrainer::PrintStatus(bool training) {
	clock_t now = clock();
	double entropy = (m_sharedError/m_sharedPredictions)/log(2);
	double perpl = pow(2, entropy);
	long long difftime = (now - m_startTime)/1000;

	std::cout << difftime/3600 <<":"<<(difftime % 3600)/60<<":"<<((difftime % 3600) % 60); 
	std::cout << " - Entr. / Perp. : " << std::setw(8) << entropy << " / " << perpl
		<< " P/s: " << std::setw(5) << (long long) (m_sharedPredictions/difftime);
		std::flush(std::cout);

	/** Print norms **/
	std::cout << std::setw(5);

	std::cout << " W_hh : " << std::setw(5) << Whh.Norm2();
	std::cout << " W_out : " << std::setw(5) << Wout.Norm2();
	std::cout << " W_in : " << std::setw(5) << Win.Norm2();
	std::cout << " b_hh : " << std::setw(5) << bhh.Norm2();
	std::cout << " b_out : " << std::setw(5) << bout.Norm2();
	if (m_includeH || m_deepO) std::cout << " H : " << std::setw(5) << H.Norm2();
	if (m_includeO) std::cout << " O : " << std::setw(5) << O.Norm2();
	if (m_deepO) std::cout << " dO : " << std::setw(5) << deepO.Norm2();
	for (size_t i = 0; i < m_layerWeights.size(); i++) {
		std::cout << " D_" << i << " : " << std::setw(5) << m_layerWeights[i]->Norm2();
	}
	for (size_t i = 0; i < m_layerBias.size(); i++) {
		std::cout << " b_" << i << " : " << std::setw(5) << m_layerBias[i]->Norm2();
	}
	std::cout << std::endl;
}

void DcgmTrainer::DoJob(AbstractTrainer<TripleReader>& parentCaller, 
		ThreadedQueue<Batch<Triple>>& workingQueue,
		bool isTraining) {
	long long loaded_seq = 0, seq_position = 0;
	Batch<Triple> wi;

	IGradientUpdater* upd = new AdagradUpdater(parentCaller.GetAlpha());
	upd->mb_ = 10;

	DcgmI dcgmi;	
	DcgmII dcgmii;

	if (m_dcgmType == DCGMI)
	{
		dcgmi.SetTau(m_tau);
		dcgmi.SetExternalRnnWeights(&Win, &Whh, &Wout);
		dcgmi.SetExternalBias(&bhh, &bout);
		dcgmi.SetContextRnnWeights(&H, &O);
		dcgmi.SetExternalWeights(m_layerDims, m_layerWeights, m_layerBias);
		dcgmi.SetDeepOutputWeight(&deepO);
		dcgmi.SetUpdater(upd);
		dcgmi.SetContextAct(m_ctxAct);
		dcgmi.SetGenerator(m_generator);
		dcgmi.SetVocabulary(m_vocabulary);
		dcgmi.SetTiled(&Whh);
		dcgmi.SetTiled(&H);
		dcgmi.SetIncludeH(m_includeH);
		dcgmi.SetIncludeO(m_includeO);
		dcgmi.SetDeepO(m_deepO);
		dcgmi.SetAct(m_act);
	}
	else
	{
		dcgmii.SetTau(m_tau);
		dcgmii.SetExternalRnnWeights(&Win, &Whh, &Wout);
		dcgmii.SetContextRnnWeights(&H, &O);
		dcgmii.SetExternalBias(&bhh, &bout);
		dcgmii.SetExternalWeights(m_layerDims, m_layerWeights, m_layerBias);
		dcgmii.SetDeepOutputWeight(&deepO);
		dcgmii.SetUpdater(upd);
		dcgmii.SetContextAct(m_ctxAct);
		dcgmii.SetGenerator(m_generator);
		dcgmii.SetVocabulary(m_vocabulary);
		dcgmii.SetTiled(&Whh);
		dcgmii.SetTiled(&H);
		dcgmii.SetIncludeH(m_includeH);
		dcgmii.SetIncludeO(m_includeO);
		dcgmii.SetDeepO(m_deepO);
		dcgmii.SetAct(m_act);
	}

	// Some regularization for the parameters!
	if (m_beta != 0) {
		upd->add_pre_filter(&Whh, new L2RegularizeFilter(m_beta));
		upd->add_pre_filter(&Wout, new L2RegularizeFilter(m_beta));
		upd->add_pre_filter(&Win, new L2RegularizeFilter(m_beta));

		for (int i = 0; i < m_layerWeights.size(); i++) {
			upd->add_pre_filter(m_layerWeights[i], new L2RegularizeFilter(m_beta));
		}
		
		for (int i = 0; i < m_layerBias.size(); i++) {
			upd->add_pre_filter(m_layerBias[i], new L2RegularizeFilter(m_beta));
		}
	
		if (m_includeO) upd->add_pre_filter(&O, new L2RegularizeFilter(m_beta));
		if (m_includeH || m_deepO) upd->add_pre_filter(&H, new L2RegularizeFilter(m_beta));
		if (m_deepO) upd->add_pre_filter(&deepO, new L2RegularizeFilter(m_beta));
	}

	while (1) {
		// GetElement job
		if (loaded_seq == 0) {
			// GetElement the work item
			wi = static_cast<Batch<Triple>&&>(workingQueue.PopFront());
			loaded_seq = wi.m_size;
			seq_position = 0;

			// poison pill
			if (loaded_seq == 0) {
				delete upd;
				// Exit here
				break;
			}
		}

		Triple& tri = wi.data[seq_position];
		int len = (int) tri[2].size();

		try {
			if (isTraining) 
				if (m_dcgmType == DCGMI)
					parentCaller.UpdateError(dcgmi.Train(tri));
				else 
					parentCaller.UpdateError(dcgmii.Train(tri));
			else 
				if (m_dcgmType == DCGMI)
					parentCaller.UpdateError(dcgmi.Test(tri));
				else	
					parentCaller.UpdateError(dcgmii.Test(tri));
		} catch (std::exception e) {
			std::cerr << e.what() << std::endl;
			exit(-1);
		}
			
		parentCaller.UpdatePredictions(len);

		// Next pair
		seq_position++;
		parentCaller.UpdateBatchesDone(1);

		if (seq_position >= loaded_seq) {
			loaded_seq = 0;
			seq_position = 0;
		}
	}
}

void DcgmTrainer::WriteModel(const std::string& outputPath) {
	if (outputPath != "") 
	{	
		std::string output;
		std::stringstream outDir;
		outDir << outputPath 
			<< "_dcgm" << ((m_dcgmType == DcgmType::DCGMI) ? "I" : "II") 
			<< "_" << m_epochs
			<< "_rnnAct" << m_act
			<< "_tau" << m_tau 
			<< "_a" << m_startAlpha
			<< "_b" << m_beta
			<< "_hid" << m_nHid 
			<< "_out" << m_nOut
			<< (m_includeH ? "_wH" : "") 
			<< (m_includeO ? "_wO" : "") 
			<< (m_deepO ? "_dO" : "");

		for (int i = 0; i < m_layerDims.size(); i++) {
			outDir << "_d" << m_layerDims[i];
		}

		output = outDir.str();
		
		std::cout << "Saving model parameters to " << output << std::endl;

		std::string make_dir = "mkdir " + output;
		system(make_dir.c_str());

		WriteVocabulary(output);
		WriteRnnWeights(output);
		WriteErrors(output);

		std::ofstream modelType(output + "/modelType");
		modelType << "DCGM-" << ((m_dcgmType == DcgmType::DCGMI) ? "I" : "II") << std::endl;
		modelType.close();
		
		std::ofstream config(output + "/config");
		config << "act=" << m_act << std::endl;
		config << "alpha=" << m_alpha << std::endl;
		config << "beta=" << m_beta << std::endl;
		config.close();

		if (m_deepO) {
			deepO.GetDense()->WriteToExternalBinary(output + "/do.bin");
		}
		
		std::ofstream d(output + "/d_num.txt");
		d << m_layerDims.size() << std::endl;
		d.close();

		for (int i = 0; i < m_layerWeights.size(); i++) {
			m_layerWeights[i]->GetDense()->WriteToExternalBinary(output + "/d_" + std::to_string(i) + ".bin");
		}
		for (int i = 0; i < m_layerBias.size(); i++) {
			m_layerBias[i]->GetDense()->WriteToExternalBinary(output + "/b_" + std::to_string(i) + ".bin");
		}
	}
}