
#pragma once

#include "FcgmTrainer.hpp"
#include "DcgmI.hpp"
#include "DcgmII.hpp"
#include "print.hpp"
#include <chrono>

class DcgmTrainer : public FcgmTrainer {
typedef enum { DCGMI, DCGMII } DcgmType;

protected:
	bool m_hasBiases;
	std::vector<int> m_layerDims;
	std::vector<Matrix*> m_layerWeights;
	std::vector<Matrix*> m_layerBias;
	DcgmType m_dcgmType;
	Activation m_ctxAct;

public:
	DcgmTrainer(Vocabulary* v, RandomGenerator* g) : FcgmTrainer(v, g) {
		m_dcgmType = DcgmType::DCGMI;
		m_hasBiases = false;
		m_ctxAct = SIGMOID;
	}

	void SetDcgmType(const std::string& t) {
		if (t == "DCGMI") m_dcgmType = DcgmType::DCGMI;
		else m_dcgmType = DcgmType::DCGMII;
	}

	void InitializeWeights(int nIn, int nHid, int nOut, const std::vector<int>& layerDims) { 
		if (layerDims.size() == 0) {
			throw std::exception("Should specify at least one layer!");
		}
		
		InitializeRnnWeights(nIn, nHid, nOut);
		m_layerDims = layerDims;

		std::cerr << "Initializing model DCGM-" << m_dcgmType << " " << m_nIn << ":" << m_nHid << ":" << m_nOut;
		for (int i = 0; i < layerDims.size(); i++)
			std::cerr << ":" << layerDims[i];
		std::cerr<<std::endl;

		// Deep forward layers of the DCGM
		for (int i = 0; i < layerDims.size(); i++) {
			int bottomLayerDim = 0;
			int currLayerDim = layerDims[i];
			if (i == 0) bottomLayerDim = m_nIn;
			else bottomLayerDim = layerDims[i-1];

			if (i == 1 && m_dcgmType == DCGMII)
			{
				// Concat here
				bottomLayerDim = layerDims[i-1]*2;
			}

			Matrix* D_i = new Matrix(currLayerDim, bottomLayerDim, "D_" + std::to_string(i));
			random_fill_policy(D_i->GetDataForUpdate(), D_i->GetSize());
			m_layerWeights.push_back(D_i);
						
			Matrix* b_i = new Matrix(currLayerDim, 1, "b_" + std::to_string(i));
			m_layerBias.push_back(b_i);
		}

		if (m_includeO) {
			O.Resize(m_nOut, layerDims[layerDims.size()-1]);
			O.SetName("O");
			random_fill_policy(O.GetDataForUpdate(), O.GetSize());
		}

		if (m_includeH || m_deepO) {
			H.Resize(m_nHid, layerDims[layerDims.size()-1]);
			H.SetName("H");
			random_fill_policy(H.GetDataForUpdate(), H.GetSize());
		}
				
		if (m_deepO) {
			deepO.Resize(m_nHid, m_nHid);
			deepO.SetName("deepO");
			random_fill_policy(deepO.GetDataForUpdate(), deepO.GetSize());
		}
	}

	void DoJob(AbstractTrainer<TripleReader>& parentCaller, 
		ThreadedQueue<Batch<Triple>>& workingQueue,
		bool isTraining=false);
	void WriteModel(const std::string& outputPath);
	void PrintStatus(bool training=true);
	void SetBiases(bool yes=false) {
		m_hasBiases = true;
	}

	void SetPretrainedCtxEmbeddings(const std::string& extFile) {
		if (m_layerWeights.size() == 0) {
			throw std::exception("You should first initialize the weights.");
		}

		Matrix* D_0 = m_layerWeights[0];
		D_0->LoadFromExternalBinary(extFile);
		
		std::cerr << "External embeddings loaded" << std::endl;
	}

	void SetContextActivation(const std::string& act) 
	{ 
		if (act == "sigmoid") m_ctxAct = SIGMOID; 
		else if (act == "tanh") m_ctxAct = TANH;
		else if (act == "relu") m_ctxAct = ReLU;
	}
	
	~DcgmTrainer() {
		for (int i = 0; i < m_layerWeights.size(); i++) {
			if (m_layerWeights[i] != nullptr)
				delete m_layerWeights[i];
		}
		for (int i = 0; i < m_layerBias.size(); i++) {
			if (m_layerBias[i] != nullptr)
				delete m_layerBias[i];
		}
	}
};