#include "Fcgm.hpp"

Fcgm::Fcgm() : Rnn() {
	m_I = nullptr;	
	m_H = nullptr;
	m_O = nullptr;
	m_nC = 0;
	m_includeH = true;
	m_includeO = true;
	m_deepO = false;
	m_modelName = "FCGM";
}
Fcgm::Fcgm(bool proprietaryMemory) : Rnn(proprietaryMemory) {
	m_I = nullptr;	
	m_H = nullptr;
	m_O = nullptr;
	m_nC = 0;
	m_includeH = true;
	m_includeO = true;
	m_deepO = false;
	m_modelName = "FCGM";
}
Fcgm::Fcgm(int nHid, int nOut) : Rnn(nHid, nOut) {
	m_I = nullptr;	
	m_H = nullptr;
	m_O = nullptr;
	m_includeH = true;
	m_deepO = false;
	m_includeO = true;
	m_nC = 0;
	m_modelName = "FCGM";
}

Layer* Fcgm::BuildContext(const Triple& triple, NeuralNet& rnn) {
	// Build the context vector
	SparseInputLayer* cmInput = rnn.AddLayer(new SparseInputLayer(m_nIn, "cm_i"));
	Layer* contextLayer = rnn.AddLayer(new HiddenLayer(m_nC, "c_o"));
	rnn.Connect(cmInput, contextLayer, new FixedWeight(m_I));

	// Does NOT need gradient and should be skipped for reset
	contextLayer->SetNeedError(false);
	contextLayer->SetSkipReset(true);

	// Paste C&M together as BOW
	std::vector<int> cmIds;	
	std::for_each(triple[0].begin(), triple[0].end(), [&](int id) {
		// Skip ss and se tokens
		if (id != 0 && id != 1) cmIds.push_back(id); 
	});

	std::for_each(triple[1].begin(), triple[1].end(), [&](int id) {
		// Skip ss and se tokens
		if (id != 0 && id != 1) cmIds.push_back(id); 
	});

	std::vector<double> cmValues(cmIds.size(), 1.0);
	cmInput->SetData(m_nIn, cmIds, cmValues);
	return contextLayer;
}

void Fcgm::AddStep(const Sequence& s, size_t end, NeuralNet& rnn, Layer*& contextLayer, Layer*& lastHiddenLayer, bool training) {
	if (end >= s.size())
		throw std::exception("Out of bound step in RNN!");

	// link to the last time step
	rnn.DeleteOutputs();

	Layer* hl = lastHiddenLayer;
	std::string id = std::to_string(end);

	// set current input
	SparseInputLayer* x_next = rnn.AddLayer(new SparseInputLayer(m_nIn, "x_" + id));
	std::vector<int> ids(1, s[end-1]);
	std::vector<double> values(1, 1);
	x_next->SetData(m_nIn, ids, values);

	HiddenLayer * h_next = rnn.AddLayer(new HiddenLayer(m_nHid, m_act, "h_" + id));

	if (hl != 0) rnn.Connect(hl, h_next, new DynamicWeight(m_Whh));
	rnn.Connect(x_next, h_next, new DynamicWeight(m_Win));

	rnn.Connect(rnn.AddLayer(new BiasLayer("bhh_" + id)), h_next, new DynamicWeight(m_Bhh)); 
	rnn.Connect(contextLayer, h_next, new DynamicWeight(m_H));

	LossLayer* y_t;
	if (!training) {
		CrossEntropyLossLayer* yCe_t = rnn.AddLayer(new CrossEntropyLossLayer(m_nOut, "pred_" + id));
		yCe_t->SetOutput(1, s[end]);
		y_t = yCe_t;
	} else {
		std::vector<int> samples = m_generator->SampleNoise(25, s[end]);
		std::vector<double> probs; probs.reserve(25);
		std::for_each(samples.begin(), samples.end(),
			[&](int x) { probs.push_back(m_generator->GetLogProb(x)); }); 

		double target_prob = -m_generator->GetLogProb(s[end]);
		samples.push_back(s[end]);
		probs.push_back(target_prob);

		NCELossLayer* yNce_t = rnn.AddLayer(new NCELossLayer(m_nOut, "pred_" + id));
		yNce_t->SetOutput(samples, probs);
		y_t = yNce_t;
	}

	rnn.Connect(h_next, y_t, new DynamicWeight(m_Wout));
	rnn.Connect(rnn.AddLayer(new BiasLayer("bout_" + id)), y_t, new DynamicWeight(m_Bout)); 

	lastHiddenLayer = h_next;
}

void Fcgm::CleanNetwork(NeuralNet& rnn, bool cleanActivations) {
	ResetVisitor v;
	v.cleanError_ = true;
	v.cleanActivation_ = cleanActivations;
	rnn.Visit(v);
}

Fcgm* Fcgm::LoadFromDisk(const std::string& dirPath) {
	Fcgm* a = new Fcgm(true);

	Matrix* Win = new Matrix();
	Matrix* Whh = new Matrix();
	Matrix* Wout = new Matrix();
	Matrix* bout = new Matrix();
	Matrix* bhh = new Matrix();
	Matrix* I = new Matrix();
	Matrix* O = new Matrix();
	Matrix* H = new Matrix();

	Vocabulary* vO = new Vocabulary(dirPath + "\\vocab");
	RandomGenerator* g = new RandomGenerator(*vO);

	Whh->LoadFromExternalBinary(dirPath + "\\w_hh.bin");
	std::cerr << Whh->GetRows() << " " << Whh->GetCols() << " " << Whh->Norm2() << std::endl;

	Win->LoadFromExternalBinary(dirPath + "\\w_in.bin");
	std::cerr << Win->GetRows() << " " << Win->GetCols() << " " << Win->Norm2() << std::endl;

	Wout->LoadFromExternalBinary(dirPath + "\\w_out.bin");
	std::cerr << Wout->GetRows() << " " << Wout->GetCols() << " " << Wout->Norm2() << std::endl;

	I->LoadFromExternalBinary(dirPath + "\\i.bin");
	std::cerr << I->GetRows() << " " << I->GetCols() << " " << I->Norm2() << std::endl;

	bool bo_found = bout->LoadFromExternalBinary(dirPath + "\\b_out.bin");
	bool bhh_found = bhh->LoadFromExternalBinary(dirPath + "\\b_hh.bin");
	if (bo_found&&bhh_found)
		a->SetExternalBias(bhh, bout);

	bool ofound = O->LoadFromExternalBinary(dirPath + "\\o.bin");
	if (ofound)
		std::cerr << O->GetRows() << " " << O->GetCols() << " " << O->Norm2() << std::endl;

	bool hfound = H->LoadFromExternalBinary(dirPath + "\\h.bin");
	if (hfound)
		std::cerr << H->GetRows() << " " << H->GetCols() << " " << H->Norm2() << std::endl;

	Config c = Config::ReadConfigFile(dirPath + "\\config");
	a->SetAct(c.GetString("act", "sigmoid"));

	std::cerr << "RnnActivation = " << c.GetString("act", "sigmoid") << std::endl;
	std::cerr << "ContextActivation = " << c.GetString("ctxAct", "sigmoid") << std::endl;

	a->SetIncludeH(hfound);
	a->SetIncludeO(ofound);
	a->SetVocabulary(vO);
	a->SetGenerator(g);
	a->SetExternalRnnWeights(Win, Whh, Wout);
	a->SetContextRnnWeights(H, O);
	a->SetExternalWeights(I);
	return a;
}

std::vector<Candidate> Fcgm::GenerateNextWord(const Triple& ctx, 
	PastInformation* info, int currentInput, int nK, bool sampleCandidates)
{
	double error = 0;
	NeuralNet rnn;
	std::vector<Matrix*> outputs;
	bool networkChanged = false;

	Layer* lastHiddenLayer = nullptr;
	Layer* ctxLayer = nullptr;

	Matrix* lastH = nullptr;
	Matrix* ctxA = nullptr;

	if (currentInput == m_vocabulary->sId) 
	{
		lastH = new Matrix();
		ctxA = new Matrix();

		info->infos.insert(std::make_pair("ctx", ctxA));
		info->infos.insert(std::make_pair("last_h", lastH));

		ctxLayer = BuildContext(ctx, rnn);
	} 
	else
	{
		ctxLayer = rnn.AddLayer(new HiddenLayer(m_nC, ID, "c_e"));
		lastHiddenLayer = rnn.AddLayer(new HiddenLayer(m_nHid, ID, "h_e"));

		// Copy in past activations
		ctxA = info->infos["ctx"];
		lastH = info->infos["last_h"];

		ctxLayer->GetActivation()->CopyFrom(*ctxA);
		lastHiddenLayer->GetActivation()->CopyFrom(*lastH);

		// And set evaluated flag
		ctxLayer->SetEvaluated(true);
		lastHiddenLayer->SetEvaluated(true);
	}

	Sequence past;
	past.push_back(currentInput);
	past.push_back(0);

	AddStep(past, past.size()-1, rnn, ctxLayer, lastHiddenLayer, false);
	outputs = rnn.Forward();

	// Save back the last activation now
	lastH->CopyFrom(*lastHiddenLayer->GetActivation());	
	ctxA->CopyFrom(*ctxLayer->GetActivation());

	// Get the max
	ThreadedFixedPQ<Candidate, std::vector<Candidate>, CandidateComparator> max(nK);

	if (!sampleCandidates)
	{
		// Take the top-k
		int w = 0;
		for (; w < outputs.back()->GetSize(); w++) {
			double currProb = outputs.back()->GetElement(w, 0);
			Candidate r(w, log(currProb));
			max.Push(r);
		}
	} 
	else
	{
		// Sample without replacement	
		std::set<int> sampledKeys;
		std::vector<Candidate> samples;

		double prev = 0;
		double maxProb = 1;

		// Can be done better but lack of time...
		while (sampledKeys.size() != nK)
		{
			double random = 0;
			double g = 0;
			maxProb -= prev;
			max_min_policy(&random, 1, 0, maxProb);		
			int w = 0;

			for (w = 0; w < outputs.back()->GetSize(); w++) {
				if (sampledKeys.find(w) != sampledKeys.end())
					continue;

				g += outputs.back()->GetElement(w, 0);
				if (g > random) {
					break;
				}
			}

			w = MIN(w, m_vocabulary->vocab_size-1);
			sampledKeys.insert(w);

			Candidate c;
			c.id = w;
			prev = outputs.back()->GetElement(w, 0);
			c.score = log(prev);
			max.Push(c);
		}
	}

	return max.PopAll();
}

double Fcgm::Test(const Triple& tri, bool stoch) {
	double error = 0;
	NeuralNet rnn;
	std::vector<Matrix*> outputs;
	bool networkChanged = false;

	Layer* ctxLayer = BuildContext(tri, rnn);
	Layer* lastHiddenLayer = nullptr;
	for (size_t step = 1; step < tri[2].size(); step++) {
		AddStep(tri[2], step, rnn, ctxLayer, lastHiddenLayer, stoch);
		outputs = rnn.Forward();
		if (stoch) {
			error += -outputs.back()->GetVecSparse()->GetValues().back();
		} else {
			// log prob.
			error += rnn.GetLoss();
		}
	}

	return error;
}

double Fcgm::Train(const Triple& triple) {
	NeuralNet rnn;
	rnn.SetBptt(m_Whh, m_tau);

	double error = 0;
	std::vector<Matrix*> outputs;
	bool networkChanged = false;

	const Sequence& R = triple[2];
	size_t len = R.size();

	Layer* ctxLayer = BuildContext(triple, rnn);
	Layer* lastHiddenLayer = nullptr;
	for (size_t step = 1; step < len; step++) 
	{
		CleanNetwork(rnn, networkChanged);
		networkChanged = false;

		AddStep(R, step, rnn, ctxLayer, lastHiddenLayer);
		rnn.Optimize();

		outputs = rnn.Forward();

		// Compute the loss
		error += rnn.GetLoss();
		rnn.Backward();
		rnn.Visit(v);

		// BPTT-block
		if (step % 10 == 0 || step == len - 1) 
		{
			// Update the parameters here!
			m_updater->UpdateParameters(v);
			networkChanged = true; // trigger to clean the rnn
		}
	}

	return error;
}