#include "FcgmTrainer.hpp"

void FcgmTrainer::PrintStatus(bool training) {
	clock_t now = clock();
	double entropy = (m_sharedError/m_sharedPredictions)/log(2);
	double perpl = pow(2, entropy);
	long long difftime = (now - m_startTime)/1000;

	std::cout << difftime/3600 <<":"<<(difftime % 3600)/60<<":"<<((difftime % 3600) % 60); 
	std::cout << " - Entr. / Perp. : " << std::setw(8) << entropy << " / " << perpl
		<< " P/s: " << std::setw(5) << (long long) (m_sharedPredictions/difftime);
		std::flush(std::cout);

	/** Print norms **/
	std::cout << std::setw(5);

	std::cout << " W_hh : " << std::setw(5) << Whh.Norm2();
	std::cout << " W_out : " << std::setw(5) << Wout.Norm2();
	std::cout << " W_in : " << std::setw(5) << Win.Norm2();
	std::cout << " H : " << std::setw(5) << H.Norm2();
	std::cout << " O : " << std::setw(5) << O.Norm2();
	std::cout << " I : " << std::setw(5) << I.Norm2();	
	std::cout << std::endl;
}

void FcgmTrainer::DoJob(AbstractTrainer<TripleReader>& parentCaller, 
		ThreadedQueue<Batch<Triple>>& workingQueue,
		bool isTraining) {
	long long loaded_seq = 0, seq_position = 0;
	Batch<Triple> wi;

	IGradientUpdater* upd = new SGDUpdater(parentCaller.GetAlpha());

	Fcgm fcgm;
	fcgm.SetTau(m_tau);
	fcgm.SetExternalRnnWeights(&Win, &Whh, &Wout);
	fcgm.SetContextRnnWeights(&H, &O);
	fcgm.SetExternalBias(&bhh, &bout);
	fcgm.SetExternalWeights(&I);
	fcgm.SetUpdater(upd);
	fcgm.SetGenerator(m_generator);
	fcgm.SetTiled(&Whh);
	fcgm.SetAct(m_act);
	fcgm.SetTiled(&H);
	fcgm.SetIncludeH(m_includeH);
	
	// Some regularization for the parameters!
	if (m_beta != 0) {
		upd->add_pre_filter(&Whh, new L2RegularizeFilter(m_beta));
		upd->add_pre_filter(&Wout, new L2RegularizeFilter(m_beta));
		upd->add_pre_filter(&Win, new L2RegularizeFilter(m_beta));
		upd->add_pre_filter(&O, new L2RegularizeFilter(m_beta));
		upd->add_pre_filter(&H, new L2RegularizeFilter(m_beta));
	}

	while (1) {
		// GetElement job
		if (loaded_seq == 0) {
			// GetElement the work item
			wi = static_cast<Batch<Triple>&&>(workingQueue.PopFront());
			loaded_seq = wi.m_size;
			seq_position = 0;

			// poison pill
			if (loaded_seq == 0) {
				delete upd;
				// Exit here
				break;
			}
		}

		Triple& tri = wi.data[seq_position];
		int len = (int) tri[2].size();

		try {
			if (isTraining) parentCaller.UpdateError(fcgm.Train(tri));
			else parentCaller.UpdateError(fcgm.Test(tri));
		} catch (std::exception& e) {
			std::cerr<<e.what()<<std::endl;
			exit(-1);
		}
		parentCaller.UpdatePredictions(len);

		// Next pair
		seq_position++;
		parentCaller.UpdateBatchesDone(1);

		if (seq_position >= loaded_seq) {
			loaded_seq = 0;
			seq_position = 0;
		}
	}
}

void FcgmTrainer::WriteRnnWeights(const std::string& output)
{
	Win.GetDense()->WriteToExternalBinary(output + "/w_in.bin");
	Wout.GetDense()->WriteToExternalBinary(output + "/w_out.bin");
	Whh.GetDense()->WriteToExternalBinary(output + "/w_hh.bin");
	bhh.GetDense()->WriteToExternalBinary(output + "/b_hh.bin");
	bout.GetDense()->WriteToExternalBinary(output + "/b_out.bin");
	
	if (m_includeO) O.GetDense()->WriteToExternalBinary(output + "/o.bin");
	if (m_includeH) H.GetDense()->WriteToExternalBinary(output + "/h.bin");
}

void FcgmTrainer::WriteModel(const std::string& outputPath) {
	if (outputPath != "") 
	{	
		std::string output;
		std::stringstream outDir;
		outDir << outputPath << "_fcgm_" << m_epochs 
			<< "_tau" << m_tau 
			<< "_a" << m_startAlpha
			<< "_b" << m_beta
			<< "_c" << m_nC
			<< "_in" << m_nHid 
			<< "_out" << m_nOut;
		
		output = outDir.str();
		
		std::cout << "Saving model parameters to " << output << std::endl;
				
		std::string make_dir = "mkdir " + output;
		system(make_dir.c_str());
		
		std::ofstream modelType(output + "/modelType");
		modelType << "FCGM" << std::endl;
		modelType.close();

		I.GetDense()->WriteToExternalBinary(output + "/i.bin");
				
		WriteVocabulary(output);
		WriteRnnWeights(output);
		WriteErrors(output);
	}
}