
#pragma once

#include "AbstractTrainer.hpp"
#include "GradientUpdater.hpp"
#include "SequenceReader.hpp"
#include "Fcgm.hpp"
#include "print.hpp"
#include <chrono>

class FcgmTrainer : public AbstractTrainer<TripleReader> {
protected:
	int m_tau;
	int m_nIn; 
	int m_nHid; 
	int m_nC;
	int m_nOut;

	bool m_includeH;
	bool m_includeO;
	bool m_deepO;

	Activation m_act;
	
	Matrix deepO;
	Matrix O;
	Matrix H;
	Matrix bout;
	Matrix bhh;
	Matrix Win;
	Matrix Wout;
	Matrix Whh;

private:
	Matrix I;

public:
	FcgmTrainer(Vocabulary* v, RandomGenerator* g) : AbstractTrainer(v, g) {
		m_includeH = true;
		m_includeO = false;
		m_act = SIGMOID;
		m_deepO = false;
		m_tau = 7;
	}

	void InitializeExternalEmbeddings(const std::string& ePath)
	{
		int founds = LoadFromExternalEmbeddings(ePath, *m_vocabulary, &I);
		I.InPlaceTranspose();
		I.SetName("W_emb");
		m_nC = I.GetRows();
		
		std::cerr << "External embeddings loaded, terms found = " << founds << std::endl;
	}

	void InitializeRnnWeights(int nIn, int nHid, int nOut) 
	{
		m_nIn = nIn;
		m_nHid = nHid;
		m_nOut = nOut;

		Win.Resize(m_nHid, m_nIn);
		Win.SetName("W_in");
		
		Wout.Resize(m_nOut, m_nHid);
		Wout.SetName("W_out");
		
		Whh.Resize(m_nHid, m_nHid);
		Whh.SetName("W_hh");
		
		bhh.Resize(m_nHid, 1);
		bhh.SetName("b_hh");
		
		bout.Resize(m_nOut, 1);
		bout.SetName("b_out");
		
		random_fill_policy(Win.GetDataForUpdate(), Win.GetSize());
		random_fill_policy(Wout.GetDataForUpdate(), Wout.GetSize());
		random_fill_policy(Whh.GetDataForUpdate(), Whh.GetSize());
		orthogonal_projection(Whh.GetRows(), Whh.GetCols(), Whh.GetDataForUpdate());
	}

	void InitializeWeights(int nIn, int nHid, int nOut, const std::string& ePath) { 
		InitializeRnnWeights(nIn, nHid, nOut);
		InitializeExternalEmbeddings(ePath);
		
		if (m_includeO) {
			O.Resize(Wout.GetRows(), I.GetRows());
			O.SetName("O");
			random_fill_policy(O.GetDataForUpdate(), O.GetSize());
		}

		if (m_includeH || m_deepO) {
			H.Resize(Whh.GetRows(), I.GetRows());
			H.SetName("H");
			random_fill_policy(H.GetDataForUpdate(), H.GetSize());
		}

		if (m_deepO) {
			deepO.Resize(Whh.GetRows(), Whh.GetCols());
			deepO.SetName("deepO");
			random_fill_policy(deepO.GetDataForUpdate(), deepO.GetSize());
		}
	}

	void SetTau(int tau) { m_tau = tau; }
	
	void SetActivation(const std::string& act) 
	{ 
		if (act == "sigmoid") m_act = SIGMOID; 
		else if (act == "tanh") m_act = TANH;
		else if (act == "relu") m_act = ReLU;
	}

	
	void SetDeepOutput(bool deepO) { m_deepO = deepO; }
	void SetIncludeH(bool incH) { m_includeH = incH; }
	void SetIncludeO(bool incO) { m_includeO = incO; }
	
	virtual void DoJob(AbstractTrainer<TripleReader>& parentCaller, 
		ThreadedQueue<Batch<Triple>>& workingQueue,
		bool isTraining=false);

	virtual void WriteModel(const std::string& outputPath);
	void WriteRnnWeights(const std::string& outputPath);

	void PrintStatus(bool training=true);

	~FcgmTrainer() {}
};