
#include "Rcgm.hpp"

/// Build the network for this triple 
/// (the time spent in building the network is likely to be 
/// irrelevant wrt the time passed in matrix multiplications).
void Rcgm::BuildNetwork(const Triple& triple, NeuralNet& network) {
	m_nceLayers.clear();
	m_ceLayers.clear();

	Layer* hiddenC = BuildRnn(nullptr, "c", triple[0], network);
	Layer* hiddenM = BuildRnn(hiddenC, "m", triple[1], network);
	Layer* hiddenR = BuildRnn(hiddenM, "r", triple[2], network);

	network.Connect(hiddenC, hiddenM, new DynamicWeight(m_Wshh));
	network.Connect(hiddenM, hiddenR, new DynamicWeight(m_Wshh));
	
	network.Connect(network.AddLayer(new BiasLayer("bs_hh_m")), hiddenM, new DynamicWeight(m_bshh));
	network.Connect(network.AddLayer(new BiasLayer("bs_hh_c")), hiddenC, new DynamicWeight(m_bshh));
	network.Connect(network.AddLayer(new BiasLayer("bs_hh_r")), hiddenR, new DynamicWeight(m_bshh));
	network.Optimize();
}

/// Builds the rnn for the sentence and returns a max-pooling 
/// layer.
Layer* Rcgm::BuildRnn(Layer* contextLayer, std::string seqName, const Sequence& seq, NeuralNet& network)
{
	HiddenLayer* lastHiddenLayer = nullptr;
	HiddenLayer* sentInputLayer = network.AddLayer(new HiddenLayer(m_nHid, "hs_" + seqName));
	MaxPoolingLayer* poolingLayer = network.AddLayer(new MaxPoolingLayer(m_nHid, "pool_" + seqName));
	
	network.Connect(poolingLayer, sentInputLayer, new FixedWeight());	

	for (int i = 0; i < seq.size() - 1; ++i) 
	{
		std::string id = seqName + "_" + std::to_string(i+1);
		int input = seq[i];
		int output = seq[i+1];

		/// set current input
		SparseInputLayer* x_next = network.AddLayer(new SparseInputLayer(m_nIn, "x_" + id));
		std::vector<int> ids(1, input);
		std::vector<double> values(1, 1);
		x_next->SetData(m_nIn, ids, values);
		///

		HiddenLayer * h_next = network.AddLayer(new HiddenLayer(m_nHid, m_act, "h_" + id));
		if (lastHiddenLayer != nullptr) network.Connect(lastHiddenLayer, h_next, new DynamicWeight(m_Whh));
		
		/// Input connection
		network.Connect(x_next, h_next, new DynamicWeight(m_Win));
		network.Connect(network.AddLayer(new BiasLayer("bhh_" + id)), h_next, new DynamicWeight(m_Bhh)); 

		/// Connect to pooling layer
		network.Connect(h_next, poolingLayer, new FixedWeight());

		/// Output layers
		NCELossLayer* yNce_t = network.AddLayer(new NCELossLayer(m_nOut, "pred_" + id));
		CrossEntropyLossLayer* yCe_t = network.AddLayer(new CrossEntropyLossLayer(m_nOut, "pred_" + id));
		
		std::vector<int> samples = m_generator->SampleNoise(25, output);
		std::vector<double> probs; probs.reserve(25);
		std::for_each(samples.begin(), samples.end(),
			[&](int x) { probs.push_back(m_generator->GetLogProb(x)); }); 

		double target_prob = -m_generator->GetLogProb(output);
		samples.push_back(output);
		probs.push_back(target_prob);

		yNce_t->SetOutput(samples, probs);
		yCe_t->SetOutput(m_nOut, output);
		///

		if (contextLayer != nullptr)
			network.Connect(contextLayer, h_next, new DynamicWeight(m_H));

		m_nceLayers.push_back(yNce_t);
		m_ceLayers.push_back(yCe_t);

		network.Connect(h_next, yNce_t, new DynamicWeight(m_Wout));
		network.Connect(h_next, yCe_t, new DynamicWeight(m_Wout));

		network.Connect(network.AddLayer(new BiasLayer("boutC_" + id)), yNce_t, new DynamicWeight(m_Bout));
		network.Connect(network.AddLayer(new BiasLayer("boutNC_" + id)), yCe_t, new DynamicWeight(m_Bout));
		lastHiddenLayer = h_next;
	}
	
	return sentInputLayer;
}

double Rcgm::Test(const Triple& tri, bool stoch) {
	double error = 0;
	NeuralNet network;
	network.SetBptt(m_Whh, m_tau);
	BuildNetwork(tri, network);

	for (size_t step = 0; step < m_ceLayers.size(); step++) 
	{
		LossLayer* currentLossLayer = m_ceLayers[step];		
		network.Forward(currentLossLayer);
		error += network.GetLoss(currentLossLayer);
	}

	return error;
}

double Rcgm::Train(const Triple& triple) {
	NeuralNet network;
	network.SetBptt(m_Whh, m_tau);
	// network.SetBarrier(m_Wsin);
	BuildNetwork(triple, network);

	double error = 0;
	std::vector<Matrix*> outputs;
	bool networkChanged = false;

	for (size_t step = 0; step < m_nceLayers.size(); step++) 
	{
		LossLayer* currentLossLayer = m_nceLayers[step];
		CleanNetwork(network, networkChanged);
		networkChanged = false;
		
		network.Forward(currentLossLayer);
		error += network.GetLoss(currentLossLayer);
		network.Backward(currentLossLayer);
		network.Visit(v);

		// BPTT-block
		if (step % 10 == 0 || step == m_nceLayers.size() - 1) 
		{
			// Update the parameters here!
			m_updater->UpdateParameters(v);
			networkChanged = true; // trigger to clean the rnn
		}
	}

	return error;
}