#include "RcgmTrainer.hpp"

void RcgmTrainer::PrintStatus(bool training) {
	clock_t now = clock();
	double entropy = (m_sharedError/m_sharedPredictions)/log(2);
	double perpl = pow(2, entropy);
	long long difftime = (now - m_startTime)/1000;

	std::cout << difftime/3600 <<":"<<(difftime % 3600)/60<<":"<<((difftime % 3600) % 60); 
	std::cout << " - Entr. / Perp. : " << std::setw(8) << entropy << " / " << perpl
		<< " P/s: " << std::setw(5) << (long long) (m_sharedPredictions/difftime);
		std::flush(std::cout);

	/** Print norms **/
	std::cout << std::setw(5);

	std::cout << " W_hh : " << std::setw(5) << Whh.Norm2();
	std::cout << " W_out : " << std::setw(5) << Wout.Norm2();
	std::cout << " W_in : " << std::setw(5) << Win.Norm2();

	std::cout << " WS_in : " << std::setw(5) << WSin.Norm2();
	std::cout << " WS_hh : " << std::setw(5) << WShh.Norm2();
	std::cout << " b_hh : " << std::setw(5) << bhh.Norm2();
	std::cout << " b_out : " << std::setw(5) << bout.Norm2();
	std::cout << " bs_hh : " << std::setw(5) << bshh.Norm2();
	
	std::cout << " H : " << std::setw(5) << H.Norm2();
	// std::cout << " O : " << std::setw(5) << O.Norm2();
	
	std::cout << std::endl;
}

void RcgmTrainer::DoJob(AbstractTrainer<TripleReader>& parentCaller, 
		ThreadedQueue<Batch<Triple>>& workingQueue,
		bool isTraining) {
	long long loaded_seq = 0, seq_position = 0;
	Batch<Triple> wi;

	IGradientUpdater* upd = new AdagradUpdater(parentCaller.GetAlpha());

	Rcgm rcgm;
	rcgm.SetTau(m_tau);
	rcgm.SetExternalRnnWeights(&Win, &Whh, &Wout);
	rcgm.SetContextRnnWeights(&H, &O);
	rcgm.SetExternalBias(&bhh, &bout);
	rcgm.SetSentenceRnnWeights(&WShh, &WSin, &bshh);
	rcgm.SetUpdater(upd);
	rcgm.SetGenerator(m_generator);
	rcgm.SetTiled(&Whh);
	rcgm.SetTiled(&H);
	
	// Some regularization for the parameters!
	if (m_beta != 0) {
		upd->add_pre_filter(&Whh, new L2RegularizeFilter(m_beta));
		upd->add_pre_filter(&Wout, new L2RegularizeFilter(m_beta));
		upd->add_pre_filter(&Win, new L2RegularizeFilter(m_beta));
		upd->add_pre_filter(&WSin, new L2RegularizeFilter(m_beta));
		upd->add_pre_filter(&WShh, new L2RegularizeFilter(m_beta));
		upd->add_pre_filter(&bshh, new L2RegularizeFilter(m_beta));
		upd->add_pre_filter(&bhh, new L2RegularizeFilter(m_beta));
		upd->add_pre_filter(&bout, new L2RegularizeFilter(m_beta));
		upd->add_pre_filter(&H, new L2RegularizeFilter(m_beta));
	}

	while (1) {
		// GetElement job
		if (loaded_seq == 0) {
			// GetElement the work item
			wi = static_cast<Batch<Triple>&&>(workingQueue.PopFront());
			loaded_seq = wi.m_size;
			seq_position = 0;

			// poison pill
			if (loaded_seq == 0) {
				delete upd;
				// Exit here
				break;
			}
		}

		Triple& tri = wi.data[seq_position];
		int len = (int) tri[0].size() + tri[1].size() + tri[2].size();

		try {
			if (isTraining) parentCaller.UpdateError(rcgm.Train(tri));
			else parentCaller.UpdateError(rcgm.Test(tri));
		} catch (std::exception& e) {
			std::cerr<<e.what()<<std::endl;
			exit(-1);
		}
		parentCaller.UpdatePredictions(len);

		// Next pair
		seq_position++;
		parentCaller.UpdateBatchesDone(1);

		if (seq_position >= loaded_seq) {
			loaded_seq = 0;
			seq_position = 0;
		}
	}
}

void RcgmTrainer::WriteModel(const std::string& outputPath) {
	if (outputPath != "") 
	{	
		std::string output;
		std::stringstream outDir;
		outDir << outputPath << "_rcgm_" << m_epochs 
			<< "_tau" << m_tau 
			<< "_act" << m_act
			<< "_a" << m_startAlpha
			<< "_b" << m_beta
			<< "_in" << m_nHid 
			<< "_out" << m_nOut;
		
		output = outDir.str();
		
		std::cout << "Saving model parameters to " << output << std::endl;

		std::string make_dir = "mkdir " + output;
		system(make_dir.c_str());
		
		WriteVocabulary(output);
		WriteRnnWeights(output);
		WriteErrors(output);

		std::ofstream modelType(output + "/modelType");
		modelType << "RCGM" << std::endl;
		modelType.close();
		
		std::ofstream config(output + "/config");
		config << "act=" << m_act << std::endl;
		config << "alpha=" << m_alpha << std::endl;
		config << "beta=" << m_beta << std::endl;
		config.close();
		
		bshh.GetDense()->WriteToExternalBinary(output + "/bs_hh.bin");
		WSin.GetDense()->WriteToExternalBinary(output + "/ws_in.bin");
		WShh.GetDense()->WriteToExternalBinary(output + "/ws_hh.bin");
		H.GetDense()->WriteToExternalBinary(output + "/h.bin");
	}
}