#include "Rnn.hpp"

Rnn::Rnn(bool proprietaryMemory) : AbstractModel(proprietaryMemory) {
	m_nIn = 0;
	m_act = SIGMOID;
	m_modelName = "RNN";
	m_nOut = 0;
	m_nHid = 0;
	m_tau = 7;	
	m_Bout = nullptr;
	m_Bhh = nullptr;
	m_Whh = nullptr;
	m_Wout = nullptr;
	m_Win = nullptr;
}
Rnn::Rnn() {
	m_nIn = 0;
	m_modelName = "RNN";
	m_act = SIGMOID;
	m_nOut = 0;
	m_nHid = 0;
	m_tau = 7;	
	m_Bout = nullptr;
	m_Bhh = nullptr;
	m_Whh = nullptr;
	m_Wout = nullptr;
	m_Win = nullptr;
}
Rnn::Rnn(int nHid, int nOut) {
	m_nIn = 0;
	m_nOut = nOut;
	m_modelName = "RNN";
	m_nHid = nHid;
	m_Bout = nullptr;
	m_Bhh = nullptr;
	m_act = SIGMOID;
	m_Whh = nullptr;
	m_Wout = nullptr;
	m_Win = nullptr;
	m_tau = 7;
}

void Rnn::AddStep(const std::vector<int>& s, size_t end, Layer*& lastHiddenLayer, NeuralNet& rnn, bool training) {
	if (end >= s.size())
		throw std::exception("Out of bound step in RNN!");
	
	// link to the last time step
	rnn.DeleteOutputs();
	
	Layer* hl = lastHiddenLayer;
	std::string id = std::to_string(end);

	// set current input
	SparseInputLayer* x_next = rnn.AddLayer(new SparseInputLayer(m_nIn, "x_" + id));
	std::vector<int> ids(1); ids[0] = s[end-1];
	std::vector<double> values(1); values[0] = 1;
	x_next->SetData(m_nIn, ids, values);

	//
	HiddenLayer * h_next = rnn.AddLayer(new HiddenLayer(m_nHid, m_act, "h_" + id));
	if (hl != 0) rnn.Connect(hl, h_next, new DynamicWeight(m_Whh));
	rnn.Connect(x_next, h_next, new DynamicWeight(m_Win));
	rnn.Connect(rnn.AddLayer(new BiasLayer("bhh_"+id)), h_next, new DynamicWeight(m_Bhh));

	LossLayer* y_t;
	if (!training) {
		CrossEntropyLossLayer* yCe_t = rnn.AddLayer(new CrossEntropyLossLayer(m_nOut, "pred_" + id));
		yCe_t->SetOutput(1, s[end]);
		y_t = yCe_t;
	} else {
		std::vector<int> samples = m_generator->SampleNoise(25, s[end]);
		std::vector<double> probs; probs.reserve(25);
		std::for_each(samples.begin(), samples.end(), 
			[&](int x) { probs.push_back(m_generator->GetLogProb(x)); }); 

		double target_prob = -m_generator->GetLogProb(s[end]);
		samples.push_back(s[end]);
		probs.push_back(target_prob);

		NCELossLayer* yNce_t = rnn.AddLayer(new NCELossLayer(m_nOut, "pred_" + id));
		yNce_t->SetOutput(samples, probs);
		y_t = yNce_t;
	}

	rnn.Connect(h_next, y_t, new DynamicWeight(m_Wout));
	rnn.Connect(rnn.AddLayer(new BiasLayer("bout_" + id)), y_t, new DynamicWeight(m_Bout));
	
	lastHiddenLayer = h_next;
}

double Rnn::Test(const Sequence& triple, bool stoch) {
	double error = 0;
	NeuralNet rnn;
	std::vector<Matrix*> outputs;
	bool networkChanged = false;

	size_t len = triple.size();
	Layer* lastHiddenLayer = nullptr;

	int sent = 0;
	for (size_t step = 1; step < len; step++) {
		AddStep(triple, step, lastHiddenLayer, rnn, stoch);
		outputs = rnn.Forward();

		// HACK to aggregate the right scores - only when we see 2 </s>
		if (triple[step] == 1) sent++;
		
		if (sent == 2)
		if (stoch) {
			error += -outputs.back()->GetVecSparse()->GetValues().back();
		} else
			error += rnn.GetLoss();
	}

	return error;
}

double Rnn::Test(const Triple& triple, bool stoch) {
	double error = 0;
	Sequence seq;
	
	std::copy(triple[0].begin(), triple[0].end(), std::back_inserter(seq));
	std::copy(triple[1].begin(), triple[1].end(), std::back_inserter(seq));
	std::copy(triple[2].begin(), triple[2].end(), std::back_inserter(seq));
	
	return Test(seq, stoch);
}

std::vector<Candidate> Rnn::GenerateNextWord(const Triple& ctx, 
	PastInformation* info, int currentInput, size_t nK)
{
	double error = 0;
	NeuralNet rnn;
	std::vector<Matrix*> outputs;
	bool networkChanged = false;
	
	Layer* lastHiddenLayer = nullptr;
	Layer* ctxLayer = nullptr;
	
	if (currentInput == m_vocabulary->sId) 
	{
		info->infos.insert(std::make_pair("last_h", new Matrix()));
		
		/// Unroll RNN for context
		for (int i = 1; i < ctx[0].size(); i++) {
			AddStep(ctx[0], i, lastHiddenLayer, rnn, false);
		}
		for (int i = 1; i < ctx[1].size(); i++) {
			AddStep(ctx[1], i, lastHiddenLayer, rnn, false);
		}
	} 
	else
	{
		lastHiddenLayer = rnn.AddLayer(new HiddenLayer(m_nHid, ID, "h_0"));
		
		// Copy in past activations
		*(lastHiddenLayer->GetActivation()) = *(info->infos["last_h"]);
		
		// And set evaluated flag
		ctxLayer->SetEvaluated(true);
		lastHiddenLayer->SetEvaluated(true);
	}
	
	Sequence past;
	past.push_back(currentInput);
	past.push_back(0);

	AddStep(past, past.size()-1, lastHiddenLayer, rnn, false);
	outputs = rnn.Forward();

	// Save back the last activation now
	*(info->infos["last_h"]) = *(lastHiddenLayer->GetActivation());	

	// Get the max
	ThreadedFixedPQ<Candidate, std::vector<Candidate>, CandidateComparator> max(nK);
	std::set<int> key;
	int w = 0;
	for (; w < outputs.back()->GetSize(); w++) {
		double currProb = outputs.back()->GetElement(w, 0);
		Candidate r(w, log(currProb));
		max.Push(r);
		key.insert(w);
	}
	
	return max.PopAll();
}

void Rnn::CleanNetwork(NeuralNet& rnn, bool cleanActivations) {
	ResetVisitor rv;
	rv.cleanError_ = true;
	rv.cleanActivation_ = cleanActivations;
	rnn.Visit(rv);
}

Rnn* Rnn::LoadFromDisk(const std::string& dirPath) {
	Rnn* a = new Rnn(true);

	Matrix* Win = new Matrix();
	Matrix* Whh = new Matrix();
	Matrix* Wout = new Matrix();
	Matrix* bout = new Matrix();
	Matrix* bhh = new Matrix();

	Vocabulary* vO = new Vocabulary(dirPath + "\\vocab");
	RandomGenerator* g = new RandomGenerator(*vO);

	bool bo_found = bout->LoadFromExternalBinary(dirPath + "\\b_out.bin");
	bool bhh_found = bhh->LoadFromExternalBinary(dirPath + "\\b_hh.bin");
	if (bo_found && bhh_found)
		a->SetExternalBias(bhh, bout);
	bhh->InPlaceTranspose();
	bout->InPlaceTranspose();

	Whh->LoadFromExternalBinary(dirPath + "\\w_hh.bin");
	Whh->InPlaceTranspose();
	std::cerr << Whh->GetRows() << " " << Whh->GetCols() << " " << Whh->Norm2() << std::endl;
	
	Win->LoadFromExternalBinary(dirPath + "\\w_in.bin");
	Win->InPlaceTranspose();
	std::cerr << Win->GetRows() << " " << Win->GetCols() << " " << Win->Norm2() << std::endl;
	
	Wout->LoadFromExternalBinary(dirPath + "\\w_out.bin");
	Wout->InPlaceTranspose();

	std::cerr << Wout->GetRows() << " " << Wout->GetCols() << " " << Wout->Norm2() << std::endl;
	
	a->SetExternalRnnWeights(Win, Whh, Wout);
	a->SetExternalBias(bhh, bout);
	a->SetVocabulary(vO);
	a->SetGenerator(g);

	// TODO: here read config file
	a->SetAct("sigmoid");

	return a;
}

double Rnn::Train(const Sequence& triple) {
	if (m_updater == nullptr || m_Win == nullptr 
		|| m_Wout == nullptr || m_Whh == nullptr) {
		throw std::exception("Rnn needs proper weights and updater to train!");
	}
	
	double error = 0;
	bool networkChanged = false;
	
	NeuralNet rnn;
	rnn.SetBptt(m_Whh, m_tau);
	
	const Sequence& sequence = triple;
	size_t len = sequence.size();
	size_t step = 0;
	
	Layer* lastHiddenLayer = nullptr;
	for (step = 1; step < len; step++) 
	{		
		CleanNetwork(rnn, networkChanged);
		networkChanged = false;

		AddStep(sequence, step, lastHiddenLayer, rnn);

		rnn.Forward();
		error += rnn.GetLoss();
		rnn.Backward();
		rnn.Visit(v);

		// BPTT-block
		if (step % 10 == 0 || step == len - 1)
		{
			// Update the parameters here!
			m_updater->UpdateParameters(v);
			networkChanged = true; // trigger to clean the rnn
		}
	}

	return error;
}