#include "RnnTrainer.hpp"

void RnnTrainer::PrintStatus(bool training) {
	clock_t now = clock();
	double entropy = (m_sharedError/m_sharedPredictions)/log(2);
	double perpl = pow(2, entropy);
	long long difftime = (now - m_startTime)/1000;

	std::cout << difftime/3600 <<":"<<(difftime % 3600)/60<<":"<<((difftime % 3600) % 60); 
	std::cout << " - Entr. / Perp. : " << std::setw(8) << entropy << " / " << perpl
		<< " P/s: " << std::setw(5) << (long long) (m_sharedPredictions/difftime);
		std::flush(std::cout);

	/** Print norms **/
	std::cout << std::setw(5);

	std::cout << " W_hh : " << std::setw(5) << Whh.Norm2();
	std::cout << " W_out : " << std::setw(5) << Wout.Norm2();
	std::cout << " W_in : " << std::setw(5) << Win.Norm2();
	std::cout << " b_out : " << std::setw(5) << bout.Norm2();
	std::cout << " b_hh : " << std::setw(5) << bhh.Norm2();
	
	std::cout << std::endl;
}

// This method should be re-implemented by the children workers!
void RnnTrainer::DoJob(AbstractTrainer<SequenceReader>& parentCaller, 
		ThreadedQueue<Batch<Sequence>>& workingQueue,
		bool isTraining) {
	long long loaded_seq = 0, seq_position = 0;
	Batch<Sequence> wi;

	IGradientUpdater* upd = new AdagradUpdater(parentCaller.GetAlpha());
	upd->mb_ = 10;

	Rnn rnn;
	rnn.SetTau(m_tau);
	rnn.SetExternalRnnWeights(&Win, &Whh, &Wout);
	rnn.SetExternalBias(&bhh, &bout);
	rnn.SetUpdater(upd);
	rnn.SetVocabulary(m_vocabulary);
	rnn.SetGenerator(m_generator);
	rnn.SetTiled(&Whh);
	rnn.SetAct(m_act);
	
	// Some regularization for the parameters!
	if (m_beta != 0) {
		upd->add_pre_filter(&Whh, new L2RegularizeFilter(m_beta));
		upd->add_pre_filter(&Wout, new L2RegularizeFilter(m_beta));
		upd->add_pre_filter(&Win, new L2RegularizeFilter(m_beta));
		upd->add_pre_filter(&bhh, new L2RegularizeFilter(m_beta));
		upd->add_pre_filter(&bout, new L2RegularizeFilter(m_beta));
	}
	
	while (1) {
		// GetElement job
		if (loaded_seq == 0) {
			// GetElement the work item
			wi = static_cast<Batch<Sequence>&&>(workingQueue.PopFront());
			loaded_seq = wi.m_size;
			seq_position = 0;

			// poison pill
			if (loaded_seq == 0) {
				delete upd;
				// Exit here
				break;
			}
		}

		Sequence& seq = wi.data[seq_position];
		int len = (int) seq.size();

		try {
			if (isTraining) parentCaller.UpdateError(rnn.Train(seq));
			else parentCaller.UpdateError(rnn.Test(seq));	
		} catch (std::exception e) {
			std::cerr<<e.what()<<std::endl;
			exit(-1);
		}
		
		parentCaller.UpdatePredictions(len);

		// Next pair
		seq_position++;
		parentCaller.UpdateBatchesDone(1);

		if (seq_position >= loaded_seq) {
			loaded_seq = 0;
			seq_position = 0;
		}
	}
}

void RnnTrainer::WriteModel(const std::string& outputPath) {
	if (outputPath != "") 
	{	
		std::string output;
		std::stringstream outDir;
		outDir << outputPath << "_rnn_" << m_epochs 
			<< "_tau" << m_tau 
			<< "_a" << m_startAlpha
			<< "_b" << m_beta
			<< "_hid" << m_nHid 
			<< "_out" << m_nOut;
		
		output = outDir.str();
		
		std::cout << "Saving model parameters to " << output << std::endl;

		std::string make_dir = "mkdir " + output;
		system(make_dir.c_str());

		m_vocabulary->WriteToFile(output + "/dict");
		
		std::ofstream modelType(output + "/modelType");
		modelType << "RNN" << std::endl;
		modelType.close();

		WriteErrors(output);
		WriteVocabulary(output);

		Win.GetDense()->WriteToExternalBinary(output + "/w_in.bin");
		Wout.GetDense()->WriteToExternalBinary(output + "/w_out.bin");
		Whh.GetDense()->WriteToExternalBinary(output + "/w_hh.bin");
		bhh.GetDense()->WriteToExternalBinary(output + "/b_hh.bin");
		bout.GetDense()->WriteToExternalBinary(output + "/b_out.bin");
	}
}