//
//  Activation.hpp
//  Nnet
//
//  Created by Alessandro Sordoni on 2014-04-23.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//

#pragma once

#include "Alg.hpp"
#include "Matrix.hpp"
#include "DenseMatrix.hpp"

typedef std::string Activation;
const std::string ID = "id";
const std::string SOFTMAX = "softmax";
const std::string TANH = "tanh";
const std::string SIGMOID = "sigmoid";
const std::string ReLU = "relu";
const std::string EXP = "exp";

class ActivationFn {
public:
    virtual void f(Matrix* p, Matrix* ap) {
        if (p->IsDense() && ap->IsDense())
            dense_f(p->GetDense(), ap->GetDense());
        else
			sparse_f(p->GetVecSparse(), ap->GetVecSparse());
	};

    virtual void df(Matrix* p, Matrix* ap) {
        if (p->IsDense() && ap->IsDense())
            dense_df(p->GetDense(), ap->GetDense());
        else
			sparse_df(p->GetVecSparse(), ap->GetVecSparse());
    };
    
    virtual void dense_f(DenseMatrix* p, DenseMatrix* ap) = 0;
    virtual void dense_df(DenseMatrix* p, DenseMatrix* ap) = 0;
    virtual void sparse_f(SparseVector* p, SparseVector* ap) {};
    virtual void sparse_df(SparseVector* p, SparseVector* ap) {};
                      
    virtual ~ActivationFn() {};
};

class IdFn : public ActivationFn {
public:
    void dense_f(DenseMatrix* p, DenseMatrix* ap) {}
	void dense_df(DenseMatrix* p, DenseMatrix* ap) {}
    void sparse_f(SparseVector* p, SparseVector* ap) {}
    ~IdFn() {};
};

class SigmoidFn : public ActivationFn {
public:
    void dense_f(DenseMatrix* p, DenseMatrix* ap) {
        __sigmoid__(p->GetSize(), p->GetDataForUpdate(), ap->GetDataForUpdate());
    }
    
    void dense_df(DenseMatrix* p, DenseMatrix* ap) {
        const double* data = p->GetDataForUpdate();
        double* adata = ap->GetDataForUpdate();
		for (int t = 0; t < p->GetSize(); t++) {
            adata[t] = adata[t]*data[t]*(1-data[t]);
        }
    }
    
    void sparse_f(SparseVector* p, SparseVector* ap) {
		NOT_IMPLEMENTED;
    }
    
    void sparse_df(SparseVector* p, SparseVector* ap) {
		NOT_IMPLEMENTED;
    }
    
    ~SigmoidFn() {}
};

class TanhFn : public ActivationFn {
public:
    void dense_f(DenseMatrix* p, DenseMatrix* ap) {
        __tanh__(p->GetSize(), p->GetDataForUpdate(), ap->GetDataForUpdate());
    }
    void dense_df(DenseMatrix* p, DenseMatrix* ap) {
        const double* data = p->GetDataForUpdate();
        double* ddata = ap->GetDataForUpdate();
        
        for (int t=0; t<ap->GetSize(); t++) {
            ddata[t] *= (1 - data[t]*data[t]);
        }
    }
    ~TanhFn() {}
};

class ReLUFn : public ActivationFn {
public:
    void dense_f(DenseMatrix* p, DenseMatrix* ap) {
		// Leave max unchanged while clipping min
        double* ddata = ap->GetDataForUpdate();
        for (int t=0; t < ap->GetSize(); t++) {
			if (ddata[t] < 0) ddata[t] = 0;
		}
    }

    void dense_df(DenseMatrix* p, DenseMatrix* ap) {
        const double* data = p->GetDataForUpdate();
        double* ddata = ap->GetDataForUpdate();
        for (int t=0; t<ap->GetSize(); t++) {
			if (data[t] == 0) ddata[t] = 0;
		}
    }

    ~ReLUFn() {}
};

class ExpFn : public ActivationFn {
public:
    void dense_f(DenseMatrix* p, DenseMatrix* ap) {
        for (int i = 0; i < p->GetSize(); i++)
            ap->GetDataForUpdate()[i] = exp(p->GetDataForUpdate()[i]);
    }
    
    // activation & error
    void dense_df(DenseMatrix* p, DenseMatrix* ap) {
        double *adata = ap->GetDataForUpdate();
		double *data = p->GetDataForUpdate();
		for (int i = 0; i < p->GetSize(); i++)
            adata[i] *= data[i];
    }
    
    void sparse_f(SparseVector* p, SparseVector* ap) {
        NOT_IMPLEMENTED;
    }
    ~ExpFn() {}
};

class SoftmaxFn : public ActivationFn {
public:
    void dense_f(DenseMatrix* p, DenseMatrix* ap) {
        __softmax__(p->GetSize(), p->GetDataForUpdate(), ap->GetDataForUpdate());
    }
    
    // activation & error
    void dense_df(DenseMatrix* p, DenseMatrix* ap) {
		NOT_IMPLEMENTED;
    }
    
    void sparse_f(SparseVector* p, SparseVector* ap) {
		NOT_IMPLEMENTED;
    }
    
    ~SoftmaxFn() {}
};
