//
//  alg.cpp
//  Nnet
//
//  Created by Alessandro Sordoni on 2014-03-25.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//

#include <stdio.h>
#include <iostream>
#include "Alg.hpp"
#include <random>
#include <time.h>

//*********************
// FILL POLICIES
double * create_tensor(int rows, int cols, void (*fill_policy)(double *, int)) {
    int m_size = rows * cols;
    double *x = (double *) malloc(m_size * sizeof(double));
    fill_policy(x, rows * cols);
    return x;
}

void no_fill_policy(double* tensor, int m_size) {
}

void max_min_policy(double* tensor, int m_size, double min, double max) {
	for (int t=0; t<m_size; t++) {
		tensor[t] += ((double)rand()/(double)RAND_MAX) * (max-min) + min;
	}
}

void one_fill_policy(double* tensor, int m_size) {
    memset(tensor, 1, m_size*sizeof(double));
}

void zero_fill_policy(double* tensor, int m_size) {
    memset(tensor, 0, m_size*sizeof(double));
}

void random_fill_policy(double * tensor, int m_size, float mean, float sigma) {
	std::default_random_engine rd(1);
	std::normal_distribution<double> distribution(mean, sigma);
	for (int t=0; t<m_size; ++t) {
        tensor[t] = distribution(rd);
    }
}
//*********************

void logadd(double& a, double b)
{
    if (a > b) a += log(1+exp(b-a)); else { a+=b; b=a-b; a=a-b; a+=log(1+exp(b-a));}
}

// Assume all dimensions are already checked out!
double vvdot(int N,  double *a,  double *b) {
    double res = 0.0;
    int i = 0;
    for (; i <= N-4; i+=4)
    {
        res += (a[i] * b[i] +
                a[i+1] * b[i+1] +
                a[i+2] * b[i+2] +
                a[i+3] * b[i+3]);
    }
    for (; i < N; i++)
    {
        res += a[i] * b[i];
    }
    return res;
}

void softmax(int N,  double *a, double *out) {
    double part = 0;
    double max = 0;

	for (int i = 0; i < N; i++) {
		if (i == 0 || max < a[i])
			max = a[i];
    }

	for (int i = 0; i < N; i++) {
        out[i] = exp(a[i] - max);
        part += out[i];
    }

#ifdef CBLAS
    __dscal__(N, out, 1.0/part, out);
#else
    for (int i = 0; i < N; i++) out[i] /= part;
#endif
}

void tacc(int N,  double f,  double *a, double *b) {
    int i = 0;
    for (; i <= N-4; i+=4)
    {
        b[i] = f*a[i]+b[i];
        b[i+1] = f*a[i+1]+b[i+1];
        b[i+2] = f*a[i+2]+b[i+2];
        b[i+3] = f*a[i+3]+b[i+3];
    }
    for (; i < N; i++)
    {
        b[i] = f*a[i]+b[i];
    }
}

void sigmoid(int N,  double *a, double *output) {
    for (int i=0; i<N; i++)
        output[i] = 1.0/(double)(1 + exp(-a[i]));
}

void gendot(int s1, int m, int s2,  double *W,  double *U, double *output, bool Wt, bool Ut) {
    int j, k, i;
	// TODO: Do the transpose case here
    for (j = 0; j < s2; j++)
	{
        for (k = 0; k < m; k++)
			for (i = 0; i < s1; i++)
				output[i*s2 + j] += W[i*m + k]*U[k*s2 + j];
	}
}

void sprod(int N,  double *a, double s, double *output) {
    int i;
    for (i = 0; i < N; i++) {
        double o = a[i] * s;
        output[i] = o;
    }
}

void truncate(int N,  double *a, double min, double max, double *output) {
    int i;
    for (i = 0; i < N; i++) {
        double o = a[i];
        if (o < min) output[i] = min;
        else if (o > max) output[i] = max;
    }
}

void ArgMax(int N,  double *v, int& max) {
    max = 0;
    double maximum = v[0];
    
    for (int i = 1; i < N; i++)
        if (v[i] > maximum) {
            maximum = v[i];
            max = i;
        }
}

void onehot(int N, int elt, double* o) {
    memset((double *) o, 0, N * sizeof(double));
    o[elt] = 1.0;
}


void mvdot(int s1, int s2,  double* M,  double* x, double* output, bool Wt) {
    int i;
	if (!Wt) {
		for (i = 0; i < s1; i++) {
			output[i] += vvdot(s2, M+i*s2, x);
		} 
	} else {
		for (int j = 0; j < s1; j++) {
			for (i = 0; i < s2; i++) {
				output[j] += M[i*s1 + j] * x[i];
			}
		}
	}
}

/***************** WRAPPERS ****************/
void __sigmoid__(int N,  double* elt, double *output) {
    sigmoid(N, elt, output);
}

void __onehot__(int N, int elt, double *o) {
    onehot(N, elt, o);
}

void __dadd__(int N,  double *a,  double s, double *output) {
    // Do not rely on CBLAS here
    int i = 0;
    for (; i <= N-4; i+=4)
    {
        output[i] = a[i]+s;
        output[i+1] = a[i+1]+s;
        output[i+2] = a[i+2]+s;
        output[i+3] = a[i+3]+s;
    }
    for (; i < N; i++)
    {
        output[i] = a[i] + s;
    }
}

// __cacc__ - accumulate on columns
// N - row number
// col_id - column id
// a - original matrix
// f - a scalar
// x - the column to be added
void __cacc__(int N, int col_id, double *a,  double f, double *x) {
    // Do not rely on CBLAS here
    int i = 0;
    for (; i <= N-8; i+=8)
    {
        a[i*N + col_id] += x[i]*f;
        a[(i+1)*N + col_id] = x[i+1]*f;
        a[(i+2)*N + col_id] = x[i+2]*f;
		a[(i+3)*N + col_id] = x[i+3]*f;
		a[(i+4)*N + col_id] = x[i+4]*f;
		a[(i+5)*N + col_id] = x[i+5]*f;
		a[(i+6)*N + col_id] = x[i+6]*f;
		a[(i+7)*N + col_id] = x[i+7]*f;
    }
    for (; i < N; i++)
    {
        a[i*N + col_id] += x[i]*f;
    }
}

void __dcopy__(int N, double *a, double *output) {
#ifdef CBLAS
	cblas_dcopy(N, a, 1, output, 1);
#else
    memcpy(output, a, N*sizeof(double));
#endif
}

void __dscal__(int N, double *a, double s, double *output) {
#ifdef CBLAS
    // Copy vector into output
    if (a != output)
        __dcopy__(N, a, output);
    // Scale vector
	cblas_dscal(N, s, output, 1);	
    //dscal(N, s, output, 1);
#else
    sprod(N, a, s, output);
#endif
}

// b = f*a + b
void __tacc__(int N, double f, double *a, double *b) {
#ifdef CBLAS
    cblas_daxpy(N, f, a, 1, b, 1);
#else
    tacc(N, f, a, b);
#endif
}

// B = A'
void __mt__(double* A,  int r,  int c, double* B)
{
    for(int n = 0; n<r*c; n++) {
        int i = n/r;
        int j = n%c;
        B[n] = A[c*j + i];
    }
}

void __gendot__(int w_rows, int w_cols, int u_cols, double *W, double *U, double *output, bool Wt, bool Ut) {
#ifdef CBLAS
    int M = w_rows;         // Output row: Fix for W^T U , if W is a vector
    int N = u_cols;         // Output columns
    int K = w_cols;         // (w_rows==1) ? u_cols : w_rows;     // Output rows
    
    //std::cerr << K << " " << M << " " << N << std::endl;
    int LDA = (Wt) ? w_rows : w_cols;                 // Leading dimension (rowmaj) of W (swap if transposed)
    int LDB = (Ut) ? w_cols : u_cols;                 // Leading dimension (rowmaj) of U (swap if transposed)
    int LDC = u_cols;

	// Conversion to column-major
	// Wt = !Wt;
	// Ut = !Ut;
	// cblas_dgemm(CblasColMajor, (Wt) ? CblasTrans : CblasNoTrans,
    //      (Ut) ? CblasTrans : CblasNoTrans,
    //      N, M, K, 1.0,
    //      U, LDB,
    //      W, LDA,
    //      1, output, LDC);
	// RowMajor
	cblas_dgemm(CblasRowMajor, (Wt) ? CblasTrans : CblasNoTrans,
          (Ut) ? CblasTrans : CblasNoTrans,
          M, N, K, 1.0,
          W, LDA,
          U, LDB,
          1, output, LDC);
#else
    gendot(w_rows, w_cols, u_cols, W, U, output, Wt, Ut);
#endif
}

/**
 ** Matrix x vector dot; watch out this will accumulate on the output!!
 **/
void __mvdot__(int s1, int s2, double *W, double *x, double *output, bool Wt) {
#ifdef CBLAS    
	int M = (Wt) ? s2 : s1;         // Output row: Fix for W^T U , if W is a vector
    int N = (Wt) ? s1 : s2;         // Output columns
    int LDA = (Wt) ? s1 : s2;
	
	// Conversion to ColumnMajor
	// Wt = !Wt;
	// ColMajor: cblas_dgemv(CblasColMajor, (Wt) ? CblasTrans : CblasNoTrans, N, M, 1.0, W, LDA, x, 1, 1, output, 1);
	cblas_dgemv(CblasRowMajor, (Wt) ? CblasTrans : CblasNoTrans, M, N, 1.0, W, LDA, x, 1, 1, output, 1);
#else
    mvdot(s1, s2, W, x, output, Wt);
#endif
}

void __tanh__(int N,  double *a, double *output) {
    for (int i=0; i<N; i++) {
        output[i] = -1.0f + 2.0f / (1.0f + exp(-2.0f * a[i]));
    }
}

/**
 ** Performs the sparse multiplication of M rows of the matrix
 ** and accumulates the output of m_size s2.
 **/
void __sparsemult__( int* idx,  double* values, int len, int rows, int cols,  double *M, double *output) {
    // create vector of ones
    for (int i = 0; i < len; i++) {
        int r = idx[i];
        for (int k = 0; k < rows; k++) {
            output[k] += *(M + k*cols + r) * values[i];
        }
    }
}

double __norm2__(int N, double *X) {
    double norm = 0.;
#ifdef CBLAS
    norm = cblas_dnrm2(N, X, 1);
#else
    norm = 0;
    for (int i = 0; i < N; i++) {
        norm += pow(X[i], 2);
    }
    norm = sqrt(norm);
#endif
    return norm;
}

void __dset__(int N, int alpha, double* tensor) {
    memset(tensor, alpha, N*sizeof(double));
}

/**
 ** Hadamard (element-wise) product
 ** alpha * diag(x) * y + beta * output
 **/
void __had__(int N, double *x, double *y, double *output) {
#ifdef CBLAS
    //TODO: Something is not working here ; WATCH OUT
    int k = 0; // Just the diagonal; 0 super-diagonal bands
    double alpha = 1.0;
    int lda = 1;
    int incx = 1;
    double beta = 0;
    int incy = 1;
    
    cblas_dsbmv(CblasRowMajor, CblasUpper, N, k, alpha, x, lda, y, incx, beta, output, incy);
#else
    int i = 0;
    for (; i <= N-4; i+=4)
    {
        output[i] = x[i]*y[i];
        output[i+1] = x[i+1]*y[i+1];
        output[i+2] = x[i+2]*y[i+2];
        output[i+3] = x[i+3]*y[i+3];
    }
    for (; i < N; i++)
    {
        output[i] = x[i]*y[i];
    }
#endif
}

/**
 ** Outer dot
 ** output is of the m_size of a matrix s1 * s2
 **/
void __vvoutdot__(int s1, int s2, double *x, double *y, double *output) {
#ifdef CBLAS
    /*  row_order lenX lenY alpha x incX y, incY, output, LDA */
	cblas_dger(CblasRowMajor, s1, s2, 1.0, x, 1, y, 1, output, s2);
#else
    __gendot__(s1, 1, s2, x, y, output);
#endif
}

void __argmax__(int N, double* v, int& max) {
#ifdef CBLAS
    max = (int) cblas_idamax(N, v, 1);
#else
    ArgMax(N, v, max);
#endif
}

void __softmax__(int N,  double *a, double *out) {
    softmax(N, a, out);
}

void __truncate__(int N,  double *a, double min, double max, double *output) {
    truncate(N, a, min, max, output);
}

double __d__(int N,  double* a,  double* b) {
    double d = 0;
    for (int i = 0; i < N; i++)
        d += pow(a[i] - b[i], 2);
    return sqrt(d);
}

double __vvdot__(int N, double *a, double *b) {
#ifdef CBLAS
    return cblas_ddot(N, a, 1, b, 1);
#else
    return vvdot(N, a, b);
#endif
}

void __svd__(int m, int n, double *M, double*& U, double*& s, double*& Vt, int &num_s) {
#ifdef CBLAS
    // num of singular values
    num_s = m < n ? m : n;
    
    int nrows_u = m;
    int ncols_u = num_s;
    int nrows_vt = num_s;
    
    int lda_t = MAX(1, m);
    int ldu_t = MAX(1, nrows_u);
    int ldvt_t = MAX(1, nrows_vt);
    
    double* a_t = NULL;
    double* u_t = NULL;
    double* vt_t = NULL;
    
    s = create_tensor(num_s, 1, &no_fill_policy);
    // Auxiliary matrices
    a_t = (double*)malloc( sizeof(double) * lda_t * MAX(1,n) );
    u_t = (double*)malloc( sizeof(double) * ldu_t * MAX(1,ncols_u) );
    vt_t = (double*)malloc( sizeof(double) * ldvt_t * MAX(1,n) );
    
    // Transpose input matrix because of row-major format
    __mt__(M, m, n, a_t);
    
    double* work;
    int info = 0;
    int lwork = 2 * MIN( 3*MAX(m,n) + MAX(m,n) , 5*MIN(m,n) );
    work = (double*)malloc(lwork * sizeof(double));
	
	dgesvd_("all", "all", &m, &n, a_t, &lda_t, s, u_t, &ldu_t, vt_t, &ldvt_t, work, &lwork, &info);
    /* Check for convergence */
    if( info > 0 ) {
        printf( "The algorithm computing SVD failed to converge.\n" );
        exit(-1);
    }
    
    U = create_tensor(nrows_u, ncols_u, &no_fill_policy);
    Vt = create_tensor(nrows_vt, n, &no_fill_policy);
    __mt__(u_t, ncols_u, nrows_u, U);
    __mt__(vt_t, nrows_vt, n, Vt);
    
    // Cleanup workspace:
    delete[] work;
    // delete[] iwork;
    delete[] a_t;
    delete[] u_t;
    delete[] vt_t;
#else
    throw std::exception("Operation not supported! (__svd__)");
#endif
}

/**
 ** Project onto the Stiefeld manifold by SVD
 **/
void orthogonal_projection( int r, int c, double *W ) {
#ifdef CBLAS
    double *U, *s, *Vt;
    int num_s = 0;
    __svd__(r, c, W, U, s, Vt, num_s);
	__dcopy__(r*c, U, W);
	__dscal__(r*c, W, 0.1, W);
	delete U;
	delete s;
    delete Vt;
#endif
}
//*********************

void normalize_rows( int r, int c, double *W ) {
    for (int i = 0; i < r; i++) {
        double norm = __norm2__(c, W + i*c);
        for (int j = 0; j < c; j++)
            W[i*c + j] /= norm;
    }
}

void normalize_cols( int r, int c, double *W ) {
	for (int col = 0; col < c; col++) {
        double cNorm = 0.;
		for (int row = 0; row < r; row++)
			cNorm += W[row*c + col]*W[row*c + col];
        
		for (int row = 0; row < r; row++)
            W[row*c + col] /= sqrt(cNorm);
    }
}