#pragma once

#include "Commons.hpp"
#include "Alg.hpp"
#include <vector>
#include <sstream>

// Columns of this matrix are sparse
// This is good if we have a gradient
// that updates each *row* of the parameter (i.e. 
// a sparse output layer - NCE!)
class ColSparseMatrix {
private:
	int m_allocated;
	int m_numRows;
	int m_numCols;
	int m_size;
	std::string m_name;

public:
	std::vector<int> rowPosToIdx;
	std::vector<int> rowIdxToPos;
	std::vector<double> rowVal;
	int m_nnz;

	void ResetValues() {
		// Do not deallocate memory!
		std::fill(rowIdxToPos.begin(), rowIdxToPos.end(), -1);
		std::fill(rowPosToIdx.begin(), rowPosToIdx.end(), 0);
		std::fill(rowVal.begin(), rowVal.end(), 0);
		m_nnz = 0;
	}

	// Swap rows and columns!
	ColSparseMatrix(int rows, int cols) {
		m_numRows = rows; 
		m_numCols = cols; 
		m_size = rows*cols; 
		m_name = "@@@";
		m_allocated = 0;
		m_nnz = 0;
		// resize the row_id
		rowIdxToPos.resize(m_numRows, -1);
		_ensure_memory();
	}

	ColSparseMatrix(int rows, int cols, std::string name)  {
		m_allocated = 0;
		m_numRows = rows; 
		m_numCols = cols; 
		m_size = rows*cols; 
		m_name = name;
		m_nnz = 0;
		// resize the row_id
		rowIdxToPos.resize(m_numRows, -1);
		_ensure_memory();
	}

	void _ensure_memory() {
		// Pre-allocate some space
		m_allocated += 50;
		rowVal.resize(m_allocated*m_numCols, 0);
		rowPosToIdx.resize(m_allocated);
	}
	virtual ~ColSparseMatrix() { ResetValues(); }
	int nnz() { return m_nnz; }

	double& GetElementForUpdate(int i, int j) {
		if (i >= m_numRows || j >= m_numCols) 
			throw std::exception("ColumnSparseMatrix_get: Element out of range");

		int row_pos = rowIdxToPos[i];
		if (row_pos == -1) { 
			double* rowVal = GetRow(i);
			return rowVal[j];	
		}

		return rowVal[IDX2R(row_pos, j, m_numCols)];
	}

	double GetElement(int i, int j) const {
		if (i >= m_numRows || j >= m_numCols) 
			throw std::exception("ColumnSparseMatrix_get: Element out of range");

		int row_pos = rowIdxToPos[i];
		if (row_pos == -1)
			return 0;
		return rowVal[IDX2R(row_pos, j, m_numCols)];
	}	

	void InPlaceScale(double f) {
		__dscal__(m_nnz*m_numCols, &rowVal[0], f, &rowVal[0]);
	}

	/** Attention!! **/
	void InPlaceAdd(double f) {
		throw new std::exception("ColSparseMatrix_ewise_add: Are you sure you want to do this?");
	}

	void InPlacePow(double n) {
		for (int i = 0; i < m_nnz*m_numCols; i++)
			rowVal[i] = pow(rowVal[i], n);
	}

	void InPlaceClip(double threshold) {
		double n = Norm2();
		if (n > threshold)
			InPlaceScale(threshold/n);
	}

	void InPlaceClip(double min, double max) {
		for (int i = 0; i < m_nnz*m_numCols; i++)
		{
			double& rowval = rowVal[i];
			if (rowval < min) rowval = min;
			else if (rowval > max) rowval = max;	
		}
	}

	double* GetRow(int c) {
		if (c >= m_numRows) 
			throw new std::exception(("ColSparseMatrix_row: Row out of range! " 
			+ std::to_string(m_numRows) + " requested is " + std::to_string(c)).c_str());

		// Get the pos of asked column
		int row_pos = rowIdxToPos[c];
		if (row_pos == -1) {
			// If it is not found
			if (m_nnz >= m_allocated) {
				// Ensure allocated m_size if 
				// we add another row
				_ensure_memory();
			}

			rowPosToIdx[m_nnz] = c;
			rowIdxToPos[c] = m_nnz;
			row_pos = m_nnz;
			m_nnz++;
		}

		return &rowVal[row_pos*m_numCols];

	}

	int ArgMax() {
		throw std::exception("ColSparseMatrix_argmax: Argmax not implemented");
	}

	void SetElement(int i, int j, double v) {
		if (i >= m_numRows || j >= m_numCols)
			throw std::exception("ColSparseMatrix_set: Element out of range");
	}

	double Norm2() {
		return __norm2__(m_nnz*m_numCols, &rowVal[0]);
	}

	int GetRows() { return m_numRows; }
	int GetSize() { return m_size; }
	int GetCols() { return m_numCols; }
	std::string GetName() { return m_name; }
	void SetName(const std::string& name) { m_name = name; }

	std::string ToString() {
		std::stringstream ss;
		ss << "ColSparseMatrix(" << GetRows() << "," << GetCols() << ") - " << GetName() << "\n";
		for (int i = 0; i < m_nnz; i++) {
			ss << rowPosToIdx[i] << " : ";
			for (int k = 0; k < m_numCols; k++) {
				ss << rowVal[IDX2R(i, k, m_numCols)] << " ";
			}
			ss << "\n";
		}
		return ss.str();
	}
};