//
//  GradientUpdater.hpp
//  Nnet
//
//  Created by Alessandro Sordoni on 2014-04-21.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//

#ifndef Nnet_GradientUpdater_hpp
#define Nnet_GradientUpdater_hpp

#include "GradientVisitor.hpp"
#include "Matrix.hpp"
#include <mutex>

/**
 **
 **/
class IGradientFilter {
public:
    virtual void apply(Matrix* matrix, Matrix* gradient) = 0;
    virtual ~IGradientFilter() {};
};

class L2RegularizeFilter : public IGradientFilter {
public:
    void apply(Matrix* matrix, Matrix* gradient) {
        updates_++;
		// Regularize each K updates
		if (updates_ % delay_ == 0) {	
			ParallelAddDD(matrix->GetDense(), matrix->GetDense(), -beta_);
			// Matrix::Add(matrix, matrix, -beta_);
		}
	}
    
    L2RegularizeFilter() {
		updates_ = 0;
		delay_ = 25;
		beta_ = 0.0000001f;
    }
	
	L2RegularizeFilter(double beta, int delay) {
		updates_ = 0;
		delay_ = delay;
		beta_ = beta;
    }

    L2RegularizeFilter(double beta) {
        beta_ = beta;
		delay_ = 25;
		updates_ = 0;
    }
    
private:
	long long delay_;
	long long updates_;
	double beta_;
};

class ClipGradientFilter : public IGradientFilter {
public:
    void apply(Matrix* matrix, Matrix* gradient) {
        // Cast to dense matrix
		double n = gradient->Norm2();
		if (n > norm_threshold_)
			gradient->InPlaceScale(norm_threshold_/n);
	}

    ClipGradientFilter(double threshold) {
        norm_threshold_ = threshold;
    }

private:
    double norm_threshold_;
};

class ScaleFilter : public IGradientFilter {
public:
    void apply(Matrix* matrix, Matrix* gradient) {
		// Scale the gradients
		gradient->InPlaceScale(scale_);
	}
	
    ScaleFilter() : scale_(1.0) {};
	ScaleFilter(double scale) : scale_(scale_) {};
	double scale_;
};

class HardClipGradientFilter : public IGradientFilter {
public:
    void apply(Matrix* matrix, Matrix* gradient) {
		gradient->InPlaceClip(-threshold_, threshold_);		
	}

    HardClipGradientFilter(double threshold) {
        threshold_ = threshold;
    }

private:
    double threshold_;
};

class PrintFilter : public IGradientFilter {
public:
    void apply(Matrix* matrix, Matrix* gradient) {
        std::cerr << matrix->ToString() << std::endl;
        std::cerr << gradient->ToString() << std::endl;
    }
    
    PrintFilter() {
    }
    
private:
};

class IHyperScheduler {
private:
    int update_rate_;
    
public:
    virtual void apply() = 0;

    IHyperScheduler() {
        update_rate_ = 50000;
    }
    IHyperScheduler(int update_rate) : update_rate_(update_rate) {}
};

/**
 ** Abstract Interface permitting of updating
 ** the gradient store in the gradient visitor
 **/
class IGradientUpdater {
public:
	double alpha_;
    double mb_;

protected:
	std::mutex update_lock;
	std::unordered_map<Matrix*, std::vector<IGradientFilter*> > pre_filters_;
    std::unordered_map<Matrix*, std::vector<IGradientFilter*> > post_filters_;
    
public:
    void add_pre_filter(Matrix* key, IGradientFilter* filter) {
        auto it = pre_filters_.find(key);
		if (it != pre_filters_.end())
			it->second.push_back(filter);
		else {
			std::vector<IGradientFilter*> fs;
			fs.push_back(filter);
			pre_filters_.insert(std::make_pair(key, fs));
		}
	}
    
    void add_post_filter(Matrix* key, IGradientFilter* filter) {
		auto it = post_filters_.find(key);
		if (it != post_filters_.end())
			it->second.push_back(filter);
		else {
			std::vector<IGradientFilter*> fs;
			fs.push_back(filter);
			post_filters_.insert(std::make_pair(key, fs));
		}
    }
    
    void UpdateParameters(GradientVisitor& gv) {
        Grads::iterator git;
        Grads& gradients = gv.ComputeGradients();
        
		for (git = gradients.begin(); git != gradients.end(); git++) {
            auto it = pre_filters_.find(git->first);
            if (it != pre_filters_.end()) {
				for (IGradientFilter* f : it->second)
					f->apply(git->first, git->second);
			}
			
			update_lock.lock();
			ApplyUpdate(git->first, git->second);
			update_lock.unlock();
            
			it = post_filters_.find(git->first);
            if (it != post_filters_.end())
				for (IGradientFilter* f : it->second)
					f->apply(git->first, git->second);
        }

		gv.CleanGradients();
    }
    
    virtual void CleanUp() {}
    virtual void schedule() {}
    virtual std::string status() { return ""; }
    
    virtual ~IGradientUpdater() {
		std::set<IGradientFilter*> to_delete;
		auto it = pre_filters_.begin();
        for (; it != pre_filters_.end(); it++) {
            for (IGradientFilter* f : it->second) 
				to_delete.insert(f);
        }
        
        it = post_filters_.begin();
        for (; it != post_filters_.end(); it++) {
            for (IGradientFilter* f : it->second) 
				to_delete.insert(f);
        }

		auto td = to_delete.begin();
		for (; td != to_delete.end(); td++)
			delete *td;
    }
    
    virtual void ApplyUpdate(Matrix* param, Matrix* gradient) = 0;
};

class SGDUpdater : public IGradientUpdater {
private:
    double updates_;
    double start_alpha_;
    
public:
    SGDUpdater(double alpha) {
        alpha_ = alpha;
        start_alpha_ = alpha_;
        updates_ = 0;
    }
    
    void ApplyUpdate(Matrix* param, Matrix* gradient) {
		Matrix::Add(param, gradient, alpha_);
    }
    
    std::string status() {
        std::stringstream status;
        status << "Alpha: " << alpha_;
        return status.str();
    }
};

class AdagradUpdater : public IGradientUpdater {
private:
    Grads past_gradients_;
    
    Matrix* get_gradient(Matrix* param, Matrix* gradient) {
        auto gf = past_gradients_.find(param);
        if (gf == past_gradients_.end()) {
            Matrix* dg = new Matrix(param->GetRows(), param->GetCols(), param->GetName() + "_adagrad");
            gf = past_gradients_.insert(std::make_pair(param, dg)).first;
        }
		return gf->second;
    }
    
public:
    AdagradUpdater(double alpha) {
        alpha_ = alpha;
    }
    
	~AdagradUpdater() {
        for (Grads::iterator it = past_gradients_.begin(); it != past_gradients_.end(); it++) {
            delete it->second; it->second = NULL;
        }
        past_gradients_.clear();
    }
    
	void ApplyUpdate(Matrix* param, Matrix* gradient) {
        Matrix* gsq = get_gradient(param, gradient);
		Matrix::AdagradUpdate(param, gradient, gsq, alpha_/mb_);
    }
};
#endif