//
//  GradientVisitor.cpp
//  Nnet
//
//  Created by Alessandro Sordoni on 2014-04-21.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//

#include "GradientVisitor.hpp"

GradientVisitor::GradientVisitor() : Visitor() {
	m_visitNodes = false; // only edges (i.e. weights)
}
GradientVisitor::~GradientVisitor() {
	destroy();
}

bool GradientVisitor::__upd_tiled(Matrix* param, DenseMatrix* t_error, DenseMatrix* t_activation) {
	auto it = m_tiledUpdateRow.find(param);
	if (it == m_tiledUpdateRow.end())
		return false;
	
	// This is actually scheduled for a tiled update!
	int& row = m_tiledUpdateRow[param];
	Matrix* tiled_errors = m_tiledUpdate[param].first;
	Matrix* tiled_activations = m_tiledUpdate[param].second;
	
	int allocatedSize = tiled_errors->GetRows();
	if (row >= allocatedSize) {
		std::string message = "Tiled dimensions for " + param->GetName() + " does not match the expected size!";
		throw std::exception(message.c_str());
	}
	
#ifdef DEBUG
	std::cerr << "Tiled for matrix " << param->GetName() << " at row: " << row << std::endl;
#endif
	__dcopy__(param->GetRows(), t_error->GetDataForUpdate(), tiled_errors->GetDense()->GetRow(row));
	__dcopy__(param->GetCols(), t_activation->GetDataForUpdate(), tiled_activations->GetDense()->GetRow(row)); 
	row += 1;
	
	return true;
}

Grads& GradientVisitor::ComputeGradients()
{
	// Update all the tiled params
	std::for_each(m_tiledUpdate.begin(), m_tiledUpdate.end(), 
		[&](const TiledUpdates::value_type& x) {
			Matrix* dg = GetOrCreateGradient(x.first, MatrixType::DENSE);	
			int rows = m_tiledUpdateRow[x.first];
			if (rows == 0) return;

			// Hack! cblas_dgemm on submatrix to speed up things!
			__gendot__(dg->GetRows(), rows, dg->GetCols(), 
				x.second.first->GetDataForUpdate(), 
				x.second.second->GetDataForUpdate(),
				dg->GetDense()->GetDataForUpdate(), true, false);
	});

	for (auto it = m_tiledUpdateRow.begin(); it != m_tiledUpdateRow.end(); it++) 
		it->second = 0;
	//for (auto it = m_tiledUpdate.begin(); it != m_tiledUpdate.end(); it++) {	
	//	it->second.second->ResetValues(); 
	//	it->second.first->ResetValues(); 
	//}

	return m_gradients; 
}

void GradientVisitor::__create_tiled(Matrix* param, int expSize) { 
	if (m_tiledUpdateRow.find(param) != m_tiledUpdateRow.end())
		return;

	m_tiledUpdateRow.insert(std::make_pair(param, 0));
	std::string errName = "TiledErr_"+param->GetName();
	std::string actName = "TiledAct_"+param->GetName();
	
	Matrix* tiledErr = new Matrix(expSize, param->GetRows(), errName, MatrixType::DENSE);
	Matrix* tiledAct = new Matrix(expSize, param->GetCols(), actName, MatrixType::DENSE);
	m_tiledUpdate.insert(std::make_pair(param, std::make_pair(tiledErr, tiledAct)));
}

void GradientVisitor::CleanGradients() {
	for (Grads::iterator it = m_gradients.begin(); it != m_gradients.end(); it++) {
		it->second->ResetValues();
	}

	// Reset the row for tiled updates
	m_visitedWeights.clear();
}

void GradientVisitor::destroy() {
	for (Grads::iterator it = m_gradients.begin(); it != m_gradients.end(); it++) {
		delete it->second; it->second = NULL;
	}

	for (auto it = m_tiledUpdate.begin(); it != m_tiledUpdate.end(); it++) {	
		delete it->second.second; 
		delete it->second.first; 
	}

	m_visitedWeights.clear();
	m_gradients.clear();
}

Matrix* GradientVisitor::GetOrCreateGradient(Matrix* param, MatrixType gradientType) {
	auto gf = m_gradients.find(param);
	Matrix* g = 0;
	if (gf == m_gradients.end()) {
		g = new Matrix(param->GetRows(), param->GetCols(), "grad_" + param->GetName(), gradientType);
		m_gradients.insert(std::make_pair(param, g));
	} else
		g = gf->second;
	return g;
}

void GradientVisitor::Visit(DynamicWeight* w)
{
	// Current parameters
	Matrix* param = w->m();
	Layer* child_hl = (Layer*) w->child();
	Layer* parent_hl = (Layer*) w->parent();

	// If this child layer is not errorated just return (its error is 0).
	// this condition is useful in BPTT and avoids to recompute dot of zero
	if (!child_hl->IsErrorated()) return;

#ifdef DEBUG
	std::cerr << "Updating: " << param->GetName() << " = d" << ((Layer*)w->child())->GetName() << 
		" * " << ((Layer*)w->parent())->GetName() << "^T" << std::endl;
#endif

	if (parent_hl->GetActivation()->IsVecSparse()) 
	{  // Handle sparse input layer
		Matrix* sg = GetOrCreateGradient(param, MatrixType::ROWSPARSE);
		Matrix::Dot(child_hl->GetError(), parent_hl->GetActivation(), sg, false, true);
	} 
	else if (parent_hl->GetActivation()->IsDense()) 
	{
		// Handle dense layers (input or not)
		if (child_hl->GetActivation()->IsVecSparse()) 
		{
			Matrix* sg = GetOrCreateGradient(param, MatrixType::COLSPARSE);
			Matrix::Dot( child_hl->GetError(), parent_hl->GetActivation(), sg, false, true);
		}
		else 
		{
			// tiled can only be here for the moment
			bool tiled = __upd_tiled(param,
				child_hl->GetError()->GetDense(),
				parent_hl->GetActivation()->GetDense());
			if (tiled) return;

			Matrix* dg = GetOrCreateGradient(param, MatrixType::DENSE);
			Matrix::Dot(child_hl->GetError(), parent_hl->GetActivation(), dg, false, true);
		}
	}
	else
	{
		throw std::exception("Layer type not recognized!");
	}
}