//

#pragma once

#include <vector>
#include <set>
#include <queue>
#include <map>
#include <string>
#include <unordered_map>
#include <iostream>

#include "Visitor.hpp"

/* forward declaration */
class Link;
class Node;
typedef std::set<Link*> Links;

class Node {
private:
    std::set<Link*> m_childrenSet;
    std::set<Link*> m_parentsSet;

protected:
	std::string m_name;

public:
    Node(std::string name) : m_childrenSet(), m_parentsSet(), m_name(name) {};
    virtual ~Node() {
        m_childrenSet.clear();
        m_parentsSet.clear();
    };
    
    bool ConnectOutput(Link* l) {
        /* add an outward link */
        return m_childrenSet.insert(l).second;
    }
    
    bool ConnectInput(Link *l) {
        /* add an inward link */
        return m_parentsSet.insert(l).second;
    }

	bool EraseInwardLink(Link* l) {
		auto it = m_parentsSet.find(l);
		if (it != m_parentsSet.end()) {
			m_parentsSet.erase(it);
			return true;
		} 
		return false;
	}

	bool EraseOutwardLink(Link* l) {
		auto it = m_childrenSet.find(l);
		if (it != m_childrenSet.end()) {
			m_childrenSet.erase(it);
			return true;
		}
		return false;
	}
    
    std::set<Link*> &GetChildren() { return m_childrenSet; }
    std::set<Link*> &GetParents() { return m_parentsSet; }
	std::string GetName() const { return m_name; }
    virtual void Visit(Visitor& c) = 0;
};

class Link {
private:
    Node *m_parentNode, *m_childrenNode;

protected:
	std::string m_name;

public:
    Link():m_parentNode(0),m_childrenNode(0),m_name("") {};	
    virtual ~Link() {};                     
        
    Node* ConnectOutput(Node *n) {			
        Node *t=m_childrenNode;
        m_childrenNode=n;
        return t;
    }
    
    Node* ConnectInput(Node *n) {			
        Node *t=m_parentNode;
        m_parentNode=n;
        return t;
    }
        
    /* accessors */
    Node* parent() const {return m_parentNode;}
    Node* child() const {return m_childrenNode;}
	std::string GetName() const { return m_name; }
    virtual void Visit(Visitor& c) = 0;
};

/*
 this class holds a set of nodes and a set of links.
 the CleanUp method deletes them all
 */
class Graph {
public:
	// Nodes have a name, not links
    std::unordered_map<std::string, Node*> V;
    std::set<Link*> E;
    
    ~Graph() {}
	void CleanUp() {
        auto l = E.begin();
        auto n = V.begin();
        for (l=E.begin();l!=E.end();l++) {
            if (*l!=0) delete *l;
        }

        for (n=V.begin();n!=V.end();n++) {
            if (n->second!=0) delete n->second;
        }

        E.clear();
        V.clear();
    }

	Node* GetNodeByName(const std::string& name) {
		auto it = V.find(name);
		if (it != V.end())
			return it->second;
		return nullptr;
	}

	bool ContainsNode(Node* n) {
		return (V.find(n->GetName()) != V.end());
	}
	
	bool ContainsLink(Link* l) {
		return (E.find(l) != E.end());
	}

	void AddNode(Node* n) {
		if (V.find(n->GetName()) == V.end()) {
			V.insert(std::make_pair(n->GetName(), n));
		} else {
			throw std::exception("Node already existing in the graph!");
		}
	}

	void RemoveNode(Node* n) {
		auto it = V.find(n->GetName());
		bool found = (it != V.end());
		if (found) {
			// Remove the node here
			delete it->second;
			V.erase(it);
		} else {
			throw std::exception("This node doesn't exists in the graph!");
		}
	}

	void RemoveLink(Link* n) {
		auto it = E.find(n);
		bool found = (it != E.end());
		if (found) {
			// Remove the node here
			delete *it;
			E.erase(it);
		} else {
			throw std::exception("This link doesn't exists in the graph!");
		}
	}
};