//
//  matrix.cpp
//  Nnet
//
//
// These implement the logic of the Dot product and matrix additions
// wrt to our implemented matrix multiplications / additions.
//
//  Created by Alessandro Sordoni on 2014-04-16.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//

#include "Matrix.hpp"
#include "DenseMatrix.hpp"
#include <ppl.h>
#define ROUND_DOWN(x, s) ((x) & ~((s)-1))

// AddDS - a = a + f*b, adds a dense matrix to a sparse matrix (useful for gradient updates)
// - a the dense matrix to update 
// - b the sparse matrix (the gradient)
// - f a multiplicative factor
void Matrix::AddDS(DenseMatrix* a, Matrix* b, double f) {
	if (b->IsColSparse()) {
		// If it is a column sparse matrix (i.e. dense rows) this will
		// be a *very* fast update (due to the row-major format of the matrices)
		ColSparseMatrix* c = b->m_colSparse;

		for (int i = 0; i < c->nnz(); i++) {
			int row_idx = c->rowPosToIdx[i];
			double* c_row = c->GetRow(row_idx);
			double* a_row = a->GetRow(row_idx);
			__tacc__(a->GetCols(), f, c_row, a_row);
		}
	} else if (b->IsRowSparse()) {
		// If it is a row sparse matrix (i.e. dense columns) this will
		// be a *less* fast update (due to the row-major format of the matrices)
		// However less updates are done in this way! (because this is the input matrix)
		RowSparseMatrix* r = b->m_rowSparse;
		double* a_data = a->GetDataForUpdate();
		
		for (int i = 0; i < r->nnz(); i++) {
			int col_idx = r->colPosToIdx[i];
			double* c_col = r->GetColumn(col_idx);

			for (int row = 0; row < a->GetRows(); row++) {
				double& a_value = a_data[IDX2R(row, col_idx, a->GetCols())];
				a_value += f*c_col[row];
			}
		}
	} else {
		SparseVector* v = b->m_vecSparse;
		double* a_data = a->GetDataForUpdate();
		for (int i = 0; i < v->nnz(); i++)
		{
			a_data[v->GetIds()[i]] += f*v->GetValues()[i];
		}
	}
}

struct daxpy_kernel
{
public:
  enum { BLOCK_SIZE = 4 }; // this defines how many elements of the array are processed by a single call to block()
  daxpy_kernel(double a) : a(a) {}
  inline void operator()(double* y, const double* x) const {
    (*y) += a * (*x);
  }
  inline void block(double* y, const double* x) const {
    // load 4 data elements at a time
    __m128d X = _mm_loadu_pd(x);
    __m128d Y = _mm_loadu_pd(y);
    // do the computations
    __m128d result = _mm_add_pd(Y, _mm_mul_pd(X, _mm_set1_pd(a)));
    // store the results
    _mm_storeu_pd(y, result);
  }
protected:
  double a;
};

void ParallelAddDD(DenseMatrix* a, DenseMatrix* b, double f) {	
	double* bData = b->GetDataForUpdate();
	double* aData = a->GetDataForUpdate();
	
	#pragma omp parallel for
	for (int i = 0; i < a->GetSize(); i++)
		aData[i] += f*bData[i];
}

// AddDD - a = a + f*b, adds a dense matrix to a dense matrix
// - a the dense matrix to update 
// - b the dense matrix (the gradient)
// - f a multiplicative factor
void Matrix::AddDD(DenseMatrix* a, DenseMatrix* b, double f) {
	__tacc__(a->GetSize(), f, b->GetDataForUpdate(), a->GetDataForUpdate());
}

void Matrix::AdagradUpdate(Matrix* a, Matrix* b, Matrix *c, double f) {
	if (!a->IsDense())
		throw std::exception("Cannot update non dense matrices!");
	
	DenseMatrix* ad = a->m_dense;
	DenseMatrix* cd = c->m_dense;
	int bCols = b->GetCols();
	int bRows = b->GetRows();

	double* aData = ad->GetDataForUpdate();
	double* cData = cd->GetDataForUpdate();

	int size = a->GetSize();

	if (b->IsDense()) {
		double* bData = b->m_dense->GetDataForUpdate();
		
		int i = 0;
		
		for (; i < size; i++) {
			cData[i] += bData[i]*bData[i];
			aData[i] += f*bData[i]/(sqrt(1e-6+cData[i]));
		}
	} else if (b->IsColSparse()) {
		ColSparseMatrix* bs = b->m_colSparse;
		int nnz = bs->nnz();
		
		// #pragma omp parallel for
		for (int i = 0; i < nnz; i++) {
			int row_idx = bs->rowPosToIdx[i];
			double* cRow = cd->GetRow(row_idx);
			double* bRow = bs->GetRow(row_idx);
			double* aRow = ad->GetRow(row_idx);
			for (int i = 0; i < bCols; i++) {
				// Update history
				cRow[i] += bRow[i]*bRow[i];
				aRow[i] += f*bRow[i]/(sqrt(1e-6+cRow[i]));
			}
		}
	} else if (b->IsRowSparse()) {
		RowSparseMatrix* bs = b->m_rowSparse;
		int nnz = bs->nnz();
		
		// #pragma omp parallel for
		for (int i = 0; i < nnz; i++) {
			int col_idx = bs->colPosToIdx[i];
			double* bCol = bs->GetColumn(col_idx);
			
			for (int row = 0; row < bRows; row++) {
				double& aVal = aData[IDX2R(row, col_idx, a->GetCols())];
				double& cVal = cData[IDX2R(row, col_idx, a->GetCols())];
				// Update history
				cVal += bCol[row]*bCol[row];
				aVal += f*bCol[row]/(sqrt(1e-6+cVal));
			}
		}
	} else {
		NOT_IMPLEMENTED;
	}
}

// Store the results in a
void Matrix::Add(Matrix* a, Matrix* b, double f) {
	if (!a || !b)
		throw std::exception("Add: null not expected!");

	if (a->GetRows() != b->GetRows() || b->GetCols() != a->GetCols())
		throw std::exception(("Add: dimensions mismatch! " + a->GetName() + " / " + std::to_string(a->GetRows()) + " - " + b->GetName() + " / " + std::to_string(b->GetRows())).c_str());

	if (a->IsDense()) {
		if (b->IsDense()) 
		{
			AddDD(a->m_dense, b->m_dense, f);
		} 
		else 
		{
			AddDS(a->m_dense, b, f);  
		}
	} else {
		if (!b->IsDense())
		{
			throw std::exception("Not supported Add!");
		}
		
		SparseVector* v = a->GetVecSparse();
		double* b_data = b->GetDataForUpdate();
		for (int i = 0; i < v->nnz(); i++)
		{
			v->GetValues()[i] += f*b_data[IDX2R(v->GetIds()[i], 0, b->GetCols())];
		}
	}
}

// Dense Dense dot product
void Matrix::DotDD(DenseMatrix* a, DenseMatrix* b, Matrix* c, int a_rows, int a_cols, int b_rows, int b_cols, int c_rows, int c_cols, bool at, bool bt) {
	if (c->IsDense()) {
		DenseMatrix* cDense = c->m_dense;

		/** Outer product **/
		if (a_cols == 1 && b_rows == 1) {
			__vvoutdot__(a_rows, b_cols, a->GetDataForUpdate(), b->GetDataForUpdate(), cDense->GetDataForUpdate());
		}
		/** Inner product **/
		else if (a_rows == 1 && b_cols == 1) {
			cDense->GetDataForUpdate()[0] += __vvdot__(a_cols, a->GetDataForUpdate(), b->GetDataForUpdate());
		}
		/** Matrix vector product, M*v **/
		else if (b_cols == 1 && a_cols != 1) {
			__mvdot__(a_rows, a_cols, a->GetDataForUpdate(), b->GetDataForUpdate(), cDense->GetDataForUpdate(), at);
		} 
		/** x^T * A = (A^T * x)^T **/
		else if (a_rows == 1 && b_cols != 1) {
			// Transpose it!
			__mvdot__(a_cols, a_rows, a->GetDataForUpdate(), b->GetDataForUpdate(), cDense->GetDataForUpdate(), true);		
		} /** General product **/
		else {
			__gendot__(a_rows, a_cols, b_cols, a->GetDataForUpdate(), b->GetDataForUpdate(), cDense->GetDataForUpdate(), at, bt);
		}
	} else 
		if (c->IsVecSparse()) {
			// Typical NCE output here
			SparseVector* cSparse = c->m_vecSparse;
			if (b_cols == 1) {
				std::vector<int>& ids = cSparse->GetIds();
				std::vector<double>& values = cSparse->GetValues();

				double* a_data = a->GetDataForUpdate();
				double* b_data = b->GetDataForUpdate();

				double value = 0;
				
				#pragma omp parallel for
				for (int i = 0; i < cSparse->nnz(); i++) {
					int c_row = ids[i];
					if (at) {
						for (int k = 0; k < a->GetRows(); k++)
							value += a_data[IDX2R(k, c_row, a->GetCols())]  * b_data[k];
					} else {
						value = __vvdot__(a->GetCols(), &a_data[c_row * a->GetCols()], b_data);	
					}
					values[i] += value;
				}
			} else {
				throw std::exception("DotDD: sparse output format not supported!");
			}
		}
}

void Matrix::DotDS(DenseMatrix* a, SparseVector* b, Matrix* c, int a_rows, int a_cols, int b_rows, int b_cols, int c_rows, int c_cols, bool at, bool bt) 
{
	if (b_cols == 1) 
	{
		/**
		**  A * b_sparse = DenseVector sum of columns of A
		**/
		if (c->IsDense()) {
			DenseMatrix* cDense = c->m_dense;
			std::vector<int>& ids = b->GetIds();
			std::vector<double>& values = b->GetValues();
			double * a_data = a->GetDataForUpdate();
			double * c_data = cDense->GetDataForUpdate();
			
			#pragma omp parallel for
			for (int i = 0; i < b->nnz(); i++) {
				int col_id = ids[i];
				double col_mult = values[i];
				
				for (int row_id = 0; row_id < a_rows; row_id++) {
					double value = ((!at) ? a_data[IDX2R(row_id, col_id, a->GetCols())] 
					: a_data[IDX2R(col_id, row_id, a->GetCols())]) * col_mult;
					c_data[IDX2R(row_id, 0, 1)] += value;
				}
			}
		} else {
			throw std::exception("Result should be dense!");
		}
	} 
	else if (b_rows == 1 && a_cols == 1) 
	{
		// Sparse outer product (dense columns)
		// TODO: Clean implementation here
		if (c->IsRowSparse()) {
			RowSparseMatrix* cSparse = c->m_rowSparse;
			double * a_data = a->GetDataForUpdate();
			std::vector<int>& ids = b->GetIds();
			std::vector<double>& values = b->GetValues();

			for (int i = 0; i < b->nnz(); i++) {
				int col_id = ids[i];
				double mult = values[i];
				double* col = cSparse->GetColumn(col_id);
				__tacc__(a_rows, mult, a_data, col);
			}
		} else {
			NOT_IMPLEMENTED;
		}
	} 
	else 
	{
		throw std::exception("DotDS: Not supported!");
	}
}

void Matrix::DotSD(SparseVector* a, DenseMatrix* b, Matrix* c, int a_rows, int a_cols,  int b_rows, int b_cols, int c_rows, int c_cols, bool at, bool bt) {
	if (a_cols == 1 && b_rows == 1) {
		// We expect a *sparse* c matrix (i.e. a dense vector)
		// Outer product x * y^T (full row, sparse column)
		// Sparse outer product (dense columns)
		if (c->IsColSparse()) {
			ColSparseMatrix* cSparse = c->m_colSparse;
			double * b_data = b->GetDataForUpdate();
			std::vector<int>& ids = a->GetIds();
			std::vector<double>& values = a->GetValues();

			for (int i = 0; i < a->nnz(); i++) {
				int row_id = ids[i];
				double row_mult = values[i];
				double* row = cSparse->GetRow(row_id);
				__tacc__(c_cols, row_mult, b->GetDataForUpdate(), row);
			}
		} else {
			throw std::exception("DotSD: not supported sparse multiplication");
		}
	} else if (a_rows == 1 && b_rows != 1) {
		// Dense result!
		throw std::exception("DotSD: not supported sparse multiplication");
	} else {
		throw std::exception("DotSD: not supported sparse multiplication");
	}
}

// Dot - dot product between two matrices
// c += a * b
// at - a is transposed
// bt - b is transposed
void Matrix::Dot(Matrix* a, Matrix* b, Matrix* c, bool at, bool bt) {
	if (!a || !b || !c)
		throw std::exception("Dot: null not expected!");

	int a_rows = (at) ? a->GetCols() : a->GetRows();
	int b_rows = (bt) ? b->GetCols() : b->GetRows();
	int a_cols = (at) ? a->GetRows() : a->GetCols();
	int b_cols = (bt) ? b->GetRows() : b->GetCols();
	int c_rows = c->GetRows();
	int c_cols = c->GetCols();

	if (a_cols != b_rows)
		throw std::exception(("Dot: input dimensions mismatch!\n" +
		a->ToString() +
		b->ToString()).c_str());
	if (a_rows != 1 && (c_rows != a_rows || c_cols != b_cols)) {
		throw std::exception(("Dot: output dimensions mismatch!\n" +
			c->ToString()).c_str());
	}

	if (a->IsDense()) {
		if (b->IsDense()) {
			DotDD(a->m_dense, b->m_dense, c, a_rows, b_rows, a_cols, b_cols, c_rows, c_cols, at, bt);
		} else if (b->IsVecSparse()) {
			DotDS(a->m_dense, b->m_vecSparse, c, a_rows, b_rows, a_cols, b_cols, c_rows, c_cols, at, bt);
		}
	} else if (a->IsVecSparse()) {
		if (b->IsDense())
			DotSD(a->m_vecSparse, b->m_dense, c, a_rows, b_rows, a_cols, b_cols, c_rows, c_cols, at, bt);
	} else 
		throw std::exception("Dot: Dot format not supported!");
	// Sparse Sparse dot product is not supported.
}