//
//  matrix.hpp
//  nnet
//
// These are very 'ad-hoc' Matrix classes for our purposes.
// The interesting point is that we have an inheritance
// both for sparse classes and dense classes wrt to the base
// object Matrix (so that it can flow around in the graph).
//
//
//  Created by Alessandro Sordoni on 2014-03-25.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//


#pragma once

#include "Alg.hpp"
#include "Commons.hpp"

#include <string>
#include <algorithm>
#include <unordered_map>
#include <sstream>
#include <vector>
#include <exception>
#include <stdexcept>

#include "DenseMatrix.hpp"
#include "SparseVector.hpp"
#include "RowSparseMatrix.hpp"
#include "ColSparseMatrix.hpp"

class Matrix;
typedef enum { DENSE, COLSPARSE, ROWSPARSE, VECSPARSE } MatrixType;


class Matrix {
protected:
	MatrixType m_matrixType;
	DenseMatrix* m_dense;
	RowSparseMatrix* m_rowSparse;
	ColSparseMatrix* m_colSparse;
	SparseVector* m_vecSparse;
		
public:
	Matrix() {
		m_rowSparse = nullptr;
		m_colSparse = nullptr;
		m_vecSparse = nullptr;
		m_matrixType = MatrixType::DENSE;
		m_dense = new DenseMatrix();
	}

	Matrix(const Matrix& deepCopy) {
		m_matrixType = deepCopy.m_matrixType;
		m_rowSparse = nullptr;
		m_colSparse = nullptr;
		m_vecSparse = nullptr;
		m_dense = nullptr;
		if (m_matrixType == MatrixType::DENSE) {
			m_dense = new DenseMatrix(*deepCopy.GetDense());
		} else {
			NOT_IMPLEMENTED;
		}
	}

    Matrix(MatrixType m) {
		m_matrixType = m;
		m_dense = nullptr;
		m_rowSparse = nullptr;
		m_colSparse = nullptr;
		m_vecSparse = nullptr;

		if (m == MatrixType::DENSE) {
			m_dense = new DenseMatrix();
		} else if (m == MatrixType::VECSPARSE) {
			m_vecSparse = new SparseVector();
		} else // TODO
			NOT_IMPLEMENTED;
	}

    Matrix(int rows, int cols, MatrixType m) {
		m_matrixType = m;
		m_dense = nullptr;
		m_rowSparse = nullptr;
		m_colSparse = nullptr;
		m_vecSparse = nullptr;

		if (m == MatrixType::DENSE) {
			m_dense = new DenseMatrix(rows, cols);		
		} else if (m == MatrixType::ROWSPARSE) {
			m_rowSparse = new RowSparseMatrix(rows, cols);
		} else if (m == MatrixType::COLSPARSE) {
			m_colSparse = new ColSparseMatrix(rows, cols);
		} else if (m == MatrixType::VECSPARSE) {
			m_vecSparse = new SparseVector(rows);
		}
	};

	Matrix(int rows, int cols, std::string name) {
		m_matrixType = MatrixType::DENSE;
		m_dense = nullptr;
		m_rowSparse = nullptr;
		m_colSparse = nullptr;
		m_vecSparse = nullptr;
		m_dense = new DenseMatrix(rows, cols, name);		
	};

    Matrix(int rows, int cols, std::string name, MatrixType m) {
		m_matrixType = m;
		m_dense = nullptr;
		m_rowSparse = nullptr;
		m_colSparse = nullptr;
		m_vecSparse = nullptr;

		if (m == MatrixType::DENSE) {
			m_dense = new DenseMatrix(rows, cols, name);		
		} else if (m == MatrixType::ROWSPARSE) {
			m_rowSparse = new RowSparseMatrix(rows, cols, name);
		} else if (m == MatrixType::COLSPARSE) {
			m_colSparse = new ColSparseMatrix(rows, cols, name);
		} else if (m == MatrixType::VECSPARSE) {
			m_vecSparse = new SparseVector(rows, name);
		}
	};
    
	bool IsDense() { return m_matrixType == MatrixType::DENSE; }
	bool IsRowSparse() { return m_matrixType == MatrixType::ROWSPARSE; }
	bool IsColSparse() { return m_matrixType == MatrixType::COLSPARSE; }
	bool IsVecSparse() { return m_matrixType == MatrixType::VECSPARSE; }
		
	void CopyFrom(const Matrix& matrix) {
		m_matrixType = matrix.m_matrixType;
		if (m_rowSparse != nullptr)
		{
			delete m_rowSparse;
			m_rowSparse = nullptr;
		}
		if (m_colSparse != nullptr)
		{
			delete m_colSparse;
			m_colSparse = nullptr;
		}
		
		if (m_vecSparse != nullptr) 
		{
			delete m_vecSparse;
			m_vecSparse = nullptr;
		}

		if (m_dense != nullptr) 
		{
			delete m_dense;
			m_dense = nullptr;
		}

		if (m_matrixType == MatrixType::DENSE) {
			m_dense = new DenseMatrix(*matrix.GetDense());
		} else {
			NOT_IMPLEMENTED;
		}
	}

	double* GetDataForUpdate() {
		if (IsDense()) {
			return m_dense->GetDataForUpdate();
		} else
			throw std::exception("Cannot call the method for non-dense matrices!");
	}
	
	// element accessors
	double GetElement(int i, int j) {
		if (IsDense()) {
			return m_dense->GetElement(i, j);
		} else if (IsRowSparse()) {
			return m_rowSparse->GetElement(i, j);
		} else if (IsColSparse()) {
			return m_colSparse->GetElement(i, j);
		} 
		
		NOT_IMPLEMENTED;
	}

	double& GetElementForUpdate(int i, int j) {
		if (IsDense()) {
			return m_dense->GetElementForUpdate(i, j);
		} else if (IsRowSparse()) {
			return m_rowSparse->GetElementForUpdate(i, j);
		} else if (IsColSparse()) {
			return m_colSparse->GetElementForUpdate(i, j);
		}

		NOT_IMPLEMENTED;
	}

	// clean the matrix
	void ResetValues() {
		if (IsDense()) 
		{
			m_dense->ResetValues();
		} else if (IsRowSparse()) {
			m_rowSparse->ResetValues();
		} else if (IsColSparse()) {
			m_colSparse->ResetValues();
		} else if (IsVecSparse()) 
		{
			m_vecSparse->ResetValues();
		}	
	}

	std::string ToString() {
		if (IsDense()) return m_dense->ToString();
		else if (IsRowSparse()) return m_rowSparse->ToString();
		else if (IsColSparse()) return m_colSparse->ToString();
		else if (IsVecSparse()) return m_vecSparse->ToString();
		NOT_IMPLEMENTED;
	}

    // index of the maximum element in the matrix
    int ArgMax() {
		if (IsDense()) return m_dense->ArgMax();
		else if (IsRowSparse()) return m_rowSparse->ArgMax();
		else if (IsColSparse()) return m_colSparse->ArgMax();
		else if (IsVecSparse()) return m_vecSparse->ArgMax();
		NOT_IMPLEMENTED;
	}

    // element-wise power
    void InPlacePow(double n) {
		if (IsDense())  m_dense->InPlacePow(n);
		else if (IsRowSparse())  m_rowSparse->InPlacePow(n);
		else if (IsColSparse())  m_colSparse->InPlacePow(n);
		else if (IsVecSparse())  m_vecSparse->InPlacePow(n);
	}
	
	void InPlaceClip(double min, double max) {
		if (IsDense()) m_dense->InPlaceClip(min, max);
		else if (IsRowSparse())  m_rowSparse->InPlaceClip(min, max);
		else if (IsColSparse())  m_colSparse->InPlaceClip(min, max);
		else if (IsVecSparse()) m_vecSparse->InPlaceClip(min, max);
	}

	void InPlaceClip(double threshold) {
		if (IsDense()) m_dense->InPlaceClip(threshold);
		else if (IsRowSparse())  m_rowSparse->InPlaceClip(threshold);
		else if (IsColSparse())  m_colSparse->InPlaceClip(threshold);
		else if (IsVecSparse()) m_vecSparse->InPlaceClip(threshold);
	}

	// 2-norm
	double Norm2() {
		if (IsDense()) return m_dense->Norm2();
		else if (IsRowSparse()) return m_rowSparse->Norm2();
		else if (IsColSparse()) return m_colSparse->Norm2();
		else if (IsVecSparse()) return m_vecSparse->Norm2();
		NOT_IMPLEMENTED;
	}

	// element-wise add
	void InPlaceAdd(double f) {
		if (IsDense()) m_dense->InPlaceAdd(f);
		else if (IsRowSparse())  m_rowSparse->InPlaceAdd(f);
		else if (IsColSparse())  m_colSparse->InPlaceAdd(f);
		else if (IsVecSparse()) m_vecSparse->InPlaceAdd(f);
	}

    // element-wise product
	void InPlaceScale(double f) {
		if (IsDense()) m_dense->InPlaceScale(f);
		else if (IsRowSparse())  m_rowSparse->InPlaceScale(f);
		else if (IsColSparse())  m_colSparse->InPlaceScale(f);
		else if (IsVecSparse()) m_vecSparse->InPlaceScale(f);	
	}

	void InPlaceTranspose() {
		if (IsDense()) m_dense->InPlaceTranspose();
		else NOT_IMPLEMENTED;	
	}

	void Resize(int rows, int cols) {
		if (IsDense()) m_dense->Resize(rows, cols);
		else if (IsVecSparse()) m_vecSparse->Resize(rows);
		else NOT_IMPLEMENTED;	
	}

    std::string GetName() {
		if (IsDense()) return m_dense->GetName();
		else if (IsRowSparse()) return m_rowSparse->GetName();
		else if (IsColSparse()) return m_colSparse->GetName();
		else if (IsVecSparse()) return m_vecSparse->GetName();
		NOT_IMPLEMENTED;
	}

	void SetName(const std::string& name) {
		if (IsDense()) return m_dense->SetName(name);
		else if (IsRowSparse()) return m_rowSparse->SetName(name);
		else if (IsColSparse()) return m_colSparse->SetName(name);
		else if (IsVecSparse()) return m_vecSparse->SetName(name);
		NOT_IMPLEMENTED;
	}
    
	int GetRows() { 
		if (IsDense()) return m_dense->GetRows();
		else if (IsRowSparse()) return m_rowSparse->GetRows();
		else if (IsColSparse()) return m_colSparse->GetRows();
		else if (IsVecSparse()) return m_vecSparse->GetRows();
		NOT_IMPLEMENTED;
    }
	
	int GetCols() { 
		if (IsDense()) return m_dense->GetCols();
		else if (IsRowSparse()) return m_rowSparse->GetCols();
		else if (IsColSparse()) return m_colSparse->GetCols();
		else if (IsVecSparse()) return m_vecSparse->GetCols();
		NOT_IMPLEMENTED;
	}

    int GetSize() {
		if (IsDense()) return m_dense->GetSize();
		else if (IsRowSparse()) return m_rowSparse->GetSize();
		else if (IsColSparse()) return m_colSparse->GetSize();
		else if (IsVecSparse()) return m_vecSparse->GetSize();
		NOT_IMPLEMENTED;
	}

    virtual ~Matrix() {
		if (m_dense != nullptr) delete m_dense;
		if (m_rowSparse != nullptr) delete m_rowSparse;
		if (m_colSparse != nullptr) delete m_colSparse;
		if (m_vecSparse != nullptr) delete m_vecSparse;
	};
	
	static void Matrix::Add(Matrix*, Matrix*, double f=1.0);
	static void Matrix::AddDD(DenseMatrix*, DenseMatrix*, double f=1.0);
	static void Matrix::AddDS(DenseMatrix*, Matrix*, double f=1.0);
	static void Matrix::Dot(Matrix*, Matrix*, Matrix*, bool at=false, bool st=false);
	static void Matrix::DotDD(DenseMatrix*, DenseMatrix*, Matrix*, 
		int a_rows, int a_cols, 
		int b_rows, int b_cols,
		int c_rows, int c_cols,
		bool at=false, bool st=false);
	static void Matrix::DotSD(SparseVector*, DenseMatrix*, Matrix*, 
		int a_rows, int a_cols, 
		int b_rows, int b_cols,
		int c_rows, int c_cols,
		bool at=false, bool st=false);
	static void Matrix::DotDS(DenseMatrix*, SparseVector*, Matrix*, 
		int a_rows, int a_cols, 
		int b_rows, int b_cols,
		int c_rows, int c_cols,
		bool at=false, bool st=false);

	static void Matrix::AdagradUpdate(Matrix*, Matrix*, Matrix*,double f=1.0);

	bool LoadFromExternalBinary(const std::string& path) {
		return GetDense()->LoadFromExternalBinary(path);
	}
	
	DenseMatrix* GetDense() const { return m_dense; }
	DenseMatrix* GetDense() { return m_dense; }
	RowSparseMatrix* GetRowSparse() { return m_rowSparse; }
	ColSparseMatrix* GetColSparse() { return m_colSparse; }
	SparseVector* GetVecSparse() { return m_vecSparse; }
};

// Matrix operations
void ParallelAddDD(DenseMatrix* a, DenseMatrix* b, double f=1.0);

#ifdef EXPERIMENTAL
void Divide(Matrix*, Matrix*, double f=1.0);
void Product(Matrix*, Matrix*, double f=1.0);
#endif
