//
//  nnet.cpp
//  nnet
//
//  Created by Alessandro Sordoni on 2014-03-25.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//
#include <cassert>
#include <stdio.h>
#include "NnetLib.hpp"

/***************************/
/**      Forward pass     **/
/***************************/
Matrix* Layer::GetValue() {
	// If we already visited this node just return its value!
	if (m_evaluated) return GetActivation();

#ifdef DEBUG
	std::cerr << "Evaluating: " << GetName() << std::endl;
#endif
	
	for (Links::iterator p=GetParents().begin(); p!=GetParents().end(); p++) {
		Weight* parent_weight = static_cast<Weight*>(*p);
		Layer*	parent_layer = static_cast<Layer*>(parent_weight->parent());		
		Matrix* parent_value = parent_layer->GetValue();
		
		// Small hack to just add from bottom layers
		if (parent_weight->m() == nullptr) 
		{
			Matrix::Add(GetActivation(), parent_value);
#ifdef DEBUG
		std::cerr << GetName() << " += " << parent_layer->GetName() << std::endl;
#endif
		} 
		else 
		{
			Matrix::Dot(parent_weight->m(), parent_value, m_activation);
#ifdef DEBUG
		std::cerr << GetName() << " += " << parent_weight->GetName() << " * " << parent_layer->GetName() << std::endl;
#endif
		}
	}

	// Do the activation of the current input
	fn_->f(m_activation, m_activation);

#ifdef DEBUG
	std::cerr << "Evaluation done!" << GetName() << std::endl;
#endif

	m_evaluated = true;
	return GetActivation();
}

Matrix* ConcatLayer::GetValue() {
	// If we already visited this node just return its value!
	if (m_evaluated) return GetActivation();
		
	for (Links::iterator p=GetParents().begin(); p!=GetParents().end(); p++) {
		Weight* parent_weight = static_cast<Weight*>(*p);
		Layer*	parent_layer = static_cast<Layer*>(parent_weight->parent());		
		Matrix* parent_value = parent_layer->GetValue();
		
		auto it = m_orderedParents.find(parent_layer);
		int size = parent_layer->GetSize();
		__dcopy__(parent_layer->GetSize(),
			parent_value->GetDataForUpdate(),
			m_activation->GetDataForUpdate() + size*it->second);
		
#ifdef DEBUG
		std::cerr << "Concatenating: " << parent_layer->GetName() << std::endl;
#endif
	}

	// Do the activation of the current input
	fn_->f(GetActivation(), GetActivation());

#ifdef DEBUG
	std::cerr << "Evaluation done!" << GetName() << std::endl;
#endif

	m_evaluated = true;
	return GetActivation();
}

Matrix* ConcatLayer::ComputeError(Layer* parent) {
	if (!m_errorated) {
		/** Execute only if the error has not been yet calculated **/
#ifdef DEBUG
		std::cerr << "# Computing error on layer = " << ToString() << std::endl;
#endif
		for (Links::iterator p = GetChildren().begin(); p != GetChildren().end(); p++) {
			Weight *child_weight = static_cast<Weight*>(*p);
			Layer *child_layer = static_cast<Layer*>(child_weight->child());
			if (!child_layer->IsNeedError()) continue;

#ifdef DEBUG
			std::cerr << " --> Poking layer for error = " << child_layer->GetName() << std::endl;
#endif
			// Accumulate the current error by summing the errors
			// coming from child layers at level t+1 \sum (dh_{t+1}^T * W)
			Matrix* child_weight_matrix = child_weight->m();
			Matrix* child_error = child_layer->ComputeError(this);
			
			// Small hack to just add from bottom layers
			if (child_weight_matrix == nullptr)
				Matrix::Add(GetError(), child_error);
			else
				Matrix::Dot(child_weight_matrix, child_error, m_error, true, false);

#ifdef DEBUG
			std::cerr << m_error->GetName() << " += " << child_weight_matrix->GetName() << "^T * " << child_error->GetName() << std::endl;
#endif
			// Clip errors with the
			// R.Pascanu method (norm clipping)
			m_error->InPlaceClip(15);
		}

		// Here the derivative of the activation function is
		// applied element wise to the *error* computed above.
		// The derivative is usually computed using the *activation*
		// from this layer. This is usually the case for the TANH and the SIGMOID
		// which are 1-a_t^2 and a_t*(1-a_t).
		fn_->df(m_activation, m_error);
#ifdef DEBUG
		std::cerr << " --> Error for layer done = " << GetName() << std::endl;
#endif

		// Fill the errors for the children
		for (auto it = m_errors.begin(); it != m_errors.end(); it++) {
			int id = m_orderedParents.find(it->first)->second;
			__dcopy__(it->first->GetSize(), GetError()->GetDataForUpdate() + id*it->first->GetSize(), 
				it->second->GetDataForUpdate());
		}

		SetErrorated(true);
	}

	// Return the right error for the right caller
	Matrix* error = nullptr;
	if (parent != this) {
		auto finder = m_errors.find(parent);
		if (finder == m_errors.end()) {
			std::string errMessage = parent->GetName() + " not registered for error!";
			throw std::exception(errMessage.c_str());
		}

		error = finder->second;
	}

	return error;
}

Matrix* MaxPoolingLayer::GetValue() {
	if (IsEvaluated())
		return GetActivation();
    
    
#ifdef DEBUG
	std::cerr << "Evaluating: " << GetName() << std::endl;
#endif
	
	int id = -1;
    for (Links::iterator p=GetParents().begin(); p != GetParents().end(); p++) {
        Weight* parentWeight = static_cast<Weight*>(*p);
        Layer* parentLayer = (Layer* ) parentWeight->parent();
        Matrix* parentActivation = parentLayer->GetValue();

        // Get parent activation;
		for (int i = 0; i < m_size; i++) {
			const double newValue = parentActivation->GetElement(i, 0);
			double& currValue = m_activation->GetElementForUpdate(i, 0);
            
            if (newValue > currValue || id == -1) {
                max_ids[i] = parentLayer; // Save current layer as the maximum
                currValue = newValue;
            }
        }

		id++;
    }
	
#ifdef DEBUG
	std::cerr << "Evaluation done!" << GetName() << std::endl;
	std::cerr << GetActivation()->ToString() << std::endl;
	std::cerr << "IDs = " << std::endl;
	for (int i = 0; i < max_ids.size(); i++)
		std::cerr << max_ids[i]->GetName() << " ";
	std::cerr << std::endl;
#endif
	
	m_evaluated = true;
	return GetActivation();
}

Matrix* MaxPoolingLayer::ComputeError(Layer* parent) {
	if (!m_errorated) {
#ifdef DEBUG
		std::cerr << "# Computing error on layer = " << ToString() << std::endl;
#endif
        for (Links::iterator p = GetChildren().begin(); p != GetChildren().end(); p++) {
            Weight *child_weight = static_cast<Weight*>(*p);
            Layer *child_layer = static_cast<Layer*>(child_weight->child());
			
			Matrix* child_weight_matrix = child_weight->m();
			Matrix* child_error = child_layer->ComputeError(this);

			if (child_weight_matrix == nullptr)
				Matrix::Add(GetError(), child_error);
			else
				Matrix::Dot(child_weight_matrix, child_error, m_error, true, false);
		}
        
		SetErrorated(true);
    }
    
	if (parent == this) return nullptr;

    auto it = errors_.find(parent);
    if (it == errors_.end()) {
        // return the right error for this caller
		it = errors_.insert(std::make_pair(parent, new Matrix(m_size, 1, MatrixType::VECSPARSE))).first;
    }

	Matrix* error = it->second;
	std::vector<int> idx; idx.reserve(m_size);
	std::vector<double> vals; vals.reserve(m_size);

	for (int i = 0; i < m_size; i++) {
		// Find in the vector if this is the error flow corresponding
        // to this parent
        if (max_ids[i] == parent) {
			idx.push_back(i);
			vals.push_back(m_error->GetElement(i, 0));
        }
	}
        
	error->GetVecSparse()->SetData(m_size, idx, vals);
    
	// Return the right error
#ifdef DEBUG
	std::cerr << "Max Pooling error = " << error->ToString() << std::endl;
#endif
    return error;
}

Matrix* DenseLayer::ComputeError(Layer* parent) {
	if (!m_errorated) {
		/** Execute only if the error has not been yet calculated **/
#ifdef DEBUG
		std::cerr << "# Computing error on layer = " << ToString() << std::endl;
#endif
		for (Links::iterator p = GetChildren().begin(); p != GetChildren().end(); p++) {
			Weight *child_weight = static_cast<Weight*>(*p);
			Layer *child_layer = static_cast<Layer*>(child_weight->child());
			if (!child_layer->IsNeedError()) continue;

#ifdef DEBUG
			std::cerr << " --> Poking layer for error = " << child_layer->GetName() << std::endl;
#endif
			// Accumulate the current error by summing the errors
			// coming from child layers at level t+1 \sum (dh_{t+1}^T * W)
			Matrix* child_weight_matrix = child_weight->m();
			Matrix* child_error = child_layer->ComputeError(this);
			
			// Small hack to just add from bottom layers
			if (child_weight_matrix == nullptr) 
			{
				Matrix::Add(GetError(), child_error);
#ifdef DEBUG
			std::cerr << GetName() << " += " << child_layer->GetName() << std::endl;
#endif
			}
			else 
			{
				Matrix::Dot(child_weight_matrix, child_error, m_error, true, false);	
#ifdef DEBUG
			std::cerr << m_error->GetName() << " += " << child_weight_matrix->GetName() << "^T * " << child_error->GetName() << std::endl;
#endif
			}
			
			// Clip errors with the
			// R.Pascanu method (norm clipping)
			m_error->InPlaceClip(15);
		}

		// Here the derivative of the activation function is
		// applied element wise to the *error* computed above.
		// The derivative is usually computed using the *activation*
		// from this layer. This is usually the case for the TANH and the SIGMOID
		// which are 1-a_t^2 and a_t*(1-a_t).
		fn_->df(m_activation, m_error);
#ifdef DEBUG
		std::cerr << " --> Error for layer done = " << GetName() << std::endl;
		std::cerr << GetError()->ToString() << std::endl;
#endif
		SetErrorated(true);
	}

	return GetError();
}

Matrix* InputLayer::ComputeError(Layer* parent) {
	return GetError();
}

Matrix* NeuralNet::Forward(LossLayer* ll) 
{
	if (ll == nullptr)
		throw std::exception("Loss layer should be non-null in forward pass!");
	return ll->GetValue();
}

std::vector<Matrix*> NeuralNet::Forward() {
	std::vector<Matrix*> outputs;
	for (auto it = O.begin(); it != O.end(); it++) {
		outputs.push_back(Forward(*it));
	}
	return outputs;
}

// In the case of BPTT, we should flag the error nodes 
// before doing the backward pass in order to tell the backward
// algorithm not to ask for error to children nodes outside of 
// the bptt scope
void NeuralNet::FlagErrorNodes(LossLayer* ll) {
	// Start the BFS search by walking only 
	// through parents
	std::set<Layer*> nodeVisited;
	std::queue<Layer*> bfsQueue;

	bfsQueue.push(ll);
	nodeVisited.insert(ll);

	// the current loss layer trivially has a gradient
	ll->SetNeedError(true);

	int bpttAdvance = 0;
	while (!bfsQueue.empty()) {
		Layer* current = bfsQueue.front();
		bfsQueue.pop();
	
		// Add into queue
		for (Links::iterator p=current->GetParents().begin(); p!=current->GetParents().end(); p++) 
		{			
			Weight* pw = static_cast<Weight*>(*p);
			Layer* pl = static_cast<Layer*>(pw->parent());
			
			if (pw->m() == m_barrierWeight)
			{
				pl->SetNeedError(true);
				/// cut off this branch
				break;
			}

			// Flag this parent node if this is 
			// out of the BPTT limit and if it is on the path
			// of the BPTT recurrent weight
			if (pw->m() == bpttWeight) 
			{
				bpttAdvance++;
				if (bpttAdvance > bpttStop)
					pl->SetNeedError(false);
				else
					pl->SetNeedError(true);
			} else 
			// Otherwise we schedule this layer 
			// to have a gradient.
			{
				pl->SetNeedError(true);
			}

#ifdef DEBUG
			std::cerr << pl->GetName() << " - Flagged " << pl->IsNeedError() << std::endl;
#endif DEBUG

			// Continue searching and flagging backwards
			if (nodeVisited.find(pl) == nodeVisited.end() && pl->IsNeedError()) {
				nodeVisited.insert(pl);
				bfsQueue.push(pl);
			}
		}
	}
}

void NeuralNet::Backward(LossLayer* ll) {
#ifdef DEBUG
	std::cerr << "----- BACKWARD PASS -----" << std::endl;
#endif DEBUG
	if (O.size() == 0) {
		throw std::exception("Should add at least one output node!");
	}
	
	// Reset the error flag first
	for (auto it = V.begin(); it != V.end(); it++)
		((Layer*) it->second)->SetNeedError(false);

	std::set<Layer*> nodeVisited;
	std::queue<Layer*> bfsQueue;

	// BFS for error propagation
	bfsQueue.push(ll);
	nodeVisited.insert(ll);
	FlagErrorNodes(ll);
	
	while (!bfsQueue.empty()) {
		Layer* current = bfsQueue.front();
		bfsQueue.pop();

		// Error propagation here
#ifdef DEBUG
		std::cerr << "START propagating from " << current->ToString() << std::endl;
#endif DEBUG
		current->ComputeError(current);

		// Add into queue
		for (Links::iterator p=current->GetParents().begin(); p!=current->GetParents().end(); p++) 
		{
			Weight* pw = static_cast<Weight*>(*p);
			Layer* pl = static_cast<Layer*>(pw->parent());
			
			// Insert if we have not seen this node and 
			// if this effectively needs an error
			// With BPTT this will have the effect of chuncking some of the recurrent
			// nodes too far away in time while keeping the others scheduled for error.
			if (nodeVisited.find(pl) == nodeVisited.end() && pl->IsNeedError()) {
				nodeVisited.insert(pl);
				bfsQueue.push(pl);
			}
		}
	}
}

void NeuralNet::Backward() {
#ifdef DEBUG
	std::cerr << "----- BACKWARD PASS -----" << std::endl;
#endif DEBUG
	if (O.size() == 0) {
		throw std::exception("Should add at least one output node!");
	}
	
	// Reset the error flag first
	for (auto it = V.begin(); it != V.end(); it++)
		((Layer*) it->second)->SetNeedError(false);

	std::set<Layer*> nodeVisited;
	std::queue<Layer*> bfsQueue;

	// BFS for error propagation
	for (LossLayer* p : O) {
		bfsQueue.push(p);
		nodeVisited.insert(p);
		FlagErrorNodes(p);
	}

	while (!bfsQueue.empty()) {
		Layer* current = bfsQueue.front();
		bfsQueue.pop();

		// Error propagation here
#ifdef DEBUG
		std::cerr << "START propagating from " << current->ToString() << std::endl;
#endif DEBUG
		current->ComputeError(current);

		// Add into queue
		for (Links::iterator p=current->GetParents().begin(); p!=current->GetParents().end(); p++) 
		{
			Weight* pw = static_cast<Weight*>(*p);
			Layer* pl = static_cast<Layer*>(pw->parent());
			
			// Insert if we have not seen this node and 
			// if this effectively needs an error
			// With BPTT this will have the effect of chuncking some of the recurrent
			// nodes too far away in time while keeping the others scheduled for error.
			if (nodeVisited.find(pl) == nodeVisited.end() && pl->IsNeedError()) {
				nodeVisited.insert(pl);
				bfsQueue.push(pl);
			}
		}
	}
}

// Returns true if the node is an internal node inserted
// by the optimizer
bool NeuralNet::IsInternal(Layer* n) {
	return m_internals.find(n) != m_internals.end();
}

// Graph optimization: group the 
// links with the same weight to store
// intermediary multiplication results
int NeuralNet::Optimize() {	
	for (auto it = V.begin(); it != V.end(); it++) {
		// Start clustering only from high-arity objects
		if (it->second->GetChildren().size() > 4) {
			Layer* alreadyInternal = nullptr;
			
			// Skip internals
			Layer* n = (Layer*) it->second;
			if (m_internals.find(n) != m_internals.end())
				continue;
				
			// Try to cluster outbound connections by weight!
			std::map<Matrix*, std::vector<Weight*>> cluster;
			for (auto lit = n->GetChildren().begin(); lit != n->GetChildren().end(); lit++) {
				Weight* w = (Weight*) *lit;				
				Layer* child = (Layer*) w->child();
				auto fit = cluster.find(w->m());

				if (w->m() == nullptr)
					continue;

				if (fit != cluster.end()) {
					fit->second.push_back(w);
				} else {
					std::vector<Weight*> toLink; toLink.push_back(w);
					cluster.insert(std::make_pair(w->m(), toLink));
				}
			}

			// Found clusters !!
			for (auto c = cluster.begin(); c != cluster.end(); c++) {
				if (c->second.size() > 2) {
					std::string key = "ILink_" + c->first->GetName() + "_" + n->GetName();
					alreadyInternal = GetInternalForLayerParam(key);
					
					if (alreadyInternal == nullptr) 
					{
						// Create and register an internal node
						Layer* newInternal = new HiddenLayer(c->first->GetRows(), key);
						AddLayer(newInternal);
						
						alreadyInternal = newInternal;
						
						// Register this node for this weight
						m_internals.insert(alreadyInternal);
						m_internMap.insert(std::make_pair(key, alreadyInternal));
						// end

						Connect((Layer*) n, newInternal, new DynamicWeight(c->first));
					}
					
					for (int i = 0; i < c->second.size(); i++) {
						Weight* linkTo = c->second[i];
						Layer* layerTo = (Layer*) linkTo->child();

						// Skip the internal
						if (layerTo == alreadyInternal) continue;
						
						// Disconnect the original link
						n->EraseOutwardLink(linkTo);
						layerTo->EraseInwardLink(linkTo);
						RemoveLink(linkTo);
						
						Connect(alreadyInternal, layerTo, new FixedWeight());
					}
				}
			}
		}
	}

	return 0;
}

// The order of the visit here is not important
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
void NeuralNet::Visit(Visitor& g) {
	g.pre_visit();
	if (g.IsVisitEdges()) {
		for (Link *l : E) {
			l->Visit(g);
		}
	}
	if (g.IsVisitNodes()) {
		for (auto it = V.begin(); it != V.end(); it++) {
			(it->second)->Visit(g);
		}
	}	
	g.post_visit();
}
