//
//  nnet.hpp
//  nnet
//
//  Created by Alessandro Sordoni on 2014-03-25.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//

#pragma warning( disable : 4250 )

#ifndef nnet_nnet_hpp
#define nnet_nnet_hpp

#include <exception>
#include <stdexcept>

#include "Alg.hpp"
#include "Matrix.hpp"
#include "Activation.hpp"
#include "Graph.hpp"
#include "Visitor.hpp"

#define my_delete(x) {delete x; x = NULL;}

// Forward declaration
// class Visitor;
class Weight;

typedef enum { DELEGATE, SHARED } Memory;
typedef enum { TRANSPOSE, NOTRANSPOSE } MatrixView;

class IOutput {
protected:
	Matrix* m_activation;

public:
    Matrix* GetActivation() {
		return m_activation;
	}

	void SetActivation(Matrix* act) {
		if (m_activation != nullptr)
			delete m_activation;
		m_activation = act;
	}

    virtual ~IOutput() {};
};
class IError {
protected:
	Matrix* m_error;

public:
    Matrix* GetError() {
		return m_error;
	}
    virtual ~IError() {};
};

class DenseError : public virtual IError {    
protected: 
	DenseError(std::string name) {
        m_error = new Matrix();
		m_error->SetName("d_" + name);
    }
    DenseError(int size, std::string name) {
        m_error = new Matrix(size, 1, "d_" + name, MatrixType::DENSE);
    }
    virtual ~DenseError() {
        my_delete(m_error);
    }
};
class SparseError : public virtual IError {
protected:
	SparseError(std::string name) {
		m_error = new Matrix(MatrixType::VECSPARSE);
		m_error->SetName("d_" + name);
    }
    SparseError(int size, std::string name) {
		m_error = new Matrix(size, 1, "d_" + name, MatrixType::VECSPARSE);
    }
    virtual ~SparseError() {
        my_delete(m_error);
    }
};

class DenseIO : public virtual IOutput {
protected:
    DenseIO(std::string name) {
		m_activation = new Matrix();
		m_activation->SetName("a_" + name);
    }
    DenseIO(int size, std::string name) {
        m_activation = new Matrix(size, 1, "a_" + name, MatrixType::DENSE);
    }
    
    virtual ~DenseIO() {
        my_delete(m_activation);
    }
};

class SparseIO : public virtual IOutput {
protected:
	SparseIO(std::string name) {
		m_activation = new Matrix(MatrixType::VECSPARSE);
		m_activation->SetName("a_" + name);
    }

    SparseIO(int size, std::string name) {
		m_activation = new Matrix(size, 1, MatrixType::VECSPARSE);
		m_activation->SetName("a_" + name);
    }
	
    virtual ~SparseIO() {
        my_delete(m_activation);
    }
};

class Layer : public virtual IOutput,
    public virtual IError,
    public Node {

private:
    ActivationFn* init_activation(Activation fn) {
        if (fn == SIGMOID) return new SigmoidFn();
        else if (fn == TANH) return new TanhFn();
        else if (fn == SOFTMAX) return new SoftmaxFn();
        else if (fn == EXP) return new ExpFn();
        else if (fn == ReLU) return new ReLUFn();
        return new IdFn();
    }
protected:
    ActivationFn* fn_; // activation function
    int m_size;        // m_size of the layer
    
	bool m_skipReset;	  // if this node should skip the reset visitor
	bool m_needError;  // if this node need an error

    bool m_evaluated;     // keep track if the node has already been evaluated
    bool m_errorated;     // keep track if the node has already been errorated
    
public:
    Layer(std::string name) : Node(name), fn_(0), m_size(0) {
        m_evaluated = false;
        m_errorated = false;
		m_skipReset = false;
		m_needError = false;
        fn_ = init_activation(ID);
    }
    
    Layer(int m_size, std::string name) : Node(name), m_size(m_size) {
        m_evaluated = false;
        m_errorated = false;
		m_skipReset = false;
		m_needError = false;
        fn_ = init_activation(ID);
    }
    
    Layer(Activation fn, std::string name) : Node(name) {
        m_evaluated = false;
        m_errorated = false;
		m_skipReset = false;
		m_needError = false;
        fn_ = init_activation(fn);
    }
    
    Layer(int m_size, Activation fn, std::string name) : Node(name), m_size(m_size)
    {
        m_evaluated = false;
        m_errorated = false;
		m_skipReset = false;
		m_needError = false;
        fn_ = init_activation(fn);
    }
    
    virtual ~Layer() { my_delete(fn_); }    
    std::string ToString() {
        std::stringstream ss;
        ss << "Layer = " << GetName() << " / Size = " << m_size;
        return ss.str();
    }
    
    virtual Matrix* GetValue();
    virtual Matrix* ComputeError(Layer* parent) = 0;
	
	void SetSkipReset(bool v) { m_skipReset = v; }
	void SetErrorated(bool v) { m_errorated = v; }
	void SetEvaluated(bool v) { m_evaluated = v; }
	void SetNeedError(bool v) { m_needError = v; }
	int GetSize() { return m_size; } 
	bool IsErrorated() { return m_errorated; }
	bool IsEvaluated() { return m_evaluated; }
	bool IsSkipReset() { return m_skipReset; }
	bool IsNeedError() { return m_needError; }

    virtual void Visit(Visitor& c) = 0;
    void CheckEvaluate() {
        if (!m_evaluated) throw std::exception("Evaluation needed beforehand");
    }
    void CheckErrorate() {
        if (!m_errorated) throw std::exception("Error computation needed beforehand");
    }
};

// A general dense layer
class DenseLayer : public DenseError, public DenseIO, public Layer {
public:
    DenseLayer(std::string name) :
        DenseError(name),
        DenseIO(name),
        Layer(name) {};
    DenseLayer(int m_size, std::string name) :
        DenseError(m_size, name),
        DenseIO(m_size, name),
        Layer(m_size, name) {};
    DenseLayer(int m_size, Activation fn, std::string name) :
        DenseError(m_size, name),
        DenseIO(m_size, name),
        Layer(m_size, fn, name) {};
    
    virtual ~DenseLayer() {};
    virtual void Visit(Visitor& c) {
		c.Visit(this);
    }

    virtual Matrix* ComputeError(Layer* parent);
};

// A hidden layer, softmax not allowed here!
class HiddenLayer : public DenseLayer {
public:
    HiddenLayer(std::string name) : DenseLayer(name) {};
    HiddenLayer(int size, std::string name) : DenseLayer(size, name) {};
    HiddenLayer(int size, Activation fn, std::string name) : DenseLayer(size, fn, name) {
        if (fn == SOFTMAX) throw std::exception("Softmax is not permitted in the hidden layers!");
    }
    
    virtual ~HiddenLayer() {};
};

// A max-pooling layer, take the max!
class ConcatLayer : public HiddenLayer {
private:
	// Need an additional structure to keep parents in order.
    std::map<Layer*, int> m_orderedParents;
	std::map<Layer*, Matrix*> m_errors;

    int m_numInputs;
	int m_registeredInputs;

public:
	ConcatLayer(int size, int numInputs, std::string name) : HiddenLayer(numInputs*size, name) {
		m_numInputs = numInputs;
		m_registeredInputs = 0;
	};
	ConcatLayer(int size, int numInputs, Activation fn, std::string name) : HiddenLayer(numInputs*size, fn, name) {
		m_numInputs = numInputs;
		m_registeredInputs = 0;
	};
        
	void RegisterParent(Layer* parent) {
		auto finder = m_orderedParents.find(parent);
		if (finder != m_orderedParents.end()) {
			// Layer already registered -- error
			throw std::exception("Adding an already registered layer to ConcatLayer");
		}

		if (m_registeredInputs >= m_numInputs) {
			throw std::exception("Cannot register another child layer. Either enlarge the ConcatLayer"
				"or revise the layers already registered to it.");
		}
		
		m_errors.insert(std::make_pair(parent, new Matrix(parent->GetSize(), 1, MatrixType::DENSE)));
		m_orderedParents.insert(std::make_pair(parent, m_registeredInputs++));
	}

    virtual ~ConcatLayer() {
		for (auto it = m_errors.begin(); it != m_errors.end(); it++) {
			delete it->second;
		}
	}

    virtual Matrix* GetValue();
    virtual Matrix* ComputeError(Layer* parent);
};

// A max-pooling layer, take the max!
class MaxPoolingLayer : public HiddenLayer {
private:
    std::vector<Layer*> max_ids;
    std::unordered_map<Layer*, Matrix*> errors_;
    
public:
	MaxPoolingLayer(std::string name) : HiddenLayer(name) {};
    MaxPoolingLayer(int size, std::string name) : HiddenLayer(size, name) {
		max_ids.resize(size);
    };
    
	virtual ~MaxPoolingLayer() {
        for (auto it = errors_.begin(); it != errors_.end(); it++) {
            my_delete(it->second);
        }
        errors_.clear();
        max_ids.clear();
    }
    
    void ResetInternal() {
        for (auto it = errors_.begin(); it != errors_.end(); it++) {
			it->second->ResetValues();
        }
	}

	virtual Matrix* GetValue();
    virtual Matrix* ComputeError(Layer* parent);
};

// A general loss-layer
class LossLayer : public Layer {
protected:
    // A multiplicative factor i.e. could be useful to do means
    // or stuff like that
    double multiplicative_factor_;
    double loss_;

public:
    LossLayer(Activation fn, std::string name) : Layer(fn, name) {
        multiplicative_factor_ = 1.0;
        loss_ = 0;
    };

    LossLayer(int m_size, Activation fn, std::string name) : Layer(m_size, fn, name) {
        multiplicative_factor_ = 1.0;
        loss_ = 0;
    };

    // Compute the error
    double GetLoss() {
        CheckEvaluate();
        ComputeError(0);
        return loss_;
    }
    
    //
    virtual void Visit(Visitor& c) = 0;
    //
    
    void multiplicative_factor(double mf) {
        if (mf == 0) throw std::exception("0 as multiplicative factor is not allowed!");
        multiplicative_factor_ = mf;
    }
    double multiplicative_factor() {
        return multiplicative_factor_;
    }
};

class CrossEntropyLossLayer : public DenseIO, public DenseError, public LossLayer {
private:
    int m_target;
    
public:
    CrossEntropyLossLayer(int size,
                     std::string name) :
    DenseIO(size, "CELoss_" + name),
    DenseError(size, "CELoss_" + name),
    LossLayer(size, SOFTMAX, "CELoss_" + name)
    {}
    
	void SetOutput(int size, int target)
	{
		m_target = target;
	}

    double GetLoss() {
        CheckEvaluate();
        ComputeError(0);
        return loss_;
    }
    
    Matrix* ComputeError(Layer* parent) {
        CheckEvaluate();
        
        if (!m_errorated) {
            __tacc__(m_size, -1.0*multiplicative_factor(), 
				m_activation->GetDense()->GetDataForUpdate(), 
				m_error->GetDense()->GetDataForUpdate());
			
			m_error->GetElementForUpdate(m_target, 0) += 1.0*multiplicative_factor();
            loss_ = -multiplicative_factor_ * log(m_activation->GetElement(m_target, 0));
			
			SetErrorated(true);
        }
        
        return GetError();
    }
    
    void Visit(Visitor& c) {
        c.Visit(this);
    }
};

class L2LossLayer : public DenseIO, public DenseError, public LossLayer {
private:
    Matrix* m_target;
    
public:
    L2LossLayer(int m_size,
                Activation fn,
                std::string name,
                double* target) :
    DenseIO(m_size, "L2LossLayer_" + name),
    DenseError(m_size, "L2LossLayer_" + name),
    LossLayer(m_size, fn, "L2LossLayer_" + name)
    {
        m_target = new Matrix(m_size, 1, "target", MatrixType::DENSE);
		__dcopy__(m_size, target, m_target->GetDense()->GetDataForUpdate());
    }
    
    Matrix* ComputeError(Layer* parent) {
        CheckEvaluate();
        
        if (!m_errorated) {
            // copy activation datas to error
			__dcopy__(m_size, m_activation->GetDense()->GetDataForUpdate(), m_error->GetDense()->GetDataForUpdate());
            
            // loss is the norm 1/2 * || activation - target ||^2
			Matrix::Add(m_error, m_target, -1);
			loss_ += .5 * pow(m_error->Norm2(), 2);
			m_error->InPlaceScale(-1);
            
            // Derivative
            fn_->df(m_activation, m_error);
            SetErrorated(true);
        }
        
        return GetError();
    }
    
    void Visit(Visitor& c) {
        c.Visit(this);
    }
};

class ConstantLossLayer : public DenseIO, public DenseError, public LossLayer {
public:
	ConstantLossLayer(int m_size, std::string name) :
    DenseIO(m_size, "ConstantLossLayer" + name),
    DenseError(m_size, "ConstantLossLayer" + name),
    LossLayer(m_size, ID, "ConstantLossLayer" + name)
    {}
    
    Matrix* ComputeError(Layer* parent) {
        CheckEvaluate();
		SetErrorated(true);
		return GetError();
    }
    
    void Visit(Visitor& c) {
        c.Visit(this);
    }
};

class NCELossLayer : public SparseIO, public SparseError, public LossLayer {
private:
    std::vector<double> pnns_;
    
public:
	NCELossLayer(std::string name) :
    SparseIO("NCELoss_" + name),
    SparseError("NCELoss_" + name),
    LossLayer(ID,  "NCELoss_" + name)
    {}

    NCELossLayer(int size, std::string name) :
    SparseIO(size,  "NCELoss_" + name),
    SparseError(size,  "NCELoss_" + name),
    LossLayer(size, ID,  "NCELoss_" + name)
    {}

	void SetOutput(std::vector<int>& idx, std::vector<double>& pnns)
	{
		std::vector<double> values(idx.size());		
		m_activation->GetVecSparse()->SetData(m_size, idx, values); // 0 the inputs
		pnns_ = pnns;
	}
    
    Matrix* ComputeError(Layer* parent) {
        CheckEvaluate();
        
        if (!m_errorated) {
			int noise_count = m_activation->GetVecSparse()->nnz() - 1;
			double log_ncount = log(noise_count);

			std::vector<double> error_vals(noise_count + 1);
			std::vector<double>& acts = m_activation->GetVecSparse()->GetValues();
			std::vector<int>& idx = m_activation->GetVecSparse()->GetIds();

			for (int i = 0; i < m_activation->GetVecSparse()->nnz(); i++) {
				double log_pnw = pnns_[i];
				if (log_pnw > 0) {
					// This is the target
					double nom_t = log_ncount - log_pnw;
					double act = acts[i]; double actd = act;
					// error on positive sample: N * pnt / (act + N * pnt)
					// positive sample: log(act / (act + N * pnt))
					logadd(actd, nom_t);
					double ert = exp(nom_t - actd);
					error_vals[i] = ert;
					// Loss here
					double den = act;
					logadd(den, log_ncount - log_pnw);
					loss_ += (act - den);
				} else {
					// Negative sample
					double actn = acts[i];
					double actnd = actn;

					// negative sample: act / (act + N * pnn);
					logadd(actn, log_ncount + log_pnw);
					double ern = exp(actnd - actn);
					error_vals[i] = -ern;

					// negative sample: log(N * pnn / (act + N * pnn))
					double nom = log_ncount + log_pnw;
					double den = nom;
					logadd(den, actnd);
					loss_ += (nom-den);
				}
			}

			// Set the error 
			m_error->GetVecSparse()->SetData(m_size, idx, error_vals);
			loss_ *= -1;
			
			SetErrorated(true);
        }
        
        return m_error;
    }
    
    void Visit(Visitor& c) {
        c.Visit(this);
    }
};

class InputLayer : public Layer {
public:
    InputLayer(std::string name) : Layer(name) {
		m_skipReset = true;
		m_needError = false;
	}
    InputLayer(int m_size, std::string name) : Layer(m_size, name) {
		m_skipReset = true;
		m_needError = false;
	}
    Matrix* ComputeError(Layer* parent);
	Matrix* GetValue() { return GetActivation(); }
};

class DenseInputLayer : public DenseIO, public DenseError, public InputLayer {
public:
	DenseInputLayer(std::string name) :
        DenseIO(name),
        DenseError(name),
        InputLayer(name) {}
    DenseInputLayer(int size, std::string name) :
        DenseIO(size,name),
        DenseError(size,name),
        InputLayer(size, name) {}
    virtual ~DenseInputLayer() {};
    
    void SetInput(int size, double* input) {
        // Directly copy input to the activation array; an input layer does not have
        // a activation function!
		m_activation->GetDense()->SetData(&input[0]);
	}
    
    void Visit(Visitor& c) {
        c.Visit(this);
    }
};

class BiasLayer : public DenseInputLayer {
public:
    BiasLayer(std::string name) : DenseInputLayer(1, name) {
        // Set the unique activation element to one
		m_activation->GetDense()->GetDataForUpdate()[0] = 1;
    };

    virtual ~BiasLayer() {}
    void Visit(Visitor& c) {
        c.Visit(this);
    }
};
class SparseInputLayer : public SparseIO, public SparseError, public InputLayer {
public:
    SparseInputLayer(int size, std::string name) :
        SparseIO(size, name),
        SparseError(size, name),
        InputLayer(size, name) {};
    virtual ~SparseInputLayer() {};
    
	void SetData(int size, std::vector<int>& ids, std::vector<double>& values) {
		GetActivation()->GetVecSparse()->SetData(size, ids, values);
	}

    void Visit(Visitor& c) {
        c.Visit(this);
    }
};

class Weight : public Link {
protected:
    Matrix* m_;  // actual shared weight
    Memory mstrategy_;
    
public:
	Weight() { m_ = nullptr; }
	Weight(Matrix* m) : Link(), m_(m), mstrategy_(SHARED) {
		m_name = m->GetName();
	};
    Weight(Matrix* m, Memory strategy) : Link(), m_(m), mstrategy_(strategy) {
		m_name = m->GetName();
	};
    
    virtual ~Weight() {
        if (mstrategy_ == DELEGATE) {
            // Handle destruction
            if (m_ != nullptr) 
				delete m_; 
			m_ = nullptr;
        }
    };

    virtual void Visit(Visitor& c) = 0;    
    virtual Matrix* m() {
        return m_;
    }

    /** Just helpers **/
    int GetRows() const { return m_->GetRows(); }
    int GetCols() const { return m_->GetCols(); }
    int GetSize() const { return m_->GetSize(); }
};

class FixedWeight : public Weight {
public:
	FixedWeight() : Weight() {};
    FixedWeight(Matrix* m) : Weight(m) {};
    FixedWeight(Matrix* m, Memory strategy) : Weight(m, strategy) {};
    virtual ~FixedWeight() {};
    
    void Visit(Visitor& c) {
        c.Visit(this);
    }
};

class DynamicWeight : public Weight {
public:
	DynamicWeight(Matrix* m) : Weight(m) {};
    DynamicWeight(Matrix* m, Memory strategy) : Weight(m, strategy) {};
    virtual ~DynamicWeight() {};
    
    virtual void Visit(Visitor& c) {
        c.Visit(this);
    }
};

// A very simple print visitor
class PrintVisitor : public Visitor {
public:
	void Visit(class DenseLayer* w) { std::cerr << "Visiting " << w->GetName() << std::endl; };
	void Visit(class HiddenLayer* w) { std::cerr << "Visiting " << w->GetName() << std::endl; };
	void Visit(class NCELossLayer* w) { std::cerr << "Visiting " << w->GetName() << std::endl; };
	void Visit(class CrossEntropyLossLayer* w) { std::cerr << "Visiting " << w->GetName() << std::endl; };
	void Visit(class SparseInputLayer* w) { std::cerr << "Visiting " << w->GetName() << std::endl; };
};

// This resets some information in the graph
class ResetVisitor : public Visitor {
public:
		void cleanLayer(Layer* l) {
			if (l->IsSkipReset()) 
				return;
#ifdef DEBUG
			std::cerr<<"Resetting the value of "<<l->GetName()<<std::endl;
#endif
			if (cleanError_ && l->GetError() != nullptr && l->IsErrorated())
			{ 
				l->GetError()->ResetValues(); 
				l->SetErrorated(false); 
			}

			if (cleanActivation_ && l->GetActivation() != nullptr && l->IsEvaluated())
			{	
				l->GetActivation()->ResetValues();  
				l->SetEvaluated(false); 
			}
		};

		bool cleanError_;
		bool cleanActivation_;
	
		ResetVisitor() : Visitor() {
			m_visitNodes = true;	// visit the nodes
			m_visitEdges = false;	// do not visit edges
			cleanError_ = true;
			cleanActivation_ = true;
		}

		void Visit(DynamicWeight* w) { };
		void Visit(FixedWeight* w) { };
		void Visit(DenseLayer* w) { cleanLayer(w); };
		void Visit(HiddenLayer* w) { cleanLayer(w); };
		void Visit(MaxPoolingLayer* w) { cleanLayer(w); };
		void Visit(ConcatLayer* w) { cleanLayer(w); };
		
		void Visit(SparseInputLayer* w) { /* Do not clean inputs! */ };
		void Visit(DenseInputLayer* w) { /* Do not clean inputs! */ };

		void Visit(CrossEntropyLossLayer* w) { cleanLayer(w); };
		void Visit(NCELossLayer* w) { cleanLayer(w); };
		void Visit(L2LossLayer* w) { cleanLayer(w); };
		void Visit(LossLayer* w) { cleanLayer(w); };
};

class NeuralNet : public Graph {
private:
    /** Map of output layers **/
    std::vector<LossLayer*> O;
    std::vector<InputLayer*> I;

	// Optimization related stuff 
	std::map<std::string, Layer*> m_internMap;
	std::set<Layer*> m_internals;
	bool IsInternal(Layer* n);
	//
		
	int mbh_;
	
	// Bptt algorithm related stuff
	int bpttStop;
	Matrix* bpttWeight;
	//

	// Barrier weight = error cannot flow behind this layer
	Matrix* m_barrierWeight;
	//
	
	void FlagErrorNodes(LossLayer* ll);

	void RegisterLossLayer(LossLayer* layer) {
		AddNode(layer);
		O.push_back(layer);
	}

	Layer* GetInternalForLayerParam(const std::string& key) 
	{
		auto it = m_internMap.find(key);
		if (it == m_internMap.end())
			return nullptr;
		return it->second;
	}

public:
	const std::vector<InputLayer*>& Inputs() { return I; }
	const std::vector<LossLayer*>& Outputs() { return O; }

	void SetBptt(Matrix* weight, int stop) {
		bpttStop = stop;
		bpttWeight = weight;
	}

	void SetBarrier(Matrix* weight) {
		m_barrierWeight = weight;
	}

	NeuralNet() : Graph(), mbh_(1) {
		bpttWeight = nullptr;
		bpttStop = 0;
	}
	// mbh - the number of minibatches to be processed at the same time
	NeuralNet(int mbh) : Graph(), mbh_(mbh) {
		bpttWeight = nullptr;
		bpttStop = 0;
	}

    SparseInputLayer* AddLayer(SparseInputLayer* layer) {
        I.push_back(layer);
		AddNode(layer);
		return layer;
    }
    
	BiasLayer* AddLayer(BiasLayer* layer) {
        I.push_back(layer);
		AddNode(layer);
        return layer;
    }
    
    DenseInputLayer* AddLayer(DenseInputLayer* layer) {
        I.push_back(layer);
		AddNode(layer);
        return layer;
    }
    
    L2LossLayer* AddLayer(L2LossLayer* layer) {
        RegisterLossLayer(layer);
        return layer;
    }
    
	ConstantLossLayer* AddLayer(ConstantLossLayer* layer) {
        RegisterLossLayer(layer);
        return layer;
    }

    CrossEntropyLossLayer* AddLayer(CrossEntropyLossLayer* layer) {
        RegisterLossLayer(layer);
        return layer;
    }
    
    NCELossLayer* AddLayer(NCELossLayer* layer) {
        RegisterLossLayer(layer);
        return layer;
    }
    
    template<class T> T* AddLayer(T* layer) {
		AddNode(layer);
        return layer;
    }
    
    std::vector<Matrix*> Forward();     /* Compute the output from the net */
	Matrix* Forward(LossLayer* ll);     /* Compute the output from the net */
    void Backward();                    /* Propagates the error back in the net */
    void Backward(LossLayer* ll);		/* Propagates the error back in the net */

	void Visit(Visitor& g);

    inline bool Connect(Layer *a, Layer *b, Weight *l) {
        /** Attach two nodes with a link
         ** a -> l -> b **/
        bool ret=true;

		if (a == 0 || b == 0 || l == 0) { 
			throw std::exception("Trying to add a Null node/link!"); 
		}
        
        // Insert if it is not there
        if (!ContainsLink(l)) E.insert(l);
		else { throw std::exception("Trying to add a Null link!"); }
		
		if (!ContainsNode(a)) throw std::exception("Should register the layer before linking!");
		if (!ContainsNode(b)) throw std::exception("Should register the layer before linking!");
        
        l->ConnectOutput(b); l->ConnectInput(a);
        ret &= a->ConnectOutput(l);
        ret &= b->ConnectInput(l);
        return ret;
    }
    
    Layer* GetLayerByName(const std::string& name) {
		Node* r = GetNodeByName(name);
        return (Layer*) r;
    }

	void DeleteLayer(Layer* l) {
		if (l == 0) return;

		// Unlink the outgoing links from the children
		if (l->GetChildren().size() > 0)
		for (Link* link : l->GetChildren()) {
			(link->child())->EraseInwardLink(link);
			RemoveLink(link);
		}
		
		// Unlink the incoming links from the parents
		if (l->GetParents().size() > 0)
		for (Link* link : l->GetParents()) {
			(link->parent())->EraseOutwardLink(link);
			RemoveLink(link);
		}

		// If this is a loss layer just find it in
		// the output set.
		LossLayer* ll = dynamic_cast<LossLayer*>(l);
		if (ll)
		{
			auto it = std::find(O.begin(), O.end(), ll);
			if (it != O.end()) O.erase(it);
		}
		
		// If this is an input layer just find it in
		// the input set
		InputLayer* il = dynamic_cast<InputLayer*>(l);
		if (il)
		{
			auto it = std::find(I.begin(), I.end(), il);
			if (it != I.end()) I.erase(it);
		}
		
		// and the node
		RemoveNode(l);
	}

	void DeleteInputs() {
		for (Layer* l : I) {
			for (Link* link : l->GetChildren()) {
				(link->child())->EraseInwardLink(link);
				// Delete both the link
				E.erase(link);
				delete link;
			}
			
			// and the node
			RemoveNode(l);
		}
		
		I.clear();
	}
	
	void DeleteOutputs() {
		for (Layer* l : O) {
			for (Link* link : l->GetParents()) {
				(link->parent())->EraseOutwardLink(link);
				// Delete both the link
				E.erase(link);
				delete link;
			}
			
			// and the node
			RemoveNode(l);
		}
		
		O.clear();
	}
	
	double GetLoss(LossLayer* ll) {
		return ll->GetLoss();
    }

    double GetLoss() {
        double error = 0.;
        for (auto it = O.begin(); it != O.end(); it++) {
            error += (*it)->GetLoss();
        }

        return error;
    }
    
	// Clear inputs / outputs and
	// call the parent CleanUp method
	// which deallocates all memory
	void format() {
		O.clear();
		I.clear();
		m_internals.clear();
		CleanUp();
	}
		
	int Optimize();

    ~NeuralNet() {
		m_internals.clear();
		format();
	}
};

#endif
