#pragma once

#include "Commons.hpp"
#include "Alg.hpp"
#include <vector>
#include <sstream>


// Rows of this matrix are sparse
// This is good if we have a gradient
// that updates each *column* of the parameter (i.e. 
// a sparse input layer!)
class RowSparseMatrix {
private:
	int m_allocated;
	int m_numRows;
	int m_numCols;
	int m_size;
	std::string m_name;

public:
	std::vector<int> colPosToIdx;
	std::vector<int> colIdxToPos;
	std::vector<double> colVal;
	int m_nnz;
	
	int GetRows() { return m_numRows; }
	int GetSize() { return m_size; }
	int GetCols() { return m_numCols; }
	std::string GetName() { return m_name; }
	void SetName(const std::string& name) { m_name = name; }

	void ResetValues() {
		// Do not deallocate memory!
		std::fill(colIdxToPos.begin(), colIdxToPos.end(), -1);
		std::fill(colPosToIdx.begin(), colPosToIdx.end(), 0);
		std::fill(colVal.begin(), colVal.end(), 0);
		m_nnz = 0;
	}

	// Swap rows and columns!
	RowSparseMatrix(int rows, int cols) {
		m_numRows = rows; 
		m_numCols = cols; 
		m_size = rows*cols; 
		m_name = "@@@";
		m_allocated = 0;
		m_nnz = 0;
		// resize the row_id
		colIdxToPos.resize(m_numCols, -1);
		_ensure_memory();
	}

	RowSparseMatrix(int rows, int cols, std::string name) {
		m_numRows = rows; 
		m_numCols = cols; 
		m_size = rows*cols; 
		m_name = name;
		m_allocated = 0;
		m_nnz = 0;
		// resize the row_id
		colIdxToPos.resize(m_numCols, -1);
		_ensure_memory();
	}

	void _ensure_memory() {
		// Pre-allocate some space
		m_allocated += 50;
		colVal.resize(m_allocated*m_numRows, 0);
		colPosToIdx.resize(m_allocated);
	}

	virtual ~RowSparseMatrix() { 
		colPosToIdx.clear();
		colIdxToPos.clear();
		colVal.clear();
	}

    int nnz() { return m_nnz; }

	double& GetElementForUpdate(int i, int j) {
        if (i >= m_numRows || j >= m_numCols) 
			throw std::exception("RowSparseMatrix_get: Element out of range");

		int colPos = colIdxToPos[j];
		if (colPos == -1) {
			double* colVal = GetColumn(j);
			return colVal[i];
		}
		return colVal[IDX2R(i, colPos, m_numRows)];
    }

	double GetElement(int i, int j) const {
        if (i >= m_numRows || j >= m_numCols) 
			throw std::exception("RowSparseMatrix_get: Element out of range");

		int colPos = colIdxToPos[j];
		if (colPos == -1) return 0;
		return colVal[IDX2R(i, colPos, m_numRows)];
    }
	   
    void InPlaceScale(double f) {
		__dscal__(m_nnz*m_numRows, &colVal[0], f, &colVal[0]);
    }

	/** Attention!! **/
    void InPlaceAdd(double f) {
		throw new std::exception("RowSparseMatrix_ewise_add: Are you sure you want to do this?");
    }

    void InPlacePow(double n) {
		for (int i = 0; i < m_nnz*m_numRows; i++)
			colVal[i] = pow(colVal[i], n);
    }
	
	void InPlaceClip(double threshold) {
		double n = Norm2();
		if (n > threshold)
			InPlaceScale(threshold/n);
	}
	
	void InPlaceClip(double min, double max) {
		for (int i = 0; i < m_nnz*m_numRows; i++)
		{
			double& rowval = colVal[i];
			if (rowval < min) rowval = min;
			else if (rowval > max) rowval = max;	
		}
	}
	
	double* GetColumn(int c) {
#ifdef BOUND_CHECK
		if (c >= m_numCols) throw new std::exception("RowSparseMatrix_column: column out of range");
#endif
		// Get the pos of asked column
		int col_pos = colIdxToPos[c];
		if (col_pos == -1) {
			// If it is not found
			if (m_nnz >= m_allocated) {
				// Ensure allocated m_size if 
				// we add another row
				_ensure_memory();
			}

			colPosToIdx[m_nnz] = c;
			colIdxToPos[c] = m_nnz;
			col_pos = m_nnz;
			m_nnz++;
		}
		
		return &colVal[col_pos*m_numRows];
	}

    int ArgMax() {
        throw std::exception("ColSparseMatrix_argmax: Argmax not implemented");
    }
		
	double Norm2() {
		return __norm2__(m_nnz*m_numRows, &colVal[0]);
    }

    std::string ToString() {
        std::stringstream ss;
        ss << "RowSparseMatrix(" << GetRows() << "," << GetCols() << ") - " << GetName() << "\n";
        for (int i = 0; i < m_nnz; i++) {
			ss << colPosToIdx[i] << " : ";
			for (int k = 0; k < m_numRows; k++) {
				ss << colVal[IDX2R(i, k, m_numRows)] << " ";
			}
			ss << "\n";
        }
        return ss.str();
    }
};
