#pragma once

#include "Alg.hpp"
#include "Commons.hpp"
#include <vector>
#include <sstream>

// Just a coo sparse vector.
class SparseVector {
	int m_numRows;
	int m_numCols;
	int m_size;
	int m_nnz;
	std::string m_name;

	std::vector<double> m_vals;
	std::vector<int> m_ids;

public:
	// Swap rows and columns!
	SparseVector() { 
		m_nnz = 0; 
		m_numRows = 0; 
		m_numCols = 1; 
		m_size = 0; 
		m_name = "@@@";
	}
	SparseVector(int rows) { 
		m_nnz = 0; 
		m_numRows = rows; 
		m_numCols = 1; 
		m_size = rows; 
		m_name = "@@@";
	}
	SparseVector(int rows, std::string name) { 
		m_nnz = 0; 
		m_numRows = rows; 
		m_numCols = 1; 
		m_name = name;
		m_size = rows;
	}

	void ResetValues() {
		m_vals.clear();
		m_ids.clear();
		m_nnz = 0;
	}

	virtual ~SparseVector() {
		ResetValues();
	}

	void Resize(int rows) {
		m_numRows = rows;
		m_size = rows;
	}

	void SetData(int rows, const std::vector<int>& ids, const std::vector<double>& values) {
		if (ids.size() != values.size())
			throw std::exception("SparseVector_set: sizes should match!");

		Resize(rows);
		m_ids = std::vector<int>(ids);
		m_vals = std::vector<double>(values);
		m_nnz = (int) m_vals.size();
	}

	int ArgMax() {
		int i = 0;
		int m = 0;
		__argmax__(m_nnz, &m_vals[0], m);
		return m_ids[i];
	}

	double Norm2() {
		return __norm2__(m_nnz, &m_vals[0]);
	}

	void InPlaceScale(double f) {
		__dscal__(m_nnz, &m_vals[0], f, &m_vals[0]);		
	}

	void InPlacePow(double n) {
		for (int i = 0; i < m_nnz; i++) m_vals[i] = pow(m_vals[i], n);
	}
	void InPlaceAdd(double f) {
		__tacc__(m_nnz, f, &m_vals[0], &m_vals[0]);
	}

	void InPlaceClip(double threshold) {
		double n = Norm2();
		if (n > threshold)
			InPlaceScale(threshold/n);
	}

	void InPlaceClip(double min, double max) {
		for (int i = 0; i < m_nnz; i++)
			if (m_vals[i] < min) m_vals[i] = min;
			else if (m_vals[i] > max) m_vals[i] = max;
	}

	int GetRows() { return m_numRows; }
	int GetSize() { return m_size; }
	int GetCols() { return m_numCols; }
	std::string GetName() { return m_name; }
	void SetName(const std::string& name) { m_name = name; }

	std::vector<int>& GetIds() { return m_ids; }
	std::vector<double>& GetValues() { return m_vals; }
	int nnz() { return m_nnz; }

	std::string ToString() {
		std::stringstream ss;
		ss << "SparseVector(" << GetRows() << "," << GetCols() << ") " << GetName() << "\n";

		for (int i = 0; i < m_nnz; i++) {
			ss << m_ids[i] << " = ";
			ss << m_vals[i] << " ";
		}

		ss << "\n";
		return ss.str();
	}
};