// Rerank.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

#include "Vocabulary.hpp"
#include "DataStructures.hpp"
#include "AbstractRankedReader.hpp"
#include "ThreadedQueue.hpp"
#include "BLReader.hpp"
#include "TGTReader.hpp"
#include "BLWriter.hpp"
#include "TGTWriter.hpp"
#include "ModelLoader.hpp"

#include <ppl.h>
#include <unordered_map>

AbstractRankedReader* r = nullptr;
AbstractWriter* writer = nullptr;

struct MatchPathSeparator
{
	bool operator()( char ch ) const
	{
		return ch == '\\' || ch == '/';
	}
};

std::string
	basename( std::string const& pathname )
{
	return std::string( 
		std::find_if( pathname.rbegin(), pathname.rend(),
		MatchPathSeparator() ).base(),
		pathname.end() );
}

int main(int argc, char* argv[])
{	
	Config conf = Config::ReadCommandLine(argc, argv);

	float lambda = conf.GetFloat("lambda", 0);

	std::string modelType = conf.GetString("modelType", "DCGM-I");
	std::string fileType = conf.GetString("fileType", "TGT");
	std::string prior = conf.GetString("prior", "noPrior");
	bool normalize = conf.GetBool("normalize");

	bool hasPrior = false;
	LengthPrior p;
	if (prior != "noPrior")
	{
		hasPrior = true;
		p.Load(prior);
	}

	if (fileType == "BL") r = new BLReader(conf.GetString("nbest"));
	else r = new TGTReader(conf.GetString("nbest"));

	AbstractModel* m = nullptr;
	try {
		m = ModelLoader::LoadFromDisk(conf.GetString("model"));
	} catch (std::exception e)
	{
		std::cerr << e.what() << std::endl;
		exit(-1);
	}

	std::ofstream outputFile;
	
	std::string output = "";
	if (conf.GetBool("output"))
	{
		output = conf.GetString("output");
	}
	else
	{
		std::string tail = basename(conf.GetString("nbest"));
		output = conf.GetString("model") + "/" + tail + "." + m->GetModelName() + ".feature";
	}

	outputFile = std::ofstream(output);
	outputFile << m->GetModelName() << std::endl;
	int numRef = 0;

	HeadRecord x;
	std::cerr << "Start Reading " << std::endl;

	try
	{
		while (r->ReadNextCandidate(x)) {
			std::cerr << "=============================== " << std::endl;
			std::cerr << "- C = " << x.ctext << std::endl;
			std::cerr << "- M = " << x.mtext << std::endl;

			std::vector<std::string> mtextTokens = Tokenize(x.mtext);
			std::vector<int> mtextIds = m->GetVocabulary()->ConvertToIds(x.mtext);
			std::vector<int> ctextIds = m->GetVocabulary()->ConvertToIds(x.ctext);

			std::vector<float> scores;

			std::mutex lck;
			concurrency::parallel_for_each(x.rlist.begin(), x.rlist.end(), [&](TailRecord& c) {
				double score = 0;
				std::vector<int> rtextIds = m->GetVocabulary()->ConvertToIds(c.rtext);

				// Build triple
				Triple t; t.resize(3); 
				t[0] = ctextIds; t[1] = mtextIds; t[2] = rtextIds;
				//
				
				score = m->Test(t, !normalize);
				c.featureValues.push_front(score);
			});

			for (int i = 0; i < x.rlist.size(); i++)
				outputFile << x.rlist[i].featureValues.front() << std::endl;

			std::cerr << "- Done = " << x.rlist.size() << std::endl;
			std::cerr << "=============================== " << std::endl;
			numRef++;
		}
	} catch (std::exception e) {
		std::cerr << e.what() << std::endl;

		delete m;
		if (r) delete r;
	}

	std::cerr << "# Total number of references reranked " << numRef << std::endl;

	outputFile.close();

	delete m;
	if (r) delete r;
	return 0;
}