
#pragma once

#include "GradientUpdater.hpp"
#include "ThreadedQueue.hpp"
#include "AbstractReader.hpp"
#include "SequenceReader.hpp"
#include "TripleReader.hpp"
#include <chrono>
#include <iostream>
#include <string>
#include <iomanip>
#include <thread>
#include <mutex>

typedef std::vector<std::thread*> WorkerPool;

template<typename T, class ReaderClass>
class AbstractTrainer {
protected:
	std::mutex m_updMutex;
	int m_delay;

	std::string m_trainPath;
	std::string m_validPath;
	std::string m_testPath;
	int m_nThreads;
	std::string m_outputPath;
	
	RandomGenerator* m_generator;
	Vocabulary* m_vocabulary;
	IGradientUpdater* m_upd;

	std::vector<float> m_histTrainErr;
	std::vector<float> m_histValidErr;

	int m_epochs;
	int m_epochSize;

	clock_t m_startTime;

	float m_alpha;
	float m_beta;
	float m_eta;
	float m_startAlpha;
	float m_trainError;
	float m_testError;
	float m_validError;

	// Threads will keep track in these variables
	float m_sharedError;
	long long m_sharedPredictions;
	long long m_sharedBatchesDone;
	// 

	ThreadedQueue<T> m_trainQueue;
	ThreadedQueue<T> m_testQueue;
	ThreadedQueue<T> m_validQueue;

	WorkerPool m_trainPool;
	WorkerPool m_validPool;
	WorkerPool m_testPool;
	std::thread m_printer;

	ReaderClass m_trainReader;
	ReaderClass m_validReader;
	ReaderClass m_testReader;

	virtual void PrintStatus(bool training=true) = 0;

	static void __dispatchPrintJob(void* obj, int delay, bool& stopPrinting) {
		AbstractTrainer<T, ReaderClass>* caller = static_cast<AbstractTrainer<T, ReaderClass>*>(obj);

		while (!stopPrinting) {
			std::this_thread::sleep_for(std::chrono::seconds(delay));			
			caller->PrintStatus();
		}
	}

	void WriteErrors(const std::string& output) 
	{
		std::ofstream error(output + "/train.err");
		std::for_each(m_histTrainErr.begin(), m_histTrainErr.end(), [&](double e) {
			error << e << std::endl;
		});
		error.close();

		error = std::ofstream(output + "/valid.err");
		std::for_each(m_histValidErr.begin(), m_histValidErr.end(), [&](double e) {
			error << e << std::endl;
		});
		error.close();

		error = std::ofstream(output + "/test.err");
		error << m_testError << std::endl;
		error.close();
	}

	void WriteVocabulary(const std::string& output) 
	{
		m_vocabulary->WriteToFile(output + "/vocab");
	}


public:
	// This method should be re-implemented by the chilren workers!
	virtual void DoJob(AbstractTrainer<T, ReaderClass>& parentCaller, 
		ThreadedQueue<T>& workingQueue,
		bool isTraining=false) {};

	void Train() {
		int epoch = 0;

		m_trainReader.SetQueue(&m_trainQueue);
		m_trainReader.OpenStream(m_trainPath);

		bool stillReading = true;
		bool notImproved = false; 

		// Set up workers
		for (int i = 0; i < m_nThreads; i++) {
			m_trainPool.push_back(new std::thread(&AbstractTrainer::__dispatchTrainJob, static_cast<void*>(this)));
		}

		m_startAlpha = m_alpha;
		m_eta = 0;

		while (epoch < m_epochs) {
			m_sharedError = 0;
			m_sharedPredictions = 0;
			m_sharedBatchesDone = 0;

			// Start Epoch
			std::cout << std::endl << "- Starting Epoch #" << std::setw(2) << epoch << " / Alpha = " << m_alpha << std::endl;
			// Dispatch the printer
			bool stopPrinting = false;
			m_startTime = clock();
			
			if (m_delay != 0)
				m_printer = std::thread(&AbstractTrainer::__dispatchPrintJob, 
				static_cast<void*>(this), m_delay, std::ref(stopPrinting));

			m_trainReader.ResetCount();
			while (m_trainReader.GetCount() < m_epochSize && stillReading) {
				stillReading = m_trainReader.ReadNextBatch();
			}
			
			std::this_thread::sleep_for(std::chrono::milliseconds(1000)); 

			// Wait for all the workers to starve
			while (!(m_trainQueue.GetWaiting() == m_nThreads)) { 
				std::this_thread::sleep_for(std::chrono::milliseconds(1000)); 
			}

			m_trainError = (float) ((m_sharedError/(float)m_sharedPredictions)/log(2));

			// Kill the printer
			if (m_delay != 0) 
			{
				stopPrinting = true;
				m_printer.join(); 
			}
			// 

			// Print the current status
			std::cout << "- End Epoch #" << std::setw(2) << epoch << ", Error = " << pow(2, m_trainError) << std::endl;

			// Start validation job; fill m_validError
			Validate();
			m_validError = (float) ((m_sharedError/(float)m_sharedPredictions)/log(2));

			// Print the current status
			std::cerr << "- Validation Error = " << m_validError << " / Perpl. = " << pow(2, m_validError) << std::endl;

			// End Epoch
			if (!stillReading) {
				// Reset count and restart stream
				m_trainReader.CloseStream();
				m_trainReader.OpenStream(m_trainPath);
				stillReading = true;
			}

			/*
			// buggy:
			if (m_validError > m_histValidErr.back() || epoch == 0) {
				notImproved = true;
			} else {
				notImproved = false;
			}
			*/
			notImproved = (epoch > 0) && (m_validError > m_histValidErr.back());
			if (notImproved)
				std::cerr << "- Not improved: Validation Error = " << m_validError << "  Old Validation Error = " << m_histValidErr.back() << std::endl;

			m_histValidErr.push_back(m_validError);
			m_histTrainErr.push_back(m_trainError);

			if (notImproved) {
				break;
			} else {
				WriteModel();
			}

			epoch++;
		}

		m_trainReader.SendKillBatch(m_nThreads);
		m_trainReader.CloseStream();

		// Join workers
		for (int i = 0; i < m_nThreads; i++) {
			m_trainPool[i]->join();
			delete m_trainPool[i];
		}
	}

	void Test() {
		m_sharedPredictions = 0;
		m_sharedError = 0;

		m_testReader.SetQueue(&m_testQueue);
		m_testReader.OpenStream(m_testPath);

		// Set up workers
		for (int i = 0; i < m_nThreads; i++) {
			m_testPool.push_back(new std::thread(&AbstractTrainer::__dispatchTestJob, static_cast<void*>(this)));
		}

		while (m_testReader.ReadNextBatch()) {}
		m_testReader.SendKillBatch(m_nThreads);

		m_testError = (float) ((m_sharedError/(float)m_sharedPredictions)/log(2));

		std::cout << "- Test Error = " << m_testError << " / Perpl. = " << pow(2, m_testError) << std::endl;

		// Join workers
		for (int i = 0; i < m_nThreads; i++) {
			m_testPool[i]->join();
			delete m_testPool[i];
		}

		m_testPool.clear();

		std::cerr << "Train finished." << std::endl;
	}

	void SetAlpha(float alpha) { m_alpha = alpha; }
	void SetBeta(float beta) { m_beta = beta; }
	void SetEpochSize(int n) { m_epochSize = n; }	
	void SetEpochs(int n) { m_epochs = n; }
	void SetNThreads(int n) { m_nThreads = n; }
	void SetTrainDataset(const std::string& path) { m_trainPath = path; }
	void SetTestDataset(const std::string& path) { m_testPath = path; }
	void SetValidDataset(const std::string& path) { m_validPath = path; }
	void UpdateError(float e) { m_updMutex.lock(); m_sharedError += e; m_updMutex.unlock(); }
	void UpdatePredictions(int p) { m_updMutex.lock(); m_sharedPredictions += p; m_updMutex.unlock(); }
	void UpdateBatchesDone(int p) {m_updMutex.lock(); m_sharedBatchesDone += p; m_updMutex.unlock(); }
	void SetDelay(int delay) { m_delay = delay; }

	float GetStartAlpha() { return m_startAlpha; }
	float GetAlpha() { return m_alpha; }
	float GetBeta() { return m_beta; }
	bool IsTrain() { return m_trainPath != ""; }
	bool IsTest() { return m_testPath != ""; }
	bool IsValid() { return m_validPath != ""; }

	AbstractTrainer(const std::string& outputPath) {
		m_alpha = 0.1f;
		m_beta = 1e-7f;
		
		m_upd = nullptr;
		m_generator = nullptr;
		m_vocabulary = nullptr;
		m_outputPath = outputPath;

		m_trainPath = "";
		m_validPath = "";
		m_testPath = "";
		m_delay = 5;

		m_nThreads = 1;
		m_epochs = 20;
		m_epochSize = 1e6;
	}

	~AbstractTrainer() {
		if (m_upd) delete m_upd;
	}
	
	void SetVocabulary(Vocabulary* v) { m_vocabulary = v; }
	void SetGenerator(RandomGenerator* g) { m_generator = g; }

private:
	static void __dispatchTrainJob(void* obj) {
		AbstractTrainer<T, ReaderClass>* caller = static_cast<AbstractTrainer<T, ReaderClass>*>(obj);
		caller->DoJob(*caller, caller->m_trainQueue, true);
	}

	static void __dispatchValidJob(void* obj) {
		AbstractTrainer<T, ReaderClass>* caller = static_cast<AbstractTrainer<T, ReaderClass>*>(obj);
		caller->DoJob(*caller, caller->m_validQueue, false);
	}

	static void __dispatchTestJob(void* obj) {
		AbstractTrainer<T, ReaderClass>* caller = static_cast<AbstractTrainer<T, ReaderClass>*>(obj);
		caller->DoJob(*caller, caller->m_testQueue, false); 
	}

	void Validate() {
		m_sharedError = 0;
		m_sharedPredictions = 0;

		m_validReader.SetQueue(&m_validQueue);
		m_validReader.OpenStream(m_validPath);

		// Set up workers
		for (int i = 0; i < m_nThreads; i++) {
			m_validPool.push_back(new std::thread(&AbstractTrainer::__dispatchValidJob, static_cast<void*>(this)));
		}

		while (m_validReader.ReadNextBatch()) {}
		m_validReader.SendKillBatch(m_nThreads);

		// Join workers
		for (int i = 0; i < m_nThreads; i++) {
			m_validPool[i]->join();
			delete m_validPool[i];
		}

		m_validPool.clear();
		m_validReader.CloseStream();
	}

	virtual void WriteModel() = 0;
};