#include "DcgmII.hpp"

/// PrepareMatrices - we stock all the necessary information for the current sequence
///					  on the gpu.
/// [in] - subBatch, inputs and outputs (time ordered)
void DcgmII::PrepareMatrices(const Batch<Triple>& batch, bool isTraining)
{
	// Find the max length
	int batchSize = (int) batch.data.size();
	int sentLen = (int) batch.data[0][2].size();

	std::vector<int> ctxCInputRows; 
	std::vector<float> ctxCInputValues;
	std::vector<int> ctxCInputCols;
	
	std::vector<int> ctxMInputRows;
	std::vector<float> ctxMInputValues;
	std::vector<int> ctxMInputCols;

	std::vector<int> inputRows; 
	std::vector<float> inputValues;
	std::vector<int> inputCols;

	std::vector<int> outputRows;
	std::vector<int> outputCols;
	std::vector<float> outputValues;

	for (int i = 0; i < batchSize; ++i)
	{
		FrequencyMap c;
		const Sequence& cseq = batch.data[i][0];
		for (int s = 0; s < cseq.size(); s++)
		{
			c.Store(cseq[s]);
		}
		
		FrequencyMap m;
		const Sequence& mseq = batch.data[i][1];
		for (int s = 0; s < mseq.size(); s++)
		{
			m.Store(mseq[s]);
		}
		
		std::vector<int>& ck = c.GetKeys();
		std::vector<float>& cv = c.GetValues();
		
		float norm = 0;
		for (int j = 0; j < ck.size(); j++)
		{
			float tfIdf = cv[j]*m_vocabulary->GetIdf(ck[j]);
			cv[j] = tfIdf;
			norm += tfIdf*tfIdf;
		}
		norm = sqrt(norm);
		
		for (int j = 0; j < ck.size(); j++)
		{
			ctxCInputRows.push_back(i);
			ctxCInputCols.push_back(ck[j]);
			ctxCInputValues.push_back(cv[j]/norm);		
		}

		std::vector<int>& mk = m.GetKeys();
		std::vector<float>& mv = m.GetValues();
		
		norm = 0;
		for (int j = 0; j < mk.size(); j++)
		{
			float tfIdf = mv[j]*m_vocabulary->GetIdf(mk[j]);
			mv[j] = tfIdf;
			norm += tfIdf*tfIdf;
		}
		norm = sqrt(norm);
		
		for (int j = 0; j < mk.size(); j++)
		{
			ctxMInputRows.push_back(i);
			ctxMInputCols.push_back(mk[j]);
			ctxMInputValues.push_back(mv[j]/norm);		
		}
	}

	for (int j = 0; j < sentLen - 1; j++)
	{
		for (int i = 0; i < batchSize; ++i) {
			// Take the last sequence
			const Sequence& seq = batch.data[i][2];

			int input = seq[j];
			int output = seq[j+1];

			// [batch_n, token_id] = 1
			inputRows.push_back(i);
			inputCols.push_back(input);
			inputValues.push_back(1);

			if (!isTraining)
			{
				/// If we evaluate we only push the real output value
				outputRows.push_back(i);
				outputCols.push_back(output);	
				outputValues.push_back(1);
			}
			else
			{
				std::vector<int> noise = m_generator->SampleNoise(25, output);
				for (int x = 0; x < noise.size(); x++) 
				{
					float logPn = m_generator->GetLogProb(noise[x]);

					outputRows.push_back(i);
					outputCols.push_back(noise[x]);
					outputValues.push_back(logPn);
				}

				// Push the target
				outputRows.push_back(i);
				outputCols.push_back(output);
				outputValues.push_back(-m_generator->GetLogProb(output));
			}
		}
	}

	int ctxCInputsNnz = (int) ctxCInputRows.size();	
	int ctxMInputsNnz = (int) ctxMInputRows.size();	
	int inputNnz = (int) inputRows.size();
	int outputNnz = (int) outputRows.size();

	/// We set the values
	m_ctxInputs->SetData(1, 1, ctxMInputsNnz, ctxMInputRows.data(), ctxMInputCols.data(), ctxMInputValues.data());
	m_ctxCInputs->SetData(1, 1, ctxCInputsNnz, ctxCInputRows.data(), ctxCInputCols.data(), ctxCInputValues.data());
	m_rnnInputs->SetData(1, 1, inputNnz, inputRows.data(), inputCols.data(), inputValues.data());
	m_rnnOutputs->SetData(1, 1, outputNnz, outputRows.data(), outputCols.data(), outputValues.data());
	/// We're done
}

void DcgmII::SetCtxIO(int seqLen, int batchSize)
{
	// Just copy all of the data
	for (int i = 0; i < m_biasLayers.size(); i++)
		m_biasLayers[i]->SetInput(batchSize);

	m_ctxInputLayer->SetInput(batchSize, m_nIn, m_ctxInputs->m_sparse->m_nz, m_ctxInputs->m_sparse->m_cooRow,
		 m_ctxInputs->m_sparse->m_cooCol, m_ctxInputs->m_sparse->m_cooVal, true, true);

	m_ctxCInputLayer->SetInput(batchSize, m_nIn, m_ctxCInputs->m_sparse->m_nz, m_ctxCInputs->m_sparse->m_cooRow,
		 m_ctxCInputs->m_sparse->m_cooCol, m_ctxCInputs->m_sparse->m_cooVal, true, true);
}

Layer* DcgmII::ComputeContext() {
	// Build the context vector
	m_ctxInputLayer = m_trainRnn->AddLayer(new SparseInputLayer(m_nIn, "m_i"));
	m_ctxCInputLayer = m_trainRnn->AddLayer(new SparseInputLayer(m_nIn, "c_i"));
	
	// Continue
	Layer* lastLayer;

	BiasLayer* bm = m_trainRnn->AddLayer(new BiasLayer(m_batchSize, "bm"));
	BiasLayer* bc = m_trainRnn->AddLayer(new BiasLayer(m_batchSize, "bc"));
	m_biasLayers.push_back(bm);
	m_biasLayers.push_back(bc);

	std::vector<BiasLayer*> biases;
	for (int i = 1; i < m_layerWeights.size(); i++)
	{
		std::string id = std::to_string(i);
		BiasLayer* b = m_trainRnn->AddLayer(new BiasLayer(m_batchSize, "b_" + id));
		m_biasLayers.push_back(b);
		biases.push_back(b);
	}
	
	for (size_t i = 0; i < m_layerWeights.size(); i++) {
		std::string id = std::to_string(i);
		if (i == 0) 
		{
			HiddenLayer* mh = m_trainRnn->AddLayer(new HiddenLayer(ID, "m_h"));
			HiddenLayer* ch = m_trainRnn->AddLayer(new HiddenLayer(ID, "c_h"));

			m_trainRnn->Connect(m_ctxInputLayer, mh, new DynamicWeight(m_layerWeights[i]));
			m_trainRnn->Connect(m_ctxCInputLayer, ch, new DynamicWeight(m_layerWeights[i]));
			
			m_trainRnn->Connect(bm, mh, new DynamicWeight(m_layerBias[i]));
			m_trainRnn->Connect(bc, ch, new DynamicWeight(m_layerBias[i]));
			
			ConcatLayer* concatInputs = m_trainRnn->AddLayer(new ConcatLayer(2, ID, "concat_cm"));
			concatInputs->RegisterParent(ch);
			concatInputs->RegisterParent(mh);
			
			// Connect now
			m_trainRnn->Connect(mh, concatInputs, new FixedWeight());
			m_trainRnn->Connect(ch, concatInputs, new FixedWeight());
			lastLayer = concatInputs;
		}
		else 
		{
			HiddenLayer* h = m_trainRnn->AddLayer(new HiddenLayer(SIGMOID, "d_" + id));
			m_trainRnn->Connect(lastLayer, h, new DynamicWeight(m_layerWeights[i]));
			m_trainRnn->Connect(biases[i-1], h, new DynamicWeight(m_layerBias[i]));
			lastLayer = h;
		}
	}
	
	for (int i = 0; i < m_maxLength - 1; ++i) 
	{
		std::string id = std::to_string(i+1);
		Layer* l = m_trainRnn->GetLayerByName("h_" + id);
		m_trainRnn->Connect(lastLayer, l, new DynamicWeight(m_H));		
	}

	m_trainRnn->Optimize();
	return lastLayer;
}

DcgmII* DcgmII::LoadFromFile(const std::string& dirPath) {
	DcgmII* a = new DcgmII(true);

	CudaMatrix* Win = new CudaMatrix();
	CudaMatrix* Whh = new CudaMatrix();
	CudaMatrix* Wout = new CudaMatrix();
	CudaMatrix* H = new CudaMatrix();
	CudaMatrix* bhh = new CudaMatrix();
	CudaMatrix* bout = new CudaMatrix();
	
	Whh->LoadFromExternalBinary(dirPath + "\\w_hh.bin");
	std::cerr << Whh->GetRows() << " " << Whh->GetCols() << " " << Whh->Norm2() << std::endl;

	Win->LoadFromExternalBinary(dirPath + "\\w_in.bin");
	std::cerr << Win->GetRows() << " " << Win->GetCols() << " " << Win->Norm2() << std::endl;
	
	Wout->LoadFromExternalBinary(dirPath + "\\w_out.bin");
	std::cerr << Wout->GetRows() << " " << Wout->GetCols() << " " << Wout->Norm2() << std::endl;
	
	H->LoadFromExternalBinary(dirPath + "\\h.bin");
	std::cerr << H->GetRows() << " " << H->GetCols() << " " << H->Norm2() << std::endl;
	
	Config c = Config::ReadConfigFile(dirPath + "\\config");
	a->SetAct(c.GetString("act", "sigmoid"));

	std::cerr << "RnnActivation = " << a->GetAct() << std::endl;
	
	bhh->LoadFromExternalBinary(dirPath + "\\b_hh.bin");
	bout->LoadFromExternalBinary(dirPath + "\\b_out.bin");
	
	std::ifstream f(dirPath + "\\d_num.txt");
	int numLayers = 0;
	f >> numLayers;
	f.close();
	std::cerr << numLayers << std::endl;

	std::vector<CudaMatrix*> layerWeights;
	std::vector<CudaMatrix*> layerBias;
	std::vector<size_t> layerDims;

	for (int i = 0; i < numLayers; i++) {
		CudaMatrix* D = new CudaMatrix();
		D->LoadFromExternalBinary(dirPath + "\\d_" + std::to_string(i) + ".bin");
		std::cerr << D->GetName() << " " << D->GetRows() << " " << D->GetCols() << " " << D->Norm2() << std::endl;
		layerWeights.push_back(D);

		CudaMatrix* b = new CudaMatrix();
		b->LoadFromExternalBinary(dirPath + "\\b_" + std::to_string(i) + ".bin");
		std::cerr << b->GetName() << " " << b->GetRows() << " " << b->GetCols() << " " << b->Norm2() << std::endl;
		layerBias.push_back(b);
		
		layerDims.push_back(D->GetRows());
	}
	
	Vocabulary* o = new Vocabulary(dirPath + "\\vocab");
	RandomGenerator* g = new RandomGenerator(o);

	a->SetGenerator(g);
	a->SetVocabulary(o);
	a->SetExternalInWeights(Win, Whh, Wout);
	a->SetContextRnnWeights(H);
	a->SetExternalBias(bhh, bout);
	a->SetExternalWeights(layerDims, layerWeights, layerBias);
	
	return a;
}