#include "DcgmTrainer.hpp"

void DcgmTrainer::PrintStatus(bool training) {
	clock_t now = clock();
	float entropy = (m_sharedError/m_sharedPredictions)/log(2);
	float perpl = pow(2, entropy);
	long long difftime = (now - m_startTime)/1000;

	std::cout << difftime/3600 <<":"<<(difftime % 3600)/60<<":"<<((difftime % 3600) % 60); 
	std::cout << " - Entr. / Perp. : " << std::setw(8) << entropy << " / " << perpl
		<< " P/s: " << std::setw(5) << (long long) (m_sharedPredictions/difftime);
	std::flush(std::cout);

	/** Print norms **/
	std::cout << std::setw(5);

	std::cout << " W_hh : " << std::setw(5) << Whh.Norm2();
	std::cout << " W_out : " << std::setw(5) << Wout.Norm2();
	std::cout << " W_in : " << std::setw(5) << Win.Norm2();
	std::cout << " H : " << std::setw(5) << H.Norm2();

	for (size_t i = 0; i < m_layerWeights.size(); i++) {
		std::cout << " D_" << i << " : " << std::setw(5) << m_layerWeights[i]->Norm2();
	}
	for (size_t i = 0; i < m_layerBias.size(); i++) {
		std::cout << " b_" << i << " : " << std::setw(5) << m_layerBias[i]->Norm2();
	}

	std::cout << " b_hh : " << std::setw(5) << bhh.Norm2();
	std::cout << " b_out : " << std::setw(5) << bout.Norm2();

	std::cout << std::endl;
}

void DcgmTrainer::DoJob(AbstractTrainer<Batch<Triple>, TripleBatchedReader>& parentCaller, 
	ThreadedQueue<Batch<Triple>>& workingQueue, bool isTraining) {
		
	long long loaded_seq = 0, seq_position = 0;
	Batch<Triple> wi;

	DcgmI dcgmi;	
	DcgmII dcgmii;
	DcgmIII dcgmiii;
	DcgmIV dcgmiv;

	if (m_upd == nullptr)
	{
		m_upd = new AdagradUpdater(parentCaller.GetAlpha(), 1);

		// Some regularization for the parameters!
		if (m_beta != 0) {
			m_upd->add_pre_filter(&Whh, new L2RegularizeFilter(m_beta));
			m_upd->add_pre_filter(&Wout, new L2RegularizeFilter(m_beta));
			m_upd->add_pre_filter(&Win, new L2RegularizeFilter(m_beta));
			m_upd->add_pre_filter(&bhh, new L2RegularizeFilter(m_beta));
			m_upd->add_pre_filter(&bout, new L2RegularizeFilter(m_beta));
			
			for (int i = 0; i < m_layerWeights.size(); i++) {
				m_upd->add_pre_filter(m_layerWeights[i], new L2RegularizeFilter(m_beta));
			}

			for (int i = 0; i < m_layerBias.size(); i++) {
				m_upd->add_pre_filter(m_layerBias[i], new L2RegularizeFilter(m_beta));
			}
			
			m_upd->add_pre_filter(&H, new L2RegularizeFilter(m_beta));
		}
	}

	if (m_dcgmType == DCGMI)
	{
		dcgmi.SetTau(m_tau);
		dcgmi.SetExternalInWeights(&Win, &Whh, &Wout);
		dcgmi.SetContextRnnWeights(&H);
		dcgmi.SetExternalWeights(m_layerDims, m_layerWeights, m_layerBias);
		dcgmi.SetUpdater(m_upd);
		dcgmi.SetGenerator(m_generator);
		dcgmi.SetVocabulary(m_vocabulary);
		dcgmi.SetAct(m_act);
		dcgmi.SetBatchSize(BUNCH_SIZE);
		dcgmi.SetExternalBias(&bhh, &bout);
	}
	else if (m_dcgmType == DCGMII)
	{
		dcgmii.SetTau(m_tau);
		dcgmii.SetExternalInWeights(&Win, &Whh, &Wout);
		dcgmii.SetContextRnnWeights(&H);
		dcgmii.SetExternalWeights(m_layerDims, m_layerWeights, m_layerBias);
		dcgmii.SetUpdater(m_upd);
		dcgmii.SetGenerator(m_generator);
		dcgmii.SetVocabulary(m_vocabulary);
		dcgmii.SetAct(m_act);
		dcgmii.SetBatchSize(BUNCH_SIZE);
		dcgmii.SetExternalBias(&bhh, &bout);
	} else if (m_dcgmType == DCGMIII)
	{
		dcgmiii.SetTau(m_tau);
		dcgmiii.SetExternalInWeights(&Win, &Whh, &Wout);
		dcgmiii.SetContextRnnWeights(&H);
		dcgmiii.SetExternalWeights(m_layerDims, m_layerWeights, m_layerBias);
		dcgmiii.SetUpdater(m_upd);
		dcgmiii.SetGenerator(m_generator);
		dcgmiii.SetVocabulary(m_vocabulary);
		dcgmiii.SetAct(m_act);
		dcgmiii.SetBatchSize(BUNCH_SIZE);
		dcgmiii.SetExternalBias(&bhh, &bout);
	} else
	{
		dcgmiv.SetTau(m_tau);
		dcgmiv.SetExternalInWeights(&Win, &Whh, &Wout);
		dcgmiv.SetExternalWeights(m_layerDims, m_layerWeights, m_layerBias);
		dcgmiv.SetUpdater(m_upd);
		dcgmiv.SetContextRnnWeights(&H);
		dcgmiv.SetGenerator(m_generator);
		dcgmiv.SetVocabulary(m_vocabulary);
		dcgmiv.SetAct(m_act);
		dcgmiv.SetBatchSize(BUNCH_SIZE);
		dcgmiv.SetExternalBias(&bhh, &bout);
	}

	while (1) {
		// GetElement job
		wi = static_cast<Batch<Triple>&&>(workingQueue.PopFront());
		loaded_seq = wi.m_size;

		// poison pill
		if (loaded_seq == 0) {
			// Exit here
			break;
		}

		// Prepare for current batch
		int batchSentSize = (int) wi.data[0][2].size();
		m_upd->mb_ = wi.m_size;

		// Do the job
		if (m_dcgmType == DCGMI)
			parentCaller.UpdateError(dcgmi.Run(wi, isTraining));
		else if (m_dcgmType == DCGMII)
			parentCaller.UpdateError(dcgmii.Run(wi, isTraining));
		else if (m_dcgmType == DCGMIII)
			parentCaller.UpdateError(dcgmiii.Run(wi, isTraining));
		else 
			parentCaller.UpdateError(dcgmiv.Run(wi, isTraining));

		parentCaller.UpdatePredictions(batchSentSize * wi.m_size);
		parentCaller.UpdateBatchesDone(1);
	}
}

void DcgmTrainer::WriteModel() {
	std::string output;
	std::stringstream outDir;
	outDir << m_outputPath 
		<< "_" << m_dcgmType
		<< "_" << m_epochs
		<< "_rnnAct" << m_act
		<< "_tau" << m_tau 
		<< "_a" << m_startAlpha
		<< "_b" << m_beta
		<< "_hid" << m_nHid 
		<< "_out" << m_nOut;

	for (int i = 0; i < m_layerDims.size(); i++) {
		outDir << "_d" << m_layerDims[i];
	}

	output = outDir.str();

	std::string make_dir = "mkdir " + output;
	system(make_dir.c_str());

	WriteErrors(output);
	WriteVocabulary(output);

	std::ofstream modelType(output + "/modelType");
	modelType << m_dcgmType;
	modelType.close();

	std::ofstream config(output + "/config");
	config << "act=" << m_act << std::endl;
	config << "alpha=" << m_alpha << std::endl;
	config << "beta=" << m_beta << std::endl;
	config.close();

	Win.WriteToExternalBinary(output + "/w_in.bin");
	Wout.WriteToExternalBinary(output + "/w_out.bin");
	Whh.WriteToExternalBinary(output + "/w_hh.bin");
	bhh.WriteToExternalBinary(output + "/b_hh.bin");
	bout.WriteToExternalBinary(output + "/b_out.bin");

	H.WriteToExternalBinary(output + "/h.bin");
	
	std::ofstream d(output + "/d_num.txt");
	d << m_layerDims.size() << std::endl;
	d.close();

	for (int i = 0; i < m_layerWeights.size(); i++) {
		m_layerWeights[i]->WriteToExternalBinary(output + "/d_" + std::to_string(i) + ".bin");
	}
	for (int i = 0; i < m_layerBias.size(); i++) {
		m_layerBias[i]->WriteToExternalBinary(output + "/b_" + std::to_string(i) + ".bin");
	}
}