#pragma once

#include "FcgmTrainer.hpp"
#include "DcgmI.hpp"
#include "DcgmII.hpp"
#include "DcgmIII.hpp"
#include "DcgmIV.hpp"
#include "print.hpp"
#include <chrono>

const std::string DCGMI = "DCGM-I";
const std::string DCGMII = "DCGM-II";
const std::string DCGMIII = "DCGM-III";
const std::string DCGMIV = "DCGM-IV";

class DcgmTrainer : public FcgmTrainer {
protected:
	bool m_deep;
	bool m_hasBiases;

	std::vector<size_t> m_layerDims;
	std::vector<CudaMatrix*> m_layerWeights;
	std::vector<CudaMatrix*> m_layerBias;
	std::string m_dcgmType;

public:
	DcgmTrainer(const std::string& outputPath) : FcgmTrainer(outputPath) {
		m_dcgmType = DCGMI;
		m_deep = false;
	}

	void SetDcgmType(const std::string& t) {
		if (t != DCGMI && t != DCGMII && t != DCGMIII && t != DCGMIV)
			throw std::exception("Dcgm type not recognized!");
		m_dcgmType = t;
	}
	
	void InitializeDcgmWeights(int nIn, int nHid, int nOut, const std::vector<size_t>& layerDims) {
		if (layerDims.size() == 0) {
			throw std::exception("Should specify at least one layer!");
		}

		long seed = 0;
		InitializeRnnWeights(nIn, nHid, nOut, seed);
		m_layerDims = layerDims;

		if (m_dcgmType == DCGMIV || m_dcgmType == DCGMIII)
		{
			// Resize output weights
			if (m_nHid != layerDims[layerDims.size()-1] && layerDims.size() != 1)
				throw std::exception("Not possible to concatenate different sizes with DcgmIV");
			
			if (m_dcgmType == DCGMIV)
			{
				Whh.Resize(m_nHid*2, m_nHid*2);
				//Whh.OrthogonalFill(0, 0.1f); // NAACL
				Whh.GaussianRandomFill(0, 0.1f);
				bhh.Resize(1, m_nHid*2); 
			}

			Wout.Resize(m_nHid*2, m_nOut);
			Wout.GaussianRandomFill(0, 0.1f);
			
			std::cerr << "Whh " << Whh.GetRows() << " " << Whh.GetCols() << std::endl;
			std::cerr << "Wout " << Wout.GetRows() << " " << Wout.GetCols() << std::endl;
		}

		std::cerr << "Initializing model " << m_dcgmType << " " << m_nIn << ":" << m_nHid << ":" << m_nOut;
		for (int i = 0; i < layerDims.size(); i++)
			std::cerr << ":" << layerDims[i];
		std::cerr<<std::endl;
		
		// Deep forward layers of the DCGM
		for (int i = 0; i < layerDims.size(); i++) {
			size_t bottomLayerDim = 0;
			size_t currLayerDim = layerDims[i];
			if (i == 0) bottomLayerDim = m_nIn;
			else bottomLayerDim = layerDims[i-1];

			if (i == 1 && (m_dcgmType == DCGMII || m_dcgmType == DCGMIII || m_dcgmType == DCGMIV))
			{
				// Concat here
				bottomLayerDim = layerDims[i-1]*2;
			}

			CudaMatrix* D_i = new CudaMatrix(bottomLayerDim, currLayerDim, MatrixType::DENSE);
			D_i->SetName("D_" + std::to_string(i));
			
			seed += 7919;

			D_i->SetSeed(seed);
			if (i != 0 && m_dcgmType == DCGMIV)
			{
				D_i->SparseInit(15);
			}
			else
				D_i->GaussianRandomFill(0.f, 0.1f);
			
			m_layerWeights.push_back(D_i);
						
			CudaMatrix* b_i = new CudaMatrix(1, currLayerDim, MatrixType::DENSE);
			b_i->SetName("b_" + std::to_string(i));
						
			m_layerBias.push_back(b_i);
		}

		m_nC = layerDims[layerDims.size()-1];
		if (layerDims.size() == 1 && m_dcgmType == DCGMII 
				|| m_dcgmType == DCGMIII 
				|| m_dcgmType == DCGMIV)
			m_nC *= 2;
	
		
		if (m_dcgmType == DCGMIV)
			H.Resize(m_nHid*2, m_nHid*2);
		else if (m_dcgmType == DCGMIII)
			H.Resize(m_nHid*2, m_nHid);
		else
			H.Resize(m_nC, m_nHid);

		H.SetName("H");
		seed += 7919;
		H.SetSeed(seed);
		if (m_dcgmType == DCGMIII || m_dcgmType == DCGMIV)
			H.SparseInit(15);
		else
			H.GaussianRandomFill(0, 0.1f);
	}

	void DoJob(AbstractTrainer<Batch<Triple>, TripleBatchedReader>& parentCaller, 
		ThreadedQueue<Batch<Triple>>& workingQueue,
		bool isTraining=false);
	void WriteModel();
	void PrintStatus(bool training=true);
	void SetBiases(bool yes=false) {
		m_hasBiases = true;
	}

	void SetPretrainedCtxEmbeddings(const std::string& extFile) {
		if (m_layerWeights.size() == 0) {
			throw std::exception("You should first initialize the weights.");
		}

		CudaMatrix* D_0 = m_layerWeights[0];
		D_0->LoadFromExternalBinary(extFile);
		
		std::cerr << "External embeddings loaded" << std::endl;
	}

	void SetDeep(bool deep)
	{
		m_deep = deep;
	}
	
	~DcgmTrainer() {
		for (int i = 0; i < m_layerWeights.size(); i++) {
			if (m_layerWeights[i] != nullptr)
				delete m_layerWeights[i];
		}
		for (int i = 0; i < m_layerBias.size(); i++) {
			if (m_layerBias[i] != nullptr)
				delete m_layerBias[i];
		}
	}
};