#include "Fcgm.hpp"

Fcgm::Fcgm(bool proprietaryMemory) : Rnn(proprietaryMemory)
{
	m_modelName = "FCGM";
	m_I = nullptr;	
	m_H = nullptr;
	m_nC = 0;

	// Initialize inputs & outputs
	m_ctxInputs = new CudaMatrix(1, 1, MatrixType::SPARSE);
	m_ctxInputs->m_sparse->EnsureSpace(100*256);
	//
}

Fcgm::Fcgm() : Rnn() {
	m_modelName = "FCGM";
	m_I = nullptr;	
	m_H = nullptr;
	m_nC = 0;
	// Initialize inputs & outputs
	m_ctxInputs = new CudaMatrix(1, 1, MatrixType::SPARSE);
	m_ctxInputs->m_sparse->EnsureSpace(100*256);
	//
}

Fcgm::Fcgm(int nHid, int nOut) : Rnn(nHid, nOut) {
	m_modelName = "FCGM";
	m_I = nullptr;	
	m_H = nullptr;
	m_nC = 0;
	
	// Initialize inputs & outputs
	m_ctxInputs = new CudaMatrix(1, 1, MatrixType::SPARSE);
	m_ctxInputs->m_sparse->EnsureSpace(100*256);
	//
}

Layer* Fcgm::ComputeContext() {
	// Build the context vector
	SparseInputLayer* cmInput = m_trainRnn->AddLayer(new SparseInputLayer("cm_i"));
	Layer* contextLayer = m_trainRnn->AddLayer(new HiddenLayer(ID, "c_o"));
	m_trainRnn->Connect(cmInput, contextLayer, new FixedWeight(m_I));

	// Does NOT need gradient and should be skipped for reset
	contextLayer->SetNeedError(false);
	contextLayer->SetSkipReset(true);
	
	for (int i = 0; i < m_maxLength - 1; ++i) 
	{
		std::string id = std::to_string(i+1);
		Layer* l = m_trainRnn->GetLayerByName("h_" + id);
		m_trainRnn->Connect(contextLayer, l, new DynamicWeight(m_H));		
	}
	
	m_trainRnn->Optimize();
	m_ctxInputLayer = cmInput;
	return contextLayer;
}

void Fcgm::SetCtxIO(int seqLen, int batchSize)
{
	// Just copy all of the data
	for (int i = 0; i < m_biasLayers.size(); i++)
		m_biasLayers[i]->SetInput(batchSize);
	
	m_ctxInputLayer->SetInput(batchSize, m_nIn, m_ctxInputs->m_sparse->m_nz, m_ctxInputs->m_sparse->m_cooRow,
		 m_ctxInputs->m_sparse->m_cooCol, m_ctxInputs->m_sparse->m_cooVal, true, true);
}

void Fcgm::CleanNetwork(NeuralNet& rnn, bool cleanActivations) {
	ResetVisitor v;
	v.cleanError_ = true;
	v.cleanActivation_ = cleanActivations;
	rnn.Visit(v);
}

Fcgm* Fcgm::LoadFromFile(const std::string& dirPath) {
	Fcgm* a = new Fcgm(true);

	CudaMatrix* Win = new CudaMatrix();
	CudaMatrix* Whh = new CudaMatrix();
	CudaMatrix* Wout = new CudaMatrix();
	CudaMatrix* I = new CudaMatrix();
	CudaMatrix* H = new CudaMatrix();
	
	Whh->LoadFromExternalBinary(dirPath + "\\w_hh.bin");
	std::cerr << Whh->GetRows() << " " << Whh->GetCols() << " " << Whh->Norm2() << std::endl;

	Win->LoadFromExternalBinary(dirPath + "\\w_in.bin");
	std::cerr << Win->GetRows() << " " << Win->GetCols() << " " << Win->Norm2() << std::endl;
	
	Wout->LoadFromExternalBinary(dirPath + "\\w_out.bin");
	std::cerr << Wout->GetRows() << " " << Wout->GetCols() << " " << Wout->Norm2() << std::endl;
	
	I->LoadFromExternalBinary(dirPath + "\\i.bin");
	std::cerr << I->GetRows() << " " << I->GetCols() << " " << I->Norm2() << std::endl;
	
	Config c = Config::ReadConfigFile(dirPath + "\\config");

	a->SetAct(c.GetString("act", "sigmoid"));
	std::cerr << "RnnActivation = " << c.GetString("act", "sigmoid") << std::endl;

	H->LoadFromExternalBinary(dirPath + "\\h.bin");
	std::cerr << H->GetRows() << " " << H->GetCols() << " " << H->Norm2() << std::endl;
	
	a->SetExternalInWeights(Win, Whh, Wout);
	a->SetContextRnnWeights(H);
	a->SetExternalWeights(I);
	return a;
}

std::vector<int> Fcgm::Generate(const Triple& cmr) {
	NOT_IMPLEMENTED;
}

/// PrepareMatrices - we stock all the necessary information for the current sequence
///					  on the gpu.
/// [in] - subBatch, inputs and outputs (time ordered)
void Fcgm::PrepareMatrices(const Batch<Triple>& batch, bool isTraining)
{
	// Find the max length
	int batchSize = (int) batch.data.size();
	int sentLen = (int) batch.data[0][2].size();

	std::vector<int> ctxInputRows; 
	std::vector<float> ctxInputValues;
	std::vector<int> ctxInputCols;

	std::vector<int> inputRows; 
	std::vector<float> inputValues;
	std::vector<int> inputCols;

	std::vector<int> outputRows;
	std::vector<int> outputCols;
	std::vector<float> outputValues;

	for (int i = 0; i < batchSize; ++i)
	{
		FrequencyMap b;
		for (int j = 0; j < 2; j++)
		{
			const Sequence& seq = batch.data[i][j];
			for (int s = 0; s < seq.size(); s++)
			{
				b.Store(seq[s]);
			}
		}

		std::vector<int> k = b.GetKeys();
		std::vector<float> v = b.GetValues();
		
		float norm = 0;
		for (int j = 0; j < k.size(); j++)
		{
			float tfIdf = v[j]*m_vocabulary->GetIdf(k[j]);
			v[j] = tfIdf;
			norm += tfIdf*tfIdf;
		}
		norm = sqrt(norm);
		
		for (int j = 0; j < k.size(); j++)
		{
			ctxInputRows.push_back(i);
			ctxInputCols.push_back(k[j]);
			ctxInputValues.push_back(v[j]/norm);
		}
	}

	for (int j = 0; j < sentLen - 1; j++)
	{
		for (int i = 0; i < batchSize; ++i) {
			// Take the last sequence
			const Sequence& seq = batch.data[i][2];

			int input = seq[j];
			int output = seq[j+1];

			// [batch_n, token_id] = 1
			inputRows.push_back(i);
			inputCols.push_back(input);
			inputValues.push_back(1);

			if (!isTraining)
			{
				/// If we evaluate we only push the real output value
				outputRows.push_back(i);
				outputCols.push_back(output);	
				outputValues.push_back(1);
			}
			else
			{
				std::vector<int> noise = m_generator->SampleNoise(25, output);
				for (int x = 0; x < noise.size(); x++) 
				{
					float logPn = m_generator->GetLogProb(noise[x]);

					outputRows.push_back(i);
					outputCols.push_back(noise[x]);
					outputValues.push_back(logPn);
				}

				// Push the target
				outputRows.push_back(i);
				outputCols.push_back(output);
				outputValues.push_back(-m_generator->GetLogProb(output));
			}
		}
	}

	int ctxInputsNnz = (int) ctxInputRows.size();	
	int inputNnz = (int) inputRows.size();
	int outputNnz = (int) outputRows.size();

	/// We set the values
	m_ctxInputs->SetData(1, 1, ctxInputsNnz, ctxInputRows.data(), ctxInputCols.data(), ctxInputValues.data());
	m_rnnInputs->SetData(1, 1, inputNnz, inputRows.data(), inputCols.data(), inputValues.data());
	m_rnnOutputs->SetData(1, 1, outputNnz, outputRows.data(), outputCols.data(), outputValues.data());
}

float Fcgm::Evaluate(const Batch<Triple>& batch, bool stochastic) {	
	float error = 0;
	bool networkChanged = true;
	
	int batchSize = (int) batch.data.size();
	int sentLen = (int) batch.data[0][2].size();
	
	if (m_trainRnn == nullptr) {
		BuildNetwork();
		// Link the context
		ComputeContext();
	}
	
	PrepareMatrices(batch, stochastic);
	SetIO(sentLen, batchSize, *m_rnnInputs, *m_rnnOutputs, stochastic);
	SetCtxIO(sentLen, batchSize);

	// Run through the RNN
	for (int i = 0; i < sentLen - 1; ++i) 
	{
		CleanNetwork(*m_trainRnn, networkChanged);
		networkChanged = false;
		
		LossLayer* toEvaluate = nullptr;
		if (stochastic) toEvaluate = m_nceLayers[i];
		else toEvaluate = m_ceLayers[i];

		CudaMatrix* output = m_trainRnn->Forward(toEvaluate);
		
		if (stochastic) {
			// Works only with 1 sentence
			// Get last nnz (it is the target by construction)
			error -= output->m_sparse->GetLastNnz();
		} else {
			error += m_trainRnn->GetLoss(toEvaluate);
		}
	}
	
	return error;
}

float Fcgm::Run(const Batch<Triple>& batch, bool isTraining) {	
	float error = 0;
	bool networkChanged = true;
	
	int batchSize = (int) batch.data.size();
	int sentLen = (int) batch.data[0][2].size();
	
	if (m_trainRnn == nullptr) {
		BuildNetwork();
		// Link the context
		ComputeContext();
	}
	
	PrepareMatrices(batch, isTraining);	
	SetIO(sentLen, batchSize, *m_rnnInputs, *m_rnnOutputs, isTraining);
	SetCtxIO(sentLen, batchSize);

	// Run through the RNN
	for (int i = 0; i < sentLen - 1; ++i) 
	{
		CleanNetwork(*m_trainRnn, networkChanged);
		networkChanged = false;
		
		LossLayer* toEvaluate = nullptr;
		if (isTraining) toEvaluate = m_nceLayers[i];
		else toEvaluate = m_ceLayers[i];

		CudaMatrix* output = m_trainRnn->Forward(toEvaluate);
		error += m_trainRnn->GetLoss(toEvaluate);
		if (isTraining)
		{
			m_trainRnn->Backward(toEvaluate);
			m_trainRnn->Visit(v);
			m_updater->UpdateParameters(v);
			networkChanged = true; // trigger to clean the rnn
		}
	}
	
	return error;
}
