#include "FcgmTrainer.hpp"

void FcgmTrainer::PrintStatus(bool training) {
	clock_t now = clock();
	float entropy = (m_sharedError/m_sharedPredictions)/log(2);
	float perpl = pow(2, entropy);
	long long difftime = (now - m_startTime)/1000;

	std::cout << difftime/3600 <<":"<<(difftime % 3600)/60<<":"<<((difftime % 3600) % 60); 
	std::cout << " - Entr. / Perp. : " << std::setw(8) << entropy << " / " << perpl
		<< " P/s: " << std::setw(5) << (long long) (m_sharedPredictions/difftime);
	std::flush(std::cout);

	/** Print norms **/
	std::cout << std::setw(5);

	std::cout << " W_hh : " << std::setw(5) << Whh.Norm2();
	std::cout << " W_out : " << std::setw(5) << Wout.Norm2();
	std::cout << " W_in : " << std::setw(5) << Win.Norm2();
	std::cout << " H : " << std::setw(5) << H.Norm2();
	std::cout << " I : " << std::setw(5) << I.Norm2();	
	std::cout << std::endl;
}

void FcgmTrainer::DoJob(AbstractTrainer<Batch<Triple>, TripleBatchedReader>& parentCaller, 
	ThreadedQueue<Batch<Triple>>& workingQueue, bool isTraining) {

	long long loaded_seq = 0, seq_position = 0;
	Batch<Triple> wi;

	if (m_upd == nullptr)
	{
		m_upd = new AdagradUpdater(parentCaller.GetAlpha(), 1);

		if (m_beta != 0) {
			m_upd->add_pre_filter(&Whh, new L2RegularizeFilter(m_beta));
			m_upd->add_pre_filter(&Wout, new L2RegularizeFilter(m_beta));
			m_upd->add_pre_filter(&Win, new L2RegularizeFilter(m_beta));
			m_upd->add_pre_filter(&H, new L2RegularizeFilter(m_beta));
		}
	}

	Fcgm fcgm;
	fcgm.SetTau(m_tau);
	fcgm.SetExternalInWeights(&Win, &Whh, &Wout);
	fcgm.SetContextRnnWeights(&H);
	fcgm.SetExternalWeights(&I);
	fcgm.SetVocabulary(m_vocabulary);
	fcgm.SetUpdater(m_upd);
	fcgm.SetGenerator(m_generator);
	fcgm.SetTiled(&Whh);
	fcgm.SetTiled(&H);

	while (1) {
		wi = static_cast<Batch<Triple>&&>(workingQueue.PopFront());
		loaded_seq = wi.m_size;

		std::cerr << "Got " << loaded_seq << std::endl;
		// poison pill
		if (loaded_seq == 0) {
			break;
		}


		// Do the job
		int batchSentSize = (int) wi.data[0][2].size();
		parentCaller.UpdateError(fcgm.Run(wi, isTraining));
		parentCaller.UpdatePredictions(batchSentSize * wi.m_size);
		parentCaller.UpdateBatchesDone(1);

		if (seq_position >= loaded_seq) {
			loaded_seq = 0;
			seq_position = 0;
		}
	}
}

void FcgmTrainer::WriteModel() {
	std::string output;
	std::stringstream outDir;
	outDir << m_outputPath << "_fcgm_" << m_epochs 
		<< "_tau" << m_tau 
		<< "_a" << m_startAlpha
		<< "_b" << m_beta
		<< "_c" << m_nC
		<< "_in" << m_nHid 
		<< "_out" << m_nOut;

	output = outDir.str();

	std::cout << "Saving model parameters to " << output << std::endl;

	std::string make_dir = "mkdir " + output;
	system(make_dir.c_str());

	WriteErrors(output);
	WriteVocabulary(output);

	std::ofstream config(output + "/config");
	config << "act=" << m_act << std::endl;
	config << "alpha=" << m_alpha << std::endl;
	config << "beta=" << m_beta << std::endl;
	config.close();

	std::ofstream modelType(output + "/modelType");
	modelType << "FCGM" << std::endl;
	modelType.close();
	
	Win.WriteToExternalBinary(output + "/w_in.bin");
	Wout.WriteToExternalBinary(output + "/w_out.bin");
	Whh.WriteToExternalBinary(output + "/w_hh.bin");
	bhh.WriteToExternalBinary(output + "/b_hh.bin");
	bout.WriteToExternalBinary(output + "/b_out.bin");
	I.WriteToExternalBinary(output + "/i.bin");
	H.WriteToExternalBinary(output + "/h.bin");
}