
#pragma once

#include "AbstractTrainer.hpp"
#include "GradientUpdater.hpp"
#include "SequenceReader.hpp"
#include "LengthBatchedReader.hpp"
#include "Fcgm.hpp"
#include "print.hpp"
#include <chrono>

class FcgmTrainer : public AbstractTrainer<Batch<Triple>, TripleBatchedReader> {
protected:
	int m_tau;
	size_t m_nIn;
	size_t m_nHid;
	size_t m_nC;
	size_t m_nOut;

	Activation m_act;
	
	CudaMatrix H;
	CudaMatrix Win;
	CudaMatrix Wout;
	CudaMatrix Whh;
	
	CudaMatrix bhh;
	CudaMatrix bout;

private:
	CudaMatrix I;

protected:
	void InitializeRnnWeights(int nIn, int nHid, int nOut, long& seed) {
		m_nIn = nIn;
		m_nHid = nHid;
		m_nOut = nOut;

		Win.Resize(m_nIn, m_nHid);
		Win.SetName("W_in");
		
		Wout.Resize(m_nHid, m_nOut);
		Wout.SetName("W_out");
		

		float sigma = 0.1f;
		if (m_act == "tanh")
		{
			sigma = 0.01f;
		}
		
		seed = 7919;
		Win.SetSeed(seed);
		Win.GaussianRandomFill(0.f, 0.1f);
		
		seed += 7919;
		Wout.SetSeed(seed);
		Wout.GaussianRandomFill(0.f, 0.1f);
		
		seed += 7919;
		Whh.Resize(m_nHid, m_nHid);
		Whh.SetName("W_hh");
		Whh.SetSeed(seed);
		//Whh.OrthogonalFill(0.f, 0.1f); // NAACL
		Whh.GaussianRandomFill(0.f, 0.1f);
				
		bhh.Resize(1, m_nHid);
		bhh.SetName("b_hh");
		
		bout.Resize(1, m_nOut);
		bout.SetName("b_out");
	}

public:
	FcgmTrainer(const std::string& outputPath) : AbstractTrainer(outputPath) {
		m_act = SIGMOID;
		m_nC = 0;
		m_nIn = 0;
		m_nOut = 0;
		m_nHid = 0;
		m_tau = 7;
	}

	void InitializeFcgmWeights(int nIn, int nHid, int nOut, const std::string& ePath) { 
		long seed = 0;
		InitializeRnnWeights(nIn, nHid, nOut, seed);
		InitializeExternalEmbeddings(ePath);
	
		if (m_nC == 0)
			throw std::exception("Should initialize embeddings before that!");
		
		H.Resize(m_nC, m_nHid);
		
		seed++;
		H.SetSeed(seed);		
		H.SetName("H");
		H.GaussianRandomFill(0.f, 0.1f);
	}

	virtual void DoJob(AbstractTrainer<Batch<Triple>, TripleBatchedReader>& parentCaller, 
		ThreadedQueue<Batch<Triple>>& workingQueue,
		bool isTraining=false);
	virtual void WriteModel();

	void PrintStatus(bool training=true);

	~FcgmTrainer() {}

	void SetTau(int tau) { m_tau = tau; }	
	void SetActivation(const std::string& act) 
	{ 
		if (act == "sigmoid") m_act = SIGMOID; 
		else if (act == "tanh") m_act = TANH;
		else if (act == "relu") m_act = ReLU;
	}

private:
	void InitializeExternalEmbeddings(const std::string& ePath) {
		int founds = LoadFromExternalEmbeddings(ePath, *m_vocabulary, &I);
		I.SetName("W_emb");
		m_nC = I.GetRows();
		
		std::cerr << "External embeddings loaded, terms found = " << founds << std::endl;
	}
};