
#include "Rcgm.hpp"

/// Build the network for this triple 
/// (the time spent in building the network is likely to be 
/// irrelevant wrt the time passed in matrix multiplications).
void Rcgm::BuildNetwork() {
	m_trainRnn = new NeuralNet();
	m_trainRnn->SetBptt(m_Whh, m_tau);
	m_trainRnn->SetBarrierWeight(m_Wsin);
	
	m_nceLayers.clear();
	m_ceLayers.clear();
	
	BiasLayer* b_hh = m_trainRnn->AddLayer(new BiasLayer(m_batchSize, "b_hh"));
	BiasLayer* b_out = m_trainRnn->AddLayer(new BiasLayer(m_batchSize, "b_out"));
	BiasLayer* bs_hh = m_trainRnn->AddLayer(new BiasLayer(m_batchSize, "bs_hh"));

	m_biasLayers.push_back(b_hh);
	m_biasLayers.push_back(b_out);
	m_biasLayers.push_back(bs_hh);

	Layer* hiddenC = BuildRnn(nullptr, "c");
	Layer* hiddenM = BuildRnn(hiddenC, "m");
	Layer* hiddenR = BuildRnn(hiddenM, "r");

	m_trainRnn->Connect(hiddenC, hiddenM, new DynamicWeight(m_Wshh));
	m_trainRnn->Connect(hiddenM, hiddenR, new DynamicWeight(m_Wshh));

	m_trainRnn->Connect(bs_hh, hiddenM, new DynamicWeight(m_bshh));
	m_trainRnn->Connect(bs_hh, hiddenC, new DynamicWeight(m_bshh));
	m_trainRnn->Connect(bs_hh, hiddenR, new DynamicWeight(m_bshh));
	
	m_trainRnn->Optimize();
}

/// Builds the rnn for the sentence and returns a max-pooling 
/// layer.
Layer* Rcgm::BuildRnn(Layer* contextLayer, std::string seqName)
{
	HiddenLayer* lastHiddenLayer = nullptr;
	HiddenLayer* sentInputLayer = m_trainRnn->AddLayer(new HiddenLayer(SIGMOID, "hs_" + seqName));
	MaxPoolingLayer* poolingLayer = m_trainRnn->AddLayer(new MaxPoolingLayer("pool_" + seqName));
	
	m_trainRnn->Connect(poolingLayer, sentInputLayer, new DynamicWeight(m_Wsin));	
	
	for (int i = 0; i < 100; ++i)
	{
		std::string id = seqName + "_" + std::to_string(i+1);

		/// set current input
		SparseInputLayer* x_next = m_trainRnn->AddLayer(new SparseInputLayer("x_" + id));
		if (seqName == "c")
			m_cInputLayers.push_back(x_next);
		else if (seqName == "m")
			m_mInputLayers.push_back(x_next);
		else
			m_inputLayers.push_back(x_next);
		///

		HiddenLayer * h_next = m_trainRnn->AddLayer(new HiddenLayer(m_act, "h_" + id));
		if (lastHiddenLayer != nullptr) m_trainRnn->Connect(lastHiddenLayer, h_next, new DynamicWeight(m_Whh));
		if (contextLayer != nullptr) m_trainRnn->Connect(contextLayer, h_next, new DynamicWeight(m_H));
		
		/// Input connection
		m_trainRnn->Connect(x_next, h_next, new DynamicWeight(m_Win));
		m_trainRnn->Connect(m_trainRnn->GetLayerByName("b_hh"), h_next, new DynamicWeight(m_Bhh)); 

		/// Connect to pooling layer
		m_trainRnn->Connect(h_next, poolingLayer, new FixedWeight());

		/// Output layers
		if (seqName == "r")
		{
			NCELossLayer* yNce_t = m_trainRnn->AddLayer(new NCELossLayer(25, "pred_" + id));
			CrossEntropyLossLayer* yCe_t = m_trainRnn->AddLayer(new CrossEntropyLossLayer("pred_" + id));
			
			m_nceLayers.push_back(yNce_t);
			m_ceLayers.push_back(yCe_t);

			m_trainRnn->Connect(h_next, yNce_t, new DynamicWeight(m_Wout));
			m_trainRnn->Connect(h_next, yCe_t, new DynamicWeight(m_Wout));

			m_trainRnn->Connect(m_trainRnn->GetLayerByName("b_out"), yNce_t, new DynamicWeight(m_Bout));
			m_trainRnn->Connect(m_trainRnn->GetLayerByName("b_out"), yCe_t, new DynamicWeight(m_Bout));
		}

		lastHiddenLayer = h_next;
	}
	
	return sentInputLayer;
}

/// PrepareMatrices - we stock all the necessary information for the current sequence
///					  on the gpu.
/// [in] - subBatch, inputs and outputs (time ordered)
void Rcgm::PrepareMatrices(const Batch<Triple>& batch, bool isTraining)
{
	// Find the max length
	int batchSize = (int) batch.data.size();
	int sentLen = (int) batch.data[0][2].size();
	
	std::vector<int> ctxCInputRows; 
	std::vector<float> ctxCInputValues;
	std::vector<int> ctxCInputCols;
	
	std::vector<int> ctxMInputRows;
	std::vector<float> ctxMInputValues;
	std::vector<int> ctxMInputCols;

	std::vector<int> inputRows; 
	std::vector<float> inputValues;
	std::vector<int> inputCols;

	std::vector<int> outputRows;
	std::vector<int> outputCols;
	std::vector<float> outputValues;

	/// Detect the maximum length of the batch C & M
	int maxCLength = 0;
	int maxMLength = 0;
	for (int i = 0; i < batchSize; ++i)
	{
		if (batch.data[i][0].size() > maxCLength)
			maxCLength = batch.data[i][0].size();
		if (batch.data[i][1].size() > maxMLength)
			maxMLength = batch.data[i][1].size();
	}

	if (maxCLength > 100 || maxMLength > 100)
	{
		throw std::exception("Sentence too long!");
	}
	
	for (int j = 0; j < 100; j++)
	{
		for (int i = 0; i < batchSize; ++i)
		{
			const Triple& data = batch.data[i];
			const size_t cLen = data[0].size();
			const size_t mLen = data[1].size();
		
			if (j < maxCLength)
			{
				if (cLen > j)
				{
					ctxCInputRows.push_back(i);
					ctxCInputCols.push_back(data[0][j]);
					ctxCInputValues.push_back(1);				
				} 
				else
				{
					// Pad with null tokens
					ctxCInputRows.push_back(i);
					ctxCInputCols.push_back(m_vocabulary->eId);
					ctxCInputValues.push_back(1);				
				}
			}

			if (j < maxMLength)
			{
				if (mLen > j)
				{
					ctxMInputRows.push_back(i);
					ctxMInputCols.push_back(data[1][j]);
					ctxMInputValues.push_back(1);				
				} 
				else
				{
					// Pad with null tokens
					ctxMInputRows.push_back(i);
					ctxMInputCols.push_back(m_vocabulary->eId);
					ctxMInputValues.push_back(1);				
				}
			}

			if (j < sentLen - 1)
			{
				/// Everyone has the same length here
				inputRows.push_back(i);
				inputCols.push_back(data[2][j]);
				inputValues.push_back(1);
				
				if (!isTraining)
				{
					/// If we evaluate we only push the real output value
					outputRows.push_back(i);
					outputCols.push_back(data[2][j+1]);	
					outputValues.push_back(1);
				}
				else
				{
					std::vector<int> noise = m_generator->SampleNoise(25, data[2][j+1]);
					for (int x = 0; x < noise.size(); x++) 
					{
						float logPn = m_generator->GetLogProb(noise[x]);
						outputRows.push_back(i);
						outputCols.push_back(noise[x]);
						outputValues.push_back(logPn);
					}

					// Push the target
					outputRows.push_back(i);
					outputCols.push_back(data[2][j+1]);
					outputValues.push_back(-m_generator->GetLogProb(data[2][j+1]));
				}
			}
		}
	}
	
	int ctxMInputsNnz = (int) ctxMInputRows.size();	
	int ctxCInputsNnz = (int) ctxCInputRows.size();	
	int inputNnz = (int) inputRows.size();
	int outputNnz = (int) outputRows.size();

	/// We set the values
	m_ctxCInputs->SetData(1, 1, ctxCInputsNnz, ctxCInputRows.data(), ctxCInputCols.data(), ctxCInputValues.data());
	m_ctxInputs->SetData(1, 1, ctxMInputsNnz, ctxMInputRows.data(), ctxMInputCols.data(), ctxMInputValues.data());
	m_rnnInputs->SetData(1, 1, inputNnz, inputRows.data(), inputCols.data(), inputValues.data());
	m_rnnOutputs->SetData(1, 1, outputNnz, outputRows.data(), outputCols.data(), outputValues.data());
}

void Rcgm::SetCtxIO(int seqLen, int batchSize)
{
	// Just copy all of the data
	for (int i = 0; i < m_biasLayers.size(); i++)
		m_biasLayers[i]->SetInput(batchSize);

	int cLen = m_ctxCInputs->m_sparse->m_nz / batchSize;
	int mLen = m_ctxInputs->m_sparse->m_nz / batchSize;

	for (int i = 0; i < max(mLen, cLen); i++)
	{
		if (i < mLen)
			m_mInputLayers[i]->SetInput(batchSize, m_nIn, batchSize, m_ctxInputs->m_sparse->m_cooRow + batchSize*i,
				m_ctxInputs->m_sparse->m_cooCol + batchSize*i, m_ctxInputs->m_sparse->m_cooVal + batchSize*i, true, true);

		if (i < cLen)
			m_cInputLayers[i]->SetInput(batchSize, m_nIn, batchSize, m_ctxCInputs->m_sparse->m_cooRow + batchSize*i,
				m_ctxCInputs->m_sparse->m_cooCol + batchSize*i, m_ctxCInputs->m_sparse->m_cooVal + batchSize*i, true, true);
	}
}

void Rcgm::SetIO(int seqLen, int effectiveBatchSize, 
				const CudaMatrix& inputs, 
				const CudaMatrix& outputs,
				bool isTraining)
{
	const std::vector<InputLayer*>& I = m_trainRnn->Inputs();

	SparseCudaMatrix* outputSparse = outputs.GetSparseMatrix();
	SparseCudaMatrix* inputSparse = inputs.GetSparseMatrix();

	for (int i = 0; i < seqLen - 1; ++i) {
		int iBase = i*effectiveBatchSize;
		int oBase = 26*iBase;

		if (isTraining)
			m_nceLayers[i]->SetOutput(effectiveBatchSize, m_nOut, 26 * effectiveBatchSize, 
				outputSparse->m_cooRow + oBase,
				outputSparse->m_cooCol + oBase, 
				outputSparse->m_cooVal + oBase, nullptr, true, true);
		else
			m_ceLayers[i]->SetOutput(effectiveBatchSize, m_nOut, effectiveBatchSize, 
				outputSparse->m_cooRow + iBase,
				outputSparse->m_cooCol + iBase,
				outputSparse->m_cooVal + iBase, true, true);
		
		m_inputLayers[i]->SetInput(effectiveBatchSize, m_nIn, effectiveBatchSize, 
			inputSparse->m_cooRow + iBase,
			inputSparse->m_cooCol + iBase, 
			inputSparse->m_cooVal + iBase, true, true);
	}
}

float Rcgm::Run(const Batch<Triple>& batch, bool isTraining) {	
	float error = 0;
	bool networkChanged = true;
	
	int batchSize = (int) batch.data.size();
	int sentLen = (int) batch.data[0][2].size();
	
	if (m_trainRnn == nullptr) {
		BuildNetwork();
	}
	
	PrepareMatrices(batch, isTraining);	
	SetIO(sentLen, batchSize, *m_rnnInputs, *m_rnnOutputs, isTraining);
	SetCtxIO(sentLen, batchSize);

	// Run through the RNN
	for (int i = 0; i < sentLen - 1; ++i) 
	{
		CleanNetwork(*m_trainRnn, networkChanged);
		networkChanged = false;
		
		LossLayer* toEvaluate = nullptr;
		if (isTraining) toEvaluate = m_nceLayers[i];
		else toEvaluate = m_ceLayers[i];

		CudaMatrix* output = m_trainRnn->Forward(toEvaluate);
		error += m_trainRnn->GetLoss(toEvaluate);
		if (isTraining)
		{
			m_trainRnn->Backward(toEvaluate);
			m_trainRnn->Visit(v);
			m_updater->UpdateParameters(v);
			networkChanged = true; // trigger to clean the rnn
		}
	}
	
	return error;
}
