#include "RcgmTrainer.hpp"

void RcgmTrainer::PrintStatus(bool training) {
	clock_t now = clock();
	double entropy = (m_sharedError/m_sharedPredictions)/log(2);
	double perpl = pow(2, entropy);
	long long difftime = (now - m_startTime)/1000;

	std::cout << difftime/3600 <<":"<<(difftime % 3600)/60<<":"<<((difftime % 3600) % 60); 
	std::cout << " - Entr. / Perp. : " << std::setw(8) << entropy << " / " << perpl
		<< " P/s: " << std::setw(5) << (long long) (m_sharedPredictions/difftime);
	std::flush(std::cout);

	/** Print norms **/
	std::cout << std::setw(5);

	std::cout << " W_hh : " << std::setw(5) << Whh.Norm2();
	std::cout << " W_out : " << std::setw(5) << Wout.Norm2();
	std::cout << " W_in : " << std::setw(5) << Win.Norm2();

	std::cout << " WS_in : " << std::setw(5) << WSin.Norm2();
	std::cout << " WS_hh : " << std::setw(5) << WShh.Norm2();
	std::cout << " b_hh : " << std::setw(5) << bhh.Norm2();
	std::cout << " b_out : " << std::setw(5) << bout.Norm2();
	std::cout << " bs_hh : " << std::setw(5) << bshh.Norm2();
	std::cout << " H : " << std::setw(5) << H.Norm2();

	std::cout << std::endl;
}

void RcgmTrainer::DoJob(AbstractTrainer<Batch<Triple>, TripleBatchedReader>& parentCaller, ThreadedQueue<Batch<Triple>>& workingQueue,
	bool isTraining) {
		long long loaded_seq = 0, seq_position = 0;
		Batch<Triple> wi;

		m_upd = new AdagradUpdater(parentCaller.GetAlpha(), 1);

		Rcgm rcgm;
		rcgm.SetVocabulary(m_vocabulary);
		rcgm.SetAct(m_act);
		rcgm.SetTau(m_tau);
		rcgm.SetExternalInWeights(&Win, &Whh, &Wout);
		rcgm.SetContextRnnWeights(&H);
		rcgm.SetExternalBias(&bhh, &bout);
		rcgm.SetSentenceRnnWeights(&WShh, &WSin, &bshh);
		rcgm.SetUpdater(m_upd);
		rcgm.SetGenerator(m_generator);

		// Some regularization for the parameters!
		if (m_beta != 0) {
			m_upd->add_pre_filter(&Whh, new L2RegularizeFilter(m_beta));
			m_upd->add_pre_filter(&Wout, new L2RegularizeFilter(m_beta));
			m_upd->add_pre_filter(&Win, new L2RegularizeFilter(m_beta));
			m_upd->add_pre_filter(&WSin, new L2RegularizeFilter(m_beta));
			m_upd->add_pre_filter(&WShh, new L2RegularizeFilter(m_beta));
			m_upd->add_pre_filter(&bshh, new L2RegularizeFilter(m_beta));
			m_upd->add_pre_filter(&bhh, new L2RegularizeFilter(m_beta));
			m_upd->add_pre_filter(&bout, new L2RegularizeFilter(m_beta));
			m_upd->add_pre_filter(&H, new L2RegularizeFilter(m_beta));
		}

		while (1) {
			// GetElement job
			wi = static_cast<Batch<Triple>&&>(workingQueue.PopFront());
			loaded_seq = wi.m_size;

			// poison pill
			if (loaded_seq == 0) {
				// Exit here
				break;
			}

			// Prepare for current batch
			int batchSentSize = (int) wi.data[0][2].size();
			m_upd->mb_ = wi.m_size;

			// Do the job
			try {
				parentCaller.UpdateError(rcgm.Run(wi, isTraining));
				parentCaller.UpdatePredictions(batchSentSize * wi.m_size);
				parentCaller.UpdateBatchesDone(1);
			} catch (std::exception e) 
			{
				std::cerr << e.what() << std::endl;
				exit(-1);
			}
		}
}

void RcgmTrainer::WriteModel() {
	std::string output;
	std::stringstream outDir;
	outDir << m_outputPath << "_rcgm_" << m_epochs 
		<< "_tau" << m_tau 
		<< "_act" << m_act
		<< "_a" << m_startAlpha
		<< "_b" << m_beta
		<< "_in" << m_nHid 
		<< "_out" << m_nOut;

	output = outDir.str();

	std::cout << "Saving model parameters to " << output << std::endl;

	std::string make_dir = "mkdir " + output;
	system(make_dir.c_str());

	WriteVocabulary(output);
	WriteErrors(output);

	std::ofstream modelType(output + "/modelType");
	modelType << "RCGM" << std::endl;
	modelType.close();

	std::ofstream config(output + "/config");
	config << "act=" << m_act << std::endl;
	config << "alpha=" << m_alpha << std::endl;
	config << "beta=" << m_beta << std::endl;
	config.close();

	Win.WriteToExternalBinary(output + "/w_in.bin");
	Wout.WriteToExternalBinary(output + "/w_out.bin");
	Whh.WriteToExternalBinary(output + "/w_hh.bin");
	bhh.WriteToExternalBinary(output + "/b_hh.bin");
	bout.WriteToExternalBinary(output + "/b_out.bin");
	bshh.WriteToExternalBinary(output + "/bs_hh.bin");
	WSin.WriteToExternalBinary(output + "/ws_in.bin");
	WShh.WriteToExternalBinary(output + "/ws_hh.bin");
	H.WriteToExternalBinary(output + "/h.bin");
}