#pragma once

#include "Vocabulary.hpp"
#include "Random.hpp"
#include "NnetLib.hpp"
#include "GradientVisitor.hpp"
#include "GradientUpdater.hpp"
#include "AbstractReader.hpp"
#include "AbstractModel.hpp"

class Rnn : public AbstractModel {
protected:
	CudaMatrix* m_Win;
	CudaMatrix* m_Wout;
	CudaMatrix* m_Whh;
	CudaMatrix* m_Bout;
	CudaMatrix* m_Bhh;
	CudaMatrix* m_dIn;
	CudaMatrix* m_dOut;
	
	Activation m_act;

	/// The network object for training
	NeuralNet* m_trainRnn;

	/// Max length of the RNN
	size_t m_maxLength;
	size_t m_batchSize;
	///

	size_t m_nIn;
	size_t m_nOut;
	size_t m_nHid;
	int m_tau;
	
	CudaMatrix* m_rnnInputs;
	CudaMatrix* m_rnnOutputs;

	std::vector<BiasLayer*> m_biasLayers;
	std::vector<SparseInputLayer*> m_inputLayers;
	std::vector<CrossEntropyLossLayer*> m_ceLayers;
	std::vector<NCELossLayer*> m_nceLayers;
	
	virtual void BuildNetwork();
	
	void DeleteRnnWeights() {
		if (m_Bhh != nullptr) delete m_Bhh;
		if (m_Bout != nullptr) delete m_Bout;
		if (m_Whh != nullptr) delete m_Whh;
		if (m_Wout != nullptr) delete m_Wout;
		if (m_Win != nullptr) delete m_Win;
	}

public:
	void CleanNetwork(NeuralNet& rnn, bool cleanError=false);
	void AddStep(const Sequence& sequence, size_t step, NeuralNet& rnn, bool training=true);

public:
	Rnn();
	Rnn(bool proprietaryMemory);
	Rnn(int nHid, int nOut);
	
	~Rnn(void) {
		if (m_trainRnn != nullptr) delete m_trainRnn;
		if (m_rnnInputs != nullptr) delete m_rnnInputs;
		if (m_rnnOutputs != nullptr) delete m_rnnOutputs;
		if (m_proprietaryMemory) 
			DeleteRnnWeights();
	}

	void SetTiled(CudaMatrix* w) {
		// tiled updates for hidden weights
		v.__create_tiled(w, m_tau + 1);
	}

	void SetTau(int tau) { m_tau = tau; }
	void SetExternalInWeights(CudaMatrix* in, CudaMatrix* hh, CudaMatrix* out) {
		m_Win = in;
		m_Wout = out;
		m_Whh = hh;
		
		m_nIn = in->GetRows();
		m_nOut = out->GetCols();
		m_nHid = hh->GetRows();
	};

	Activation GetAct(){
		return m_act;
	}
	
	void SetExternalDeepConnections(CudaMatrix* in, CudaMatrix* out) {
		NOT_IMPLEMENTED;
	};

	void SetExternalBias(CudaMatrix* bhh, CudaMatrix* bout) {
		m_Bhh = bhh;
		m_Bout = bout;
	};
	
	void SetBatchSize(int batchSize) { m_batchSize = batchSize; }
	float Run(const Batch<Sequence>& seq, bool isTraining=true);
	
	// The following are not supported for the moment
	float Evaluate(const Sequence& seq, bool stoch=false);
	float Evaluate(const Sequence& ctx, const Sequence& seq, bool stoch=false);
	
	Sequence Generate(const Sequence& previous);
	static Rnn* LoadFromFile(const std::string& dirPath);
	
	void DeleteWeights() {
		if (m_Whh != nullptr) delete m_Whh;
		if (m_Wout != nullptr) delete m_Wout;
		if (m_Win != nullptr) delete m_Win;
	}

	void SetGenerator(RandomGenerator* gen) {
		m_generator = gen;
	}
	void SetUpdater(IGradientUpdater* upd) {
		m_updater = upd;
	}
	void SetAct(Activation fn) {
		m_act = fn;
	}

	void PrepareMatrices(const Batch<Sequence>& batch, bool isTraining);

protected:
	void SetIO(int seqLen, int batchSize, 
		const CudaMatrix& inputs,
		const CudaMatrix& outputs,
		bool training=true);
};