#include "RnnTrainer.hpp"

void RnnTrainer::PrintStatus(bool training) {
	clock_t now = clock();
	float entropy = (m_sharedError/m_sharedPredictions)/log(2);
	float perpl = pow(2, entropy);
	long long difftime = (now - m_startTime)/1000;

	std::cout << difftime/3600 <<":"<<(difftime % 3600)/60<<":"<<((difftime % 3600) % 60); 
	std::cout << " - Entr. / Perp. : " << std::setw(8) << entropy << " / " << perpl
		<< " P/s: " << std::setw(5) << (long long) (m_sharedPredictions/difftime);
	std::flush(std::cout);

	/** Print norms **/
	std::cout << std::setw(5);

	std::cout << " W_hh : " << std::setw(5) << Whh.Norm2();
	std::cout << " W_out : " << std::setw(5) << Wout.Norm2();
	std::cout << " W_in : " << std::setw(5) << Win.Norm2();
	std::cout << " b_out : " << std::setw(5) << bout.Norm2();
	std::cout << " b_hh : " << std::setw(5) << bhh.Norm2();

	std::cout << std::endl;
}

/// PrepareMatrices - we stock all the necessary information for the current sequence
///					  on the gpu.
/// [in] - subBatch, inputs and outputs (time ordered)
void RnnTrainer::PrepareMatrices(int sentLen, 
								 const Batch<Sequence>& batch, 
								 CudaMatrix& inputs, CudaMatrix& outputs, bool isTraining)
{
	// Find the max length
	int batchSize = (int) batch.data.size();

	std::vector<int> inputRows; inputRows.reserve(sentLen*batchSize); 
	std::vector<float> inputValues; inputValues.reserve(sentLen*batchSize);
	std::vector<int> inputCols; inputCols.reserve(sentLen*batchSize);

	std::vector<int> outputRows; outputRows.reserve(sentLen*batchSize*26);
	std::vector<int> outputCols; outputCols.reserve(sentLen*batchSize*26);
	std::vector<float> outputValues; outputValues.reserve(sentLen*batchSize*26);

	for (int j = 0; j < sentLen - 1; j++)
	{
		for (int i = 0; i < batchSize; ++i) {
			const Sequence& seq = batch.data[i];

			int input = seq[j];
			int output = seq[j+1];

			// [token_id, batch_n] = 1
			inputRows.push_back(i);
			inputCols.push_back(input);
			inputValues.push_back(1);

			if (!isTraining)
			{
				/// If we evaluate we only push the real output value
				outputRows.push_back(i);	
				outputCols.push_back(output);
				outputValues.push_back(1);
			}
			else
			{
				std::vector<int> noise = m_generator->SampleNoise(25, output);
				for (int x = 0; x < noise.size(); x++) 
				{
					float logPn = m_generator->GetLogProb(noise[x]);

					outputRows.push_back(i);
					outputCols.push_back(noise[x]);
					outputValues.push_back(logPn);
				}

				// Push the target
				outputRows.push_back(i);
				outputCols.push_back(output);
				outputValues.push_back(-m_generator->GetLogProb(output));
			}
		}
	}

	int inputNnz = (int) inputRows.size();
	int outputNnz = (int) outputRows.size();

	/// We set the values
	inputs.SetData(1, 1, inputNnz, inputRows.data(), inputCols.data(), inputValues.data());
	outputs.SetData(1, 1, outputNnz, outputRows.data(), outputCols.data(), outputValues.data());
}

// This method should be re-implemented by the children workers!
void RnnTrainer::DoJob(AbstractTrainer<Batch<Sequence>, SequenceBatchedReader>& parentCaller, 
	ThreadedQueue<Batch<Sequence>>& workingQueue, bool isTraining) {
	long long loaded_seq = 0, seq_position = 0;
	Batch<Sequence> wi;

	if (m_upd==nullptr) {
		m_upd = new AdagradUpdater(parentCaller.GetAlpha(), 1);

		if (m_beta != 0) 
		{
			m_upd->add_pre_filter(&Whh, new L2RegularizeFilter(m_beta));
			m_upd->add_pre_filter(&Wout, new L2RegularizeFilter(m_beta));
			m_upd->add_pre_filter(&Win, new L2RegularizeFilter(m_beta));
		}
	}

	Rnn rnn;
	rnn.SetTau(m_tau);
	rnn.SetExternalInWeights(&Win, &Whh, &Wout);
	rnn.SetExternalBias(&bhh, &bout);
	rnn.SetUpdater(m_upd);
	rnn.SetBatchSize(BUNCH_SIZE);
	rnn.SetGenerator(m_generator);
	rnn.SetAct(m_act);

	while (1) {
		// GetElement job
		wi = static_cast<Batch<Sequence>&&>(workingQueue.PopFront());
		loaded_seq = wi.m_size;

		// poison pill
		if (loaded_seq == 0) {
			// Exit here
			break;
		}

		// Prepare for current batch
		int batchSentSize = (int) wi.data[0].size();
		m_upd->mb_ = wi.m_size;

		// Do the job
		parentCaller.UpdateError(rnn.Run(wi, isTraining));
		parentCaller.UpdatePredictions(batchSentSize * wi.m_size);
		parentCaller.UpdateBatchesDone(1);
	}
}

void RnnTrainer::WriteModel() {
	std::string output;
	std::stringstream outDir;
	outDir << m_outputPath << "_rnn_" << m_epochs 
		<< "_tau" << m_tau 
		<< "_a" << m_startAlpha
		<< "_b" << m_beta
		<< "_hid" << m_nHid 
		<< "_out" << m_nOut;

	output = outDir.str();

	std::cout << "Saving model parameters to " << output << std::endl;

	std::string make_dir = "mkdir " + output;
	system(make_dir.c_str());

	WriteErrors(output);
	WriteVocabulary(output);				

	std::ofstream config(output + "/config");
	config << "act=" << m_act << std::endl;
	config << "alpha=" << m_alpha << std::endl;
	config << "beta=" << m_beta << std::endl;
	config.close();

	std::ofstream modelType(output + "/modelType");
	modelType << "RNN" << std::endl;
	modelType.close();

	Win.WriteToExternalBinary(output + "/w_in.bin");
	Wout.WriteToExternalBinary(output + "/w_out.bin");
	Whh.WriteToExternalBinary(output + "/w_hh.bin");
	bhh.WriteToExternalBinary(output + "/b_hh.bin");
	bout.WriteToExternalBinary(output + "/b_out.bin");
}