//
//  vocabulary.h
//  Rnn
//
//  Created by Alessandro Sordoni on 3/11/2014.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//

#pragma once

#include <vector>
#include <unordered_map>
#include <sstream>
#include <fstream>
#include <algorithm>
#include "Utils.hpp"
#include "DataStructures.hpp"

struct vocab_word {
    int id;
    std::string word;
    long long cn;
	float idf;
    vocab_word(int id,
               std::string& word,
               long long cn,
			   float idf) :
    id(id), word(word), cn(cn), idf(idf) {}
};

class Vocabulary {
public:
	long long train_words;
	long long vocab_size;
	int sId;
	int eId;
	int uId;
	
	std::vector<vocab_word> vocab;
	std::unordered_map<std::string, int> inv_vocab;
	Vocabulary() : train_words(0), vocab_size(0), sId(0), eId(1), hasUnk(false) {};
	Vocabulary(const std::string& path) : train_words(0), vocab_size(0), sId(0), eId(1), hasUnk(false) {
		LoadFromFile(path);
	};

public:
	bool hasUnk;
	 
	void LoadFromFile(const std::string& vocab_path) {
		if (!check_exists(vocab_path)) {
			std::cerr << vocab_path << " : Vocabulary file not found!" << std::endl;
			exit(-1);
		}

		std::ifstream vocif((vocab_path).c_str());
		std::string line;
    
		// Whitespace separated tuples of the form:
		while (std::getline(vocif, line)) {
			std::vector<std::string> strs = Tokenize(line);
			int id = (int) vocab.size();
			std::string& word = strs[0];
			long long cn = _atoi64(strs[1].c_str());
			float idf = 0;
			if (strs.size() > 2)
				idf = atof(strs[2].c_str());
			vocab_word vw(id, word, cn, idf);
        
			train_words += vw.cn;
			vocab.push_back(vw);
			inv_vocab.insert(std::make_pair(vw.word, vw.id));
		}
    
		if (inv_vocab.find("<s>") == inv_vocab.end() 
			|| inv_vocab.find("</s>") == inv_vocab.end())
			throw std::exception("This dictionary doesn't contain SE tokens!");
		else {
			sId = inv_vocab["<s>"];
			eId = inv_vocab["</s>"];
		}

		if (inv_vocab.find("<unk>") != inv_vocab.end())
		{
			uId = inv_vocab["<unk>"];
			hasUnk = true;
		}
		
		vocif.close();
		vocab_size = vocab.size();
	}

public:
	static void Load(Vocabulary& v, const std::string& vocab_path) {
		v.LoadFromFile(vocab_path);
	}
	
	std::vector<int> ConvertToIds(const std::string& line, bool addSE=true) {
		std::vector<int> ids;

		std::vector<std::string> tokens = Tokenize(line);
		std::for_each(tokens.begin(), tokens.end(), [&](const std::string& x) {
			int id = to_id(x);
			if (id != -1)
				ids.push_back(id);
		});

		if (addSE) {
			ids.insert(ids.begin(), sId);
			ids.push_back(eId);
		}

		return ids;
	}
	
	std::string ConvertToString(std::vector<int> ids) {
		std::string line;
		std::for_each(ids.begin(), ids.end(), [&](int x) {
			std::string s = to_word(x);
			line += s + " ";
		});
		return line.substr(0, line.size()-1);
	}
	
	void WriteToFile(std::string fileName)
	{
		std::ofstream w(fileName);
		for (auto it = vocab.begin(); it != vocab.end(); it++)
		{
			w << it->word << " " << it->cn << " " << it->idf << std::endl;
		}
		w.close();
	}

	float GetIdf(int id)
	{
		if (vocab[id].idf != 0)
			return vocab[id].idf;
		else
			throw std::exception("Idf not found in vocabulary!");
	}
	
	int to_id(const std::string& word) {
		if (inv_vocab.find(word) != inv_vocab.end())
			return inv_vocab[word];
		else if (hasUnk)
			return uId;
		return -1;
	}

	std::string to_word(int id) {
		if (id >= vocab_size) throw new std::exception("Vocabulary_to_word: index out of range");
		return vocab[id].word;
	}

	int GetSId() const { return sId; }
	int GetEId() const { return eId; }
};