//
//  CudaMatrix.cu
//  CudaMatrix
//
//  Created by Alessandro Sordoni on 2014-07-29.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//

#include "CudaMatrix.hpp"

extern bool sync;

CudaMatrix::CudaMatrix() 
{
    m_matrixType = DENSE;
    m_dense = new DenseCudaMatrix();
    m_sparse = NULL;
}

CudaMatrix::CudaMatrix(size_t rows, size_t cols, MatrixType m) 
{
    m_matrixType = m;
	m_sparse = NULL;
	m_dense = NULL;
	if (m == MatrixType::DENSE)
	{
		m_dense = new DenseCudaMatrix();
		m_dense->Resize(rows, cols);
	}
	else
	{
		m_sparse = new SparseCudaMatrix();
		m_sparse->Resize(rows, cols);
	}
}

void CudaMatrix::Resize(size_t rows, size_t cols, bool noFill)
{
    if (IsDense())
    {
        m_dense->Resize(rows, cols, noFill);
    } 
    else
    {
        m_sparse->Resize(rows, cols, noFill);
    }
}

void CudaMatrix::SetMatrixType(MatrixType matrixType)
{
    if (matrixType != m_matrixType)
    {
        if (matrixType == DENSE)
        {
            if (m_sparse != NULL)
            {
                delete m_sparse;
                m_sparse = NULL;
            }
            
            m_dense = new DenseCudaMatrix();
        } else 
        {
            if (m_dense != NULL)
            {
                delete m_dense;
                m_dense = NULL;
            }
            
            m_sparse = new SparseCudaMatrix();
        }
    
        m_matrixType = matrixType;
    }
}

void CudaMatrix::Dot(CudaMatrix& A, CudaMatrix& B, CudaMatrix& C, bool at, bool bt)
{	
	if (A.IsDense())
    {
        if (B.IsDense()) 
        {
            if (C.IsDense())
                A.m_dense->Dot(*(B.m_dense), *(C.m_dense), at, bt);
            else if (C.IsSparse())
                A.m_dense->Dot(*(B.m_dense), *(C.m_sparse), at, bt);
        } 
        else 
        {
            if (C.IsDense())
                A.m_dense->Dot(*B.m_sparse, *C.m_dense, at, bt);
            else if (C.IsSparse())
                A.m_dense->Dot(*B.m_sparse, *C.m_sparse, at, bt);
        }
    } else if (A.IsSparse())
    {
        if (B.IsSparse())
            throw std::exception();
		else if (C.IsDense())
			A.m_sparse->Dot(*B.m_dense, *C.m_dense, at, bt);
		else if (C.IsSparse())
			A.m_sparse->Dot(*B.m_dense, *C.m_sparse, at, bt);
	}
}

void CudaMatrix::MaxPoolingFwd(size_t poolingChannels, size_t channelRows, size_t channelCols, 
		float* poolingAct, float* channelAct, float* channelErrorMask)
{
	size_t channelSize = channelRows*channelCols;
    int numBlocks = (int) ceil(1.0f*channelSize/THREADS_PER_BLOCK);
	cudaEvent_t done;       
	if (sync) CUDA_CALL(cudaEventCreate(&done));

    /// Execute kernel
	GpuMaxPooling<<<numBlocks, THREADS_PER_BLOCK>>>(poolingChannels, channelSize, poolingAct, channelAct, channelErrorMask);
	
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void CudaMatrix::SetData(size_t rows,
    size_t cols, int nz, int* cooRow, int* cooCol, float* cooVal, bool onDevice, bool noCopy)
{
    if (m_matrixType == SPARSE)
        m_sparse->SetData(rows, cols, nz, cooRow, cooCol, cooVal, onDevice, noCopy);
}

void CudaMatrix::SetData(size_t rows, size_t cols, float* data, bool onDevice, bool noCopy)
{
    if (m_matrixType == DENSE)
        m_dense->SetData(rows, cols, data, onDevice);
}

void CudaMatrix::EwiseProduct(CudaMatrix& A, CudaMatrix& B, CudaMatrix& C)
{
	if (!A.IsDense()  || !B.IsDense() || !C.IsDense())
		NOT_IMPLEMENTED;
	A.m_dense->EwiseProduct(*B.m_dense, *C.m_dense);
}

void
CudaMatrix::Add(CudaMatrix& A, CudaMatrix& B, float f)
{
    if (A.IsDense())
    {
		if (B.IsDense())
			A.m_dense->AddTo(*B.m_dense, f);
		else
			A.m_dense->AddTo(*B.m_sparse, f);
	} 
	else
	{
		if (B.IsDense())
			A.m_sparse->AddTo(*B.m_dense, f);
		if (B.IsSparse())
			NOT_IMPLEMENTED;
	}
}

void
CudaMatrix::AssignCE(const DenseCudaMatrix& A, SparseCudaMatrix& T, DenseCudaMatrix& E)
{
    /// ATTENTION HERE FOR THE NUMBER OF BLOCKS!
	E.Resize(A.m_rows, A.m_cols);
	int numBlocks = (int) A.m_rows;
	
	cudaEvent_t done;       
	if (sync) CUDA_CALL(cudaEventCreate(&done));

    /// Execute kernel
	assert(T.m_nz == A.m_rows);
	GpuAssignCE<<<numBlocks, THREADS_PER_BLOCK>>>(A.m_rows, A.m_cols, A.m_dData, E.m_dData, T.m_cooCol);
	
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void
CudaMatrix::AssignCELoss(const DenseCudaMatrix& A, SparseCudaMatrix& T, float* loss)
{
	cudaEvent_t done;       
	if (sync) CUDA_CALL(cudaEventCreate(&done));

	float* dLoss = NULL;
	cudaMalloc(&dLoss, sizeof(float));
    
    int numBlocks = 1;
	GpuAssignCELoss<<<numBlocks, THREADS_PER_BLOCK>>>(A.m_rows, A.m_cols, A.m_dData, T.m_cooCol, dLoss);
	
	cudaMemcpy(loss, dLoss, sizeof(float), cudaMemcpyDeviceToHost);
	cudaFree(dLoss);
	
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void CudaMatrix::Fill(float value) 
{
	if (IsDense()) 
	{
		m_dense->Fill(value);
	} 
	else
	{
		m_sparse->Fill(value);
	}
}

void 
CudaMatrix::AssignNCE(int nnoise,
                      const SparseCudaMatrix& A,
                      const SparseCudaMatrix& P,
                      SparseCudaMatrix& E) {

	E.SetData(A.m_rows, A.m_cols, A.m_nz, A.m_cooRow, A.m_cooCol, A.m_cooVal, true);
	E.Reset();
	
	cudaEvent_t done;       
	if (sync) CUDA_CALL(cudaEventCreate(&done));
    int numBlocks = (int) ceil(1.0*A.m_nz/THREADS_PER_BLOCK);
	
	GpuAssignNCE<<<numBlocks,THREADS_PER_BLOCK>>>(nnoise,
                 A.m_nz,
                 A.m_cooVal,
                 P.m_cooVal,
                 E.m_cooVal);
	
	/// Scale by the minibatch
	// E.InPlaceScale(1.0/A.m_rows);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void 
CudaMatrix::AssignNCELoss(int nnoise,
                          const SparseCudaMatrix& A,
                          const SparseCudaMatrix& P,
                          float* loss) 
{
	cudaEvent_t done;       
	if (sync) CUDA_CALL(cudaEventCreate(&done));

	float* dLoss = NULL;
	cudaMalloc(&dLoss, sizeof(float));
    
    int numBlocks = 1;
    GpuAssignNCELoss<<<numBlocks,THREADS_PER_BLOCK>>>(nnoise,
                 A.m_nz,
                 A.m_cooVal,
                 P.m_cooVal,
                 dLoss);
	
	cudaMemcpy(loss, dLoss, sizeof(float), cudaMemcpyDeviceToHost);
	cudaFree(dLoss);
	
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void CudaMatrix::AdagradUpdate(DenseCudaMatrix& Param, DenseCudaMatrix& Grad, DenseCudaMatrix& Gsq, 
							   float beta, float mb)
{
	int blocksPerGrid = (int) ceil(1.0f*Param.m_size/THREADS_PER_BLOCK);
	
	cudaEvent_t done;
	if (sync) CUDA_CALL(cudaEventCreate(&done));
	GpuAdagrad<<<blocksPerGrid, THREADS_PER_BLOCK>>>(
            Gsq.m_dData, 
            Grad.m_dData, 
            Param.m_dData, 
            beta, mb,
            Param.m_size);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void CudaMatrix::AdadeltaUpdate(DenseCudaMatrix& Gsq, DenseCudaMatrix& Xsq, 
								DenseCudaMatrix& Grad, DenseCudaMatrix& Param, 
								float beta, float mb)
{
	int blocksPerGrid = ceil(1.0*Param.m_size/THREADS_PER_BLOCK);
	
	cudaEvent_t done;
	if (sync) CUDA_CALL(cudaEventCreate(&done));
	GpuAdadelta<<<blocksPerGrid, THREADS_PER_BLOCK>>>(
            Gsq.m_dData, 
            Xsq.m_dData,
			Grad.m_dData, 
            Param.m_dData, 
            beta, mb,
            Param.m_size);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}