//
//  CudaMatrix.hpp
//  CudaMatrix
//
//  Created by Alessandro Sordoni on 2014-07-01.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//

#pragma once

#include <iostream>
#include "CudaOps.hpp"
#include "DenseCudaMatrix.hpp"
#include "SparseCudaMatrix.hpp"


typedef enum { DENSE, SPARSE } MatrixType;

class CudaMatrix {
public:
    MatrixType m_matrixType;
    DenseCudaMatrix* m_dense;
    SparseCudaMatrix* m_sparse;
    
public:
    CudaMatrix();
    CudaMatrix(size_t rows, size_t cols, MatrixType m);
    ~CudaMatrix() {
        if (m_dense) delete m_dense;
        if (m_sparse) delete m_sparse;
    }
    
    void SetMatrixType(MatrixType matrixType);
    void Resize(size_t rows, size_t cols, bool noFill=false);
    void Reset() { if (IsDense()) m_dense->Reset(); else m_sparse->Reset(); }
	void SetSeed(long seed) { if (IsDense()) m_dense->SetSeed(seed); }

	void Fill(float value);
	void UniformRandomFill(float min, float max) { 
		if (IsDense()) m_dense->UniformRandomFill(min, max); }
	void GaussianRandomFill(float mean, float sigma) { 
		if (IsDense()) m_dense->GaussianRandomFill(mean, sigma); }
	void SparseInit(int nb) { if (IsDense()) m_dense->SparseInit(nb); }
	void OrthogonalFill(float mean, float sigma) { if (IsDense()) m_dense->OrthogonalFill(mean, sigma); }

    void SetData(size_t rows, size_t cols, int nz, int* cooRow, int* cooCol, float* cooVal, bool onDevice=false, bool noCopy=false);
    void SetData(size_t rows, size_t cols, float* data, bool onDevice=false, bool noCopy=false);
    
    size_t GetRows() { if (IsDense()) return m_dense->m_rows; return m_sparse->m_rows; }
    size_t GetCols() { if (IsDense()) return m_dense->m_cols; return m_sparse->m_cols; }
    size_t GetSize() { if (IsDense()) return m_dense->m_size; return m_sparse->m_size; }
    std::string GetName() { if (IsDense()) return m_dense->m_name; return m_sparse->m_name; }
    void SetName(std::string name) { if (IsDense()) m_dense->SetName(name); else m_sparse->SetName(name); }
    
    bool IsDense() const { return m_matrixType==DENSE; }
    bool IsSparse() const { return m_matrixType==SPARSE; }
    
	void CopyAtPosition(const CudaMatrix& B, const long sourceSize, long sourceCol=0, long targetCol=0) {
		if (IsDense() && B.IsDense()) m_dense->CopyAtPosition(*(B.GetDenseMatrix()), sourceSize, sourceCol, targetCol);
		else { NOT_IMPLEMENTED; }
	}

    DenseCudaMatrix* GetDenseMatrix() const { return m_dense; }
    DenseCudaMatrix* GetDenseMatrixForUpdate() { return m_dense; }
    SparseCudaMatrix* GetSparseMatrix() const { return m_sparse; }
    SparseCudaMatrix* GetSparseMatrixForUpdate() { return m_sparse; }
    
	void InPlaceTranspose() { 
		if (IsDense()) {
			DenseCudaMatrix Ct = m_dense->Transpose();
			*m_dense = Ct;
		}
	}

	/// Activation 
    void InPlaceSigmoid() { if (IsDense()) m_dense->InPlaceSigmoid(); else m_sparse->InPlaceSigmoid(); }
    void InPlaceSoftmax() { if (IsDense()) m_dense->InPlaceSoftmax(); /* TODO : add sparse here */ }
    void InPlaceSigmoidDerivative(const CudaMatrix& C) {
		if (IsDense()) m_dense->InPlaceSigmoidDerivative(*(C.GetDenseMatrix()));
		else m_sparse->InPlaceSigmoidDerivative(*(C.GetSparseMatrix()));
    }
    void InPlaceRelu() { if (IsDense()) m_dense->InPlaceRelu(); else NOT_IMPLEMENTED; }
    void InPlaceReluDerivative(const CudaMatrix& C) {
		if (IsDense()) m_dense->InPlaceReluDerivative(*(C.GetDenseMatrix()));
		else NOT_IMPLEMENTED;
    }
    void InPlaceTanh() { if (IsDense()) m_dense->InPlaceTanh(); else m_sparse->InPlaceTanh(); }
    void InPlaceTanhDerivative(const CudaMatrix& C) {
		if (IsDense()) m_dense->InPlaceTanhDerivative(*(C.GetDenseMatrix()));
		else m_sparse->InPlaceTanhDerivative(*(C.GetSparseMatrix()));
    }
	/// Done
	
	void InPlaceClip(float min, float max) { if (IsDense()) m_dense->InPlaceClip(min, max); else m_sparse->InPlaceClip(min, max); }
	void InPlaceClip(float v) { if (IsDense()) m_dense->InPlaceClip(v); else m_sparse->InPlaceClip(v); }
	void InPlaceScale(float v) { if (IsDense()) m_dense->InPlaceScale(v); else m_sparse->InPlaceScale(v); }
	float Norm2() { if (IsDense()) return m_dense->Norm2(); else return m_sparse->Norm2(); }
    
    void Print() { if (IsDense()) m_dense->Print(); else m_sparse->Print(); }
    
	static void EwiseProduct(CudaMatrix& A, CudaMatrix& B, CudaMatrix& C);
    static void Dot(CudaMatrix& A, CudaMatrix& B, CudaMatrix& C, bool at=false, bool bt=false);
    static void Add(CudaMatrix& A, CudaMatrix& B, float f=1.0);
    static void AssignNCE(int nnoise,
                        const SparseCudaMatrix& A,
                        const SparseCudaMatrix& P,
                        SparseCudaMatrix& E);
    static void AssignNCELoss(int nnoise,
                        const SparseCudaMatrix& A,
                        const SparseCudaMatrix& P,
						float* loss);
	
    static void AssignCE(const DenseCudaMatrix& A, SparseCudaMatrix& T, DenseCudaMatrix& E);
    static void AssignCELoss(const DenseCudaMatrix& A, SparseCudaMatrix& T, float* loss);
	
	static void AdagradUpdate(DenseCudaMatrix& Gsq, DenseCudaMatrix& Grad, DenseCudaMatrix& Param, float beta, float mb=1);
	static void AdadeltaUpdate(DenseCudaMatrix& Gsq, DenseCudaMatrix& Xsq, DenseCudaMatrix& Grad, DenseCudaMatrix& Param, float beta, float mb=1);
	
	static void MaxPoolingFwd(size_t poolingChannels, size_t channelRows, size_t channelCols, 
		float* poolingAct, float* channelAct, float* channelErrorMask);

	bool LoadFromExternalBinary(std::string fileName, bool rowMajor=false) {
		if (m_dense) 
			return m_dense->LoadFromExternalBinary(fileName);
		NOT_IMPLEMENTED;
	}
	bool WriteToExternalBinary(std::string fileName) {
		if (m_dense) 
			return m_dense->WriteToExternalBinary(fileName);
		NOT_IMPLEMENTED;
	}
};