#pragma once

#include "CudaMatrix.hpp"
#include "CudaOps.hpp"

class CudaMatrixGroup
{
	size_t m_currentIdx;
	size_t m_groupSize;
	size_t m_groupNRows;
	size_t m_groupNCols;
	size_t m_allocated;
	bool m_handleDestruction;

	// Contiguous array
	CudaMatrix** m_matrix;
	float* m_arrayData;

public:
	CudaMatrixGroup(int groupSize, bool handleDestruction=false)
	{
		m_handleDestruction = handleDestruction;
		m_currentIdx = 0;
		m_allocated = 0;
		m_groupNRows = 0;
		m_groupNCols = 0;
		m_groupSize = groupSize;
		m_arrayData = nullptr;
	}

	void AddMatrix(CudaMatrix* matrix)
	{
		if (!matrix->IsDense())
		{
			NOT_IMPLEMENTED;
		}

		if (m_currentIdx>=m_groupSize) 
			throw std::exception("Not enough space in group!");

		if (m_groupNRows != matrix->GetRows() 
			|| m_groupNCols != matrix->GetCols())
		{
			m_groupNRows = matrix->GetRows();
			m_groupNCols = matrix->GetCols();
			
			
			if (m_arrayData != nullptr)
				CUDA_CALL(cudaFree(m_arrayData));
			
			CUDA_CALL(cudaMalloc(&m_arrayData, m_groupSize*matrix->GetSize()*sizeof(float)));
			m_matrix = (CudaMatrix**) malloc(m_groupSize*sizeof(CudaMatrix*));
		}
		
		// Add the matrix to the group
		m_matrix[m_currentIdx] = matrix;
		CUDA_CALL(cudaMemcpy(m_arrayData + matrix->GetSize()*m_currentIdx++, 
			matrix->m_dense->m_dData, matrix->GetSize()*sizeof(float), cudaMemcpyDeviceToDevice));
	}

	size_t GetRows() { return m_groupNRows; }
	size_t GetCols() { return m_groupNCols; }
	size_t GetNums() { return m_groupSize; }

	void SetMatrix(CudaMatrix* matrix, size_t position)
	{
		if (!matrix->IsDense())
		{
			NOT_IMPLEMENTED;
		}

		if (position>=m_groupSize) 
			throw std::exception("Not enough space in group!");
		
		if (m_groupNRows == 0 && m_groupNCols == 0)
		{
			m_groupNRows = matrix->GetRows();
			m_groupNCols = matrix->GetCols();

			if (m_allocated < matrix->GetSize()*m_groupSize) 
			{
				if (m_arrayData != nullptr)
					CUDA_CALL(cudaFree(m_arrayData));
			}

			CUDA_CALL(cudaMalloc(&m_arrayData, m_groupSize*matrix->GetSize()*sizeof(float)));
			m_matrix = (CudaMatrix**) malloc(m_groupSize*sizeof(CudaMatrix*));
		}
		
		if (m_groupNRows != matrix->GetRows() || m_groupNCols != matrix->GetCols())
			throw std::exception("Adding different matrix size to a group is not allowed!");
		
		// Add the matrix to the group
		m_matrix[position] = matrix;
		CUDA_CALL(cudaMemcpy(m_arrayData + matrix->GetSize()*position, 
			matrix->m_dense->m_dData, matrix->GetSize()*sizeof(float), cudaMemcpyDeviceToDevice));
	}
	
	float* GetGroupData() const { return m_arrayData; }
	float* GetGroupDataForUpdate() { return m_arrayData; }

	CudaMatrix* operator[](size_t idx) {
		return m_matrix[idx];
	}

	~CudaMatrixGroup()
	{
		if (m_handleDestruction) {
			for (int i = 0; i < m_currentIdx; i++)
				delete m_matrix[i];
		}
		
		if (m_arrayData!=nullptr) cudaFree(m_arrayData);
		if (m_matrix != nullptr) delete m_matrix;
	}
};