//
//  CudaOps.cu.c
//  CudaMatrix
//
//  Created by Alessandro Sordoni on 2014-07-23.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//

#include "CudaOps.hpp"

void CUBLAS_CALL(cublasStatus_t x)
{
    if(x!=CUBLAS_STATUS_SUCCESS) 
    { 
        std::cerr << "CUBLAS failed" << std::endl;
        throw std::exception("CUBLAS fail");
    }
}

void CUDA_CALL(cudaError_t x)
{
    if(x != cudaSuccess) 
    { 
        const char* errmsg = cudaGetErrorString(x);
        std::cerr << "CUDA failed: " << errmsg << std::endl;
        throw std::exception(errmsg);
    }
}

void CURAND_CALL(curandStatus_t x)
{
	if(x!=CURAND_STATUS_SUCCESS) 
    { 
        std::cerr << "CURAND failed" << std::endl;
        throw std::exception("CURAND fail");
    }
}

static __inline__ __device__ float atomicAdd(double* address, double val)
{
    unsigned long long int* address_as_ull = (unsigned long long int*)address;
    unsigned long long int old = *address_as_ull, assumed;

    do {
        assumed = old;
        old = atomicCAS(address_as_ull, assumed, __double_as_longlong(val + __longlong_as_double(assumed)));
    } while (assumed != old);

    return __longlong_as_double(old);
}

__global__ void GpuDotSDS(int aNnz,
    const int* aRowptr,
    const int* aColptr,
    const float* aVal,
	bool aTranspose,
    int bRows,
    int bCols,
	const float* b,
    bool bTranspose,
    int* cRowPtr,
	int* cColPtr,
	float* cVal,
    int numrowc)
{
}

__global__ void GpuSDAdd(
	int aRows, 
	const float* A, 
	int bNz,
	int* bRow, 
	int* bCol, 
	float* bVal, 
	float f)
{
	int tid = blockDim.x*blockIdx.x + threadIdx.x;
	if (tid >= bNz)
		return;

	int row = bRow[tid];
	int col = bCol[tid];
	float v = A[IDX2C(row, col, aRows)];

	if (v == 0)
		return;
	
	bVal[tid] += f*v;
}

__global__ void GpuProd(const long size, float* a, const float* b, const float* c)
{
	int tid = blockDim.x*blockIdx.x + threadIdx.x;
	if (tid >= size)
		return;
	
	a[tid] = b[tid]*c[tid];
}

__global__ void GpuDSAdd(
	int aRows, 
	float* A, 
	int bNz,
	const int* bRow, 
	const int* bCol, 
	const float* bVal, 
	float f)
{
	int tid = blockDim.x*blockIdx.x + threadIdx.x;
	if (tid >= bNz)
		return;

	int row = bRow[tid];
	int col = bCol[tid];
	float v = bVal[tid];

	if (v == 0)
		return;
	
	atomicAdd(&A[IDX2C(row, col, aRows)], f*v);
}

__global__ void GpuDotDSD(int aRows, int aCols, float* aData, 
	bool at, bool bt, 
	int bNnz, 
	int* bCooRow, 
    int* bCooCol, 
	float* bCooVal, 
	float* cData,
	int cRows)
{
	// We suppose that b is not transposed!
	/// Each block is responsible for one coo entry
	int id = blockIdx.x;

	/// Just a safety check
	if (id >= bNnz) return;

	int row = bCooRow[id];
	int col = bCooCol[id];
	float value = bCooVal[id];

	if (value == 0) return;

	if (at)
	{
		// Add the a'th row to c's k col
		int loadPerThread = (aCols+blockDim.x-1)/blockDim.x;
		int start = threadIdx.x*loadPerThread;
		int end = min(aCols, loadPerThread * (threadIdx.x + 1));
	
		for (int i=start; i < end; ++i)
		{
			float v = value*aData[IDX2C((bt)?col:row, i, aRows)];
			atomicAdd(&cData[IDX2C(i, (bt)?row:col, cRows)], v);
		}	
	}
	else
	{
		// Add the a'th row to c's k col
		int loadPerThread = (aRows+blockDim.x-1)/blockDim.x;
		int start = threadIdx.x*loadPerThread;
		int end = min(aRows, loadPerThread * (threadIdx.x + 1));
	
		for (int i=start; i < end; ++i)
		{
			float v = value*aData[IDX2C(i, (bt)?col:row, aRows)];
			atomicAdd(&cData[IDX2C(i, (bt)?row:col, cRows)], v);
		}
	}
}

__global__ void GpuMinMaxClip(const long n, float* a, float min, float max)
{
    int id = blockDim.x * blockIdx.x + threadIdx.x;
    if (id>=n)
        return;

	if (a[id] < min)
		a[id] = min;
	if (a[id] > max)
		a[id] = max;
}

__global__ void GpuAdadelta(
    float* gsq,
	float* xsq,
    float* gradient,
    float* weight,
    float learnRate,
	float mb,
    const size_t n)
{
    int id = blockDim.x * blockIdx.x + threadIdx.x;
    if (id>=n)
        return;
	
	if (gradient[id] == 0)
	{
		gsq[id] = learnRate*gsq[id];
		xsq[id] = learnRate*xsq[id];
		return;
	}
	else
	{
		gradient[id] /= mb;
		gsq[id] = learnRate*gsq[id] + (1-learnRate)*gradient[id]*gradient[id];
		float deltaT = sqrt((xsq[id]+1e-6f)/(gsq[id]+1e-6f))*gradient[id];
		xsq[id] = learnRate*xsq[id] + (1-learnRate)*deltaT*deltaT;
		weight[id] += deltaT;
	}
}

__global__ void GpuAdagrad(
    float* gsq,
    float* gradient,
    float* weight,
    float learnRate,
    float mb,
	const size_t n)
{
    int id = blockDim.x * blockIdx.x + threadIdx.x;
    if (id>=n)
        return;

	if (gradient[id] == 0)
		return;
	
	gradient[id] /= mb;
	gsq[id] += gradient[id] * gradient[id];
	gradient[id] *= (rsqrtf(1e-6f+gsq[id]));
    weight[id] += learnRate * gradient[id];
}

__global__ void GpuDotStDD(float alpha,
    const float* val,
    const int* rowptr,
    const int* colptr,
    int nnz,
    const float* b,
    int bRows,
    int bCols,
    float* c,
    int cRows)
{
	/// We suppose that b is not transposed!
	/// Each block is responsible for one coo entry
	int id = blockIdx.x;
	/// Just a safety check
	if (id >= nnz) return;

	int row = rowptr[id];
	int col = colptr[id];
	float value = val[id];
	if (value == 0) return;	
	
	int loadPerThread = (bCols+blockDim.x-1)/blockDim.x;
	int start = threadIdx.x*loadPerThread;
	int end = min(bCols, loadPerThread * (threadIdx.x + 1));
	
	for (int i=start; i < end; ++i)
	{
		float v = 0;
		v = value*b[IDX2C(row, i, bRows)];
		atomicAdd(&c[IDX2C(col, i, cRows)], v);
	}
}

__global__ void GpuDotSDD(float alpha,
    const float* val,
    const int* rowptr,
    const int* colptr,
    int nnz,
    const float* b,
    int bRows,
    int bCols,
	bool bTranspose,
    float* c,
    int cRows)
{
	/// Each block is responsible for one coo entry
	int id = blockIdx.x;

	/// Just a safety check
	if (id >= nnz) return;

	int row = rowptr[id];
	int col = colptr[id];
	float value = val[id];

	if (value == 0) return;

	if (bTranspose)
	{
		int loadPerThread = (bRows+blockDim.x-1)/blockDim.x;
		int start = threadIdx.x*loadPerThread;
		int end = min(bRows, loadPerThread * (threadIdx.x + 1));
	
		for (int i=start; i < end; ++i)
		{
			float v = value*b[IDX2C(i, col, bRows)];
			atomicAdd(&c[IDX2C(row, i, cRows)], v);
		}
	}
	else
	{
		int loadPerThread = (bCols+blockDim.x-1)/blockDim.x;
		int start = threadIdx.x*loadPerThread;
		int end = min(bCols, loadPerThread * (threadIdx.x + 1));
	
		for (int i=start; i < end; ++i)
		{
			float v = value*b[IDX2C(col, i, bRows)];
			atomicAdd(&c[IDX2C(row, i, cRows)], v);
		}
	}
}

__global__ void GpuDDSMul(size_t aRows, size_t aCols,
                            size_t bRows, size_t bCols,
                            float* A, float* B, int* rows,
                            int* cols, float* vals)
{
    // Current case is the block number
    int row = rows[blockIdx.x];
    int col = cols[blockIdx.x];

    __shared__ float temp[THREADS_PER_BLOCK];
	temp[threadIdx.x] = 0;
	
	int loadPerThread = (aCols+blockDim.x-1)/blockDim.x;
	int start = threadIdx.x*loadPerThread;
	int end = min(aCols, loadPerThread * (threadIdx.x + 1));

	float temporary = 0;
	for (int i=start; i < end; ++i)
	{
		temporary += A[IDX2C(row,i,aRows)]*B[IDX2C(i,col,bRows)];
	}

    temp[threadIdx.x] = temporary;

    __syncthreads();
    
    // contiguous range pattern
    for(int offset = blockDim.x / 2;
        offset > 0;
        offset >>= 1)
    {
        if(threadIdx.x < offset)
        {
            // add a partial sum upstream to our own
            temp[threadIdx.x] += temp[threadIdx.x + offset];
        }
        
        // wait until all threads in the block have
        // updated their partial sums
        __syncthreads();
    }
    
    // thread 0 writes the final result
    if(threadIdx.x == 0)
    {
		// Suppose that valA[id] is not zero
		vals[blockIdx.x] += temp[0];
    }
}

// Row-wise softmax
__global__ void GpuRescaleToRange(float *a,
	size_t size, float min, float max)
{
    int id = blockDim.x * blockIdx.x + threadIdx.x;
    if (id>=size)
        return;    
    a[id]=a[id]*(max-min)+min;
}

// Col-wise softmax
__global__ void GpuAssignCE(size_t aRows,
                            size_t aCols,
                            float* A,
                            float* E,
                            const int* target)
{
    /// !!!!! One block per column !!!!!
	int loadPerThread = (aRows+blockDim.x-1)/blockDim.x;
	int start = threadIdx.x*loadPerThread;
	int end = min(aRows, loadPerThread * (threadIdx.x + 1));
	
	for (int i=start; i < end; ++i)
	{
		size_t absPos = IDX2C(i,blockIdx.x,aRows);
		E[absPos] = -A[absPos];
		
        if (i == target[blockIdx.x]) E[absPos] += 1;
	}
}

///--------------------------------------------------------------------
/// The access in this kernel is not coalesced so it would be better
// to change the multiplication order in the future
__global__ void GpuInplaceSoftmax(size_t aRows, size_t aCols, float* A)
{
	__shared__ float rowMax[1];
	__shared__ float partials[THREADS_PER_BLOCK];    
    rowMax[0]=-10000000;
    partials[threadIdx.x]=-10000000;
	 
	int loadPerThread = (aCols+blockDim.x-1)/blockDim.x;
	int start = threadIdx.x*loadPerThread;
	int end = min(aCols, loadPerThread * (threadIdx.x + 1));
	
	for (int i=start; i < end; ++i)
	{
		partials[threadIdx.x]=max(partials[threadIdx.x],A[IDX2C(blockIdx.x,i,aRows)]);
    }
    __syncthreads();
	
    //256
    if (threadIdx.x<256)
    {
        partials[threadIdx.x]=max(partials[threadIdx.x+256],partials[threadIdx.x]);
    }
    __syncthreads();

    //128
    if (threadIdx.x<128)
    {
        partials[threadIdx.x]=max(partials[threadIdx.x+128],partials[threadIdx.x]);
    }
    __syncthreads();

    //64
    if (threadIdx.x<64)
    {
        partials[threadIdx.x]=max(partials[threadIdx.x+64],partials[threadIdx.x]);
    }
    __syncthreads();

    //32
    if (threadIdx.x<32)
    {
        partials[threadIdx.x]=max(partials[threadIdx.x+32],partials[threadIdx.x]);
    }
    __syncthreads();

    //16
    if (threadIdx.x<16)
    {
        partials[threadIdx.x]=max(partials[threadIdx.x+16],partials[threadIdx.x]);
    }
    __syncthreads();

    //8
    if (threadIdx.x<8)
    {
        partials[threadIdx.x]=max(partials[threadIdx.x+8],partials[threadIdx.x]);
    }
    __syncthreads();

    //4
    if (threadIdx.x<4)
    {
        partials[threadIdx.x]=max(partials[threadIdx.x+4],partials[threadIdx.x]);
    }
    __syncthreads();

    if (threadIdx.x==0)
    {
        rowMax[0] = max(max(partials[0],partials[1]),max(partials[2],partials[3]));        
    }
    __syncthreads();
	
	if (threadIdx.x == 0)
		rowMax[0] = partials[0];
	
	__syncthreads();

	//end of finding max
    //now start finding sums
    partials[threadIdx.x]=0.0f;
    
	__shared__ float rowSum[1];
    rowSum[0]=0.0f;

	for (int i=start; i < end; ++i)
	{
        float tmp=A[IDX2C(blockIdx.x,i,aRows)]-rowMax[0];
		tmp = expf(tmp);
        A[IDX2C(blockIdx.x,i,aRows)]=tmp;
        partials[threadIdx.x]+=tmp;
    }
    __syncthreads();

    if (threadIdx.x<256)
    {
        partials[threadIdx.x]+=partials[threadIdx.x+256];
    }
    __syncthreads();

    if (threadIdx.x<128)
    {
        partials[threadIdx.x]+=partials[threadIdx.x+128];
    }
    __syncthreads();

    if (threadIdx.x<64)
    {
        partials[threadIdx.x]+=partials[threadIdx.x+64];
    }
    __syncthreads();

    if (threadIdx.x<32)
    {
        partials[threadIdx.x]+=partials[threadIdx.x+32];
    }
    __syncthreads();

    if (threadIdx.x<16)
    {
        partials[threadIdx.x]+=partials[threadIdx.x+16];
    }
    __syncthreads();

    if (threadIdx.x<8)
    {
        partials[threadIdx.x]+=partials[threadIdx.x+8];
    }
    __syncthreads();

    if (threadIdx.x<4)
    {
        partials[threadIdx.x]+=partials[threadIdx.x+4];
    }
    __syncthreads();

    if (threadIdx.x==0)
    {
        rowSum[0] = partials[0]+partials[1]+partials[2]+partials[3];        
    }
    __syncthreads();

	for (int i=start; i < end; ++i)
	{        
        A[IDX2C(blockIdx.x,i,aRows)]/=rowSum[0];
    }
}
static __inline__ __device__ void logadd(float& a, float b)
{
    if (a > b) a += logf(1+expf(b-a)); else { a+=b; b=a-b; a=a-b; a+=logf(1+expf(b-a));}
}
//////------------------------

__global__ void GpuMaxPooling(const long numChannels, const long channelSize, 
	float* poolingAct, float* channelAct, float* channelErrorMask)
{
	int id = blockDim.x * blockIdx.x + threadIdx.x;
	if (id>=channelSize)
		return;
	
	int maxChannel = -1;
	float max = -10000000;
	for (int i = 0; i < numChannels; i++) {
		/// Reset the mask to zero for each channel
		/// TODO: channelErrorMask[channelSize*i + id] = 0.f;

		if (channelAct[channelSize*i + id] > max) {
			maxChannel = i;
			max = channelAct[id + channelSize*i];
		}
	}

	poolingAct[id] = max;
	/// Now set the mask for this maxChannel
	/// TODO: channelErrorMask[channelSize*maxChannel + id] = 1.0f;
}

__global__ void GpuSetValue(    
    float* a,
    const float v,
    const size_t N)
{
    int id = blockDim.x * blockIdx.x + threadIdx.x;
    if (id>=N)
        return;

    a[id]=v;
};

__global__ void GpuAssignCELoss(size_t rows,
                            size_t cols,
                            float* A,
                            int* T,
							float* e)
{
	// Runs on one block
    __shared__ float temp[THREADS_PER_BLOCK];
	temp[threadIdx.x] = 0;

	int loadPerThread = (rows+blockDim.x-1)/blockDim.x;
	int start = threadIdx.x*loadPerThread;
	int end = min(rows, loadPerThread * (threadIdx.x + 1));
	
	for (int i=start; i < end; ++i)
	{
		/// printf("ThreadIdx = %d - Row index = %d - Cols = %d\n", threadIdx.x, T[i], cols);
		temp[threadIdx.x] += log(A[IDX2C(i, T[i], rows)]);
    }
    
    __syncthreads();
    
    // contiguous range pattern
    for(int offset = blockDim.x / 2;
        offset > 0;
        offset >>= 1)
    {
        if(threadIdx.x < offset)
        {
            // add a partial sum upstream to our own
            temp[threadIdx.x] += temp[threadIdx.x + offset];
        }
        
        // wait until all threads in the block have
        // updated their partial sums
        __syncthreads();
    }
    
    // thread 0 writes the final result
    if(threadIdx.x == 0)
    {
		e[0] = -temp[0];
    }
}

__global__ void GpuAssignNCELoss(int nnoise,
                             int nz,
                             const float* valA, // values of activation
                             const float* valP, // values of probabilities
                             float* e)
{
	// Runs on one block
    __shared__ float temp[THREADS_PER_BLOCK];
	temp[threadIdx.x] = 0;
	
    int total = nz;
    int loadPerThread = (total+blockDim.x-1)/blockDim.x;

    // find out the items this thread is responsible for
    int start = loadPerThread * threadIdx.x;
    int end = min(total, loadPerThread * (threadIdx.x + 1));

    for (int i = start; i < end; i++)
    {
		// add to objective
        float logPn = valP[i];
        if (logPn > 0) // positive sample: log(pw / (pw + k * pnw))
        {
            float den = valA[i];
            logadd(den, log((float)(nnoise)) - logPn);
            temp[threadIdx.x] += (valA[i] - den);
        }
        else // negative sample: log(k * pnw / (pw + k * pnw))
        {
            float nom = log((float)(nnoise)) + logPn;
            float den = nom;
            logadd(den, valA[i]);
            temp[threadIdx.x] += (nom-den);
        }
    }
    
    __syncthreads();
    
    // contiguous range pattern
    for(int offset = blockDim.x / 2;
        offset > 0;
        offset >>= 1)
    {
        if(threadIdx.x < offset)
        {
            // add a partial sum upstream to our own
            temp[threadIdx.x] += temp[threadIdx.x + offset];
        }
        
        // wait until all threads in the block have
        // updated their partial sums
        __syncthreads();
    }
    
    // thread 0 writes the final result
    if(threadIdx.x == 0)
    {
		e[0] = -temp[0];
    }
}

__global__ void GpuAssignNCE(int nnoise,
                             int nz,
                             const float* valA, // values of activation
                             const float* valP, // values of probabilities
                             float* valE) // values of error to be computed
{
    int id = blockIdx.x * blockDim.x + threadIdx.x;
    if (id >= nz)
        return;
    
	float logPn = valP[id];
	float ac = valA[id];

	float er = 0;
	if (logPn > 0) // positive sample: k * pnw / (pw + k * pnw)
    {
		float nom = log((float)(nnoise)) - logPn;
        logadd(ac, nom);
        er = expf(nom-ac);
	}
	else // negative sample: pw / (pw + k * pnw);
    {
		logadd(ac, log((float)(nnoise)) + logPn);
        er = -expf(valA[id]-ac);
    }

    valE[id] = er;
}

__global__ void GpuInplaceRelu(float* A, const long n) {
    int id = blockIdx.x * blockDim.x + threadIdx.x;
    if (id >= n)
		return;

	if (A[id] < 0)
		A[id] = 0;
}

__global__ void GpuInplaceReluDerivative(const float* A, float* E, const long n) {
    int id = blockIdx.x * blockDim.x + threadIdx.x;
    if (id >= n)
		return;

	if (A[id] == 0)
		E[id] = 0;
}

__global__ void GpuInplaceSigmoidDerivative(const float* A, float* E, const long n) {
    int id = blockIdx.x * blockDim.x + threadIdx.x;
    if (id >= n)
		return;

	E[id] *= (A[id]*(1-A[id]));
}

__global__ void GpuInplaceSigmoid(float *A, const long n)
{
    int id = blockIdx.x * blockDim.x + threadIdx.x;
    if (id < n)
        A[id] = 1.0/(1.0+expf(-A[id]));
}

__global__ void GpuInplaceTanh(float* c, const long n)
{
    int id = blockDim.x * blockIdx.x + threadIdx.x;
    if (id>=n)
        return;
    c[id]=tanh(c[id]);
}

__global__ void GpuInplaceTanhDerivative(const float* A, float* E, const long n)
{
	int id = blockDim.x * blockIdx.x + threadIdx.x;
    if (id>=n)
		return;
	E[id]*=(1-A[id]*A[id]);
}