//
//  DenseCudaMatrix.cpp
//  CudaMatrix
//
//  Created by Alessandro Sordoni on 2014-07-13.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//

#include "DenseCudaMatrix.hpp"
#include <chrono>
#include <thread>

#ifdef SYNC
bool sync = true;
#else
bool sync = false;
#endif

void DenseCudaMatrix::EwiseProduct(DenseCudaMatrix& B, DenseCudaMatrix& C)
{
	assert(B.m_rows == C.m_rows && B.m_cols == C.m_cols);
	Resize(B.m_rows, B.m_cols);

	int numBlocks = (int) ceil(1.0*(m_size)/THREADS_PER_BLOCK);
	
	cudaEvent_t done;       
    if (sync) CUDA_CALL(cudaEventCreate(&done));
	GpuProd<<<numBlocks,THREADS_PER_BLOCK>>>(m_size, m_dData, B.m_dData, C.m_dData);
	if (sync) CUDA_CALL(cudaEventRecord(done));
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}


/// This method is for NCE estimation
/// where we compute only and only the entries
/// that we really need. (V x H) * (H x b)
void 
DenseCudaMatrix::Dot(DenseCudaMatrix& B, SparseCudaMatrix& C, bool at, bool bt)
{
	cudaEvent_t done;       
    if (sync) CUDA_CALL(cudaEventCreate(&done));
	
	if (at || bt) NOT_IMPLEMENTED;
	assert(m_cols == B.m_rows); 

	// Ensure space
	C.Resize(m_rows, B.m_cols);
    GpuDDSMul<<<C.m_nz,THREADS_PER_BLOCK>>>(m_rows, m_cols, B.m_rows, B.m_cols,
                                m_dData, B.m_dData, C.m_cooRow, 
                                C.m_cooCol, C.m_cooVal);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}
///

///
void 
DenseCudaMatrix::Dot(SparseCudaMatrix& B, DenseCudaMatrix& C, bool at, bool bt)
{
	cudaEvent_t done;       
    if (sync) CUDA_CALL(cudaEventCreate(&done));
	
	// Ensure space
	if (at) {
		if (bt) 
		{
			assert(m_rows == B.m_cols); 
			C.Resize(m_cols, B.m_rows);
		}
		else
		{
			assert(m_rows == B.m_rows); 
			C.Resize(m_cols, B.m_cols);
		}
	}
	else 
	{
		if (bt) 
		{
			assert(m_cols == B.m_cols); 
			C.Resize(m_rows, B.m_rows);
		}
		else
		{
			assert(m_cols == B.m_rows); 
			C.Resize(m_rows, B.m_cols);
		}
	}
	
	GpuDotDSD<<<B.m_nz, THREADS_PER_BLOCK>>>(m_rows, m_cols,
                     m_dData, at, bt, B.m_nz, 
					 B.m_cooRow, 
                     B.m_cooCol, 
					 B.m_cooVal, 
					 C.m_dData,
					 C.m_rows);
	
	/*DenseCudaMatrix Ct;
    B.Dot(*this, Ct, !bt, !at, true);
	C.AddTo(Ct.Transpose(), 1);*/
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}
///

DenseCudaMatrix DenseCudaMatrix::Transpose()
{
    DenseCudaMatrix C;
    C.AssignTransposeOf(*this);
    return C;
}

DenseCudaMatrix& DenseCudaMatrix::AssignTransposeOf(const DenseCudaMatrix& a) 
{
    Resize(a.m_cols, a.m_rows);
    int m = (int) a.m_cols;
    int n = (int) a.m_rows;
    float alpha=1;
    float beta=0;
		
	cudaEvent_t done;    
    if (sync) CUDA_CALL(cudaEventCreate(&done));
    cublasSgeam(m_cubHandle, CUBLAS_OP_T, CUBLAS_OP_T, m, n, &alpha, a.m_dData, (int) a.m_rows,
		&beta, a.m_dData, (int) a.m_rows, m_dData, (int) a.m_cols);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));

    return *this;
}

///
void
DenseCudaMatrix::AddTo(const DenseCudaMatrix& B, float f)
{
	Resize(B.m_rows, B.m_cols);
	
	cudaEvent_t done;
    if (sync) CUDA_CALL(cudaEventCreate(&done));
	cublasSaxpy(m_cubHandle, (int) m_size, &f, B.m_dData, 1, m_dData, 1);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}
///

///
void
DenseCudaMatrix::AddTo(const SparseCudaMatrix& B, float f)
{
	// Make space
	Resize(B.m_rows, B.m_cols);
	int numBlocks = (int) ceil(1.0*B.m_nz/THREADS_PER_BLOCK);

	cudaEvent_t done;
    if (sync) CUDA_CALL(cudaEventCreate(&done));
	GpuDSAdd<<<numBlocks, THREADS_PER_BLOCK>>>(m_rows, m_dData, B.m_nz, B.m_cooRow, B.m_cooCol, B.m_cooVal, f);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}
///

void
DenseCudaMatrix::InPlaceSigmoid()
{
    int numBlocks = (int) ceil(1.0*m_size/THREADS_PER_BLOCK);

	cudaEvent_t done;
    if (sync) CUDA_CALL(cudaEventCreate(&done));
	GpuInplaceSigmoid<<<numBlocks, THREADS_PER_BLOCK>>>(m_dData, m_size);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void
DenseCudaMatrix::InPlaceTanh()
{
	int numBlocks = (int) ceil(1.0*m_size/THREADS_PER_BLOCK);

	cudaEvent_t done;       
	if (sync) CUDA_CALL(cudaEventCreate(&done));

	GpuInplaceTanh<<<numBlocks, THREADS_PER_BLOCK>>>(m_dData, m_size);

	if (sync) CUDA_CALL(cudaEventRecord(done));        
	if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void
DenseCudaMatrix::InPlaceTanhDerivative(const DenseCudaMatrix& C)
{
	assert(C.m_rows == m_rows && C.m_cols == m_cols);

	int numBlocks = (int) ceil(1.0*m_size/THREADS_PER_BLOCK);

	cudaEvent_t done;       
	if (sync) CUDA_CALL(cudaEventCreate(&done));

	GpuInplaceTanhDerivative<<<numBlocks, THREADS_PER_BLOCK>>>(C.m_dData, m_dData, m_size);

	if (sync) CUDA_CALL(cudaEventRecord(done));        
	if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void
DenseCudaMatrix::InPlaceRelu()
{    
	int numBlocks = (int) ceil(1.0*m_size/THREADS_PER_BLOCK);
    /// ATTENTION HERE FOR THE NUMBER OF BLOCKS!
	cudaEvent_t done;
    if (sync) CUDA_CALL(cudaEventCreate(&done));
    GpuInplaceRelu<<<numBlocks, THREADS_PER_BLOCK>>>(m_dData, m_size);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void
DenseCudaMatrix::InPlaceReluDerivative(const DenseCudaMatrix& C)
{
	assert(C.m_rows == m_rows && C.m_cols == m_cols);
    int numBlocks = (int) ceil(1.0*m_size/THREADS_PER_BLOCK);

	cudaEvent_t done;
    if (sync) CUDA_CALL(cudaEventCreate(&done));
    GpuInplaceReluDerivative<<<numBlocks, THREADS_PER_BLOCK>>>(C.m_dData, m_dData, m_size);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void
DenseCudaMatrix::InPlaceSoftmax()
{
	int numBlocks = (int) m_rows;
	// int numBlocks = (int) m_cols;

    /// ATTENTION HERE FOR THE NUMBER OF BLOCKS!
	cudaEvent_t done;
    if (sync) CUDA_CALL(cudaEventCreate(&done));
    GpuInplaceSoftmax<<<numBlocks, THREADS_PER_BLOCK>>>(m_rows, m_cols, m_dData);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void
DenseCudaMatrix::InPlaceSigmoidDerivative(const DenseCudaMatrix& C)
{
	if (C.m_rows != m_rows || C.m_cols != m_cols)
	{
		std::cerr << C.m_name << " " << m_name << " " << C.m_rows << " " << m_rows << std::endl;
		throw std::exception("Cols&Rows mismatch!!");
	}
	assert(C.m_rows == m_rows && C.m_cols == m_cols);
    int numBlocks = (int) ceil(1.0*m_size/THREADS_PER_BLOCK);

	cudaEvent_t done;
    if (sync) CUDA_CALL(cudaEventCreate(&done));
    GpuInplaceSigmoidDerivative<<<numBlocks, THREADS_PER_BLOCK>>>(C.m_dData, m_dData, m_size);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void
DenseCudaMatrix::Dot(DenseCudaMatrix& B, DenseCudaMatrix& C, bool at, bool bt, bool doSync)
{
    ///
    float alpha = 1.0;
    float beta = 1.0;
    	
    size_t m = (at) ? m_cols : m_rows;
    size_t n = (bt) ? B.m_rows : B.m_cols;
    size_t k = (at) ? m_rows : m_cols;

    int lda = (int) m_rows;
    int ldb = (int) B.m_rows;
	
	/// Ensure space
    C.Resize(m, n);

	cudaEvent_t done;
	if (doSync || sync) CUDA_CALL(cudaEventCreate(&done));
    		
	CUBLAS_CALL(cublasSgemm(m_cubHandle,
        (at) ? CUBLAS_OP_T : CUBLAS_OP_N,
        (bt) ? CUBLAS_OP_T : CUBLAS_OP_N,
        (int) m, (int) n, (int) k,
        &alpha,
        m_dData, lda,
        B.m_dData, ldb,
        &beta, 
        C.m_dData, (int) C.m_rows));
    ///
	
	if (doSync || sync) CUDA_CALL(cudaEventRecord(done));        
    if (doSync || sync) CUDA_CALL(cudaEventSynchronize(done));
	if (doSync || sync) CUDA_CALL(cudaEventDestroy(done));
}

///
void 
DenseCudaMatrix::Dot(SparseCudaMatrix& B, SparseCudaMatrix& C, bool at, bool bt)
{
}

void DenseCudaMatrix::SparseInit(int nb)
{
	float* index = (float*) malloc(m_rows*m_cols*sizeof(float));
	memset(index, 0, m_size*sizeof(float));

	float* dIndex;
	CUDA_CALL(cudaMalloc(&dIndex, m_rows*m_cols*sizeof(float)));

	srand(1);
	for (int i = 0; i < m_cols; i++)
	{
		for (int j = 0; j < 15; j++)
		{
			int row = ((float)rand()/(float)(RAND_MAX)) * m_rows;
			index[IDX2C(row, i, m_rows)] = 1;
		}
	}
	
	CUDA_CALL(cudaMemcpy(dIndex, index, sizeof(float)*m_size, cudaMemcpyHostToDevice));
	
	/* Create pseudo-random number generator */
	cudaEvent_t done;
    CUDA_CALL(cudaEventCreate(&done));

	curandGenerator_t curGen;
	curandCreateGenerator(&curGen, CURAND_RNG_PSEUDO_XORWOW);
	curandSetPseudoRandomGeneratorSeed(curGen, m_seed);       
	curandSetGeneratorOrdering(curGen, CURAND_ORDERING_PSEUDO_SEEDED);
	curandGenerateNormal(curGen, m_dData, m_size, 0, 1);
	curandDestroyGenerator(curGen);
	
	CUDA_CALL(cudaEventRecord(done));        
    CUDA_CALL(cudaEventSynchronize(done));
	CUDA_CALL(cudaEventDestroy(done));

	int numBlocks = ceil(1.0f*m_size/THREADS_PER_BLOCK);
	GpuProd<<<numBlocks, THREADS_PER_BLOCK>>>(m_size, m_dData, m_dData, dIndex); 
	
	CUDA_CALL(cudaFree(dIndex));
	delete index;
}

void DenseCudaMatrix::GaussianRandomFill(float mean, float sigma)
{
	std::cerr << "gaussian fill : " << m_rows << " x " << m_cols << std::endl;
	/* Create pseudo-random number generator */        
	cudaEvent_t done;
    CUDA_CALL(cudaEventCreate(&done));

	curandGenerator_t curGen;		
	CURAND_CALL(curandCreateGenerator(&curGen,CURAND_RNG_PSEUDO_XORWOW));   
	
	CURAND_CALL(curandSetPseudoRandomGeneratorSeed(curGen, (m_seed != 0) ? m_seed : time(NULL)));     
	CURAND_CALL(curandSetGeneratorOrdering(curGen, CURAND_ORDERING_PSEUDO_BEST));

	CURAND_CALL(curandGenerateNormal(curGen, m_dData, m_size, mean, sigma));
	CURAND_CALL(curandDestroyGenerator(curGen));
	
	float norm = Norm2();
	if (norm == 0)
		throw std::exception("Norm2 = 0 not expected, random fill failed!");

	CUDA_CALL(cudaEventRecord(done));        
    CUDA_CALL(cudaEventSynchronize(done));
	CUDA_CALL(cudaEventDestroy(done));
}

bool DenseCudaMatrix::OrthogonalFill(float mean, float sigma)
{
	std::cerr << "orthogonal fill : " << m_rows << " x " << m_cols << std::endl;
	this->GaussianRandomFill(mean, sigma);

	float *hData = nullptr;
	hData = (float*) malloc(m_size*sizeof(float));
	cudaMemcpy(hData, m_dData, m_size*sizeof(float), cudaMemcpyDeviceToHost);
	
	// num of singular values
    int num_s = m_rows < m_cols ? m_rows : m_cols;
    
    int nrows_u = m_rows;
    int ncols_u = num_s;
    int nrows_vt = num_s;
    
    int lda_t = (int) max(1, m_rows);
    int ldu_t = (int) max(1, nrows_u);
    int ldvt_t = (int) max(1, nrows_vt);
    
    float* u_t = NULL;
    float* vt_t = NULL;
    
    // Auxiliary matrices
	float* s = (float*)malloc(num_s*sizeof(float));
    u_t = (float*)malloc( sizeof(float) * ldu_t * max(1,ncols_u) );
    vt_t = (float*)malloc( sizeof(float) * ldvt_t * max(1,m_cols) );
    
    float* work;
    int info = 0;
    int lwork = 2 * min( 3*max(m_rows,m_cols) + max(m_rows,m_cols), 5*min(m_rows,m_cols) );
    work = (float*)malloc(lwork * sizeof(float));
	
	sgesvd_("all", "all", &nrows_u, &ncols_u, hData, &lda_t,s,u_t,&ldu_t,vt_t,&ldvt_t, work, &lwork, &info);
	
    /* Check for convergence */
    if( info > 0 ) {
        printf( "The algorithm computing SVD failed to converge.\n" );
        exit(-1);
    }
	if( info < 0 ) {
		printf( "WARNING: SVD: the %d-th argument had an illegal value\n", -info);
	}
    
	cudaMemcpy(m_dData, u_t, m_size*sizeof(float), cudaMemcpyHostToDevice);
	InPlaceScale(sigma);
	
    // Cleanup workspace:
    delete[] work;
    // delete[] iwork;
	delete s;
    delete[] hData;
    delete[] u_t;
    delete[] vt_t;
	return (info == 0);
}

void 
DenseCudaMatrix::UniformRandomFill(float min, float max) 
{
	cudaEvent_t done;
    CUDA_CALL(cudaEventCreate(&done));

	/* Create pseudo-random number generator */        
	curandGenerator_t curGen;
	
	CURAND_CALL(curandCreateGenerator(&curGen,CURAND_RNG_PSEUDO_XORWOW));   
	CURAND_CALL(curandSetPseudoRandomGeneratorSeed(curGen, m_seed));     
	CURAND_CALL(curandSetGeneratorOrdering(curGen,CURAND_ORDERING_PSEUDO_SEEDED));
	CURAND_CALL(curandGenerateUniform(curGen,m_dData, m_size));
	
	CURAND_CALL(curandDestroyGenerator(curGen));
    
	CUDA_CALL(cudaEventRecord(done));        
    CUDA_CALL(cudaEventSynchronize(done));
	CUDA_CALL(cudaEventDestroy(done));

	int blocksPerGrid =(int)ceil(1.0*m_size/THREADS_PER_BLOCK);
	
    if (sync) CUDA_CALL(cudaEventCreate(&done));
	
	GpuRescaleToRange<<<blocksPerGrid,THREADS_PER_BLOCK>>>(m_dData,m_size,min,max);
	
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void 
DenseCudaMatrix::InPlaceScale(float f) {
	float alpha = (float) f;
	cudaEvent_t done;
    if (sync) CUDA_CALL(cudaEventCreate(&done));
	cublasSscal(m_cubHandle, (int) m_size, &alpha, m_dData, 1);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

float 
DenseCudaMatrix::Norm2() {
	float norm = 1;
	cudaEvent_t done;
    if (sync) CUDA_CALL(cudaEventCreate(&done));
	cublasSnrm2(m_cubHandle, (int) m_size, m_dData, 1, &norm);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
	return norm;
}

void DenseCudaMatrix::InPlaceClip(float v)
{	
	float n = Norm2();
	if (n > v)
		InPlaceScale(v/n);
}

void DenseCudaMatrix::InPlaceClip(float min, float max)
{
	int numBlocks = (int) ceil(1.0*m_size/THREADS_PER_BLOCK);

	cudaEvent_t done;
    if (sync) CUDA_CALL(cudaEventCreate(&done));
	GpuMinMaxClip<<<numBlocks, THREADS_PER_BLOCK>>>(m_size, m_dData, min, max);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void DenseCudaMatrix::CopyAtPosition(DenseCudaMatrix& C, 
									 const long sourceSize, 
									 long sourceCol, long targetCol)
{
	// Concatenate a matrix into the other
	cudaEvent_t done;       
	if (sync) CUDA_CALL(cudaEventCreate(&done));

	assert(m_rows == C.m_rows);
	assert(m_size >= sourceSize*(targetCol+1));
	assert(C.m_size >= sourceSize*(sourceCol+1));
	
	// Copy the slice
	cudaMemcpy(m_dData + targetCol*sourceSize, 
		C.m_dData + sourceCol*sourceSize,
		sourceSize*sizeof(float), cudaMemcpyDeviceToDevice);
	
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void DenseCudaMatrix::Fill(float v) 
{
	int blocksPerGrid =(int)ceil(1.0*m_size/THREADS_PER_BLOCK);
	cudaEvent_t done;
    if (sync) CUDA_CALL(cudaEventCreate(&done));
	GpuSetValue<<<blocksPerGrid,THREADS_PER_BLOCK>>>(m_dData, v, m_size);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}