//
//  DenseCudaMatrix.hpp
//  CudaMatrix
//
//  Created by Alessandro Sordoni on 2014-07-13.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//

#pragma once

#include <iostream>
#include <fstream>

#include "CudaOps.hpp"
// Blas for SVD
#include <cblas.h>

#include <complex>
#define lapack_complex_double std::complex<double>
#define lapack_complex_float std::complex<float>
#include <lapacke.h>

#include "SparseCudaMatrix.hpp"

class SparseCudaMatrix;

class DenseCudaMatrix {
private:
	long m_seed;

public:
    float* m_dData;
	size_t m_rows;
    size_t m_cols;
    size_t m_size;
	size_t m_allocated;
    std::string m_name;
    cublasHandle_t m_cubHandle;
    
public:
    DenseCudaMatrix()
    {
        m_name = "";
		m_seed = 0;
        m_rows = 0;
		m_allocated = 0;
        m_cols = 0;
        m_size = 0;
        CUBLAS_CALL(cublasCreate(&m_cubHandle));
        m_dData = nullptr;
    }
    
    DenseCudaMatrix(const DenseCudaMatrix& deepCopy)
    {
        m_name = "";
		m_seed = 0;
		m_seed = 0;
        m_rows = 0;
        m_cols = 0;
        m_size = 0;
        CUBLAS_CALL(cublasCreate(&m_cubHandle));
        m_dData = nullptr;
        
        Resize(deepCopy.m_rows, deepCopy.m_cols);
        CUDA_CALL(cudaMemcpy(m_dData, deepCopy.m_dData, m_size*sizeof(float), cudaMemcpyDeviceToDevice));
    }
    
    DenseCudaMatrix(size_t rows, size_t cols)
    {
		m_name = "";
		m_seed = 0;
        m_rows = 0;
        m_cols = 0;
        m_size = 0;
		m_dData = nullptr;
        CUBLAS_CALL(cublasCreate(&m_cubHandle));
		Resize(rows, cols);
    }
    
	void SetSeed(long seed)
	{
		m_seed = seed;
	}

    void SetName(const std::string& name)
    {
        m_name = name;
    }
    
    void SetData(size_t rows, size_t cols, const float* data, bool onDevice=false) {
        Resize(rows, cols);
        cudaMemcpy(m_dData, data, m_size*sizeof(float), (onDevice)
                   ? cudaMemcpyDeviceToDevice
                   : cudaMemcpyHostToDevice);
    }
    
    void Reset() 
	{ 
		CUDA_CALL(cudaMemset(m_dData, 0, m_size*sizeof(float)));
	}

	void EwiseProduct(DenseCudaMatrix& B, DenseCudaMatrix& C);
	void CopyAtPosition(DenseCudaMatrix& C, const long sourceSize, long sourcePos, long targetPos);
    
	void Print() {
        float* hData = (float*)malloc(sizeof(float)*m_size);
        cudaMemcpy(hData, m_dData, m_size*sizeof(float),cudaMemcpyDeviceToHost);
        std::cerr<<"Matrix: "<<m_name<<" "<<m_rows<<" "<<m_cols<<std::endl;
        for (int i = 0; i < m_rows; i++) {
            for (int j = 0; j < m_cols; j++)
                std::cerr<<hData[IDX2C(i,j,m_rows)]<<" ";
            std::cerr << std::endl;
        }
        delete hData;
    }
    
    void Resize(size_t rows, size_t cols, bool noFill=false) {
        size_t newSize = rows*cols;
		bool sizeChanged = false;

		if (newSize != m_size) 
		{
			sizeChanged = true;
		}
		
		if (newSize > m_allocated)
        {
			// Should realloc
            if (m_dData != NULL)
                CUDA_CALL(cudaFree(m_dData));
            
			CUDA_CALL(cudaMalloc(&m_dData, newSize*sizeof(float)));
			m_allocated = newSize;
		}
        
		// We reset the values if and only if the size of the array has changed
		if (sizeChanged) 
		{
			m_rows = rows;
			m_cols = cols;
			m_size = newSize;

			if (!noFill) CUDA_CALL(cudaMemset(m_dData, 0, m_size*sizeof(float)));
		}
	}
    
	void SparseInit(int nb);
	void GaussianRandomFill(float mean, float sigma);
	bool OrthogonalFill(float mean, float sigma);
	void UniformRandomFill(float min, float max);
	
	float* CopyToHostData() const {
		float* hData = (float*) malloc(m_size*sizeof(float));
		cudaMemcpy(hData, m_dData, m_size*sizeof(float), cudaMemcpyDeviceToHost);
		return hData;
	};

    //
	bool LoadFromExternalBinary(std::string fileName, bool rowMajor=false) {
		std::ifstream input(fileName, std::ios::binary);
		if (!input.good()) return false;

		if (!rowMajor)
			input >> *this;
		else {
			
		}
		
		input.close();
		return true;
	}

	bool WriteToExternalBinary(std::string fileName) {
		std::ofstream output(fileName, std::ios::binary);
		if (!output.good()) return false;
		output << *this;
		output.close();
		return true;
	}

	friend std::ostream& operator<<(std::ostream& stream, const DenseCudaMatrix& us)
	{
		stream<<us.m_name<<" "<<us.m_rows<<" "<<us.m_cols<<std::endl;
		float* hData = us.CopyToHostData();
		for (int i=0;i<us.m_size;++i)
		{
			stream.write(reinterpret_cast<char*>(&hData[i]), sizeof(float));
		}

		stream<<std::endl;
		return stream;
	}

	friend std::istream& operator>>(std::istream& stream, DenseCudaMatrix& us)
	{
		size_t rows, cols;
		std::string name;

		stream>>name>>rows>>cols;
		stream.get();

		us.SetName(name);
		us.Resize(rows, cols);

		float *hData = (float*) malloc(us.m_size*sizeof(float));
		for (int i=0;i<us.m_size;++i)
		{
			float value = 0;
			stream.read(reinterpret_cast<char*>(&value), sizeof(float));
			hData[i] = value;
		}

		us.SetData(rows, cols, hData);
		delete hData;

		return stream;
	}

	float GetElement(size_t i, size_t j) 
	{
		if (i>=m_rows && j>=m_cols) throw std::exception("Out of range!");

		float v = 0;
		const float *src = m_dData + IDX2C(i,j,m_rows);
		cudaMemcpy(&v, src, sizeof(float), cudaMemcpyDeviceToHost);
		return v;
	}
    
	void InPlaceSigmoid();
    void InPlaceSigmoidDerivative(const DenseCudaMatrix& C);
	
	void InPlaceRelu();
	void InPlaceReluDerivative(const DenseCudaMatrix& C);

	void InPlaceTanh();
	void InPlaceTanhDerivative(const DenseCudaMatrix& C);

    void InPlaceSoftmax();
    
	void InPlaceClip(float min, float max);
	void InPlaceClip(float v);
    void InPlaceScale(float v);
	float Norm2();

	void Fill(float v);


    DenseCudaMatrix Transpose();
    DenseCudaMatrix& AssignTransposeOf(const DenseCudaMatrix& a);
    
    // Addition
    void AddTo(const DenseCudaMatrix& B, float f);
    void AddTo(const SparseCudaMatrix& B, float f);
        
    // NCE
    void Dot(DenseCudaMatrix& B, SparseCudaMatrix& C, bool at=false, bool bt=false);
    void Dot(DenseCudaMatrix& B, DenseCudaMatrix& C, bool at=false, bool bt=false,bool doSync=false);
    void Dot(SparseCudaMatrix& B, SparseCudaMatrix& C, bool at=false, bool bt=false);
    void Dot(SparseCudaMatrix& B, DenseCudaMatrix& C, bool at=false, bool bt=false);
    
    ~DenseCudaMatrix() {
        if (m_cubHandle != NULL) cublasDestroy(m_cubHandle);
        if (m_dData != NULL) cudaFree(m_dData);
    }
};