//
//  GradientUpdater.hpp
//  Nnet
//
//  Created by Alessandro Sordoni on 2014-04-21.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//

#ifndef Nnet_GradientUpdater_hpp
#define Nnet_GradientUpdater_hpp

#include "GradientVisitor.hpp"
#include "CudaMatrix.hpp"
#include <mutex>

/**
 **
 **/
class IGradientFilter {
public:
    virtual void apply(CudaMatrix* matrix, CudaMatrix* gradient) = 0;
    virtual ~IGradientFilter() {};
};

class L2RegularizeFilter : public IGradientFilter {
public:
    void apply(CudaMatrix* matrix, CudaMatrix* gradient) {
        updates_++;
		// Regularize each K updates
		if (updates_ % delay_ == 0) {	
			CudaMatrix::Add(*gradient, *matrix, -beta_);
		}
	}
    
    L2RegularizeFilter() {
		updates_ = 0;
		delay_ = 10;
		beta_ = 0.0000001f;
    }
	
	L2RegularizeFilter(float beta, int delay) {
		updates_ = 0;
		delay_ = delay;
		beta_ = beta;
    }

    L2RegularizeFilter(float beta) {
        beta_ = beta;
		delay_ = 10;
		updates_ = 0;
    }
    
private:
	long long delay_;
	long long updates_;
	float beta_;
};

class ClipGradientFilter : public IGradientFilter {
public:
    void apply(CudaMatrix* matrix, CudaMatrix* gradient) {
        // Cast to dense matrix
		float n = gradient->Norm2();
		if (n > norm_threshold_)
			gradient->InPlaceScale(norm_threshold_/n);
	}

    ClipGradientFilter(float threshold) {
        norm_threshold_ = threshold;
    }

private:
    float norm_threshold_;
};

class ScaleFilter : public IGradientFilter {
public:
    void apply(CudaMatrix* matrix, CudaMatrix* gradient) {
		// Scale the gradients
		gradient->InPlaceScale(scale_);
	}
	
    ScaleFilter() : scale_(1.0) {};
	ScaleFilter(float scale) : scale_(scale_) {};
	float scale_;
};

class HardClipGradientFilter : public IGradientFilter {
public:
    void apply(CudaMatrix* matrix, CudaMatrix* gradient) {
		NOT_IMPLEMENTED;
	}

    HardClipGradientFilter(float threshold) {
        threshold_ = threshold;
    }

private:
    float threshold_;
};

class PrintFilter : public IGradientFilter {
public:
    void apply(CudaMatrix* matrix, CudaMatrix* gradient) {
        matrix->Print();
        gradient->Print();
    }
    
    PrintFilter() {
    }
    
private:
};

class IHyperScheduler {
private:
    int update_rate_;
    
public:
    virtual void apply() = 0;

    IHyperScheduler() {
        update_rate_ = 50000;
    }
    IHyperScheduler(int update_rate) : update_rate_(update_rate) {}
};

/**
 ** Abstract Interface permitting of updating
 ** the gradient store in the gradient visitor
 **/
class IGradientUpdater {
public:
	float alpha_;
    float mb_;

protected:
	std::mutex update_lock;
	std::unordered_map<CudaMatrix*, std::vector<IGradientFilter*> > pre_filters_;
    std::unordered_map<CudaMatrix*, std::vector<IGradientFilter*> > post_filters_;
    
public:
    void add_pre_filter(CudaMatrix* key, IGradientFilter* filter) {
        auto it = pre_filters_.find(key);
		if (it != pre_filters_.end())
			it->second.push_back(filter);
		else {
			std::vector<IGradientFilter*> fs;
			fs.push_back(filter);
			pre_filters_.insert(std::make_pair(key, fs));
		}
	}
    
    void add_post_filter(CudaMatrix* key, IGradientFilter* filter) {
		auto it = post_filters_.find(key);
		if (it != post_filters_.end())
			it->second.push_back(filter);
		else {
			std::vector<IGradientFilter*> fs;
			fs.push_back(filter);
			post_filters_.insert(std::make_pair(key, fs));
		}
    }
    
    void UpdateParameters(GradientVisitor& gv) {
        Grads::iterator git;
        Grads& gradients = gv.ComputeGradients();
        
		for (git = gradients.begin(); git != gradients.end(); git++) {
            auto it = pre_filters_.find(git->first);
            if (it != pre_filters_.end()) {
				for (IGradientFilter* f : it->second)
					f->apply(git->first, git->second);
			}
			
			update_lock.lock();
			ApplyUpdate(git->first, git->second);
			update_lock.unlock();
            
			it = post_filters_.find(git->first);
            if (it != post_filters_.end())
				for (IGradientFilter* f : it->second)
					f->apply(git->first, git->second);
        }

		gv.CleanGradients();
    }
    
    virtual void CleanUp() {}
    virtual void schedule() {}
    virtual std::string status() { return ""; }
    
    virtual ~IGradientUpdater() {
		std::set<IGradientFilter*> to_delete;
		auto it = pre_filters_.begin();
        for (; it != pre_filters_.end(); it++) {
            for (IGradientFilter* f : it->second) 
				to_delete.insert(f);
        }
        
        it = post_filters_.begin();
        for (; it != post_filters_.end(); it++) {
            for (IGradientFilter* f : it->second) 
				to_delete.insert(f);
        }

		auto td = to_delete.begin();
		for (; td != to_delete.end(); td++)
			delete *td;
    }
    
    virtual void ApplyUpdate(CudaMatrix* param, CudaMatrix* gradient) = 0;
};

class SGDUpdater : public IGradientUpdater {
private:
    float updates_;
    float start_alpha_;
    
public:
    SGDUpdater(float alpha, float mb) {
        alpha_ = alpha;
		mb_ = mb;
        start_alpha_ = alpha_;
        updates_ = 0;
    }
    
    void ApplyUpdate(CudaMatrix* param, CudaMatrix* gradient) {
		CudaMatrix::Add(*param, *gradient, alpha_/mb_);
    }
    
    std::string status() {
        std::stringstream status;
        status << "Alpha: " << alpha_;
        return status.str();
    }
};

class AdagradUpdater : public IGradientUpdater {
private:
    Grads past_gradients_;
    
    CudaMatrix* get_gradient(CudaMatrix* param, CudaMatrix* gradient) {
        auto gf = past_gradients_.find(param);
        if (gf == past_gradients_.end()) {
            CudaMatrix* dg = new CudaMatrix(param->GetRows(), param->GetCols(), MatrixType::DENSE);
			dg->SetName(param->GetName() + "_adagrad");
            gf = past_gradients_.insert(std::make_pair(param, dg)).first;
        }
		return gf->second;
    }
    
public:
    AdagradUpdater(float alpha, float mb) {
        alpha_ = alpha;
		mb_ = mb;
    }
    
	~AdagradUpdater() {
        for (Grads::iterator it = past_gradients_.begin(); it != past_gradients_.end(); it++) {
            delete it->second; it->second = NULL;
        }
        past_gradients_.clear();
    }
    
	void ApplyUpdate(CudaMatrix* param, CudaMatrix* gradient) {
        CudaMatrix* gsq = get_gradient(param, gradient);
		CudaMatrix::AdagradUpdate(*(param->m_dense), *(gradient->m_dense), *(gsq->m_dense), alpha_, mb_);
    }
};

class AdaDeltaUpdater : public IGradientUpdater {
private:
	std::unordered_map<CudaMatrix*, std::pair<CudaMatrix*, CudaMatrix*>> m_gs;
    
    std::pair<CudaMatrix*, CudaMatrix*>& GetStructures(CudaMatrix* param) {
		auto gf = m_gs.find(param);
        if (gf == m_gs.end()) {
            CudaMatrix* sqG = new CudaMatrix(param->GetRows(), param->GetCols(), MatrixType::DENSE);
			CudaMatrix* mSqG = new CudaMatrix(param->GetRows(), param->GetCols(), MatrixType::DENSE);
			
			sqG->SetName(param->GetName() + "_sqG_adadelta");
			mSqG->SetName(param->GetName() + "_mSqG_adadelta");
            gf = m_gs.insert(std::make_pair(param, std::make_pair(sqG, mSqG))).first;
        }
        
		return gf->second;
    }
    
public:
    AdaDeltaUpdater(float alpha, float mb) {
        alpha_ = alpha;
		mb_ = mb;
	}
    
    ~AdaDeltaUpdater() {
        for (auto it = m_gs.begin(); it != m_gs.end(); it++) {
            delete it->second.first; 
			delete it->second.second;
			it->second.first = NULL;
			it->second.second = NULL;
        }

		m_gs.clear();                
    }
    
    void ApplyUpdate(CudaMatrix* param, CudaMatrix* gradient) {
		std::pair<CudaMatrix*, CudaMatrix*> p = GetStructures(param);
		CudaMatrix::AdadeltaUpdate(*p.first->m_dense, *p.second->m_dense, 
			*gradient->m_dense, *param->m_dense, alpha_, mb_);
    }
};

/*class MomentumUpdater : public IGradientUpdater {
private:
    float mu_;
    
    Grads past_gradients_;
    
    Matrix* get_or_increment_gradient(Matrix* param, Matrix* gradient) {
        auto gf = past_gradients_.find(param);
        
        Matrix *g = 0;
        if (gf == past_gradients_.end()) {
			// Only use dense matrices for gradients here!
			g = new Matrix(param->GetRows(), param->GetCols(), param->GetName() + "_momentum_grad");
            // Add the gradient with the alpha_ learning rate!
			Matrix::Add(g, gradient, alpha_);
			// Insert into past gradients
			past_gradients_.insert(std::make_pair(param, g));
        } else {
			// mu * g_t-1 + alpha * g_t
            gf->second->InPlaceScale(mu_);
            Matrix::Add(gf->second, gradient, alpha_);
            g = gf->second;
        }
        
        return g;
    }
    
public:
    MomentumUpdater(float alpha, float mu) {
        alpha_ = alpha;
        mu_ = mu;
    }
 
    ~MomentumUpdater() {
        CleanUp();
    }
    
    void CleanUp() {
        for (Grads::iterator it = past_gradients_.begin(); it != past_gradients_.end(); it++) {
			if (it->second) delete it->second; 
			it->second = NULL;
        }
        
        past_gradients_.clear();
    }
    
    void ApplyUpdate(Matrix* param, Matrix* gradient) {
        Matrix* g = get_or_increment_gradient(param, gradient);
		Matrix::Add(param, g);
    }
};*/
#endif
