//
//  nnet.cpp
//  nnet
//
//  Created by Alessandro Sordoni on 2014-03-25.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//
#include <cassert>
#include <stdio.h>
#include "NnetLib.hpp"

/***************************/
/**      Forward pass     **/
/***************************/
CudaMatrix* Layer::GetValue() {
	// If we already visited this node just return its value!
	if (m_evaluated) return GetActivation();

#ifdef DEBUG
	std::cerr << "Evaluating: " << GetName() << std::endl;
#endif

	for (Links::iterator p=GetParents().begin(); p!=GetParents().end(); p++) {
		Weight* parent_weight = static_cast<Weight*>(*p);
		Layer*	parent_layer = static_cast<Layer*>(parent_weight->parent());		
		CudaMatrix* parent_value = parent_layer->GetValue();
	
		// If the parent has no rows or no cols - i.e. input layer without inputs
		if (parent_value->GetRows() == 0 || parent_value->GetCols() == 0)
		{
			m_activation->Resize(0, 0);
			m_evaluated = true;
			return m_activation;
		}

		// Small hack to just add from bottom layers
		if (parent_weight->m() == nullptr) 
		{
#ifdef DEBUG
			std::cerr << GetName() << " += " << parent_layer->GetName() << std::endl;
			if (m_activation->GetSize() <= 30)
				m_activation->Print();
#endif
			CudaMatrix::Add(*m_activation, *parent_value);
		} 
		else
		{
			CudaMatrix::Dot(*(parent_value), *(parent_weight->m()), *(m_activation));
			// CudaMatrix::Dot(*(parent_weight->m()), *(parent_value), *(m_activation));
#ifdef DEBUG
			std::cerr << GetName() << " += " << parent_weight->GetName() << " * " << parent_layer->GetName() << std::endl;
			if (m_activation->GetSize() <= 30) {
				parent_weight->m()->Print();
				m_activation->Print();
			}
#endif
		}
	}

	// Do the activation of the current input
	fn_->f(m_activation);

#ifdef DEBUG
	std::cerr << "Evaluation done! " << GetName() << std::endl;
	if (m_activation->GetSize() <= 30)
		m_activation->Print();
#endif

	m_evaluated = true;
	return m_activation;
}

CudaMatrix* ConcatLayer::GetValue() {
	// If we already visited this node just return its value!
	if (m_evaluated) return GetActivation();

	for (Links::iterator p=GetParents().begin(); p!=GetParents().end(); p++) {
		Weight* parent_weight = static_cast<Weight*>(*p);
		Layer*	parent_layer = static_cast<Layer*>(parent_weight->parent());		
		CudaMatrix* parent_value = parent_layer->GetValue();

#ifdef DEBUG
		std::cerr << "Concatenating: " << parent_layer->GetName() << std::endl;
#endif
		// Make some space
		m_activation->Resize(parent_value->GetRows(), m_numInputs*parent_value->GetCols());

		// Find the parent id and put it at the right location
		auto it = m_orderedParents.find(parent_layer);
		m_activation->CopyAtPosition(*parent_value, parent_value->GetSize(), 0, it->second);
	}

	// Do the activation of the current input
	fn_->f(m_activation);

#ifdef DEBUG
	std::cerr << "Evaluation done!" << GetName() << std::endl;
	m_activation->Print();
#endif

	m_evaluated = true;
	return m_activation;
}

CudaMatrix* ConcatLayer::ComputeError(Layer* parent) {
	if (!m_errorated) {
		/** Execute only if the error has not been yet calculated **/
#ifdef DEBUG
		std::cerr << "# Computing error on layer = " << ToString() << std::endl;
#endif
		for (Links::iterator p = GetChildren().begin(); p != GetChildren().end(); p++) {
			Weight *child_weight = static_cast<Weight*>(*p);
            Layer *child_layer = static_cast<Layer*>(child_weight->child());
			if (!child_layer->IsNeedError()) continue;

			CudaMatrix* child_weight_matrix = child_weight->m();
			CudaMatrix* child_error = child_layer->ComputeError(this);
			
			// Skip these nodes
			if (child_layer == nullptr) 
				throw std::exception("Null error not expected!");

			// Small hack to just add from bottom layers
			if (child_weight_matrix == nullptr)
				CudaMatrix::Add(*m_error, *child_error);
			else
				CudaMatrix::Dot(*child_error, *child_weight_matrix, *m_error, false, true);
			
			m_error->InPlaceClip(-50,50);
		}

		// Here the derivative of the activation function is
		// applied element wise to the *error* computed above.
		// The derivative is usually computed using the *activation*
		// from this layer. This is usually the case for the TANH and the SIGMOID
		// which are 1-a_t^2 and a_t*(1-a_t).
		fn_->df(m_activation, m_error);
#ifdef DEBUG
		std::cerr << " --> Error for layer done = " << GetName() << std::endl;
		m_error->Print();
#endif

		// Fill the errors for the children
		SetErrorated(true);
	}

	if (parent == this) return nullptr;

	// We now have to distribute the errors back into the children
	auto parentIndexIt = m_orderedParents.find(parent);
	auto errorIt = m_parentErrors.find(parent);
	if (errorIt == m_parentErrors.end())
	{
		// Insert a brand new matrix for the children error
		errorIt = m_parentErrors.insert(std::make_pair(parent, 
			new CudaMatrix(m_error->GetRows(), m_error->GetCols()/m_numInputs, MatrixType::DENSE))).first;
	}

	CudaMatrix* parentError = errorIt->second;
	parentError->Resize(m_error->GetRows(), m_error->GetCols()/m_numInputs);
	parentError->CopyAtPosition(*m_error, parentError->GetSize(), parentIndexIt->second, 0);

#ifdef DEBUG
	parentError->Print();
#endif

	return parentError;
}

CudaMatrix* MaxPoolingLayer::GetValue() {
	if (IsEvaluated())
		return GetActivation();

#ifdef DEBUG
	std::cerr << "Evaluating: " << GetName() << std::endl;
#endif
	size_t actCols = 0;
	int id = 0;

	for (Links::iterator p=GetParents().begin(); p != GetParents().end(); p++) {
		Weight* parentWeight = static_cast<Weight*>(*p);
		Layer* parentLayer = (Layer* ) parentWeight->parent();
		CudaMatrix* parentActivation = parentLayer->GetValue();
		
		// If the parent has no rows or no cols - i.e. input layer without inputs
		if (parentActivation->GetRows() == 0 || parentActivation->GetCols() == 0)
		{
			continue;
		}
		
		// Make some space
		actCols = parentActivation->GetCols();
		m_stackedActivations->Resize(parentActivation->GetRows(), GetParents().size()*actCols);
		m_stackedActivations->CopyAtPosition(*parentActivation, parentActivation->GetSize(), 0, id++);
	}

	m_activation->Resize(m_stackedActivations->GetRows(), actCols);
	CudaMatrix::MaxPoolingFwd(id, m_activation->GetRows(), actCols,
		m_activation->GetDenseMatrixForUpdate()->m_dData, m_stackedActivations->GetDenseMatrixForUpdate()->m_dData, nullptr);
	
#ifdef DEBUG	
	std::cerr << "Evaluation done!" << GetName() << std::endl;
	if (m_activation->GetSize() <= 30)
		m_activation->Print();
#endif

	m_evaluated = true;
	return GetActivation();
}

CudaMatrix* MaxPoolingLayer::ComputeError(Layer* parent) {
	NOT_IMPLEMENTED;
}

CudaMatrix* DenseLayer::ComputeError(Layer* parent) {
	if (!m_errorated) {
		/** Execute only if the error has not been yet calculated **/
#ifdef DEBUG
		std::cerr << "# Computing error on layer = " << ToString() << std::endl;
#endif
		for (Links::iterator p = GetChildren().begin(); p != GetChildren().end(); p++) {
			Weight *child_weight = static_cast<Weight*>(*p);
			Layer *child_layer = static_cast<Layer*>(child_weight->child());
			if (!child_layer->IsNeedError()) continue;

#ifdef DEBUG
			std::cerr << " --> Poking layer for error = " << child_layer->GetName() << std::endl;
#endif
			// Accumulate the current error by summing the errors
			// coming from child layers at level t+1 \sum (dh_{t+1}^T * W)
			CudaMatrix* child_weight_matrix = child_weight->m();
			CudaMatrix* child_error = child_layer->ComputeError(this);
			
			if (child_layer == nullptr) 
				throw std::exception("Null error not expected!");
			
			// Small hack to just add from bottom layers
			if (child_weight_matrix == nullptr) 
			{
				CudaMatrix::Add(*(m_error), *child_error);
#ifdef DEBUG
				std::cerr << "e_" << GetName() << " += e_" << child_layer->GetName() << std::endl;
#endif
			}
			else 
			{
				CudaMatrix::Dot(*child_error, *child_weight_matrix, *m_error, false, true);
				// CudaMatrix::Dot(*child_weight_matrix, *child_error, *m_error, true, false);	
				//#ifdef DEBUG
				//std::cerr << "e_ += " << child_weight_matrix->GetName() << "^T * " << child_layer->GetName() << std::endl;
				//child_error->Print();
				//child_weight_matrix->Print();
				//m_error->Print();
				//#endif
			}

#ifdef DEBUG
			std::cerr << " --> Error for layer done = " << GetName() << std::endl;
			m_error->Print();
#endif

			// Clip errors
			m_error->InPlaceClip(-50, 50);
		}

		// Here the derivative of the activation function is
		// applied element wise to the *error* computed above.
		// The derivative is usually computed using the *activation*
		// from this layer. This is usually the case for the TANH and the SIGMOID
		// which are 1-a_t^2 and a_t*(1-a_t).
		fn_->df(m_activation, m_error);

#ifdef DEBUG
		std::cerr << " --> Error for layer done = " << GetName() << std::endl;
		m_error->Print();
#endif
		SetErrorated(true);
	}

	return GetError();
}

CudaMatrix* InputLayer::ComputeError(Layer* parent) {
	return m_error;
}

CudaMatrix* NeuralNet::Forward(LossLayer* ll) 
{
	if (ll == nullptr)
		throw std::exception("Loss layer should be non-null in forward pass!");
	return ll->GetValue();
}

std::vector<CudaMatrix*> NeuralNet::Forward() {
	std::vector<CudaMatrix*> outputs;
	for (auto it = O.begin(); it != O.end(); it++) {
		outputs.push_back((*it)->GetValue());
	}
	return outputs;
}

// In the case of BPTT, we should flag the error nodes 
// before doing the backward pass in order to tell the backward
// algorithm not to ask for error to children nodes outside of 
// the bptt scope
void NeuralNet::FlagErrorNodes(LossLayer* ll) {
	// Reset the flag first
	for (auto it = V.begin(); it != V.end(); it++)
		((Layer*) it->second)->SetNeedError(false);

	// Start the BFS search by walking only 
	// through parents
	std::set<Layer*> nodeVisited;
	std::queue<Layer*> bfsQueue;

	bfsQueue.push(ll);
	nodeVisited.insert(ll);

	// the current loss layer trivially has a gradient
	ll->SetNeedError(true);

	int bpttAdvance = 0;
	while (!bfsQueue.empty()) {
		Layer* current = bfsQueue.front();
		bfsQueue.pop();

		// Add into queue
		for (Links::iterator p=current->GetParents().begin(); p!=current->GetParents().end(); p++) 
		{			
			Weight* pw = static_cast<Weight*>(*p);
			Layer* pl = static_cast<Layer*>(pw->parent());

			// If it is a barrier node then it means that the error cannot 
			// flow behind this weight
			if (pw->m() == m_barrierWeight && m_barrierWeight != nullptr)
			{
				pl->SetNeedError(false);
				/// cut off this branch
				break;
			}

			// Flag this parent node if this is 
			// out of the BPTT limit and if it is on the path
			// of the BPTT recurrent weight
			if (pw->m() == bpttWeight && bpttWeight != nullptr) 
			{
				bpttAdvance++;
				if (bpttAdvance > bpttStop)
					pl->SetNeedError(false);
				else
					pl->SetNeedError(true);
			} else 
				// Otherwise we schedule this layer 
				// to have a gradient.
			{
				pl->SetNeedError(true);
			}

			// Continue searching and flagging backwards
			if (nodeVisited.find(pl) == nodeVisited.end() && pl->IsNeedError()) {
				nodeVisited.insert(pl);
				bfsQueue.push(pl);
			}
		}
	}
}

void NeuralNet::Backward(LossLayer* ll) {
#ifdef DEBUG
	std::cerr << "----- BACKWARD PASS FROM LL -----" << std::endl;
#endif DEBUG

	if (ll == nullptr)
		throw std::exception("Loss layer should be non-null");

	std::set<Layer*> nodeVisited;
	std::queue<Layer*> bfsQueue;

	// BFS for error propagation
	bfsQueue.push(ll);
	nodeVisited.insert(ll);

	// Flagging of relevant error computations
	FlagErrorNodes(ll);

	while (!bfsQueue.empty()) {
		Layer* current = bfsQueue.front();
		bfsQueue.pop();

		// Error propagation here
#ifdef DEBUG
		std::cerr << "START propagating from " << current->ToString() << std::endl;
#endif DEBUG
		current->ComputeError(current);

		// Add into queue
		for (Links::iterator p=current->GetParents().begin(); p!=current->GetParents().end(); p++) 
		{
			Weight* pw = static_cast<Weight*>(*p);
			Layer* pl = static_cast<Layer*>(pw->parent());

			if (nodeVisited.find(pl) == nodeVisited.end() && pl->IsNeedError()) {
				nodeVisited.insert(pl);
				bfsQueue.push(pl);
			}
		}
	}
}

void NeuralNet::Backward() {
#ifdef DEBUG
	std::cerr << "----- BACKWARD PASS -----" << std::endl;
#endif DEBUG
	if (O.size() == 0) {
		throw std::exception("Should add at least one output node!");
	}

	std::set<Layer*> nodeVisited;
	std::queue<Layer*> bfsQueue;

	// BFS for error propagation
	for (Layer* p : O) {
		bfsQueue.push(p);
		nodeVisited.insert(p);
	}

	// Only for the last one
	FlagErrorNodes(O.back());

	while (!bfsQueue.empty()) {
		Layer* current = bfsQueue.front();
		bfsQueue.pop();

		// Error propagation here
#ifdef DEBUG
		std::cerr << "START propagating from " << current->ToString() << std::endl;
#endif DEBUG
		current->ComputeError(current);

		// Add into queue
		for (Links::iterator p=current->GetParents().begin(); p!=current->GetParents().end(); p++) 
		{
			Weight* pw = static_cast<Weight*>(*p);
			Layer* pl = static_cast<Layer*>(pw->parent());

			// Insert if we have not seen this node and 
			// if this effectively needs an error
			// With BPTT this will have the effect of chuncking some of the recurrent
			// nodes too far away in time while keeping the others scheduled for error.
			if (nodeVisited.find(pl) == nodeVisited.end() && pl->IsNeedError()) {
				nodeVisited.insert(pl);
				bfsQueue.push(pl);
			}
		}
	}
}

// Returns true if the node is an internal node inserted
// by the optimizer
bool NeuralNet::IsInternal(Layer* n) {
	return m_internals.find(n) != m_internals.end();
}

// Graph optimization: group the 
// links with the same weight to store
// intermediary multiplication results
int NeuralNet::Optimize() {	
	for (auto it = V.begin(); it != V.end(); it++) {
		// Start clustering only from high-arity objects
		if (it->second->GetChildren().size() > 4) {
#ifdef DEBUG
			std::cerr << "Inspecting " << it->second->GetName() << " found " << it->second->GetChildren().size() << std::endl;
#endif			
			Layer* alreadyInternal = nullptr;

			// Skip internals
			Layer* n = (Layer*) it->second;
			if (m_internals.find(n) != m_internals.end())
				continue;

			// Try to cluster outbound connections by weight!
			std::map<CudaMatrix*, std::vector<Weight*>> cluster;
			for (auto lit = n->GetChildren().begin(); lit != n->GetChildren().end(); lit++) {
				Weight* w = (Weight*) *lit;				
				Layer* child = (Layer*) w->child();
				auto fit = cluster.find(w->m());
				
				// TODO: Fixed weights handle
				if (w->m() == nullptr)
					continue;

				if (fit != cluster.end()) {
					fit->second.push_back(w);
				} else {
					std::vector<Weight*> toLink; toLink.push_back(w);
					cluster.insert(std::make_pair(w->m(), toLink));
				}
			}

			// Found clusters !!
			for (auto c = cluster.begin(); c != cluster.end(); c++) {
				if (c->second.size() > 1) {
					std::string key = "Internal[" + n->GetName() + "," + c->first->GetName() + "]";
					alreadyInternal = GetInternalForLayerParam(key);

					if (alreadyInternal == nullptr) 
					{
						// Create and register an internal node
						Layer* newInternal = new HiddenLayer(ID, key);
						AddLayer(newInternal);

						alreadyInternal = newInternal;

						// Register this node for this weight
						m_internals.insert(alreadyInternal);
						m_internMap.insert(std::make_pair(key, alreadyInternal));
						// end

						Connect((Layer*) n, newInternal, new DynamicWeight(c->first));
					}

					for (int i = 0; i < c->second.size(); i++) {
						Weight* linkTo = c->second[i];
						Layer* layerTo = (Layer*) linkTo->child();

						// Skip the internal
						if (layerTo == alreadyInternal) continue;

						// Disconnect the original link
						n->EraseOutwardLink(linkTo);
						layerTo->EraseInwardLink(linkTo);
						RemoveLink(linkTo);

						Connect(alreadyInternal, layerTo, new FixedWeight());
					}
				}
			}
		}
	}

	return 0;
}

// The order of the visit here is not important
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
void NeuralNet::Visit(Visitor& g) {
	g.pre_visit();
	if (g.IsVisitEdges()) {
		for (Link *l : E) {
			l->Visit(g);
		}
	}
	if (g.IsVisitNodes()) {
		for (auto it = V.begin(); it != V.end(); it++) {
			(it->second)->Visit(g);
		}
	}	
	g.post_visit();
}
