//
//  nnet.hpp
//  nnet
//
//  Created by Alessandro Sordoni on 2014-03-25.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//

#pragma warning( disable : 4250 )
#pragma once

#include <exception>
#include <stdexcept>
#include <sstream>

#include "CudaMatrix.hpp"
#include "CudaMatrixGroup.hpp"
#include "Activation.hpp"
#include "Graph.hpp"
#include "Visitor.hpp"

#define my_delete(x) {delete x; x = NULL;}

// Forward declaration
// class Visitor;
class Weight;

typedef enum { DELEGATE, SHARED } Memory;
typedef enum { TRANSPOSE, NOTRANSPOSE } MatrixView;

class IOutput {
protected:
	CudaMatrix* m_activation;

public:
    CudaMatrix* GetActivation() {
		return m_activation;
	}
    virtual ~IOutput() {};
};
class IError {
protected:
	CudaMatrix* m_error;

public:
    CudaMatrix* GetError() {
		return m_error;
	}
    virtual ~IError() {};
};

class DenseError : public virtual IError {    
protected: 
	DenseError(std::string name) {
        m_error = new CudaMatrix();
		m_error->SetName("d_" + name);
    }
	
	DenseError(size_t size, std::string name) {
        m_error = new CudaMatrix(1, size, MatrixType::DENSE);
		m_error->SetName("d" + name);
    }

    DenseError(size_t batch, size_t size, std::string name) {
        m_error = new CudaMatrix(batch, size, MatrixType::DENSE);
		m_error->SetName("d" + name);
    }
    virtual ~DenseError() {
        my_delete(m_error);
    }
};
class SparseError : public virtual IError {
protected:
	SparseError(std::string name) {
		m_error = new CudaMatrix();
		m_error->SetMatrixType(MatrixType::SPARSE);
		m_error->SetName("d_" + name);
    }
    
	SparseError(size_t size, std::string name) {
		m_error = new CudaMatrix(1, size, MatrixType::SPARSE);
		m_error->SetName("d_" + name);
    }

	SparseError(size_t batch, size_t size, std::string name) {
		m_error = new CudaMatrix(batch, size, MatrixType::SPARSE);
		m_error->SetName("d_" + name);
    }

    virtual ~SparseError() {
        my_delete(m_error);
    }
};

class DenseIO : public virtual IOutput {
protected:
    DenseIO(std::string name) {
		m_activation = new CudaMatrix();
		m_activation->SetName("a_" + name);
    }
    DenseIO(size_t batch, size_t size, std::string name) {
		m_activation = new CudaMatrix(batch, size, MatrixType::DENSE);
		m_activation->SetName("a_" + name);
    }
    DenseIO(size_t size, std::string name) {
		m_activation = new CudaMatrix(1, size, MatrixType::DENSE);
		m_activation->SetName("a_" + name);
    }
    
    virtual ~DenseIO() {
        my_delete(m_activation);
    }
};

class SparseIO : public virtual IOutput {
protected:
	SparseIO(std::string name) {
		m_activation = new CudaMatrix();
		m_activation->SetMatrixType(MatrixType::SPARSE);
		m_activation->SetName("a_" + name);
		//m_activation->Fill(0.f);
    }

    SparseIO(size_t batch, size_t size, std::string name) {
		m_activation = new CudaMatrix(batch, size, MatrixType::SPARSE);
		m_activation->SetName("a_" + name);
	}
	
    SparseIO(size_t size, std::string name) {
		m_activation = new CudaMatrix(1, size, MatrixType::SPARSE);
		m_activation->SetName("a_" + name);
	}
	
    virtual ~SparseIO() {
        my_delete(m_activation);
    }
};

class Layer : public virtual IOutput,
    public virtual IError,
    public Node {

private:
    ActivationFn* init_activation(std::string fn) {
        if (fn == "sigmoid") return new SigmoidFn();
        else if (fn == "tanh") return new TanhFn();
        else if (fn == "softmax") return new SoftmaxFn();
        else if (fn == "exp") return new ExpFn();
        else if (fn == "relu") return new ReLUFn();
        return new IdFn();
    }
protected:
    ActivationFn* fn_;		// activation function
    size_t m_batch;			// batched input
	size_t m_size;			// m_size of the layer
    
	bool m_skipReset;		// if this node should skip the reset visitor
	bool m_needError;		// if this node need an error

    bool m_evaluated;		// keep track if the node has already been evaluated
    bool m_errorated;		// keep track if the node has already been errorated
    
public:
    Layer(std::string name) : Node(name), fn_(0), m_size(0) {
        m_evaluated = false;
        m_errorated = false;
		m_skipReset = false;
		m_needError = false;
        fn_ = init_activation(ID);
    }
	
	Layer(Activation fn, std::string name) : Node(name), fn_(0), m_size(0) {
        m_evaluated = false;
        m_errorated = false;
		m_skipReset = false;
		m_needError = false;
        fn_ = init_activation(fn);
    }
    
    Layer(size_t size, std::string name) : Node(name), m_size(size), m_batch(1) {
        m_evaluated = false;
        m_errorated = false;
		m_skipReset = false;
		m_needError = false;
        fn_ = init_activation(ID);
    }
    
    Layer(size_t batch, size_t size, std::string name) : Node(name), m_size(size), m_batch(batch) {
        m_evaluated = false;
        m_errorated = false;
		m_skipReset = false;
		m_needError = false;
        fn_ = init_activation(ID);
    }
    
    Layer(size_t size, Activation fn, std::string name) : Node(name), m_batch(1), m_size(size)
    {
        m_evaluated = false;
        m_errorated = false;
		m_skipReset = false;
		m_needError = false;
        fn_ = init_activation(fn);
    }

	Layer(size_t batch, size_t size, Activation fn, std::string name) : 
		Node(name), m_batch(batch), m_size(size)
    {
        m_evaluated = false;
        m_errorated = false;
		m_skipReset = false;
		m_needError = false;
        fn_ = init_activation(fn);
    }
    
    virtual ~Layer() { my_delete(fn_); }    
    std::string ToString() {
        std::stringstream ss;
        ss << "Layer = " << GetName() << " / Size = " << m_size;
        return ss.str();
    }
    
    virtual CudaMatrix* GetValue();
    virtual CudaMatrix* ComputeError(Layer* parent) = 0;
	
	void SetSkipReset(bool v) { m_skipReset = v; }
	void SetErrorated(bool v) { m_errorated = v; }
	void SetEvaluated(bool v) { m_evaluated = v; }
	void SetNeedError(bool v) { m_needError = v; }
	
	size_t GetSize() { return m_size; } 
	size_t GetBatch() { return m_batch; }
	size_t GetTotalSize() { return m_size*m_batch; }

	bool IsErrorated() { return m_errorated; }
	bool IsEvaluated() { return m_evaluated; }
	bool IsSkipReset() { return m_skipReset; }
	bool IsNeedError() { return m_needError; }

    virtual void Visit(Visitor& c) = 0;
    void CheckEvaluate() {
        if (!m_evaluated) throw std::exception("Evaluation needed beforehand");
    }
    void CheckErrorate() {
        if (!m_errorated) throw std::exception("Error computation needed beforehand");
    }
};

// A general dense layer
class DenseLayer : public DenseError, public DenseIO, public Layer {
public:
	DenseLayer(std::string name) :
        DenseError(name),
        DenseIO(name),
        Layer(name) {};
	DenseLayer(Activation fn, std::string name) :
        DenseError(name),
        DenseIO(name),
        Layer(fn, name) {};
	DenseLayer(size_t size, std::string name) :
        DenseError(size, name),
        DenseIO(size, name),
        Layer(size, name) {};
	DenseLayer(size_t batch, size_t size, std::string name) :
        DenseError(batch, size, name),
        DenseIO(batch, size, name),
        Layer(batch, size, name) {};
	DenseLayer(size_t size, Activation fn, std::string name) :
        DenseError(size, name),
        DenseIO(size, name),
        Layer(size, fn, name) {};
	DenseLayer(size_t batch, size_t size, Activation fn, std::string name) :
        DenseError(batch, size, name),
        DenseIO(batch, size, name),
        Layer(batch, size, fn, name) {};
  

    virtual ~DenseLayer() {};
    virtual void Visit(Visitor& c) {
		c.Visit(this);
    }

    virtual CudaMatrix* ComputeError(Layer* parent);
};

// A hidden layer, softmax not allowed here!
class HiddenLayer : public DenseLayer {
public:
    HiddenLayer(size_t size, std::string name) : DenseLayer(size, name) {};
	HiddenLayer(Activation fn, std::string name) : DenseLayer(fn, name) {};
    HiddenLayer(size_t batch, size_t size, std::string name) : DenseLayer(batch, size, name) {};
    HiddenLayer(size_t size, Activation fn, std::string name) : DenseLayer(size, fn, name) {
        if (fn == SOFTMAX) throw std::exception("Softmax is not permitted in the hidden layers!");
    }
	HiddenLayer(size_t batch, size_t size, Activation fn, std::string name) : DenseLayer(batch, size, fn, name) {
        if (fn == SOFTMAX) throw std::exception("Softmax is not permitted in the hidden layers!");
    }
    
    virtual ~HiddenLayer() {};
};

class ConcatLayer : public HiddenLayer {
private:
	// Need an additional structure to keep parents in order.
    std::map<Layer*, int> m_orderedParents;
	std::map<Layer*, CudaMatrix*> m_parentErrors;

    int m_numInputs;
	int m_registeredInputs;

public:
    ConcatLayer(int numInputs, std::string name) : HiddenLayer(numInputs, name) {
		m_numInputs = numInputs;
		m_registeredInputs = 0;
	};

	ConcatLayer(int numInputs, Activation fn, std::string name) : HiddenLayer(numInputs, fn, name) {
		m_numInputs = numInputs;
		m_registeredInputs = 0;
	};
        
	void RegisterParent(Layer* parent) {
		auto finder = m_orderedParents.find(parent);
		if (finder != m_orderedParents.end()) {
			// Layer already registered -- error
			throw std::exception("Adding an already registered layer to ConcatLayer");
		}

		if (m_registeredInputs >= m_numInputs) {
			throw std::exception("Cannot register another child layer. Either enlarge the ConcatLayer"
				"or revise the layers already registered to it.");
		}
		
		m_orderedParents.insert(std::make_pair(parent, m_registeredInputs++));
	}

    virtual ~ConcatLayer() {
		for (auto it = m_parentErrors.begin(); it != m_parentErrors.end(); it++)
			if (it->second != nullptr) delete it->second;
	}

    virtual CudaMatrix* GetValue();
    virtual CudaMatrix* ComputeError(Layer* parent);
};

// A max-pooling layer, take the max!
class MaxPoolingLayer : public HiddenLayer {
private:
	CudaMatrix* m_stackedActivations;
	std::unordered_map<Layer*, int> m_backMap;
    
public:
    MaxPoolingLayer(std::string name) : HiddenLayer(ID, name) {
		m_stackedActivations = new CudaMatrix();
		m_stackedActivations->SetName(name + ".StackedActs");
	};
    
    virtual ~MaxPoolingLayer() {
		if (m_stackedActivations) delete m_stackedActivations;
    }
    
    virtual CudaMatrix* GetValue();
    virtual CudaMatrix* ComputeError(Layer* parent);
};

// A general loss-layer
class LossLayer : public Layer {
protected:
    // A multiplicative factor i.e. could be useful to do means
    // or stuff like that
    float multiplicative_factor_;
    float loss_;

public:
	LossLayer(Activation fn, std::string name) : Layer(fn, name) {
        multiplicative_factor_ = 1.0;
        loss_ = 0;
    };
    LossLayer(size_t size, Activation fn, std::string name) : Layer(size, fn, name) {
        multiplicative_factor_ = 1.0;
        loss_ = 0;
    };
    LossLayer(size_t batch, size_t size, Activation fn, std::string name) : Layer(batch, size, fn, name) {
        multiplicative_factor_ = 1.0;
        loss_ = 0;
    };

    // Compute the error
    float GetLoss() {
        CheckEvaluate();
        ComputeError(0);
        return loss_;
    }
    
    //
    virtual void Visit(Visitor& c) = 0;
    //
    
    void multiplicative_factor(float mf) {
        if (mf == 0) throw std::exception("0 as multiplicative factor is not allowed!");
        multiplicative_factor_ = mf;
    }

    float multiplicative_factor() {
        return multiplicative_factor_;
    }
};

class CrossEntropyLossLayer : public DenseIO, public DenseError, public LossLayer {
private:
    CudaMatrix* m_targets;

public:
	CrossEntropyLossLayer(std::string name) :
    DenseIO("CELoss_" + name),
    DenseError("CELoss_" + name),
    LossLayer(SOFTMAX, "CELoss_" + name)
    {
		m_targets = new CudaMatrix();
		m_targets->SetMatrixType(MatrixType::SPARSE);
		m_targets->SetName(m_name + "_targets");
	}
	
	~CrossEntropyLossLayer() {
		if (m_targets) delete m_targets;
	}
    
    float GetLoss() {
        CheckEvaluate();
        ComputeError(0);
        return loss_;
    }

	void SetOutput(size_t rows, size_t cols, int nnz, int* cooRows, int* cooCols, float* cooVal, bool onDevice=false, bool noCopy=false)
	{
		m_targets->SetData(rows, cols, nnz, cooRows, cooCols, cooVal, onDevice, false);
	}
    
    CudaMatrix* ComputeError(Layer* parent) {
        CheckEvaluate();
        
        if (!m_errorated) {
			CudaMatrix::AssignCE(*(m_activation->m_dense), *(m_targets->m_sparse), *(m_error->m_dense));
			CudaMatrix::AssignCELoss(*(m_activation->m_dense), *(m_targets->m_sparse), &loss_);
			SetErrorated(true);
        }
        
        return GetError();
    }
    
    void Visit(Visitor& c) {
        c.Visit(this);
    }
};

class NCELossLayer : public SparseIO, public SparseError, public LossLayer {
private:
	int m_nnoise;
	CudaMatrix probs;
    
public:
	NCELossLayer(int noise, std::string name) :
    SparseIO("NCELoss_" + name),
    SparseError("NCELoss_" + name),
    LossLayer(ID, "NCELoss_" + name)
    {
		probs.SetMatrixType(MatrixType::SPARSE);
		probs.SetName("NCELoss_"+name+"_probs");
		m_nnoise = noise;
	}

    NCELossLayer(size_t batch,
				size_t size,
				std::string name,
				int nnoise,
				int nnz,
				int* cooRow,
				int* cooCol,
				float* cooProbs,
				float* cooFreq) :
    SparseIO(batch, size, "NCELoss_" + name),
    SparseError(batch, size, "NCELoss_" + name),
    LossLayer(batch, size, ID, "NCELoss_" + name)
    {
		m_nnoise = nnoise;
		probs.SetMatrixType(MatrixType::SPARSE);
		probs.SetData(batch, size, nnz, cooRow, cooCol, cooProbs);		
		
		m_activation->SetData(batch, size, nnz, probs.m_sparse->m_cooRow, 
			probs.m_sparse->m_cooCol, probs.m_sparse->m_cooVal, true);
		m_activation->Fill(1);
	}

	NCELossLayer(size_t size,
				std::string name,
				int nnoise,
				int nnz,
				int* cooRow,
				int* cooCol,
				float* cooProbs,
				float* cooFreq) :
    SparseIO(1, size, "NCELoss_" + name),
    SparseError(1, size, "NCELoss_" + name),
    LossLayer(1, size, ID, "NCELoss_" + name)
    {
		m_nnoise = nnoise;
		probs.SetMatrixType(MatrixType::SPARSE);
		probs.SetData(1, size, nnz, cooRow, cooCol, cooProbs);
		m_activation->SetData(1, size, nnz, cooRow, cooCol, cooFreq);
	}
    
	void SetOutput(size_t rows, size_t cols, int nnz, int* cooRow, int* cooCol, float* cooProbs, float* cooFreq, 
		bool onDevice=false, bool noCopy=false)
	{
		m_activation->SetData(rows, cols, nnz, cooRow, cooCol, cooProbs, onDevice, noCopy);
		// We need only the values for the probabilities
		probs.m_sparse->SetCooValues(nnz, m_activation->m_sparse->m_cooVal, true);
		m_activation->Reset();
	}

    CudaMatrix* ComputeError(Layer* parent) {
        CheckEvaluate();
        
        if (!m_errorated) {
			CudaMatrix::AssignNCELoss(m_nnoise, *(m_activation->m_sparse), *(probs.m_sparse), &loss_);
			CudaMatrix::AssignNCE(m_nnoise, *(m_activation->m_sparse), *(probs.m_sparse), *(m_error->m_sparse));
			SetErrorated(true);
        }
        
        return GetError();
    }
    
    void Visit(Visitor& c) {
        c.Visit(this);
    }
};

class InputLayer : public Layer {
public:
	InputLayer(std::string name) : Layer(name) {
		m_skipReset = true;
		m_needError = false;
	}
    InputLayer(size_t size, std::string name) : Layer(size, name) {
		m_skipReset = true;
		m_needError = false;
	}
	InputLayer(size_t batch, size_t size, std::string name) : Layer(batch, size, name) {
		m_skipReset = true;
		m_needError = false;
	}
    CudaMatrix* ComputeError(Layer* parent);
	CudaMatrix* GetValue() { return GetActivation(); }
};

class DenseInputLayer : public DenseIO, public DenseError, public InputLayer {
public:
	DenseInputLayer(std::string name) :
        DenseIO(name),
        DenseError(name),
        InputLayer(name) {}
    DenseInputLayer(size_t size, std::string name) :
        DenseIO(size, name),
        DenseError(size, name),
        InputLayer(size, name) {}
    DenseInputLayer(size_t batch, size_t size, std::string name) :
        DenseIO(batch, size, name),
        DenseError(batch, size, name),
        InputLayer(batch, size, name) {}
    virtual ~DenseInputLayer() {};
    
    void SetInput(size_t rows, size_t cols, float* hData) {
		CudaMatrix* act = GetActivation();
		act->SetData(rows, cols, hData);
	}
    
    void Visit(Visitor& c) {
        c.Visit(this);
    }
};

class BiasLayer : public DenseInputLayer {
public:
    BiasLayer(size_t batch, std::string name) : DenseInputLayer(batch, 1, name) {
		m_activation->Fill(1.0);
	};

	void SetInput(size_t m_batch) {
		bool noFill = true;
		m_activation->Resize(m_batch, 1, noFill);
	}

    virtual ~BiasLayer() {}
    void Visit(Visitor& c) {
        c.Visit(this);
    }
};

class SparseInputLayer : public SparseIO, public SparseError, public InputLayer {
public:
    SparseInputLayer(std::string name) :
        SparseIO(name),
        SparseError(name),
        InputLayer(name) {};
    SparseInputLayer(size_t size, std::string name) :
        SparseIO(size, name),
        SparseError(size, name),
        InputLayer(size, name) {};
    SparseInputLayer(size_t batch, size_t size, std::string name) :
        SparseIO(batch, size, name),
        SparseError(batch, size, name),
        InputLayer(batch, size, name) {};
    virtual ~SparseInputLayer() {};
    
	void SetInput(size_t rows, size_t cols, int nnz, int* cooRow, int* cooCol, float* cooVal, bool onDevice=false, bool noCopy=false)
	{	
		CudaMatrix* act = GetActivation();
		act->SetData(rows, cols, nnz, cooRow, cooCol, cooVal, onDevice, noCopy);
	}

    void Visit(Visitor& c) {
        c.Visit(this);
    }
};

class Weight : public Link {
protected:
    CudaMatrix* m_;  // actual shared weight
    Memory mstrategy_;
    
public:
	Weight() { m_ = nullptr; }
	Weight(CudaMatrix* m) : Link(), m_(m), mstrategy_(SHARED) {
		m_name = m->GetName();
	};
    Weight(CudaMatrix* m, Memory strategy) : Link(), m_(m), mstrategy_(strategy) {
		m_name = m->GetName();
	};
    
    virtual ~Weight() {
        if (mstrategy_ == DELEGATE) {
            // Handle destruction
            if (m_ != nullptr) 
				delete m_; 
			m_ = nullptr;
        }
    };

    virtual void Visit(Visitor& c) = 0;    
    virtual CudaMatrix* m() {
        return m_;
    }

    /** Just helpers **/
    size_t GetRows() const { return m_->GetRows(); }
    size_t GetCols() const { return m_->GetCols(); }
    size_t GetSize() const { return m_->GetSize(); }
};

class FixedWeight : public Weight {
public:
	FixedWeight() : Weight() {};
    FixedWeight(CudaMatrix* m) : Weight(m) {};
    FixedWeight(CudaMatrix* m, Memory strategy) : Weight(m, strategy) {};
    virtual ~FixedWeight() {};
    
    void Visit(Visitor& c) {
        c.Visit(this);
    }
};

class DynamicWeight : public Weight {
public:
	DynamicWeight(CudaMatrix* m) : Weight(m) {};
    DynamicWeight(CudaMatrix* m, Memory strategy) : Weight(m, strategy) {};
    virtual ~DynamicWeight() {};
    
    virtual void Visit(Visitor& c) {
        c.Visit(this);
    }
};

// A very simple print visitor
class PrintVisitor : public Visitor {
public:
	void Visit(class DenseLayer* w) { std::cerr << "Visiting " << w->GetName() << std::endl; };
	void Visit(class HiddenLayer* w) { std::cerr << "Visiting " << w->GetName() << std::endl; };
	void Visit(class NCELossLayer* w) { std::cerr << "Visiting " << w->GetName() << std::endl; };
	void Visit(class CrossEntropyLossLayer* w) { std::cerr << "Visiting " << w->GetName() << std::endl; };
	void Visit(class SparseInputLayer* w) { std::cerr << "Visiting " << w->GetName() << std::endl; };
};

// This resets some information in the graph
class ResetVisitor : public Visitor {
public:
		void cleanLayer(Layer* l) {
			if (l->IsSkipReset()) 
				return;
#ifdef DEBUG
			std::cerr<<"Resetting the value of "<<l->GetName()<<std::endl;
#endif
			if (cleanError_ && l->GetError() != nullptr && l->IsErrorated())
			{ 
				l->GetError()->Reset(); 
				l->SetErrorated(false); 
			}

			if (cleanActivation_ && l->GetActivation() != nullptr && l->IsEvaluated())
			{	
				l->GetActivation()->Reset();  
				l->SetEvaluated(false); 
			}
		};

		bool cleanError_;
		bool cleanActivation_;
	
		ResetVisitor() : Visitor() {
			m_visitNodes = true;	// visit the nodes
			m_visitEdges = false;	// do not visit edges
			cleanError_ = true;
			cleanActivation_ = true;
		}

		void Visit(DynamicWeight* w) { };
		void Visit(FixedWeight* w) { };
		void Visit(DenseLayer* w) { cleanLayer(w); };
		void Visit(HiddenLayer* w) { cleanLayer(w); };
		
		void Visit(SparseInputLayer* w) { /* Do not clean inputs! */ };
		void Visit(DenseInputLayer* w) { /* Do not clean inputs! */ };

		void Visit(CrossEntropyLossLayer* w) { cleanLayer(w); };
		void Visit(NCELossLayer* w) { cleanLayer(w); };
		void Visit(LossLayer* w) { cleanLayer(w); };
};

class NeuralNet : public Graph {
private:
    /** Map of output layers **/
    std::vector<LossLayer*> O;
	std::vector<InputLayer*> I;

	// Optimization related stuff 
	std::map<std::string, Layer*> m_internMap;
	std::set<Layer*> m_internals;
	bool IsInternal(Layer* n);
	//
		
	int mbh_;

	// Barrier weight = error cannot flow behind this layer
	CudaMatrix* m_barrierWeight;
	//
	
	// Bptt algorithm related stuff
	int bpttStop;
	CudaMatrix* bpttWeight;
	//
		
	void RegisterLossLayer(LossLayer* layer) {
		AddNode(layer);
		O.push_back(layer);
	}

public:
	const std::vector<InputLayer*>& Inputs() { return I; }
	const std::vector<LossLayer*>& Outputs() { return O; }

	Layer* GetInternalForLayerParam(const std::string& key) 
	{
		auto it = m_internMap.find(key);
		if (it == m_internMap.end())
			return nullptr;
		return it->second;
	}

	void SetBptt(CudaMatrix* weight, int stop) {
		if (stop <= 0)
			throw std::exception("Bptt stop cannot be less than 1!");

		bpttStop = stop;
		bpttWeight = weight;
	}

	NeuralNet() : Graph(), mbh_(1) {
		bpttWeight = nullptr;
		m_barrierWeight = nullptr;
		bpttStop = 0;
	}
	// mbh - the number of minibatches to be processed at the same time
	NeuralNet(int mbh) : Graph(), mbh_(mbh) {
		bpttWeight = nullptr;
		m_barrierWeight = nullptr;
		bpttStop = 0;
	}

    SparseInputLayer* AddLayer(SparseInputLayer* layer) {
		I.push_back(layer);
		AddNode(layer);
		return layer;
    }
    
	BiasLayer* AddLayer(BiasLayer* layer) {
        I.push_back(layer);
		AddNode(layer);
        return layer;
    }
    
    DenseInputLayer* AddLayer(DenseInputLayer* layer) {
        I.push_back(layer);
		AddNode(layer);
        return layer;
    }
    
    CrossEntropyLossLayer* AddLayer(CrossEntropyLossLayer* layer) {
        RegisterLossLayer(layer);
        return layer;
    }
    
    NCELossLayer* AddLayer(NCELossLayer* layer) {
        RegisterLossLayer(layer);
        return layer;
    }
    
    template<class T> T* AddLayer(T* layer) {
		AddNode(layer);
        return layer;
    }
    
    std::vector<CudaMatrix*> Forward();		/* Compute the output from the net */
    CudaMatrix* Forward(LossLayer* ll);		/* Compute the output from a specific loss layer */
	
	void SetBarrierWeight(CudaMatrix* mat) {
		m_barrierWeight = mat;
	}

	void Backward();						/* Propagates the error back in the net */
    void Backward(LossLayer* ll);			/* Propagates the error back in the net from a specific loss layer */
	void Visit(Visitor& g);

    inline bool Connect(Layer *a, Layer *b, Weight *l) {
        /** Attach two nodes with a link
         ** a -> l -> b **/
        bool ret=true;

		if (a == 0 || b == 0 || l == 0) { 
			throw std::exception("Trying to add a Null node/link!"); 
		}
        
        // Insert if it is not there
        if (!ContainsLink(l)) E.insert(l);
		else { throw std::exception("Trying to add a Null link!"); }
		
		if (!ContainsNode(a)) throw std::exception("Should register the layer before linking!");
		if (!ContainsNode(b)) throw std::exception("Should register the layer before linking!");
        
        l->ConnectOutput(b); l->ConnectInput(a);
        ret &= a->ConnectOutput(l);
        ret &= b->ConnectInput(l);
        return ret;
    }
    
    Layer* GetLayerByName(const std::string& name) {
		Node* r = GetNodeByName(name);
        return (Layer*) r;
    }

	void DeleteLayer(Layer* l) {
		if (l == 0) return;

		// Unlink the outgoing links from the children
		if (l->GetChildren().size() > 0)
		for (Link* link : l->GetChildren()) {
			(link->child())->EraseInwardLink(link);
			RemoveLink(link);
		}
		
		// Unlink the incoming links from the parents
		if (l->GetParents().size() > 0)
		for (Link* link : l->GetParents()) {
			(link->parent())->EraseOutwardLink(link);
			RemoveLink(link);
		}

		// If this is a loss layer just find it in
		// the output set.
		LossLayer* ll = dynamic_cast<LossLayer*>(l);
		if (ll)
		{
			auto it = std::find(O.begin(), O.end(), ll);
			if (it != O.end()) O.erase(it);
		}
		
		// If this is an input layer just find it in
		// the input set
		InputLayer* il = dynamic_cast<InputLayer*>(l);
		if (il)
		{
			auto it = std::find(I.begin(), I.end(), il);
			if (it != I.end()) I.erase(it);
		}
		
		// and the node
		RemoveNode(l);
	}

	void DeleteInputs() {
		for (Layer* l : I) {
			for (Link* link : l->GetChildren()) {
				(link->child())->EraseInwardLink(link);
				// Delete both the link
				E.erase(link);
				delete link;
			}
			
			// and the node
			RemoveNode(l);
		}
		
		I.clear();
	}
	
	void DeleteOutputs() {
		for (Layer* l : O) {
			for (Link* link : l->GetParents()) {
				(link->parent())->EraseOutwardLink(link);
				// Delete both the link
				E.erase(link);
				delete link;
			}
			
			// and the node
			RemoveNode(l);
		}
		
		O.clear();
	}

	float GetLoss(LossLayer* ll) {
        float error = 0.;
        error += ll->GetLoss();
        return error;
    }

    float GetLoss() {
        float error = 0.;
        for (auto it = O.begin(); it != O.end(); it++) {
            error += (*it)->GetLoss();
        }

        return error;
    }
    
	// Clear inputs / outputs and
	// call the parent CleanUp method
	// which deallocates all memory
	void format() {
		O.clear();
		I.clear();
		m_internals.clear();
		CleanUp();
	}
		
	int Optimize();

    ~NeuralNet() {
		m_internals.clear();
		format();
	}
		
private:
	void FlagErrorNodes(LossLayer* ll);	
};