//
//  SparseCudaMatrix.cu
//  CudaMatrix
//
//  Created by Alessandro Sordoni on 2014-07-13.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//

#include "SparseCudaMatrix.hpp"

extern bool sync;

void SparseCudaMatrix::Dot(DenseCudaMatrix& B, DenseCudaMatrix& C, bool at, bool bt, bool ensureSpace) {
	float alpha=1;

	int m = (int) m_rows;
	int n = (int) B.m_cols;
	int k = (int) m_cols;
	int l = (int) B.m_rows;
	
	int nnz = m_nz;

	cudaEvent_t done;       
	if (sync) CUDA_CALL(cudaEventCreate(&done));

	if (at)
	{
		if (bt) NOT_IMPLEMENTED;
		
		C.Resize(k,n);
		GpuDotStDD<<<nnz, THREADS_PER_BLOCK>>>(
			alpha,
			m_cooVal,
			m_cooRow,
			m_cooCol,
			nnz,
			B.m_dData,
			l,
			n,
			C.m_dData,
			k);
	} 
	else
	{
		if (bt) C.Resize(m,l);
		else C.Resize(m,n);

		GpuDotSDD<<<nnz, THREADS_PER_BLOCK>>>(
			alpha,
			m_cooVal,
			m_cooRow,
			m_cooCol,
			nnz,
			B.m_dData,
			l,
			n,
			bt,
			C.m_dData,
			m);
	}

	if (sync) CUDA_CALL(cudaEventRecord(done));        
	if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void SparseCudaMatrix::Dot(DenseCudaMatrix& B, SparseCudaMatrix& C, bool at, bool bt, bool ensureSpace) {
	NOT_IMPLEMENTED;
}

void SparseCudaMatrix::InPlaceClip(float min, float max) 
{
	int numBlocks = (int) ceil(1.0*m_nz/THREADS_PER_BLOCK);

	cudaEvent_t done;
    if (sync) CUDA_CALL(cudaEventCreate(&done));
	GpuMinMaxClip<<<numBlocks, THREADS_PER_BLOCK>>>(m_nz, m_cooVal, min, max);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void
SparseCudaMatrix::AddTo(const DenseCudaMatrix& B, float f)
{
	assert(m_rows == B.m_rows && m_cols == B.m_cols);
	int numBlocks = (int) ceil(1.0*m_nz/THREADS_PER_BLOCK);

	cudaEvent_t done;
    if (sync) CUDA_CALL(cudaEventCreate(&done));
	GpuSDAdd<<<numBlocks, THREADS_PER_BLOCK>>>(B.m_rows, B.m_dData, m_nz, m_cooRow, m_cooCol, m_cooVal, f);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
    if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void
SparseCudaMatrix::InPlaceTanh()
{
	int numBlocks = (int) ceil(1.0*m_nz/THREADS_PER_BLOCK);

	cudaEvent_t done;       
	if (sync) CUDA_CALL(cudaEventCreate(&done));

	GpuInplaceTanh<<<numBlocks, THREADS_PER_BLOCK>>>(m_cooVal, m_nz);

	if (sync) CUDA_CALL(cudaEventRecord(done));        
	if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void
SparseCudaMatrix::InPlaceTanhDerivative(const SparseCudaMatrix& C)
{
	int numBlocks = (int) ceil(1.0*m_nz/THREADS_PER_BLOCK);

	cudaEvent_t done;       
	if (sync) CUDA_CALL(cudaEventCreate(&done));

	GpuInplaceTanhDerivative<<<numBlocks, THREADS_PER_BLOCK>>>(C.m_cooVal, m_cooVal, m_nz);

	if (sync) CUDA_CALL(cudaEventRecord(done));        
	if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void
SparseCudaMatrix::InPlaceSigmoid()
{
	int numBlocks = (int) ceil(1.0*m_nz/THREADS_PER_BLOCK);

	cudaEvent_t done;       
	if (sync) CUDA_CALL(cudaEventCreate(&done));

	GpuInplaceSigmoid<<<numBlocks, THREADS_PER_BLOCK>>>(m_cooVal, m_nz);

	if (sync) CUDA_CALL(cudaEventRecord(done));        
	if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void
SparseCudaMatrix::InPlaceSigmoidDerivative(const SparseCudaMatrix& C)
{
	int numBlocks = (int) ceil(1.0*m_nz/THREADS_PER_BLOCK);

	cudaEvent_t done;       
	if (sync) CUDA_CALL(cudaEventCreate(&done));

	GpuInplaceSigmoidDerivative<<<numBlocks, THREADS_PER_BLOCK>>>(C.m_cooVal, m_cooVal, m_nz);

	if (sync) CUDA_CALL(cudaEventRecord(done));        
	if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void SparseCudaMatrix::InPlaceScale(float f) {
	float alpha = (float) f;
	cudaEvent_t done;       
	if (sync) CUDA_CALL(cudaEventCreate(&done));
	cublasSscal(m_cubHandle, m_nz, &alpha, m_cooVal, 1);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
	if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

float SparseCudaMatrix::Norm2() {
	float norm = 0;
	cudaEvent_t done;       
	if (sync) CUDA_CALL(cudaEventCreate(&done));
	cublasSnrm2(m_cubHandle, m_nz, m_cooVal, 1, &norm);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
	if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
	return norm;
}

void SparseCudaMatrix::InPlaceClip(float v)
{	
	float n = Norm2();
	if (n > v)
		InPlaceScale(v/n);
}

void SparseCudaMatrix::SetCooValues(int nz, const float *cooVal, bool onDevice)
{
	if (nz > m_allocated)
	{
		if (m_cooVal != NULL)
			cudaFree(m_cooVal);
		CUDA_CALL(cudaMalloc(&m_cooVal, nz*sizeof(float)));
	}

	m_nz = nz;
	cudaMemcpy(m_cooVal, cooVal, nz*sizeof(float), (onDevice) ? cudaMemcpyDeviceToDevice : cudaMemcpyHostToDevice);
}

void SparseCudaMatrix::EnsureSpace(int nz)
{
	if (nz > m_allocated) {
		if (m_cooRow != NULL)
			cudaFree(m_cooRow);
		if (m_cooCol != NULL)
			cudaFree(m_cooCol);
		if (m_cooVal != NULL)
			cudaFree(m_cooVal);

		CUDA_CALL(cudaMalloc(&m_cooRow, nz*sizeof(int)));
		CUDA_CALL(cudaMalloc(&m_cooCol, nz*sizeof(int)));
		CUDA_CALL(cudaMalloc(&m_cooVal, nz*sizeof(float)));
		m_allocated = nz;
	}
}

void SparseCudaMatrix::SetData(size_t rows, size_t cols, int nz,
	const int* cooRow, const int* cooCol, const float* cooVal, bool onDevice, bool noCopy)
{
	Resize(rows, cols);
	if (!noCopy) EnsureSpace(nz);

	m_nz = nz;
	cudaEvent_t done;       
	if (sync) CUDA_CALL(cudaEventCreate(&done));

	if (noCopy) 
	{
		m_cooRow = (int*) cooRow;
		m_cooCol = (int*) cooCol;
		m_cooVal = (float*) cooVal;
	}
	else
	{
		cudaMemcpy(m_cooRow, cooRow, nz*sizeof(int), (onDevice) ? cudaMemcpyDeviceToDevice : cudaMemcpyHostToDevice);
		cudaMemcpy(m_cooCol, cooCol, nz*sizeof(int), (onDevice) ? cudaMemcpyDeviceToDevice : cudaMemcpyHostToDevice);
		cudaMemcpy(m_cooVal, cooVal, nz*sizeof(float), (onDevice) ? cudaMemcpyDeviceToDevice : cudaMemcpyHostToDevice);
	}

	// cusparseXcoo2csr(m_handle, m_cooRow, m_nz, (int) rows, m_rowPtr, CUSPARSE_INDEX_BASE_ZERO);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
	if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}

void SparseCudaMatrix::Fill(float v) 
{
	int blocksPerGrid =(int)ceil(1.0*m_nz/THREADS_PER_BLOCK);

	cudaEvent_t done;       
	if (sync) CUDA_CALL(cudaEventCreate(&done));
	GpuSetValue<<<blocksPerGrid,THREADS_PER_BLOCK>>>(m_cooVal, v, m_nz);
	if (sync) CUDA_CALL(cudaEventRecord(done));        
	if (sync) CUDA_CALL(cudaEventSynchronize(done));
	if (sync) CUDA_CALL(cudaEventDestroy(done));
}