//
//  SparseCudaMatrix.hpp
//  CudaMatrix
//
//  Created by Alessandro Sordoni on 2014-07-13.
//  Copyright (c) 2014 Alessandro Sordoni. All rights reserved.
//

#pragma once

#include <string>

#include "CudaOps.hpp"
#include "DenseCudaMatrix.hpp"

class DenseCudaMatrix;

class SparseCudaMatrix {
public:
    int* m_cooRow;
    int* m_cooCol;
    float* m_cooVal;
    int m_nz;
	int m_allocated;
    int* m_rowPtr;
    
    std::string m_name;
    size_t m_rows;
    size_t m_cols;
    size_t m_size;
    
	cublasHandle_t m_cubHandle;
    cusparseHandle_t m_handle;
    cusparseMatDescr_t m_desc;
    
public:
    SparseCudaMatrix() {
        m_cooRow = NULL;
        m_cooCol = NULL;
        m_cooVal = NULL;
        m_rowPtr = NULL;
        m_name = "";
        m_rows = 0;
        m_cols = 0;
        m_size = 0;
        m_nz = 0;
        m_allocated = 0;
        
        cublasCreate(&m_cubHandle);
        cusparseCreate(&m_handle);
        cusparseCreateMatDescr(&m_desc);
        cusparseSetMatType(m_desc, CUSPARSE_MATRIX_TYPE_GENERAL);
        cusparseSetMatIndexBase(m_desc, CUSPARSE_INDEX_BASE_ZERO);
    }
    
    void SetName(const std::string& name) { m_name = name; }
	std::string GetName() { return m_name; }
    
    void Resize(size_t rows, size_t cols, bool noFill=false) {
        if (rows > m_rows)
        {
			// do not use this for the moment
            //if (m_rowPtr != NULL)
            //   cudaFree(m_rowPtr);
			//cudaMalloc(&m_rowPtr, (rows+1)*sizeof(int));
        }
        
        m_rows = rows;
        m_cols = cols;
        m_size = rows*cols;
    }
    
	float GetLastNnz() 
	{
		float h = 0;
		cudaMemcpy(&h, &m_cooVal[m_nz-1], 1*sizeof(float), cudaMemcpyDeviceToHost);
		return h;	
	}

    void Reset() { CUDA_CALL(cudaMemset(m_cooVal, 0, m_nz*sizeof(float))); }
	void EnsureSpace(int nz);
    void SetData(size_t rows, size_t cols, int nz, const int* cooRow, const int* cooCol,
                 const float* cooVal, bool onDevice=false, bool forceCopy=false);
	void SetCooValues(int nz, const float *cooVal, bool onDevice=false);

    void Dot(DenseCudaMatrix& B, DenseCudaMatrix& C, bool at=false, bool bt=false, bool ensureSpace=false);
	void Dot(DenseCudaMatrix& B, SparseCudaMatrix& C, bool at=false, bool bt=false, bool ensureSpace=false);

    void AddTo(const DenseCudaMatrix& B, float f=1.0);

    void InPlaceSigmoid();
    void InPlaceSigmoidDerivative(const SparseCudaMatrix& C);
    void InPlaceTanh();
    void InPlaceTanhDerivative(const SparseCudaMatrix& C);
    
    void Print() {
        float* hVal = (float*)malloc(sizeof(float)*m_nz);
        int* hRow = (int*)malloc(sizeof(int)*m_nz);
        int* hCol = (int*)malloc(sizeof(int)*m_nz);
        cudaMemcpy(hVal, m_cooVal, m_nz*sizeof(float),cudaMemcpyDeviceToHost);
        cudaMemcpy(hRow, m_cooRow, m_nz*sizeof(int),cudaMemcpyDeviceToHost);
        cudaMemcpy(hCol, m_cooCol, m_nz*sizeof(int),cudaMemcpyDeviceToHost);
        std::cerr<<"Matrix: "<<m_name<<" "<<m_rows<<" "<<m_cols<<std::endl;
        for (int i = 0; i < m_nz; i++)
            std::cerr<<"("<<hRow[i]<<","<<hCol[i]<<")="<<hVal[i]<<" ";
        std::cerr<<std::endl;
        delete hVal;
        delete hRow;
        delete hCol;
    }
    
	void InPlaceClip(float min, float max);
    void InPlaceClip(float v);
    void InPlaceScale(float v);
	float Norm2();
	void Fill(float v);

    ~SparseCudaMatrix() {
        if (m_cubHandle) cublasDestroy(m_cubHandle);
        if (m_handle) cusparseDestroy(m_handle);
		
		if (m_allocated > 0) 
		{
			if (m_cooRow) cudaFree(m_cooRow);
			if (m_cooCol) cudaFree(m_cooCol);
			if (m_cooVal) cudaFree(m_cooVal);
			if (m_rowPtr) cudaFree(m_rowPtr);
		}
	}
};
