#!/usr/bin/python
# Versatile implementation of Bleu4.
#  
# @author Alessandro Sordoni

import sys
import math
import copy
import re
import operator
import collections
import numpy

def GetCloserRefLength(refsTokens, candTokens):
	refsLengths = [len(x) for x in refsTokens]
	refsLengths.sort()
	
	candLen = len(candTokens)
	differences = []
	for x in refsLengths:
		differences.append(numpy.abs(x-candLen))
	
	minLen = min(differences)
	minRefLength = [refsLengths[i] for i, v in enumerate(differences) if v == minLen]
	return min(minRefLength)

def BuildNgramDict(tokensLists):
	""" Build a ngram dictionary from list of tokenized strings
	"""
	ngramDict = collections.defaultdict(int)
	for tokenList in tokensLists:
		currDict = collections.defaultdict(int)
		refLen = len(tokenList)
		
		for i in range(refLen):
			#### Unigrams
			currDict[tokenList[i]] += 1
			#### Bigrams
			if i+2 <= refLen:
				currDict[" ".join(tokenList[i:i+2])] += 1
			#### Trigrams
			if i+3 <= refLen:
				currDict[" ".join(tokenList[i:i+3])] += 1
			#### Fourgrams
			if i+4 <= refLen:
				currDict[" ".join(tokenList[i:i+4])] += 1
			
		### Store maximum occurrence
		for key, value in currDict.items():
			ngramDict[key] = max(ngramDict[key], value)
		
	return ngramDict

class SuffStats:
	def __init__(self, listOfStrings):
		if not isinstance(listOfStrings, list):
			listOfStrings = [listOfStrings]
		
		self.listOfTokenized = [x.strip().split(" ") for x in listOfStrings]
		self.ngramDict = BuildNgramDict(self.listOfTokenized)

class Bleu4:
	def __init__(self):
		# Statistics are
		# - 1-gramcount,
		# - 2-gramcount,
		# - 3-gramcount,
		# - 4-gramcount,
		# - 1-grammatch,
		# - 2-grammatch,
		# - 3-grammatch,
		# - 4-grammatch,
		# - reflen
		self.statistics = []
		self.precisions = []
		self.lp = 1
		
		self.numberOfUpdates = 0
		self.p = re.compile('[^a-zA-Z0-9\t ]+')
		
	def GetNgramCount(self, s, k):
		return max(len(s) - k + 1, 0)

	def SumStatistics(self):
		if len(self.statistics) == 0:
			return [0, 0, 0, 0, 0, 0, 0, 0, 0]
		
		statMatrix = numpy.array(self.statistics)
		statCumulative = numpy.sum(statMatrix, axis=0)
		return statCumulative
	
	def Get1GramCount(self):
		stats = self.SumStatistics()
		return stats[0]
	
	def Get2GramCount(self):
		stats = self.SumStatistics()
		return stats[1]
	
	def Get3GramCount(self):
		stats = self.SumStatistics()
		return stats[2]
	
	def Get4GramCount(self):
		stats = self.SumStatistics()
		return stats[3]
	
	def Get1GramMatch(self):
		stats = self.SumStatistics()
		return stats[4]
	
	def Get2GramMatch(self):
		stats = self.SumStatistics()
		return stats[5]
	
	def Get3GramMatch(self):
		stats = self.SumStatistics()
		return stats[6]
	
	def Get4GramMatch(self):
		stats = self.SumStatistics()
		return stats[7]
	
	def GetRefLen(self):
		stats = self.SumStatistics()
		return stats[8]
	
	def UpdateNgramMultiRefStatsFromSuffStats(self, refsSuffStats, candSuffStats):
		newStats = [0, 0, 0, 0, 0, 0, 0, 0, 0]
		
		refLength = GetCloserRefLength(refsSuffStats.listOfTokenized, candSuffStats.listOfTokenized[0])
		newStats[8] = refLength
		
		for key, value in candSuffStats.ngramDict.items():
			n = key.count(" ") ## Length of n gram
			newStats[n + 4] += min(value, refsSuffStats.ngramDict[key])
		
		newStats[0] = self.GetNgramCount(candSuffStats.listOfTokenized[0], 1)
		newStats[1] = self.GetNgramCount(candSuffStats.listOfTokenized[0], 2)
		newStats[2] = self.GetNgramCount(candSuffStats.listOfTokenized[0], 3)
		newStats[3] = self.GetNgramCount(candSuffStats.listOfTokenized[0], 4)
		
		self.statistics.append(newStats)
	
	### !!!!!!!!!!!!!!!! Old Method (DEPRECATED)
	def UpdateNgramMultiRefStat(self, refsTokens, candTokens):
		newStats = [0, 0, 0, 0, 0, 0, 0, 0, 0]
		
		refLength = GetCloserRefLength(refsTokens, candTokens)
		newStats[8] = refLength
		
		for kn in range(4):
			refsDict = []
			for refTokens in refsTokens:
				refDict = collections.defaultdict(int)
				lenRef = len(refTokens)
				
				for i in range(lenRef):
					if i+kn+1 <= lenRef:
						refDict[" ".join(refTokens[i:i+kn+1])] += 1
				refsDict.append(refDict)
			
			candDict = collections.defaultdict(int)
			for i in range(len(candTokens)):
				if i+kn+1 <= len(candTokens):
					candDict[" ".join(candTokens[i:i+kn+1])] += 1
			
			for key, value in candDict.items():
				maxJointCount = 0
				minCount = 0
				for refDict in refsDict:
					if key in refDict:
						minCount = min(refDict[key], value)
					maxJointCount = max(minCount, maxJointCount)
				newStats[kn + 4] += maxJointCount
			newStats[kn] = self.GetNgramCount(candTokens, kn+1)
			
		self.statistics.append(newStats)
	### !!!!!!!!!!!!!!!! END DEPRECATED
	
	def UpdateNgramMultiRefStats(self, refs, cands, candsScore=[]):
		""" refs is a list of strings
			cands is a (list) of strings
			k is a list of integers
		"""
		if not isinstance(cands, list):
			cands = [cands]
			candsScore = [1]
		
		if len(cands) != len(candsScore):
			print cands
			print candsScore
			print >> sys.stderr, "Number of candidates and scores should be equal"
			sys.exit(-1)
				
		newStats = [0, 0, 0, 0, 0, 0, 0, 0, 0]
		
		refsSuffStats = SuffStats(refs)
		
		# Normalize scores
		scores = candsScore
		if len(candsScore) != 1:
			# log rank + 1
			scores = range(len(scores))
			for i in range(len(scores)):
				scores[i] = numpy.exp(1.0/(numpy.log(scores[i]+2)/numpy.log(2)))
			#gamma = 1
			#scores = numpy.exp(numpy.array(candsScore))
			scores = scores/numpy.sum(scores)
		
		for idx, candidate in enumerate(cands):
			# Closer ref
			candSuffStats = SuffStats(candidate)
			refLength = GetCloserRefLength(refsSuffStats.listOfTokenized, candSuffStats.listOfTokenized[0])
			
			# Update reflen
			newStats[8] += refLength*scores[idx]
			
			for key, value in candSuffStats.ngramDict.items():
				n = key.count(" ") ## Length of n gram
				newStats[n + 4] += min(value, refsSuffStats.ngramDict[key])*scores[idx]
		
			newStats[0] += self.GetNgramCount(candSuffStats.listOfTokenized[0], 1)*scores[idx]
			newStats[1] += self.GetNgramCount(candSuffStats.listOfTokenized[0], 2)*scores[idx]
			newStats[2] += self.GetNgramCount(candSuffStats.listOfTokenized[0], 3)*scores[idx]
			newStats[3] += self.GetNgramCount(candSuffStats.listOfTokenized[0], 4)*scores[idx]
		
		self.statistics.append(newStats)
	
	def PrintStatistics(self):
		print "P@1 = %f, P@2 = %f, P@3 = %f, P@4 = %f, BP = %f" % (self.precisions[0], self.precisions[1], self.precisions[2], self.precisions[3], self.lp)
	
	def GetB4(self, smoothing=0, lpPenalty=1):
		self.precisions = []
		statMatrix = numpy.array(self.statistics)
		
		### Sum over the rows
		statCumulative = numpy.sum(statMatrix, axis=0)
		
		for i in range(4):
			### Compute the precisions
			nom = float(statCumulative[i+4]+smoothing)
			denom = float(statCumulative[i]+2*smoothing)
			self.precisions.append(nom/denom)
		
		bleu4 = 0
		bleu0 = False
		for i in range(4):
			if self.precisions[i] != 0:
				bleu4 += numpy.log(self.precisions[i])
			else:
				bleu0 = True
		bleu4 = numpy.exp(bleu4*0.25)
		
		### Length penalty
		self.lp = 1
		statCumulative[-1] = statCumulative[-1]*lpPenalty
		if statCumulative[0] < statCumulative[-1]:
			self.lp = numpy.exp(1 - float(statCumulative[-1])/float(statCumulative[0]))
			bleu4 *= self.lp
		
		# print "# Updates %d" % self.numberOfUpdates
		# self.PrintStatistics()
		
		if bleu0:
			bleu4 = 0
		
		return bleu4
	
	def ResetStats(self):
		self.statistics = []
		self.precisions = []
		self.numberOfUpdates = 0
		self.lp = 0
	
	def UpdateMultiRefStats(self, ref, cand, scores=[]):
		if not isinstance(ref, list):
			print >> sys.stderr, "Should specify a reference list!"
			sys.exit(-1)
		
		#refTokens = [x.strip().split(" ") for x in ref]
		#candTokens = cand.strip().split(" ")
		
		#self.UpdateNgramMultiRefStat(refTokens, candTokens)
		self.UpdateNgramMultiRefStats(ref, cand, scores)

def ComputeBLEU4(refPath, candPath):
	refReader = open(refPath)
	candReader = open(candPath)
	
	bleuComputer = Bleu4()
	bleu4 = 0
	done = 0
	
	refs = refReader.readlines()
	cands = candReader.readlines()
	if len(refs) != len(cands):
		print "Error! Mismatched number of lines!"
		sys.exit(-1)
	
	for ref, cand in zip(refs, cands):
		bleuComputer.UpdateStats(ref, cand)
	
	return bleuComputer.GetSentenceB4()

##
## here is the main function
##
if __name__ == '__main__':
	if len(sys.argv) != 3:
		print "arguments missing"
		sys.exit()
		
	bleu4 = ComputeBLEU4(sys.argv[1], sys.argv[2])
	print "BLEU4 is %f" % (bleu4)