#!/usr/bin/python

## Compute the NDCG@k score of
## two lists. The two files should be in the same format CID:text \t MID:text \t score:RID:text \t score:RID:text \t score:RID:text \t etc...
## The first file is the PREL FILE which contain the true REL
## The second file is the MODEL FILE from the retrieval model we want to evaluate
## The third parameter is the cutoff k.

import sys
import math
import copy
import re
import operator
import ir_reader

def bin(f):
	if f < .2:
		return 1
	if f < .4:
		return 2
	if f < .6:
		return 3
	if f < .8:
		return 4
	return 5

def get_ndcg(k, rank_scores):
	'''This is a function to get the value of DCG@k given a [(rel_id, rank), ...] list. '''
	dcg = 0
		
	for i in range(len(rank_scores)):
		rel_i = bin(rank_scores[i][0])
		rank_i = rank_scores[i][1]
		
		if rank_i <= k:
			dcg += (math.pow(2, rel_i)-1)/math.log(rank_i + 1,2)
		else:
			break   ### it cannot be further down in the list!
	return dcg
	
def extract_scores_and_id(tailRecords):
	scores = [float(tail.score) for tail in tailRecords]
	texts = [tail.rtext for tail in tailRecords]
	return scores, texts

def build_reli_ranki(text_to_reli, ranked_texts):
	#### We take the rank from ranked texts and the rel to reli
	reli_ranki = []
	#### print ranked_texts	
	for i in range(len(ranked_texts)):
		key_text = ranked_texts[i][0]
		key_rank = i+1
		
		if key_text in text_to_reli:
			reli_ranki.append((text_to_reli[key_text], key_rank))
		### else it is zero but we don't append it!
	return reli_ranki

def avg_ndcg(cutoff, prelRecords, retReader):
	k = int(cutoff)
			
	ndcg = 0
	done = 0
	
	for retID, retValues in retRecords.items():
		ret_scores, ret_texts = extract_scores_and_id(retValues.rlist)
		
		### Gather the results corresponding to this ID
		if retID not in prelRecords:
			### Silent here... print "Error key not found in prels ! ", retID
			continue
		
		relValues = prelRecords[retID]
		rel_scores, rel_texts = extract_scores_and_id(relValues.rlist)
		
		rel_zipped = zip(rel_texts, rel_scores)
		
		#### dictionary of relevant text -> relevant score
		rel_map = dict(rel_zipped)
		
		#### prels => (text, score) > (text, score) > ... > (text, score)
		rel_zip_ranked = sorted(rel_zipped, key=operator.itemgetter(1), reverse=True)
		
		#### build true (rel_i, rank_i) array
		true_rel_rank = build_reli_ranki(rel_map, rel_zip_ranked)
		
		ret_zipped = zip(ret_texts, ret_scores)
		
		#### model => (text, score) > (text, score) > ... > (text, score)
		ret_zip_ranked = sorted(ret_zipped, key=operator.itemgetter(1), reverse=True)
		
		#### build retrieved (rel_i, rank_i) array
		model_rel_rank = build_reli_ranki(rel_map, ret_zip_ranked)
		
		### Get max DCG for this list
		max_dcg = get_ndcg(k, true_rel_rank)
		dcg = get_ndcg(k, model_rel_rank)
		
		ndcg += dcg/max_dcg
		done += 1
	return ndcg/done
##
## here is the main function
##
if __name__ == '__main__':
	
	if len(sys.argv) != 3:
		print "arguments missing"
		sys.exit()
	
	prelReader = ir_reader.IRReader(sys.argv[1])
	retReader = ir_reader.IRReader(sys.argv[2])
	
	prelRecords = prelReader.GetRecords()
	retRecords = retReader.GetRecords()
	
	print "Read ", len(prelRecords), " prels"
	print "Read ", len(retRecords), " rets"
	
	ndcg1 = avg_ndcg(1, prelRecords, retRecords)
	ndcg5 = avg_ndcg(5, prelRecords, retRecords)
	ndcg10 = avg_ndcg(10, prelRecords, retRecords)
	ndcg100 = avg_ndcg(100, prelRecords, retRecords)
	
	print "NDCG@%d is %f" % (1, ndcg1)
	print "NDCG@%d is %f" % (5, ndcg5)
	print "NDCG@%d is %f" % (10, ndcg10)
	print "NDCG@%d is %f" % (100, ndcg100)
	